(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import name.
Require Import expressionSpi.
Require Import messageSpi.
Require Import formulaeSpi.
Require Import evalSpi.
Require Import process.
Require Import actionType.
Require Import Arith.
Require Import deBruijnType.
Require Import tactics.
Require Import missing.
Require Import ordmissing.
Require Import actionSpi.

Require Import calculi.

Module ActionSpiNC <: ActionType with Module A := AgentSpi.
  Module A := AgentSpi.

  Definition constraint := list Name.t.

  Fixpoint remove (n:Name.t) (l:constraint) {struct l} : constraint := 
    match l with
      | nil => nil
      | cons x xs => 
	(match Name.Ord.compare n x with
	   | EQ _ => remove n xs
	   | _ => cons x (remove n xs)
	 end)
    end.

  Lemma remove_map : forall (f:nat->Name.t->Name.t)(n:Name.t)(l:list Name.t)(d:nat),(injective (f d))->(remove (f d n) (deBruijn.map_list f d l))=(deBruijn.map_list f d (remove n l)).
    intros.
    induction l;simpl;try reflexivity.
    repeat (case Name.Ord.compare;intro);simpl;rewrite IHl;auto;try (rewrite <- e in l0;rewrite H0 in l0;elim (Name.Ord.lt_not_eq n n l0);auto);rewrite (H n a) in l0;try (rewrite <- e;trivial);try (elim (Name.Ord.lt_not_eq (f d a) (f d a) l0));try (elim (Name.Ord.lt_not_eq a a l0));auto.
  Qed.

  Lemma remove_app : forall (n:Name.t)(l1 l2:constraint),(remove n (app l1 l2))=(app (remove n l1) (remove n l2)).
    induction l1;simpl;intros;try reflexivity.
    case Name.Ord.compare;intros;rewrite IHl1;try reflexivity.
  Qed.

  Lemma fn_remove : forall (n:nat)(l:constraint),deBruijn.mem_fn Name.deBruijn_fn_base_list n 0 (remove (Name.Ref n) l) = false.
    induction l;try reflexivity.
    destruct a;simpl.
    case (OrderedNat.compare n n0);intro;auto;simpl;unfold DeBruijnNat.mem_fn;simpl;case eq_nat_dec;intro;try (subst n0;elim (OrderedNat.lt_not_eq n n);auto);auto.
  Qed.

  Lemma remove_nothing : forall (n:nat)(l:constraint),(deBruijn.mem_fn Name.deBruijn_fn_base_list n 0 l = false)->(remove (Name.Ref n) l = l).
    induction l;simpl;intros;try reflexivity.
    destruct a.
    generalize H;case (OrderedNat.compare n n0);intro;auto;simpl;unfold DeBruijnNat.mem_fn;simpl;case eq_nat_dec;intro;try (subst n0;elim (OrderedNat.lt_not_eq n n);auto;fail);intros;auto.    
    rewrite IHl;auto.
    rewrite deBruijn.fold_list_true in H0;discriminate H0.
    elim n1;auto.
    rewrite IHl;auto.
  Qed.

  Module OrdListName := OrderedList Name.Ord.

  Lemma OrdListEq_equal : forall (x y:constraint),(OrdListName.eq x y)<->x=y.
    split.
    generalize y;clear y.
    induction x.
    simpl.
    destruct y;try tauto.
    simpl.
    destruct y;try tauto.
    intros;and_elim.
    rewrite H.
    rewrite (IHx y);auto.
    intro.
    subst y;auto.
  Qed.

  Definition new_c (l:constraint) := deBruijn.low1 Name.deBruijn_base_list 0 (remove (Name.Ref 0) l).

  Lemma new_c_nil : new_c nil = nil.
    reflexivity.
  Qed.

  Lemma new_c_cons_0 : forall (xs:list Name.t),(new_c (cons (Name.Ref 0) xs))=(new_c xs).
    intros.
    reflexivity.
  Qed.

  Lemma new_c_cons_S : forall (n:nat)(xs:list Name.t),(new_c (cons (Name.Ref (S n)) xs))=(cons (Name.Ref n) (new_c xs)).
    intros.
    reflexivity.
  Qed.

  Lemma new_c_app : forall (l l':constraint),(new_c (app l l'))=(app (new_c l) (new_c l')).
    induction l;simpl;auto.
    intros.
    destruct a.
    destruct n.
    repeat (rewrite new_c_cons_0).
    auto.
    repeat (rewrite new_c_cons_S).
    rewrite IHl.
    simpl.
    auto.
  Qed.

  Lemma lift_app : forall (d k:nat)(x y:list Name.t),(deBruijn.lift Name.deBruijn_base_list d k (x++y))=(deBruijn.lift Name.deBruijn_base_list d k x)++(deBruijn.lift Name.deBruijn_base_list d k y).
    simpl;intros.
    rewrite deBruijn.map_list_app.
    auto.
  Qed.

  Lemma swapR_app : forall (d k:nat)(x y:list Name.t),(deBruijn.swapR Name.deBruijn_base_list d k (x++y))=(deBruijn.swapR Name.deBruijn_base_list d k x)++(deBruijn.swapR Name.deBruijn_base_list d k y).
    simpl;intros.
    rewrite deBruijn.map_list_app.
    auto.
  Qed.

  Lemma low1_app : forall (d:nat)(x y:list Name.t),(deBruijn.low1 Name.deBruijn_base_list d (x++y))=(deBruijn.low1 Name.deBruijn_base_list d x)++(deBruijn.low1 Name.deBruijn_base_list d y).
    simpl.
    intros.
    rewrite deBruijn.map_list_app.
    auto.
  Qed.

  Lemma new_c_scope_1 : forall (l l':constraint),(app (new_c l) l')=(new_c (app l (deBruijn.lift Name.deBruijn_base_list 0 1 l'))).
    intros.
    unfold new_c.
    rewrite remove_app.
    rewrite low1_app.
    rewrite (remove_nothing 0 (deBruijn.lift Name.deBruijn_base_list 0 1 l')).
    rewrite (deBruijn.low1_lift_simpl Name.deBruijn_base_list 0 0 l').
    rewrite deBruijn.lift_0;auto.
    apply (deBruijn.lift_not_fn Name.deBruijn_fn_base_list 0 1 0).
    auto.
  Qed.

  Lemma new_c_scope_2 : forall (l l':constraint),(app l (new_c l'))=(new_c (app (deBruijn.lift Name.deBruijn_base_list 0 1 l) l')).
    intros.
    unfold new_c.
    rewrite remove_app.
    rewrite low1_app.
    rewrite (remove_nothing 0 (deBruijn.lift Name.deBruijn_base_list 0 1 l)).
    rewrite (deBruijn.low1_lift_simpl Name.deBruijn_base_list 0 0 l).
    rewrite deBruijn.lift_0;auto.
    apply (deBruijn.lift_not_fn Name.deBruijn_fn_base_list 0 1 0).
    auto.
  Qed.
 
  Lemma new_c_swap : forall (l:constraint),new_c (new_c (deBruijn.swapR Name.deBruijn_base_list 0 1 l)) = (new_c (new_c l)).
    induction l;try reflexivity.
    replace (cons a l) with ((cons a nil) ++ l);try reflexivity.
    rewrite swapR_app.
    unfold new_c.
    repeat (rewrite remove_app || rewrite low1_app).
    fold (new_c (deBruijn.swapR Name.deBruijn_base_list 0 1 l)).
    fold (new_c (new_c (deBruijn.swapR Name.deBruijn_base_list 0 1 l))).
    rewrite IHl.
    fold (new_c l).
    fold (new_c (new_c l)).
    destruct a.
    destruct n;try reflexivity.
    destruct n;try reflexivity.
  Qed.

  Lemma lift_new_c : forall (d k:nat)(l:constraint),(deBruijn.lift Name.deBruijn_base_list d k (new_c l))=(new_c (deBruijn.lift Name.deBruijn_base_list (S d) k l)).
    intros.
    unfold new_c.
    rewrite <- (deBruijn.lift_low1_comm Name.deBruijn_fn_base_list 0 d k).
    simpl.
    generalize (remove_map (fun d => deBruijn.lift Name.deBruijn_base d k) (Name.Ref 0)).
    intros.
    rewrite <- H.
    simpl.
    unfold DeBruijnNat.lift;simpl.
    auto.
    apply deBruijn.lift_injective.
    apply fn_remove.
  Qed.

  Lemma swapR_new_c : forall (d k:nat)(l:constraint),(deBruijn.swapR Name.deBruijn_base_list d k (new_c l))=(new_c (deBruijn.swapR Name.deBruijn_base_list (S d) k l)).
    intros.
    unfold new_c.
    rewrite <- (deBruijn.swapR_low1_comm Name.deBruijn_fn_base_list 0 d k).
    simpl.
    generalize (remove_map (fun d => deBruijn.swapR Name.deBruijn_base d k) (Name.Ref 0)).
    intros.
    rewrite <- H.
    simpl.
    unfold DeBruijnNat.swapR;simpl.
    auto.
    apply deBruijn.swapR_injective.
    apply fn_remove.
  Qed.

  Lemma low1_new_c : forall (d:nat)(l:constraint),(deBruijn.mem_fn Name.deBruijn_fn_base_list 0 (S d) l = false)->(deBruijn.low1 Name.deBruijn_base_list d (new_c l))=(new_c (deBruijn.low1 Name.deBruijn_base_list (S d) l)).
    intros.
    induction l.
    reflexivity.
    assert (deBruijn.mem_fn Name.deBruijn_fn_base 0 (S d) a = false).
    simpl in H.
    simpl.
    destruct (Name.mem_fn 0 (S d) a);auto.
    rewrite deBruijn.fold_list_true in H.
    auto.
    assert (deBruijn.mem_fn Name.deBruijn_fn_base_list 0 (S d) l = false).
    simpl in H.
    simpl in H0.
    rewrite H0 in H;auto.
    destruct a.
    destruct n.
    rewrite new_c_cons_0.
    simpl.
    unfold DeBruijnNat.low1.
    simpl.
    rewrite new_c_cons_0.
    simpl in H.
    auto.
    rewrite new_c_cons_S.
    simpl.
    unfold DeBruijnNat.low1.
    case le_lt_dec;intro.
    assert (d < n).
    simpl in H0.
    unfold DeBruijnNat.mem_fn in H0.
    rewrite plus_0_r in H0.
    generalize H0;clear H0.
    case eq_nat_dec;intros;try omega.
    discriminate H0.
    case le_lt_dec;intro;try_absurd_arith.
    simpl.
    destruct n.
    inversion H2.
    simpl.
    rewrite new_c_cons_S.
    simpl in IHl.
    rewrite IHl;auto.
    case le_lt_dec;intro;try_absurd_arith.
    rewrite new_c_cons_S.
    simpl in IHl.
    rewrite IHl;auto.
  Qed.

  Lemma mem_fn_new_c : forall (d k:nat)(l:constraint),(deBruijn.mem_fn Name.deBruijn_fn_base_list k d (new_c l))=(deBruijn.mem_fn Name.deBruijn_fn_base_list k (S d) l).
    induction l.
    reflexivity.
    destruct a.
    destruct n.
    rewrite new_c_cons_0.
    rewrite IHl.
    reflexivity.
    rewrite new_c_cons_S.
    simpl.
    unfold DeBruijnNat.mem_fn.
    case eq_nat_dec;intro.
    case eq_nat_dec;intro;try_absurd_arith.
    repeat (rewrite deBruijn.fold_list_true);auto.
    case eq_nat_dec;intro;try_absurd_arith.
    auto.
  Qed.

  Inductive kind : Set :=
    | In : kind
    | Out : kind
    | Silent : kind.

  Definition t := (ActionSpi.t * (constraint))%type.

  Definition equal (x y:t) := x = y.

  Lemma equal_refl : forall (x:t),(equal x x).
    unfold equal;auto.
  Qed.

  Lemma equal_sym : forall (x y:t),(equal x y)->(equal y x).
    unfold equal;auto.
  Qed.

  Lemma equal_trans : forall (x y z:t),(equal x y)->(equal y z)->(equal x z).
    unfold equal;intros;congruence.
  Qed.

  Module OrdAct := OrderedPair ActionSpi.Ord OrdListName.

  Module Ord <: OrderedType with Definition t := t with Definition eq := @eq t.
    Definition t := t.
    Definition eq := @eq t.
    Definition lt (a b:t) := OrdAct.lt a b.

    Lemma OrdAct_eq : forall (a b:t),(eq a b)<->(OrdAct.eq a b).
      split;intros.
      rewrite H.
      auto.
      destruct a;destruct b;simpl in H.
      and_elim.
      rewrite H.
      elim (OrdListEq_equal c c0);intros.
      rewrite H1;auto.
      unfold eq.
      auto.
    Qed.

    Lemma eq_refl : forall (x:t),(eq x x).
      unfold eq;reflexivity.
    Qed.

    Lemma eq_sym : forall (x y:t),(eq x y)->(eq y x).
      unfold eq;auto.
    Qed.

    Lemma eq_trans : forall (x y z:t),(eq x y)->(eq y z)->(eq x z).
      unfold eq;intros;congruence.
    Qed.

    Lemma lt_trans : forall (x y z:t),(lt x y)->(lt y z)->(lt x z).
      apply OrdAct.lt_trans.
    Qed.

    Lemma lt_not_eq : forall (x y:t),(lt x y)->~(eq x y).
      intros.
      red;intro.
      elim (OrdAct.lt_not_eq x y);auto.
      elim (OrdAct_eq x y);auto.
    Qed.

    Definition compare : forall (x y:t),(Compare lt eq x y).
      intros.
      case (OrdAct.compare x y);intro.
      apply LT;auto.
      apply EQ.
      elim (OrdAct_eq x y);auto.
      apply GT;auto.
    Defined.

    Hint Immediate eq_sym.
    Hint Resolve eq_refl eq_trans lt_not_eq lt_trans.
  End Ord.

  Definition deBruijn_base := deBruijn.deBruijn_base_pair ActionSpi.t ActionSpi.deBruijn_base (constraint) Name.deBruijn_base_list.
  Definition deBruijn_fn_base := deBruijn.deBruijn_fn_base_pair ActionSpi.t ActionSpi.deBruijn_base ActionSpi.deBruijn_fn_base constraint Name.deBruijn_base_list Name.deBruijn_fn_base_list.

  Lemma lift_action : forall (d k:nat)(a:ActionSpi.t)(l:constraint),(deBruijn.lift deBruijn_base d k (a,l))=(deBruijn.lift ActionSpi.deBruijn_base d k a,deBruijn.lift Name.deBruijn_base_list d k l).
    intros;reflexivity.
  Qed.

  Lemma swapR_action : forall (d k:nat)(a:ActionSpi.t)(l:constraint),(deBruijn.swapR deBruijn_base d k (a,l))=(deBruijn.swapR ActionSpi.deBruijn_base d k a,deBruijn.swapR Name.deBruijn_base_list d k l).
    intros;reflexivity.
  Qed.

  Lemma low1_action : forall (d:nat)(a:ActionSpi.t)(l:constraint),(deBruijn.low1 deBruijn_base d (a,l))=(deBruijn.low1 ActionSpi.deBruijn_base d a,deBruijn.low1 Name.deBruijn_base_list d l).
    intros;reflexivity.
  Qed.

  Definition get_kind (a:t) := 
    let (a,_) := a in 
      match ActionSpi.get_kind a with
	| ActionSpi.In => In
	| ActionSpi.Out => Out
	| ActionSpi.Silent => Silent
      end.

  Lemma lift_kind : forall (d k:nat)(x:t),(get_kind (deBruijn.lift deBruijn_base d k x))=(get_kind x).
    intros.
    unfold get_kind.
    destruct x.
    rewrite lift_action.
    rewrite ActionSpi.lift_kind.
    auto.
  Qed.

  Lemma swapR_kind : forall (d k:nat)(x:t),(get_kind (deBruijn.swapR deBruijn_base d k x))=(get_kind x).
    intros.
    unfold get_kind.
    destruct x.
    rewrite swapR_action.
    rewrite ActionSpi.swapR_kind.
    auto.
  Qed.

  Definition mk_input (e:A.P.E.t) : option t :=
    match (EvalSpi.evalc e) with
      | Some (MessageSpi.MNam a) => Some (ActionSpi.Inp a,cons a nil)
      | _ => None
    end.

  Definition mk_output (e:A.P.E.t) (f:A.P.E.t) : option (t * A.P.E.t) :=
    match (EvalSpi.evalc e,EvalSpi.evalc f) with
      | (Some (MessageSpi.MNam a),Some m) => Some ((ActionSpi.Outp a,cons a nil),EvalSpi.expr_of_mesg m)
      | _ => None
    end.

  Definition mk_silent : t := (ActionSpi.Tau,nil).

  Lemma mk_input_kind : forall (e:A.P.E.t)(a:t),(mk_input e = Some a)->(get_kind a = In).
    unfold mk_input.
    intros.
    destruct (EvalSpi.evalc e).
    destruct t0;try (discriminate H).
    injection H;intros;subst a;reflexivity.
    discriminate H.
  Qed.

  Lemma mk_output_kind : forall (e f:A.P.E.t)(a:t)(m:A.P.E.t),(mk_output e f = Some (a,m))->(get_kind a = Out).
    unfold mk_output.
    intros.
    destruct (EvalSpi.evalc e);try (discriminate H).
    destruct t0;try (discriminate H).
    destruct (EvalSpi.evalc f);try (discriminate H).
    injection H;intros;subst a.
    reflexivity.
  Qed.

  Lemma mk_silent_kind : get_kind mk_silent = Silent.
    reflexivity.
  Qed.

  Definition mk_if (phi:A.P.F.t) (a:t) : option t :=
    let (alpha,l) := a in
    let (b,l') := EvalSpi.eval_nc phi in
      if b then Some(alpha,app l l') else None.

  Lemma mk_if_kind : forall (phi:A.P.F.t)(a b:t),(mk_if phi a = Some b)->(get_kind b = get_kind a).
    unfold mk_if;intros.
    destruct (EvalSpi.eval_nc phi).
    destruct a.
    destruct b0.
    injection H;intros;subst b;auto.
    discriminate H.
  Qed.

  Definition mk_res (a:t) : option t :=
    let (alpha,l) := a in
      match ActionSpi.mk_res alpha with
	| None => None
	| Some beta => Some(beta,new_c l)
      end.

  Lemma mk_res_kind : forall (a b:t),(mk_res a = Some b)->(get_kind b = get_kind a).
    unfold mk_res;intros.
    destruct a.
    generalize (ActionSpi.mk_res_kind t0).
    destruct (ActionSpi.mk_res t0);intros.
    simpl.
    injection H;intros.
    subst b.
    simpl.
    rewrite (H0 t1);auto.
    discriminate H.
  Qed.

  Definition mk_interact (a:t) (b:t) : option t :=
    match (a,b) with
      | ((alpha,l),(beta,l')) => 
	(match ActionSpi.mk_interact alpha beta with
	   | None => None
	   | Some(tau) =>
	     match ActionSpi.get_kind alpha with 
	       | ActionSpi.In => Some(tau,app l l')
	       | ActionSpi.Out => Some(tau,app l' l)
	       | ActionSpi.Silent => None
	     end
	 end)
    end.

  Lemma mk_interact_kind : forall (a b c:t),(mk_interact a b = Some c)->(get_kind c = Silent).
    destruct a as [alpha la];destruct b as [beta lb];unfold mk_interact.
    intros.
    generalize (ActionSpi.mk_interact_kind alpha beta).
    destruct (ActionSpi.mk_interact);intros;destruct (ActionSpi.get_kind alpha);try (discriminate H).
    injection H;intros;subst c;simpl;rewrite H0;auto.
    injection H;intros;subst c;simpl;rewrite H0;auto.
  Qed.

  Lemma mk_interact_exg : forall (a b c:t),(mk_interact a b = Some c)->exists c':t,equal c c' /\ (mk_interact b a = Some c').
    destruct a as [alpha la];destruct b as [beta lb];unfold mk_interact;intros.
    generalize (ActionSpi.mk_interact_exg alpha beta);intros.
    destruct alpha;destruct beta;try (discriminate H).
    destruct (ActionSpi.mk_interact (ActionSpi.Inp t0) (ActionSpi.Outp t1));try (discriminate H).
    elim (H0 t2);auto;intros;and_elim.
    rewrite H2.
    injection H;intros;subst c.
    exists (x,la ++ lb).
    split;auto.
    rewrite H1;apply equal_refl.
    simpl in H.
    destruct (ActionSpi.mk_interact (ActionSpi.Outp t0) (ActionSpi.Inp t1));try (discriminate H).
    elim (H0 t2);auto;intros;and_elim.
    rewrite H2.
    injection H;intros;subst c.
    exists (x,lb ++ la).
    split;auto.
    rewrite H1;apply equal_refl.
  Qed.

  (* equal properties *)
  Lemma mk_if_equal : forall (phi:A.P.F.t)(a b:t),(mk_if phi a = Some b)->forall (a':t),(equal a a')->(exists b':t,equal b b' /\ (mk_if phi a' = Some b')).
    intros.
    rewrite H0 in H.
    exists b.
    split;auto.
    apply equal_refl.
  Qed.

  Lemma mk_res_equal : forall (a b:t),(mk_res a = Some b)->forall (a':t),(equal a a')->(exists b':t,equal b b' /\ (mk_res a' = Some b')).
    intros.
    rewrite H0 in H.
    exists b.
    split;auto.
    apply equal_refl.
  Qed.

  Lemma mk_interact_equal : forall (a b c:t),(mk_interact a b = Some c)->forall (a' b':t),(equal a a')->(equal b b')->(exists c':t,equal c c' /\ (mk_interact a' b' = Some c')).
    intros.
    rewrite H0 in H.
    rewrite H1 in H.
    exists c.
    split;auto.
    apply equal_refl.
  Qed.

  Lemma mk_res_lift_1 : forall (a:t),exists b:t,(equal b a)/\(mk_res (deBruijn.lift deBruijn_base 0 1 a) = Some b).
    unfold mk_res;intros.
    destruct a as [alpha la].
    rewrite lift_action.
    generalize (ActionSpi.mk_res_lift_1 alpha).
    destruct (ActionSpi.mk_res (deBruijn.lift ActionSpi.deBruijn_base 0 1 alpha)).
    intros.
    ex_elim H beta.
    injection H0;clear H0;intros;subst t0.
    exists (beta,new_c (deBruijn.lift Name.deBruijn_base_list 0 1 la)).
    split;auto.
    unfold equal.
    rewrite H.
    generalize (new_c_scope_1 nil la).
    rewrite new_c_nil;unfold app.
    intro.
    rewrite <- H0.
    auto.
    intros.
    ex_elim H b.
    discriminate H0.
  Qed.

  Lemma mk_res_swapR : forall (a b:t),(mk_res a = Some b)->forall (c:t),(mk_res b = Some c)->(exists b':t,exists c':t,(equal c' c)/\(mk_res (deBruijn.swapR deBruijn_base 0 1 a) = Some b')/\(mk_res b' = Some c')).
    destruct a as [alpha la].
    destruct b as [beta lb].
    unfold mk_res.
    generalize (ActionSpi.mk_res_swapR alpha).
    destruct (ActionSpi.mk_res alpha).
    intros.
    generalize (H t0 (refl_equal (Some t0)));clear H;intro.
    injection H0;clear H0;intros;subst t0.
    destruct (ActionSpi.mk_res beta);try (discriminate H1).
    elim (H t0 (refl_equal (Some t0)));clear H.
    intro beta';intros.
    ex_elim H gamma'.
    rewrite swapR_action.
    rewrite H2.
    exists (beta',new_c (deBruijn.swapR Name.deBruijn_base_list 0 1 la)).
    rewrite H3.
    exists (gamma',new_c (new_c (deBruijn.swapR Name.deBruijn_base_list 0 1 la))).
    unfold equal.
    injection H1;clear H1;intros;subst c.
    split;auto.
    rewrite H.
    rewrite <- H0.
    rewrite new_c_swap.
    auto.
    intros.
    discriminate H0.
  Qed.

  Lemma mk_interact_res_lift : forall (a a':t),(mk_res a = Some a')->forall (b c'':t),(mk_interact a' b = Some c'')->(exists c:t,exists c':t,(equal c' c'')/\(mk_interact a (deBruijn.lift deBruijn_base 0 1 b) = Some c)/\(mk_res c = Some c')).
    unfold mk_res.
    destruct a as [alpha la].
    generalize (ActionSpi.mk_interact_res_lift alpha);intros.
    unfold mk_interact.
    generalize (ActionSpi.mk_res_kind alpha);intro.
    destruct (ActionSpi.mk_res alpha);try (discriminate H0).
    generalize (H t0 (refl_equal (Some t0)));clear H;intros.
    injection H0;clear H0;intros;subst a'.
    unfold mk_interact in H1.
    destruct b as [beta lb].
    generalize (H beta);clear H;intros.
    destruct (ActionSpi.mk_interact t0 beta);try (discriminate H1).
    elim (H t1);clear H;auto;intros.
    ex_elim H c'.
    rewrite lift_action.
    rewrite H0.
    generalize (H2 t0 (refl_equal (Some t0)));clear H2.
    intro.
    rewrite <- H2.
    destruct (ActionSpi.get_kind t0);try (discriminate H1);injection H1;clear H1;intros.
    subst c''.
    exists (x,la ++ deBruijn.lift Name.deBruijn_base_list 0 1 lb).
    rewrite H3.
    exists (c',new_c (la++deBruijn.lift Name.deBruijn_base_list 0 1 lb)).
    rewrite H.
    split;auto.
    rewrite new_c_scope_1.
    apply equal_refl.
    subst c''.
    exists (x,(deBruijn.lift Name.deBruijn_base_list 0 1 lb)++la).
    rewrite H3.
    exists (c',new_c ((deBruijn.lift Name.deBruijn_base_list 0 1 lb)++la)).
    split;auto.
    rewrite H.
    rewrite new_c_scope_2.
    apply equal_refl.
  Qed.

  Lemma mk_interact_res_lift_inv : forall (a b c:t),(mk_interact a (deBruijn.lift deBruijn_base 0 1 b) = Some c)->forall (c'':t),(mk_res c = Some c'')->exists a':t,(mk_res a = Some a')/\exists c':t,(mk_interact a' b = Some c').
    intros a b c.
    unfold mk_interact.
    destruct a as [alpha la].
    destruct b as [beta lb].
    rewrite lift_action.
    generalize (ActionSpi.mk_interact_res_lift_inv alpha beta).
    intros.
    destruct (ActionSpi.mk_interact alpha (deBruijn.lift ActionSpi.deBruijn_base 0 1 beta));try (discriminate H0).
    generalize (H t0 (refl_equal (Some t0)));clear H;intros.
    generalize (ActionSpi.mk_res_kind alpha);intro.
    destruct (ActionSpi.get_kind alpha);try (discriminate H0).
    injection H0;clear H0;intros;subst c.
    unfold mk_res in H1.
    destruct (ActionSpi.mk_res t0);try (discriminate H1).
    injection H1;clear H1;intros;subst c''.
    elim (H t1);clear H;auto;intros;and_elim.
    ex_elim H0 c'.
    unfold mk_res.
    rewrite H.
    exists (x,new_c la).
    split;auto.
    rewrite H2;auto.
    rewrite H0.
    exists (c',new_c la ++ lb);auto.
    injection H0;clear H0;intros;subst c.
    unfold mk_res in H1.
    destruct (ActionSpi.mk_res t0);try (discriminate H1).
    injection H1;clear H1;intros;subst c''.
    elim (H t1);clear H;auto;intros;and_elim.
    ex_elim H0 c'.
    unfold mk_res.
    rewrite H.
    exists (x,new_c la).
    split;auto.
    rewrite H2;auto.
    rewrite H0.
    exists (c',lb ++ new_c la);auto.
  Qed.
    
  (* lift and actions *)
  Lemma lift_mk_silent : forall (d k:nat),(equal mk_silent (deBruijn.lift deBruijn_base d k mk_silent)).
    intros.
    unfold mk_silent, equal.
    reflexivity.
  Qed.

  Lemma lift_mk_input : forall (e:A.P.E.t)(inp:t),(mk_input e = Some inp)->forall (d k:nat),(exists inp':t,mk_input (deBruijn.lift A.P.E.deBruijn_base d k e) = Some inp' /\ (equal inp' (deBruijn.lift deBruijn_base d k inp))).
    intros e inp.
    unfold mk_input.
    intros.
    rewrite EvalSpi.evalc_lift.
    destruct (EvalSpi.evalc e);simpl;try (discriminate H).
    destruct t0;simpl;try (discriminate H).
    exists (ActionSpi.Inp (Name.lift d k t0),cons (Name.lift d k t0) nil).
    injection H;intros;subst inp.
    split;auto.
    unfold equal;auto.
  Qed.

  Lemma lift_mk_output : forall (e f:A.P.E.t)(out:t)(m:A.P.E.t),(mk_output e f = Some (out,m))->forall (d k:nat),(exists out':t,mk_output (deBruijn.lift A.P.E.deBruijn_base d k e) (deBruijn.lift A.P.E.deBruijn_base d k f) = Some (out',deBruijn.lift A.P.E.deBruijn_base d k m) /\ (equal out' (deBruijn.lift deBruijn_base d k out))).
    intros e f out m.
    unfold mk_output.
    intros.
    repeat (rewrite EvalSpi.evalc_lift).
    destruct (EvalSpi.evalc e);simpl;try (discriminate H).
    destruct t0;try (discriminate H).
    simpl.
    destruct (EvalSpi.evalc f);simpl;try (discriminate H).
    injection H;intros.
    subst out.
    subst m.
    rewrite EvalSpi.map_expr_of_msg.
    exists (ActionSpi.Outp (Name.lift d k t0),cons (Name.lift d k t0) nil).
    split;auto.
    unfold equal.
    auto.
  Qed.

  Lemma lift_mk_if : forall (phi:A.P.F.t)(mu mu':t),(mk_if phi mu = Some mu')->forall (d k:nat),(exists mu'':t,(equal mu'' (deBruijn.lift deBruijn_base d k mu')) /\ mk_if (deBruijn.lift A.P.F.deBruijn_base d k phi) (deBruijn.lift deBruijn_base d k mu) = Some mu'').
    intros phi mu mu';unfold mk_if;intros.
    destruct mu as [a la].
    destruct mu' as [b lb].
    repeat (rewrite lift_action).
    rewrite EvalSpi.eval_nc_lift.
    destruct (EvalSpi.eval_nc phi);try (discriminate H).
    destruct b0;try (discriminate H).
    simpl.
    injection H;intros;subst b;subst lb.
    rewrite <- deBruijn.map_list_app.
    exists (deBruijn.lift deBruijn_base d k (a,la ++ l)).
    rewrite lift_action.
    unfold equal;auto.
  Qed.
    
  Lemma lift_mk_res : forall (mu mu':t),(mk_res mu = Some mu')->forall (d k:nat),(exists mu'':t,(equal mu'' (deBruijn.lift deBruijn_base d k mu'))/\(mk_res (deBruijn.lift deBruijn_base (S d) k mu))=(Some mu'')).
    unfold mk_res.
    destruct mu as [a la].
    destruct mu' as [b lb].
    intros.
    repeat (rewrite lift_action).
    generalize (ActionSpi.lift_mk_res a);intros.
    destruct (ActionSpi.mk_res a);try (discriminate H).
    elim (H0 t0 (refl_equal (Some t0)) d k);clear H0;intros.
    and_elim.
    rewrite H1.
    exists (x,new_c (deBruijn.lift Name.deBruijn_base_list (S d) k la)).
    split;auto.
    unfold equal.
    injection H;clear H;intros;subst b;subst lb.
    rewrite <- H0.
    rewrite lift_new_c.
    auto.
  Qed.

  Lemma lift_mk_interact : forall (a b c:t),(mk_interact a b = Some c)->forall (d k:nat),(exists c':t,(equal c' (deBruijn.lift deBruijn_base d k c))/\(mk_interact (deBruijn.lift deBruijn_base d k a) (deBruijn.lift deBruijn_base d k b) = Some c')).
    destruct a as [alpha la].
    destruct b as [beta lb].
    unfold mk_interact.
    generalize (ActionSpi.lift_mk_interact alpha beta);intros.
    destruct (ActionSpi.mk_interact alpha beta);try (discriminate H0).
    repeat (rewrite lift_action).
    elim (H t0 (refl_equal (Some t0)) d k);clear H;intros;and_elim.
    rewrite H1.
    generalize (ActionSpi.lift_kind d k alpha).
    intro.
    rewrite H2.
    repeat (rewrite <- lift_app).
    destruct (ActionSpi.get_kind alpha);try (discriminate H0);injection H0;clear H0;intros;subst c.
    exists (x,deBruijn.lift Name.deBruijn_base_list d k (la ++ lb)).
    split;auto.
    rewrite lift_action.
    unfold equal.
    rewrite H;auto.
    exists (x,deBruijn.lift Name.deBruijn_base_list d k (lb ++ la)).
    split;auto.
    rewrite lift_action.
    unfold equal.
    rewrite H;auto.
  Qed.

  Lemma lift_mk_input_inv : forall (d k:nat)(e:A.P.E.t)(inp':t),(mk_input (deBruijn.lift A.P.E.deBruijn_base d k e) = Some inp')->(exists inp:t,mk_input e= Some inp).
    intros d k e inp'.
    unfold mk_input.
    rewrite EvalSpi.evalc_lift.
    intro.
    destruct (EvalSpi.evalc e);simpl;try (discriminate H).
    simpl in H.
    destruct t0;simpl in H;try (discriminate H).
    injection H;intros;subst inp'.
    exists (ActionSpi.Inp t0,cons t0 nil).
    auto.
  Qed.
    
  Lemma lift_mk_output_inv : forall (d k:nat)(e f:A.P.E.t)(out':t)(m':A.P.E.t),(mk_output (deBruijn.lift A.P.E.deBruijn_base d k e) (deBruijn.lift A.P.E.deBruijn_base d k f) = Some (out',m'))->(exists out:t,exists m:A.P.E.t,mk_output e f = Some (out,m) /\ m' = deBruijn.lift A.P.E.deBruijn_base d k m).
    intros d k e f out' m'.
    unfold mk_output.
    repeat (rewrite EvalSpi.evalc_lift).
    intro.
    destruct (EvalSpi.evalc e);try (discriminate H).
    simpl in H.
    destruct t0;simpl in H;try (discriminate H).
    destruct (EvalSpi.evalc f);try (discriminate H).
    simpl in H.
    injection H;intros.
    subst out'.
    exists (ActionSpi.Outp t0,cons t0 nil).
    exists (EvalSpi.expr_of_mesg t1).
    split;auto.
    subst m'.
    simpl.
    rewrite EvalSpi.map_expr_of_msg.
    auto.
  Qed.
    
  Lemma lift_mk_if_inv : forall (d k:nat)(phi:A.P.F.t)(mu mu'':t),(mk_if (deBruijn.lift A.P.F.deBruijn_base d k phi) (deBruijn.lift deBruijn_base d k mu) = Some mu'')->(exists mu':t,(mk_if phi mu = Some mu')).
    intros d k phi mu mu'';unfold mk_if.
    destruct mu as [a la].
    destruct mu'' as [c lc].
    rewrite lift_action.
    rewrite EvalSpi.eval_nc_lift.
    intro H.
    destruct (EvalSpi.eval_nc phi);try (discriminate H).
    destruct b;try (discriminate H).
    simpl in H.
    injection H;intros.
    exists (a,la++l).
    auto.
  Qed.

  Lemma lift_mk_res_inv : forall (d k:nat)(mu mu'':t),(mk_res (deBruijn.lift deBruijn_base (S d) k mu) = Some mu'')->(exists mu':t,mk_res mu = Some mu').
    intros d k a c.
    destruct a as [a la].
    destruct c as [c lc].
    rewrite lift_action.
    unfold mk_res.
    generalize (ActionSpi.lift_mk_res_inv d k a).
    intros.
    destruct (ActionSpi.mk_res (deBruijn.lift ActionSpi.deBruijn_base (S d) k a));try (discriminate H0).
    injection H0;intros.
    elim (H t0);auto;intros.
    rewrite H3.
    exists (x,new_c la).
    auto.
  Qed.

  Lemma lift_mk_interact_inv : forall (d k:nat)(a b c':t),(mk_interact (deBruijn.lift deBruijn_base d k a) (deBruijn.lift deBruijn_base d k b) = Some c')->(exists c:t,mk_interact a b = Some c).
    intros d k.
    destruct a as [a la].
    destruct b as [b lb].
    repeat (rewrite lift_action).
    unfold mk_interact.
    generalize (ActionSpi.lift_mk_interact_inv d k a b).
    intros.
    destruct (ActionSpi.mk_interact (deBruijn.lift ActionSpi.deBruijn_base d k a) (deBruijn.lift ActionSpi.deBruijn_base d k b));try (discriminate H0).
    elim (H t0);auto;intros.
    rewrite H1.
    rewrite (ActionSpi.lift_kind d k a) in H0.
    destruct (ActionSpi.get_kind a);try (discriminate H0).
    exists (x,la++lb);auto.
    exists (x,lb++la);auto.
  Qed.
    
  (* swapR and actions *)
  Lemma swapR_mk_silent : forall (d k:nat),(equal mk_silent (deBruijn.swapR deBruijn_base d k mk_silent)).
    intros.
    unfold mk_silent, equal.
    reflexivity.
  Qed.

  Lemma swapR_mk_input : forall (e:A.P.E.t)(inp:t),(mk_input e = Some inp)->forall (d k:nat),(exists inp':t,mk_input (deBruijn.swapR A.P.E.deBruijn_base d k e) = Some inp' /\ (equal inp' (deBruijn.swapR deBruijn_base d k inp))).
    intros e inp.
    unfold mk_input.
    intros.
    rewrite EvalSpi.evalc_swapR.
    destruct (EvalSpi.evalc e);simpl;try (discriminate H).
    destruct t0;simpl;try (discriminate H).
    exists (ActionSpi.Inp (Name.swapR d k t0),cons (Name.swapR d k t0) nil).
    injection H;intros;subst inp.
    split;auto.
    unfold equal;auto.
  Qed.

  Lemma swapR_mk_output : forall (e f:A.P.E.t)(out:t)(m:A.P.E.t),(mk_output e f = Some (out,m))->forall (d k:nat),(exists out':t,mk_output (deBruijn.swapR A.P.E.deBruijn_base d k e) (deBruijn.swapR A.P.E.deBruijn_base d k f) = Some (out',deBruijn.swapR A.P.E.deBruijn_base d k m) /\ (equal out' (deBruijn.swapR deBruijn_base d k out))).
    intros e f out m.
    unfold mk_output.
    intros.
    repeat (rewrite EvalSpi.evalc_swapR).
    destruct (EvalSpi.evalc e);simpl;try (discriminate H).
    destruct t0;try (discriminate H).
    simpl.
    destruct (EvalSpi.evalc f);simpl;try (discriminate H).
    injection H;intros.
    subst out.
    subst m.
    rewrite EvalSpi.map_expr_of_msg.
    exists (ActionSpi.Outp (Name.swapR d k t0),cons (Name.swapR d k t0) nil).
    split;auto.
    unfold equal.
    auto.
  Qed.

  Lemma swapR_mk_if : forall (phi:A.P.F.t)(mu mu':t),(mk_if phi mu = Some mu')->forall (d k:nat),(exists mu'':t,(equal mu'' (deBruijn.swapR deBruijn_base d k mu')) /\ mk_if (deBruijn.swapR A.P.F.deBruijn_base d k phi) (deBruijn.swapR deBruijn_base d k mu) = Some mu'').
    intros phi mu mu';unfold mk_if;intros.
    destruct mu as [a la].
    destruct mu' as [b lb].
    repeat (rewrite swapR_action).
    rewrite EvalSpi.eval_nc_swapR.
    destruct (EvalSpi.eval_nc phi);try (discriminate H).
    destruct b0;try (discriminate H).
    simpl.
    injection H;intros;subst b;subst lb.
    rewrite <- deBruijn.map_list_app.
    exists (deBruijn.swapR deBruijn_base d k (a,la ++ l)).
    rewrite swapR_action.
    unfold equal;auto.
  Qed.
    
  Lemma swapR_mk_res : forall (mu mu':t),(mk_res mu = Some mu')->forall (d k:nat),(exists mu'':t,(equal mu'' (deBruijn.swapR deBruijn_base d k mu'))/\(mk_res (deBruijn.swapR deBruijn_base (S d) k mu))=(Some mu'')).
    unfold mk_res.
    destruct mu as [a la].
    destruct mu' as [b lb].
    intros.
    repeat (rewrite swapR_action).
    generalize (ActionSpi.swapR_mk_res a);intros.
    destruct (ActionSpi.mk_res a);try (discriminate H).
    elim (H0 t0 (refl_equal (Some t0)) d k);clear H0;intros.
    and_elim.
    rewrite H1.
    exists (x,new_c (deBruijn.swapR Name.deBruijn_base_list (S d) k la)).
    split;auto.
    unfold equal.
    injection H;clear H;intros;subst b;subst lb.
    rewrite <- H0.
    rewrite swapR_new_c.
    auto.
  Qed.

  Lemma swapR_mk_interact : forall (a b c:t),(mk_interact a b = Some c)->forall (d k:nat),(exists c':t,(equal c' (deBruijn.swapR deBruijn_base d k c))/\(mk_interact (deBruijn.swapR deBruijn_base d k a) (deBruijn.swapR deBruijn_base d k b) = Some c')).
    destruct a as [alpha la].
    destruct b as [beta lb].
    unfold mk_interact.
    generalize (ActionSpi.swapR_mk_interact alpha beta);intros.
    destruct (ActionSpi.mk_interact alpha beta);try (discriminate H0).
    repeat (rewrite swapR_action).
    elim (H t0 (refl_equal (Some t0)) d k);clear H;intros;and_elim.
    rewrite H1.
    generalize (ActionSpi.swapR_kind d k alpha).
    intro.
    rewrite H2.
    repeat (rewrite <- swapR_app).
    destruct (ActionSpi.get_kind alpha);try (discriminate H0);injection H0;clear H0;intros;subst c.
    exists (x,deBruijn.swapR Name.deBruijn_base_list d k (la ++ lb)).
    split;auto.
    rewrite swapR_action.
    unfold equal.
    rewrite H;auto.
    exists (x,deBruijn.swapR Name.deBruijn_base_list d k (lb ++ la)).
    split;auto.
    rewrite swapR_action.
    unfold equal.
    rewrite H;auto.
  Qed.
End ActionSpiNC.


