(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Export processType.
Require Import deBruijnType.

Module Type AgentType.
  Declare Module P : ProcessType.

  Inductive abstraction : Set := 
    | Abs : P.t -> abstraction.
  
  Inductive concretion : Set :=
    | Conc : nat -> P.E.t -> P.t -> concretion.
  
  Inductive t : Set :=
    | Pr : P.t -> t
    | F : abstraction -> t
    | C : concretion -> t.

  Declare Module Ord : OrderedType with Definition t := t with Definition eq := @eq t.

  Parameter deBruijn_base : deBruijn.deBruijn_base t.

  Parameter deBruijn_subst : deBruijn.deBruijn_subst t deBruijn_base P.E.t P.E.deBruijn_base.

  Parameter low_subst_Pr : forall (d k:nat)(es:list P.E.t)(p:P.t),deBruijn.low_subst deBruijn_subst d k es (Pr p) = Pr (deBruijn.low_subst P.deBruijn_subst d k es p).
  Parameter low_subst_F : forall (d k:nat)(es:list P.E.t)(p:P.t),deBruijn.low_subst deBruijn_subst d k es (F (Abs p)) = F (Abs (deBruijn.low_subst P.deBruijn_subst (S d) k es p)).
  Parameter low_subst_C : forall (d k:nat)(es:list P.E.t)(n:nat)(m:P.E.t)(p:P.t),deBruijn.low_subst deBruijn_subst d k es (C (Conc n m p)) = C (Conc n (deBruijn.low_subst P.E.deBruijn_subst (d+n) k es m) (deBruijn.low_subst P.deBruijn_subst (d+n) k es p)).

(*   Parameter deBruijn_fn_base : deBruijn.deBruijn_fn t deBruijn_base. *)

  Parameter new_abs : abstraction -> abstraction.
  Parameter new_conc : concretion -> concretion.
  Parameter new : t -> t.

  Parameter new_abs_def : forall (p:P.t),(new_abs (Abs p))=(Abs (P.New (deBruijn.swapR P.deBruijn_base 0 1 p))).
  Parameter new_conc_def : forall (n:nat)(m:P.E.t)(p:P.t),(new_conc (Conc n m p))=(if deBruijn.mem_fn P.E.deBruijn_fn_base 0 n m then (Conc (S n) m p) else (Conc n (deBruijn.low1 P.E.deBruijn_base n m)  (P.New (deBruijn.swapR P.deBruijn_base 0 n p)))).

  Parameter new_Pr : forall (p:P.t),(new (Pr p))=(Pr (P.New p)).
  Parameter new_F : forall (f:abstraction),(new (F f))=(F (new_abs f)).
  Parameter new_C : forall (c:concretion),(new (C c))=(C (new_conc c)).

  Parameter abs_par_l : P.t -> abstraction -> abstraction.
  Parameter conc_par_l : P.t -> concretion -> concretion.
  Parameter par_l : P.t -> t -> t.

  Parameter abs_par_l_def : forall (p:P.t)(q:P.t),(abs_par_l p (Abs q))=(Abs (P.Par (deBruijn.lift P.deBruijn_base 0 1 p) q)).
  Parameter conc_par_l_def : forall (p:P.t)(n:nat)(m:P.E.t)(q:P.t),(conc_par_l p (Conc n m q))=(Conc n m (P.Par (deBruijn.lift P.deBruijn_base 0 n p) q)).

  Parameter par_l_Pr : forall (p q:P.t),(par_l p (Pr q))=(Pr (P.Par p q)).
  Parameter par_l_F : forall (p:P.t)(f:abstraction),(par_l p (F f))=(F (abs_par_l p f)).
  Parameter par_l_C : forall (p:P.t)(c:concretion),(par_l p (C c))=(C (conc_par_l p c)).

  Parameter abs_par_r : abstraction -> P.t -> abstraction.
  Parameter conc_par_r : concretion -> P.t -> concretion.
  Parameter par_r : t -> P.t -> t.

  Parameter abs_par_r_def : forall (p:P.t)(q:P.t),(abs_par_r (Abs p) q)=(Abs (P.Par p (deBruijn.lift P.deBruijn_base 0 1 q))).
  Parameter conc_par_r_def : forall (n:nat)(m:P.E.t)(p:P.t)(q:P.t),(conc_par_r (Conc n m p) q)=(Conc n m (P.Par p (deBruijn.lift P.deBruijn_base 0 n q))).

  Parameter par_r_Pr : forall (p q:P.t),(par_r (Pr p) q)=(Pr (P.Par p q)).
  Parameter par_r_F : forall (f:abstraction)(q:P.t),(par_r (F f) q)=(F (abs_par_r f q)).
  Parameter par_r_C : forall (c:concretion)(q:P.t),(par_r (C c) q)=(C (conc_par_r c q)).

  Parameter pseudo_appl_l : abstraction -> concretion -> t.
  Parameter pseudo_appl_r : concretion -> abstraction -> t.

  Parameter pseudo_appl_l_def : forall (p:P.t)(n:nat)(m:P.E.t)(q:P.t),(pseudo_appl_l (Abs p) (Conc n m q))=(Pr (P.mk_nu n (P.Par (deBruijn.low_subst P.deBruijn_subst 0 n (cons m nil) p) q))).
  Parameter pseudo_appl_r_def : forall (n:nat)(m:P.E.t)(p:P.t)(q:P.t),(pseudo_appl_r (Conc n m p) (Abs q))=(Pr (P.mk_nu n (P.Par p (deBruijn.low_subst P.deBruijn_subst 0 n (cons m nil) q)))).

  Parameter lift_Pr : forall (d k:nat)(p:P.t),(deBruijn.lift deBruijn_base d k (Pr p))=(Pr (deBruijn.lift P.deBruijn_base d k p)).
  Parameter lift_F : forall (d k:nat)(p:P.t),(deBruijn.lift deBruijn_base d k (F (Abs p)))=(F (Abs (deBruijn.lift P.deBruijn_base (S d) k p))).
  Parameter lift_C : forall (d k:nat)(n:nat)(e:P.E.t)(p:P.t),(deBruijn.lift deBruijn_base d k (C (Conc n e p)))=(C (Conc n (deBruijn.lift P.E.deBruijn_base (d+n) k e) (deBruijn.lift P.deBruijn_base (d+n) k p))).

  Parameter lift_new : forall (d k:nat)(a:t),(deBruijn.lift deBruijn_base d k (new a))=(new (deBruijn.lift deBruijn_base (S d) k a)).
  Parameter lift_par_l : forall (p:P.t)(a:t)(d k:nat),(deBruijn.lift deBruijn_base d k (par_l p a))=(par_l (deBruijn.lift P.deBruijn_base d k p) (deBruijn.lift deBruijn_base d k a)).
  Parameter lift_par_r : forall (a:t)(q:P.t)(d k:nat),(deBruijn.lift deBruijn_base d k (par_r a q))=(par_r (deBruijn.lift deBruijn_base d k a) (deBruijn.lift P.deBruijn_base d k q)).

  Parameter lift_pseudo_appl_l : forall (p:P.t)(n:nat)(m:P.E.t)(q:P.t)(d k:nat),(deBruijn.lift deBruijn_base d k (pseudo_appl_l (Abs p) (Conc n m q)))=(pseudo_appl_l (Abs (deBruijn.lift P.deBruijn_base (S d) k p)) (Conc n (deBruijn.lift P.E.deBruijn_base (d+n) k m) (deBruijn.lift P.deBruijn_base (d+n) k q))).
  Parameter lift_pseudo_appl_r : forall (n:nat)(m:P.E.t)(p q:P.t)(d k:nat),(deBruijn.lift deBruijn_base d k (pseudo_appl_r (Conc n m p) (Abs q)))=(pseudo_appl_r (Conc n (deBruijn.lift P.E.deBruijn_base (d+n) k m) (deBruijn.lift P.deBruijn_base (d+n) k p)) (Abs (deBruijn.lift P.deBruijn_base (S d) k q))).

  Parameter swapR_Pr : forall (d k:nat)(p:P.t),(deBruijn.swapR deBruijn_base d k (Pr p))=(Pr (deBruijn.swapR P.deBruijn_base d k p)).
  Parameter swapR_F : forall (d k:nat)(p:P.t),(deBruijn.swapR deBruijn_base d k (F (Abs p)))=(F (Abs (deBruijn.swapR P.deBruijn_base (S d) k p))).
  Parameter swapR_C : forall (d k:nat)(n:nat)(e:P.E.t)(p:P.t),(deBruijn.swapR deBruijn_base d k (C (Conc n e p)))=(C (Conc n (deBruijn.swapR P.E.deBruijn_base (d+n) k e) (deBruijn.swapR P.deBruijn_base (d+n) k p))).

  Parameter swapR_new : forall (d k:nat)(a:t),(deBruijn.swapR deBruijn_base d k (new a))=(new (deBruijn.swapR deBruijn_base (S d) k a)).

  Parameter swapR_par_l : forall (p:P.t)(a:t)(d k:nat),(deBruijn.swapR deBruijn_base d k (par_l p a))=(par_l (deBruijn.swapR P.deBruijn_base d k p) (deBruijn.swapR deBruijn_base d k a)).
  Parameter swapR_par_r : forall (a:t)(q:P.t)(d k:nat),(deBruijn.swapR deBruijn_base d k (par_r a q))=(par_r (deBruijn.swapR deBruijn_base d k a) (deBruijn.swapR P.deBruijn_base d k q)).

  Parameter swapR_pseudo_appl_l : forall (p:P.t)(n:nat)(m:P.E.t)(q:P.t)(d k:nat),(deBruijn.swapR deBruijn_base d k (pseudo_appl_l (Abs p) (Conc n m q)))=(pseudo_appl_l (Abs (deBruijn.swapR P.deBruijn_base (S d) k p)) (Conc n (deBruijn.swapR P.E.deBruijn_base (d+n) k m) (deBruijn.swapR P.deBruijn_base (d+n) k q))).
  Parameter swapR_pseudo_appl_r : forall (n:nat)(m:P.E.t)(p q:P.t)(d k:nat),(deBruijn.swapR deBruijn_base d k (pseudo_appl_r (Conc n m p) (Abs q)))=(pseudo_appl_r (Conc n (deBruijn.swapR P.E.deBruijn_base (d+n) k m) (deBruijn.swapR P.deBruijn_base (d+n) k p)) (Abs (deBruijn.swapR P.deBruijn_base (S d) k q))).

  Parameter low_subst_new : forall (d k:nat)(es:list P.E.t)(a:t),(deBruijn.low_subst deBruijn_subst d k es (new a))=(new (deBruijn.low_subst deBruijn_subst (S d) k es a)).

  Parameter low_subst_par_l : forall (d k:nat)(es:list P.E.t)(p:P.t)(a:t),(deBruijn.low_subst deBruijn_subst d k es (par_l p a))=(par_l (deBruijn.low_subst P.deBruijn_subst d k es p) (deBruijn.low_subst deBruijn_subst d k es a)).
  Parameter low_subst_par_r : forall (d k:nat)(es:list P.E.t)(a:t)(p:P.t),(deBruijn.low_subst deBruijn_subst d k es (par_r a p))=(par_r (deBruijn.low_subst deBruijn_subst d k es a) (deBruijn.low_subst P.deBruijn_subst d k es p)).

  Parameter low_subst_pseudo_appl_l : forall (d k:nat)(es:list P.E.t)(p:P.t)(n:nat)(m:P.E.t)(q:P.t),deBruijn.low_subst deBruijn_subst d k es (pseudo_appl_l (Abs p) (Conc n m q)) = pseudo_appl_l (Abs (deBruijn.low_subst P.deBruijn_subst (S d) k es p)) (Conc n (deBruijn.low_subst P.E.deBruijn_subst (d+n) k es m) (deBruijn.low_subst P.deBruijn_subst (d+n) k es q)).
  Parameter low_subst_pseudo_appl_r : forall (d k:nat)(es:list P.E.t)(n:nat)(m:P.E.t)(p:P.t)(q:P.t),deBruijn.low_subst deBruijn_subst d k es (pseudo_appl_r (Conc n m p) (Abs q)) = pseudo_appl_r (Conc n (deBruijn.low_subst P.E.deBruijn_subst (d+n) k es m) (deBruijn.low_subst P.deBruijn_subst (d+n) k es p)) (Abs (deBruijn.low_subst P.deBruijn_subst (S d) k es q)).

  Inductive well_formed : t -> Prop :=
    | WFProc : forall (p:P.t),(well_formed (Pr p))
    | WFAbs : forall (p:P.t),(well_formed (F (Abs p)))
    | WFConc : forall (n:nat)(m:P.E.t)(p:P.t),(forall (i:nat),i < n->(deBruijn.mem_fn P.E.deBruijn_fn_base i 0 m = true))->(well_formed (C (Conc n m p))).

  Parameter well_formed_new : forall (a:t),(well_formed a)->(well_formed (new a)).

  Parameter well_formed_par_r : forall (a:t)(q:P.t),(well_formed a)->(well_formed (par_r a q)).
  Parameter well_formed_par_l : forall (p:P.t)(a:t),(well_formed a)->(well_formed (par_l p a)).
  Parameter well_formed_pseudo_appl_l : forall (f:abstraction)(c:concretion),(well_formed (pseudo_appl_l f c)).
  Parameter well_formed_pseudo_appl_r : forall (f:abstraction)(c:concretion),(well_formed (pseudo_appl_r c f)).
  Parameter well_formed_concretion_inv : forall (n:nat)(m:P.E.t)(p:P.t),(well_formed (C (Conc (S n) m p)))->(well_formed (C (Conc n m p)))/\(C (Conc (S n) m p))=new (C (Conc n m p)).
  Parameter well_formed_lift : forall (a:t),(well_formed a)->forall (d k:nat),(well_formed (deBruijn.lift deBruijn_base d k a)).
  Parameter well_formed_swapR : forall (a:t),(well_formed a)->forall (d k:nat),(well_formed (deBruijn.swapR deBruijn_base d k a)).
End AgentType.


