(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import deBruijnType.
Require Import name.
Require Import messageSpi.
Require Import expressionSpi.
Require Import formulaeSpi.
Require Import missing.
Require Import ordmissing.
Require Import tactics.
Require Import FSets.
Require Import hedges.
Require Import calculi.
Require Import reductionSpi.
Require Import actionSpi.
Require Import evalSpi.
Require Import substSpi.
Require Import Arith.
Require Import workAround.
Require Import LHbisimulation.

Module Example(H:Hedge).
  Module bisimulation := LateHedgedBisimulation(H).
  Import bisimulation.

  (** Example 1 *)
  (** We show that (^k)'c<enc_s(M,k)> = (^k)'c<enc_s(N,k)> for any M and N *)
  (** The initial hedge is I((c,c),(M,M),(N,N)) = h0 in the following *)

  Definition Ps (c:Name.t)(M:MessageSpi.t) :=
    AgentSpi.P.New 
    (AgentSpi.P.Output 
      (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 c)) 
      (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0)))) 
      AgentSpi.P.Zero).

  (** the proof has been split in several small lemmas *)
  Lemma instantiate_message_Ps : forall (k:Name.t)(M:MessageSpi.t),(instantiate_message (cons k nil) (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0))))=(MessageSpi.MEnc_s M (MessageSpi.MNam k)).
    intros.
    unfold instantiate_message.
    rewrite MessageSpi.low_subst_MEnc_s.
    generalize (deBruijn.low_subst_lift_nil MessageSpi.deBruijn_subst 0 0 0 (map MessageSpi.j (k :: nil)) M).
    repeat (rewrite plus_0_l).
    lazy beta delta [map length] iota.
    fold MessageSpi.deBruijn_base.
    rewrite (deBruijn.lift_0 MessageSpi.deBruijn_base 0 M).
    intro.
    rewrite H.
    assert (deBruijn.low_subst MessageSpi.deBruijn_subst 0 0 (MessageSpi.j k :: nil) (MessageSpi.MNam (Name.Ref 0)) = MessageSpi.MNam (deBruijn.lift Name.deBruijn_base 0 0 k)).
    reflexivity.
    rewrite (deBruijn.lift_0 Name.deBruijn_base) in H0.
    rewrite <- H0.
    auto.
  Qed.

  (** the only transition is - 'c(^k)<enc_s(M,k)> -> *)
  Lemma Ps_no_silent_trans : forall (c:Name.t)(M:MessageSpi.t)(p':AgentSpi.P.t),~(RedSpiCong.R.reduction (Ps c M) ActionSpi.Tau (AgentSpi.Pr p')).
    intros.
    red;intro.
    unfold Ps in H.
    inversion H.
    inversion H2.
    subst a.
    unfold ActionSpi.A.new in H1.
    rewrite ActionSpi.A.new_conc_def in H1.
    generalize H1.
    case deBruijn.mem_fn;intro.
    discriminate H9.
    discriminate H9.
  Qed.

  Lemma Ps_no_input_trans : forall (c:Name.t)(M:MessageSpi.t)(a:Name.t)(p':AgentSpi.P.t),~(RedSpiCong.R.reduction (Ps c M) (ActionSpi.Inp a) (AgentSpi.F (AgentSpi.Abs p'))).
    intros.
    red;intro.
    unfold Ps in H.
    inversion H.
    inversion H2.
    subst a0.
    unfold ActionSpi.A.new in H1.
    rewrite ActionSpi.A.new_conc_def in H1.
    generalize H1.
    case deBruijn.mem_fn;intro.
    discriminate H9.
    discriminate H9.
  Qed.

  Lemma Ps_output_trans_unique : forall (c:Name.t)(M:MessageSpi.t)(a:Name.t)(n:nat)(m:MessageSpi.t)(p':AgentSpi.P.t),(RedSpiCong.R.reduction (Ps c M) (ActionSpi.Outp a) (AgentSpi.C (AgentSpi.Conc n (EvalSpi.expr_of_mesg m) p')))->a=c /\ n=1 /\ p'=AgentSpi.P.Zero /\ m=(MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0))).
    intros.
    unfold Ps in H.
    inversion H.
    inversion H2.
    unfold ActionSpi.mk_output in H10.
    simpl in H10.
    rewrite EvalSpi.evalc_msg in H10.
    injection H10;clear H10;intros.
    subst m0.
    subst p0.
    subst mu0.
    subst mu.
    unfold ActionSpi.mk_res in H4.
    simpl in H4.
    assert (0 < 1);try omega.
    generalize (deBruijn.lift_not_fn Name.deBruijn_fn_base 0 1 0 H6 c);intro.
    simpl in H8.
    rewrite H8 in H4.
    injection H4;clear H4;intros.
    clear H6.
    generalize (deBruijn.low1_lift_lift Name.deBruijn_base 0 0 0 c).
    intros.
    simpl in H6.
    rewrite H6 in H4.
    generalize (deBruijn.lift_0 Name.deBruijn_base 0 c).
    intros.
    simpl in H10.
    rewrite H10 in H4.
    subst a.
    clear H8;clear H6;clear H10.
    subst mu'.
    subst a0.
    clear H2.
    subst p.
    subst e.
    subst f.
    unfold ActionSpi.A.new in H1.
    rewrite ActionSpi.A.new_conc_def in H1.
    rewrite ExpressionSpi.mem_fn_EEnc_s in H1.
    assert (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 0 (ExpressionSpi.ENam (Name.Ref 0)) = true).
    reflexivity.
    rewrite H0 in H1.
    rewrite orb_b_true in H1.
    injection H1;clear H1;intros.
    subst n.
    subst p'.
    clear H.
    repeat (split;trivial).
    clear H0.
    assert (ExpressionSpi.EEnc_s (EvalSpi.expr_of_mesg (MessageSpi.t_map (fun d : nat => Name.lift d 1) 0 M)) (ExpressionSpi.ENam (Name.Ref 0)) = (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0)))));try reflexivity.
    rewrite H in H2;clear H.
    symmetry.
    apply EvalSpi.expr_of_mesg_injective.
    auto.
  Qed.

  Lemma Ps_output_trans_exists : forall (c:Name.t)(M:MessageSpi.t),(RedSpiCong.R.reduction (Ps c M) (ActionSpi.Outp c) (AgentSpi.C (AgentSpi.Conc 1 (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0)))) AgentSpi.P.Zero))).
    intros.
    replace (AgentSpi.C (AgentSpi.Conc 1 (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0)))) AgentSpi.P.Zero)) with (AgentSpi.new (AgentSpi.C (AgentSpi.Conc 0 (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0)))) AgentSpi.P.Zero))).
    unfold Ps.
    eapply RedSpiCong.R.RedRes.
    eapply RedSpiCong.R.RedOutput.
    unfold ActionSpi.mk_output.
    lazy beta delta [EvalSpi.evalc] iota;fold EvalSpi.evalc.
    rewrite EvalSpi.evalc_msg.
    reflexivity.
    unfold ActionSpi.mk_res.
    assert (0<1);try omega.
    generalize (deBruijn.lift_not_fn Name.deBruijn_fn_base 0 1 0 H c);intro.
    simpl.
    simpl in H0.
    rewrite H0.
    generalize (deBruijn.low1_lift_lift Name.deBruijn_base 0 0 0 c).
    intros.
    simpl in H1.
    rewrite H1.
    rewrite (deBruijn.lift_0 Name.deBruijn_base).
    auto.
    unfold AgentSpi.new.
    rewrite AgentSpi.new_conc_def.
    lazy beta delta [EvalSpi.expr_of_mesg] iota;fold EvalSpi.expr_of_mesg.
    rewrite ExpressionSpi.mem_fn_EEnc_s.
    assert (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 0 (ExpressionSpi.ENam (Name.Ref 0)) = true).
    reflexivity.
    rewrite H.
    rewrite orb_b_true.
    auto.
  Qed.

  (** we now study h0 and its derivatives *)
  Definition h0_seed (c:Name.t)(M N:MessageSpi.t) := 
    (H.add (MessageSpi.MNam c,MessageSpi.MNam c) (H.add (M,M) (H.singleton (N,N)))).

  Lemma h0_seed_reflexive : forall (c:Name.t)(M N:MessageSpi.t),HOp.is_reflexive (HOp.relation_of_hedge (h0_seed c M N)).
    intros.
    unfold HOp.is_reflexive.
    unfold HOp.relation_of_hedge.
    intros.
    elim (HOp.HedgeProp.Add_add (H.add (M, M) (H.singleton (N, N))) (MessageSpi.MNam c,MessageSpi.MNam c) (m,n)).
    intros.
    case (H0 H);clear H0;clear H1;intros.
    simpl in H0.
    and_elim.
    rewrite <- H0.
    rewrite <- H1.
    auto.
    elim (HOp.HedgeProp.Add_add (H.singleton (N, N)) (M,M) (m,n)).
    intros.
    case (H1 H0);clear H1;clear H2;intros.
    simpl in H1.
    and_elim.
    rewrite <- H1;rewrite <- H2.
    auto.
    generalize (H.singleton_1 H1).
    intros.
    simpl in H2;and_elim.
    rewrite <- H2.
    rewrite <- H3.
    auto.
  Qed.

  Lemma h0_seed_Subset : forall (c:Name.t)(M N:MessageSpi.t),(H.Subset (h0_seed c M N) (h0_seed c N M)).
    unfold H.Subset, h0_seed;intros.
    elim (HOp.HedgeProp.Add_add (H.add (M, M) (H.singleton (N, N))) (MessageSpi.MNam c,MessageSpi.MNam c) a).
    intros.
    case (H0 H);clear H0;clear H1;intros.
    apply H.add_1.
    auto.
    elim (HOp.HedgeProp.Add_add (H.singleton (N, N)) (M,M) a).
    intros.
    case (H1 H0);clear H1;clear H2;intros.
    apply H.add_2.
    apply H.add_2.
    apply H.singleton_2.
    auto.
    apply H.add_2.
    apply H.add_1.
    apply H.singleton_1.
    auto.
  Qed.

  Lemma h0_seed_Equal : forall (c:Name.t)(M N:MessageSpi.t),(H.Equal (h0_seed c M N) (h0_seed c N M)).
    intros.
    apply HOp.HedgeProp.subset_antisym.
    apply h0_seed_Subset.
    apply h0_seed_Subset.
  Qed.
  
  Definition h0 (c:Name.t)(M N:MessageSpi.t) := 
    HOp.hedge_irreducible (H.add (MessageSpi.MNam c,MessageSpi.MNam c) (H.add (M,M) (H.singleton (N,N)))).

  Lemma h0_Equal : forall (c:Name.t)(M N:MessageSpi.t),(H.Equal (h0 c M N) (h0 c N M)).
    intros.
    unfold h0.
    apply HOp.hedge_irreducible_morphism.
    fold (h0_seed c M N).
    fold (h0_seed c N M).
    apply h0_seed_Equal.
  Qed.

  Lemma h0_reflexive : forall (c:Name.t)(M N:MessageSpi.t),HOp.is_reflexive (HOp.relation_of_hedge (h0 c M N)).
    unfold h0.
    intros.
    fold (h0_seed c M N).
    apply HOp.is_reflexive_irreducible.
    apply h0_seed_reflexive.
  Qed.

  Lemma h0_transpose : forall (c:Name.t)(M N:MessageSpi.t),(H.Equal (HOp.hedge_transpose (h0 c M N)) (h0 c M N)).
    intros.
    apply HOp.same_impl_Equal.
    eapply same_relation_trans.
    generalize (HOp.same_transpose_transp (h0 c M N));intro.
    apply H.
    apply same_relation_sym.
    apply HOp.is_reflexive_transp.
    apply h0_reflexive.
  Qed.

  Lemma h0_is_left_consistent : forall (c:Name.t)(M N:MessageSpi.t),HOp.is_left_consistent (h0 c M N).
    intros.
    unfold h0.
    fold (h0_seed c M N).
    apply HOp.is_reflexive_left_consistent.
    apply h0_seed_reflexive.
  Qed.

  Definition h1 (c:Name.t)(M N:MessageSpi.t)(k l:nat) := H.add (MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k)),MessageSpi.MEnc_s N (MessageSpi.MNam (Name.Ref l))) (h0 c M N).

  Lemma in_h1_inv : forall (c:Name.t)(M N:MessageSpi.t)(k l:nat)(m n:MessageSpi.t),H.In (m,n) (h1 c M N k l)->(m,n)=(MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k)),MessageSpi.MEnc_s N (MessageSpi.MNam (Name.Ref l))) \/ (H.In (m,n) (h0 c M N)).
    intros.
    unfold h1 in H.
    elim (HOp.HedgeProp.Add_add (h0 c M N) (MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k)),MessageSpi.MEnc_s N (MessageSpi.MNam (Name.Ref l))) (m,n)).
    intros.
    case (H0 H);clear H0;clear H1.
    simpl.
    intros.
    left.
    and_elim.
    rewrite <- H0.
    rewrite <- H1;auto.
    intros.
    right;auto.
  Qed.

  Lemma h1_same : forall (c:Name.t)(M N:MessageSpi.t)(k l:nat),same_relation _ (HOp.relation_of_hedge (h1 c M N k l)) (union _ (HOp.relation_of_hedge (H.singleton (MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k)), MessageSpi.MEnc_s N (MessageSpi.MNam (Name.Ref l))))) (HOp.relation_of_hedge (h0 c M N))).
    intros.
    split;unfold inclusion;intros.
    case (in_h1_inv _ _ _ _ _ _ _ H).
    intros.
    left.
    unfold HOp.relation_of_hedge.
    apply H.singleton_2.
    simpl.
    injection H0;intros.
    subst x;subst y.
    split;auto.
    intros.
    right.
    apply H0.
    case H.
    intros.
    unfold HOp.relation_of_hedge in H0.
    generalize (H.singleton_1 H0).
    simpl.
    intros.
    and_elim.
    unfold HOp.relation_of_hedge.      
    unfold h1.
    apply H.add_1.
    simpl.
    split;auto.
    intro.
    unfold HOp.relation_of_hedge.      
    unfold h1.
    apply H.add_2.
    apply H0.
  Qed.

  Lemma h1_Equal : forall (c:Name.t)(M N:MessageSpi.t)(k l:nat),H.Equal (h1 c M N k l) (H.union (H.singleton (MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k)), MessageSpi.MEnc_s N (MessageSpi.MNam (Name.Ref l)))) (h0 c M N)).
    intros.
    apply HOp.same_impl_Equal.
    eapply same_relation_trans.
    apply h1_same.
    apply same_relation_sym.
    apply HOp.same_union_union.
  Qed.

  Lemma h1_transpose : forall (c:Name.t)(M N:MessageSpi.t)(k l:nat),(H.Equal (HOp.hedge_transpose (h1 c M N k l)) (h1 c N M l k)).
    intros.
    eapply HOp.HedgeProp.equal_trans.
    apply HOp.hedge_transpose_morphism.
    apply h1_Equal.
    eapply HOp.HedgeProp.equal_trans.
    apply HOp.hedge_transpose_union.
    apply HOp.HedgeProp.equal_sym.
    eapply HOp.HedgeProp.equal_trans.
    apply h1_Equal.
    apply HOp.HedgeOtherProp.Equal_union_Equal_Equal.
    split;intros.
    destruct a as [m n].
    elim (HOp.hedge_transpose_prop (H.singleton (MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k)),MessageSpi.MEnc_s N (MessageSpi.MNam (Name.Ref l)))) n m).
    intros.
    apply H0;clear H0;clear H1.
    apply H.singleton_2.
    simpl.
    generalize (H.singleton_1 H);simpl;intros.
    tauto.
    destruct a as [m n].
    elim (HOp.hedge_transpose_prop (H.singleton (MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k)),MessageSpi.MEnc_s N (MessageSpi.MNam (Name.Ref l)))) n m).
    intros.
    generalize (H1 H);clear H0;clear H1.
    intro.
    apply H.singleton_2.
    simpl.
    generalize (H.singleton_1 H0);simpl;intros.
    tauto.
    apply HOp.HedgeProp.equal_sym.
    eapply HOp.HedgeProp.equal_trans.
    apply h0_transpose.
    apply h0_Equal.
  Qed.

  Lemma in_syn_h1_inv : forall (c:Name.t)(M N:MessageSpi.t)(k l:nat),(HOp.is_fresh_left_spec k (h0 c M N))->forall (m n:MessageSpi.t),(HOp.hedge_synthesis_relation (HOp.relation_of_hedge (h1 c M N k l)) m n)->deBruijn.mem_fn MessageSpi.deBruijn_fn_base k 0 m = false->(HOp.hedge_synthesis_relation (HOp.relation_of_hedge (h0 c M N)) m n).
    intros.
    apply (HOp.mem_fn_left_synthesis_union (HOp.relation_of_hedge (H.singleton (MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k)),MessageSpi.MEnc_s N (MessageSpi.MNam (Name.Ref l)))))).
    unfold HOp.relation_of_hedge.
    intros.
    generalize (H.singleton_1 H2).
    intros.
    simpl in H3.
    and_elim.
    exists k.
    split;auto.
    rewrite <- H3.
    rewrite MessageSpi.mem_fn_MEnc_s.
    apply orb_true_intro.
    right.
    simpl.
    apply (Name.fn_0_name k).
    elim (HOp.synthesis_morphism _ _ (h1_same c M N k l)).
    intros.
    apply H2.
    apply H0.
  Qed.

  Lemma h0_le_h_h1 : forall (c:Name.t)(M N:MessageSpi.t)(k l:nat),(HOp.le_h (HOp.relation_of_hedge (h0 c M N)) (HOp.relation_of_hedge (h1 c M N k l))).
    intros.
    apply HOp.incl_impl_le_h.      
    elim (h1_same c M N k l);intros.
    unfold inclusion;intros.
    apply H0.
    right;auto.
  Qed.

  Lemma h1_name_left_name : forall (c:Name.t)(M N:MessageSpi.t)(k l:nat),(HOp.is_fresh_left_spec k (h0 c M N))->(HOp.name_left_name (HOp.relation_of_hedge (h1 c M N k l))).
    intros.
    generalize (in_h1_inv c M N k l);intro.
    generalize (h0_is_left_consistent c M N).
    unfold HOp.is_left_consistent.
    unfold HOp.is_left_consistent_relation.
    intros;and_elim.
    unfold HOp.name_left_name.
    intros.
    case (H0 _ _ H10);intro.
    rewrite H11 in H12;discriminate H12.
    eapply H1;eauto.
    apply H12.
  Qed.

  Lemma h1_injective_left : forall (c:Name.t)(M N:MessageSpi.t)(k l:nat),(HOp.is_fresh_left_spec k (h0 c M N))->(HOp.injective_left (HOp.relation_of_hedge (h1 c M N k l))).
    intros.
    generalize (in_h1_inv c M N k l);intro.
    generalize (h0_is_left_consistent c M N).
    unfold HOp.is_left_consistent.
    unfold HOp.is_left_consistent_relation.
    intros;and_elim.
    unfold HOp.injective_left.
    intros.
    case (H0 _ _ H10);case (H0 _ _ H11);intros.
    injection H13;injection H14;intros.
    congruence.
    injection H14;clear H14;intros.
    subst m'.
    elim H.
    unfold HOp.is_left_name_spec.
    exists  m.
    exists n'.
    split;auto.
    rewrite H15.
    rewrite MessageSpi.mem_fn_MEnc_s.
    rewrite MessageSpi.mem_fn_MNam.
    apply orb_true_intro.
    right.
    apply Name.fn_0_name.
    injection H13;clear H13;intros.
    subst m'.
    elim H.
    exists  m.
    exists n.
    split;auto.
    rewrite H15.
    rewrite MessageSpi.mem_fn_MEnc_s.
    rewrite MessageSpi.mem_fn_MNam.
    apply orb_true_intro.
    right.
    apply Name.fn_0_name.
    eapply H2;eauto.
    apply H14.
    apply H13.
  Qed.

  Lemma h1_op_synthesis_left : forall (c:Name.t)(M N:MessageSpi.t)(k l:nat),(HOp.is_fresh_left_spec k (h0 c M N))->(HOp.op_synthesis_left (HOp.relation_of_hedge (h1 c M N k l))).
    intros.
    generalize (in_h1_inv c M N k l);intro.
    generalize (h0_is_left_consistent c M N).
    unfold HOp.is_left_consistent.
    unfold HOp.is_left_consistent_relation.
    intros;and_elim.
    unfold HOp.op_synthesis_left.
    intros.
    case (H0 _ _ H10);intros.
    injection H12;clear H12;intros.
    subst m.
    discriminate H13.
    red;intro.
    assert (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k 0 m' = false).
    case (Bool_elim_bool (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k 0 m'));intro;auto.
    elim H.
    exists m.
    exists n.
    split.
    apply H12.
    rewrite H11.
    rewrite MessageSpi.mem_fn_MOp.
    auto.
    generalize (in_syn_h1_inv _ _ _ _ _ H _ _ H13 H14).
    intros.
    apply (H3 _ _ H12 _ _ H11 _ H15).
  Qed.

  Lemma h1_pair_free_left : forall (c:Name.t)(M N:MessageSpi.t)(k l:nat),(HOp.is_fresh_left_spec k (h0 c M N))->(HOp.pair_free_left (HOp.relation_of_hedge (h1 c M N k l))).
    intros.
    generalize (in_h1_inv c M N k l);intro.
    generalize (h0_is_left_consistent c M N).
    unfold HOp.is_left_consistent.
    unfold HOp.is_left_consistent_relation.
    intros;and_elim.
    unfold HOp.pair_free_left.
    intros.
    red;intro.
    case (in_h1_inv _ _ _ _ _ _ _ H10).
    intros.
    discriminate H11.
    intros.
    apply (H4 _ _ _ H11).
  Qed.

  Lemma h1_enc_s_synthesis_left : forall (c:Name.t)(M N:MessageSpi.t)(k l:nat),(HOp.is_fresh_left_spec k (h0 c M N))->(HOp.enc_s_synthesis_left (HOp.relation_of_hedge (h1 c M N k l))).
    intros.
    generalize (in_h1_inv c M N k l);intro.
    generalize (h0_is_left_consistent c M N).
    unfold HOp.is_left_consistent.
    unfold HOp.is_left_consistent_relation.
    intros;and_elim.
    unfold HOp.enc_s_synthesis_left.
    intros.
    case (in_h1_inv _ _ _ _ _ _ _ H10);intros.
    injection H12;clear H12;intros.
    rewrite H11 in H13.
    injection H13;clear H13;intros.
    subst k0.
    right.
    intro l0;intros.
    red;intro.
    generalize (HOp.hedge_synthesis_spec_impl_hedge_in_synthesis _ _ _ H13).
    intro.
    rewrite HOp.name_left_synthesis in H15.
    generalize (H.mem_2 H15).
    intro.
    case (in_h1_inv _ _ _ _ _ _ _ H16);intros.
    discriminate H17.
    apply H.
    exists (MessageSpi.MNam (Name.Ref k)).
    exists l0.
    split.
    apply H17.
    simpl.
    apply (Name.fn_0_name k).
    case (H5 _ _ H12 _ _ H11);intro.
    left.
    intros.
    red;intro.
    apply (H13 n').
    assert (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k 0 m' = false).
    case (Bool_elim_bool (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k 0 m'));intro;auto.
    elim H.
    exists m.
    exists n.
    split.
    apply H12.
    rewrite H11.
    rewrite MessageSpi.mem_fn_MEnc_s.
    rewrite H15.
    apply orb_true_b.
    generalize (in_syn_h1_inv _ _ _ _ _ H _ _ H14 H15).
    intros.
    apply H16.
    right.
    intros.
    red;intro.
    apply (H13 l0).
    assert (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k 0 k0 = false).
    case (Bool_elim_bool (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k 0 k0));intro;auto.
    elim H.
    exists m.
    exists n.
    split.
    apply H12.
    rewrite H11.
    rewrite MessageSpi.mem_fn_MEnc_s.
    rewrite H15.
    apply orb_b_true.
    generalize (in_syn_h1_inv _ _ _ _ _ H _ _ H14 H15).
    intros.
    apply H16.
  Qed.

  Lemma h1_enc_s_analysis_left : forall (c:Name.t)(M N:MessageSpi.t)(k l:nat),(HOp.is_fresh_left_spec k (h0 c M N))->(HOp.enc_s_analysis_left (HOp.relation_of_hedge (h1 c M N k l))).
    intros.
    generalize (in_h1_inv c M N k l);intro.
    generalize (h0_is_left_consistent c M N).
    unfold HOp.is_left_consistent.
    unfold HOp.is_left_consistent_relation.
    intros;and_elim.
    unfold HOp.enc_s_analysis_left.
    intros.
    case (in_h1_inv _ _ _ _ _ _ _ H10);intros.
    injection H13;clear H13;intros.
    rewrite H11 in H14.
    injection H14;clear H14;intros.
    subst k0.
    generalize (HOp.hedge_synthesis_spec_impl_hedge_in_synthesis _ _ _ H12).
    intro.
    rewrite HOp.name_left_synthesis in H14.
    generalize (H.mem_2 H14).
    intro.
    case (in_h1_inv _ _ _ _ _ _ _ H16);intros.
    discriminate H17.
    elim H.
    exists (MessageSpi.MNam (Name.Ref k)).
    exists l0.
    split.
    apply H17.
    simpl.
    apply (Name.fn_0_name k).
    assert (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k 0 k0 = false).
    case (Bool_elim_bool (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k 0 k0));intro;auto.
    elim H.
    exists m.
    exists n.
    split.
    apply H13.
    rewrite H11.
    rewrite MessageSpi.mem_fn_MEnc_s.
    rewrite H14.
    apply orb_b_true.
    generalize (in_syn_h1_inv _ _ _ _ _ H _ _ H12 H14).
    intro.
    elim (H6 _ _ H13 _ _ H11 _ H15).
    intro n';intros.
    and_elim.
    exists n'.
    split;auto.
    apply h0_le_h_h1.
    auto.
  Qed.

  Lemma h1_enc_a_synthesis_left : forall (c:Name.t)(M N:MessageSpi.t)(k l:nat),(HOp.is_fresh_left_spec k (h0 c M N))->(HOp.enc_a_synthesis_left (HOp.relation_of_hedge (h1 c M N k l))).
    intros.
    generalize (in_h1_inv c M N k l);intro.
    generalize (h0_is_left_consistent c M N).
    unfold HOp.is_left_consistent.
    unfold HOp.is_left_consistent_relation.
    intros;and_elim.
    unfold HOp.enc_a_synthesis_left.
    intros.
    case (in_h1_inv _ _ _ _ _ _ _ H10);intros.
    injection H12;clear H12;intros.
    rewrite H11 in H13.
    discriminate H13.
    case (H7 _ _ H12 _ _ H11);intro.
    left.
    intros.
    red;intro.
    apply (H13 n').
    assert (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k 0 m' = false).
    case (Bool_elim_bool (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k 0 m'));intro;auto.
    elim H.
    exists m.
    exists n.
    split.
    apply H12.
    rewrite H11.
    rewrite MessageSpi.mem_fn_MEnc_a.
    rewrite H15.
    apply orb_true_b.
    generalize (in_syn_h1_inv _ _ _ _ _ H _ _ H14 H15).
    intros.
    apply H16.
    right.
    intros.
    red;intro.
    apply (H13 l0).
    assert (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k 0 k0 = false).
    case (Bool_elim_bool (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k 0 k0));intro;auto.
    elim H.
    exists m.
    exists n.
    split.
    apply H12.
    rewrite H11.
    rewrite MessageSpi.mem_fn_MEnc_a.
    rewrite H15.
    apply orb_b_true.
    generalize (in_syn_h1_inv _ _ _ _ _ H _ _ H14 H15).
    intros.
    apply H16.
  Qed.

  Lemma h1_enc_a_analysis_left : forall (c:Name.t)(M N:MessageSpi.t)(k l:nat),(HOp.is_fresh_left_spec k (h0 c M N))->(HOp.enc_a_analysis_left (HOp.relation_of_hedge (h1 c M N k l))).
    intros.
    generalize (in_h1_inv c M N k l);intro.
    generalize (h0_is_left_consistent c M N).
    unfold HOp.is_left_consistent.
    unfold HOp.is_left_consistent_relation.
    intros;and_elim.
    unfold HOp.enc_a_analysis_left.
    intros.
    case (in_h1_inv _ _ _ _ _ _ _ H10);intros.
    injection H14;clear H14;intros.
    rewrite H11 in H15.
    discriminate H15.
    assert (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k 0 k0 = false).
    case (Bool_elim_bool (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k 0 k0));intro;auto.
    elim H.
    exists m.
    exists n.
    split.
    apply H14.
    rewrite H11.
    rewrite MessageSpi.mem_fn_MEnc_a.
    rewrite H15.
    apply orb_b_true.
    rewrite (MessageSpi.mem_fn_inverse_key 0 k _ _ H12) in H15.
    generalize (in_syn_h1_inv _ _ _ _ _ H _ _ H13 H15).
    intro.
    elim (H8 _ _ H14 _ _ H11 _ H12 _ H16).
    intro l0;intros.
    ex_elim H17 n'.
    exists l0.
    exists n'.
    split;auto.
    split;auto.
    apply h0_le_h_h1.
    auto.
  Qed.

  Lemma h1_injective_inv_left : forall (c:Name.t)(M N:MessageSpi.t)(k l:nat),(HOp.is_fresh_left_spec k (h0 c M N))->(HOp.injective_inv_left (HOp.relation_of_hedge (h1 c M N k l))).
    intros.
    generalize (in_h1_inv c M N k l);intro.
    generalize (h0_is_left_consistent c M N).
    unfold HOp.is_left_consistent.
    unfold HOp.is_left_consistent_relation.
    intros;and_elim.
    unfold HOp.injective_inv_left.
    intros.
    case (H0 _ _ H10);case (H0 _ _ H11);intros.
    injection H13;injection H14;intros.
    subst m;subst m'.
    discriminate H12.
    injection H14;clear H14;intros.
    subst m.
    discriminate H12.
    injection H13;clear H13;intros.
    subst m'.
    destruct m;try (discriminate H12).
    destruct o;try (discriminate H12).
    eapply H9;eauto.
    apply H14.
    apply H13.
  Qed.

  Lemma h1_is_left_consistent : forall (c:Name.t)(M N:MessageSpi.t)(k l:nat),(HOp.is_fresh_left_spec k (h0 c M N))->(HOp.is_left_consistent (h1 c M N k l)).
    intros.
    unfold HOp.is_left_consistent.
    unfold HOp.is_left_consistent_relation.
    repeat split.
    apply h1_name_left_name;auto.
    apply h1_injective_left;auto.
    apply h1_op_synthesis_left;auto.
    apply h1_pair_free_left;auto.
    apply h1_enc_s_synthesis_left;auto.
    apply h1_enc_s_analysis_left;auto.
    apply h1_enc_a_synthesis_left;auto.
    apply h1_enc_a_analysis_left;auto.
    apply h1_injective_inv_left;auto.
  Qed.

  Lemma irr_h1_equal_h1 : forall (c:Name.t)(M N:MessageSpi.t)(k l:nat),(HOp.is_fresh_left_spec k (h0 c M N))->(H.Equal (h1 c M N k l) (HOp.hedge_irreducible (h1 c M N k l))).
    intros.
    apply HOp.cond_irr_impl_irreducible.
    unfold HOp.cond_irr.
    apply HOp.is_left_consistent_impl_cond_irr.
    apply h1_is_left_consistent.
    auto.
  Qed.

  Lemma h1_transpose_is_left_consistent : forall (c:Name.t)(M N:MessageSpi.t)(k l:nat),(HOp.is_fresh_right_spec l (h0 c M N))->(HOp.is_left_consistent (HOp.hedge_transpose (h1 c M N k l))).
    intros.
    unfold HOp.is_left_consistent.
    eapply HOp.same_impl_is_left_consistent_relation_impl.
    apply same_relation_sym.
    apply HOp.Equal_impl_same.
    apply h1_transpose.
    apply h1_is_left_consistent.
    elim (HOp.same_is_fresh_left _ _ (HOp.Equal_impl_same _ _ (h0_Equal c M N)) l);intros.
    apply H0.
    elim (HOp.is_reflexive_fresh_left _ (h0_reflexive c M N) l).
    intros.
    apply H3.
    apply H.
  Qed.

  (** c is a detectable channel *)
  Lemma c_in_h0 : forall (c:Name.t)(M N:MessageSpi.message),(H.In (MessageSpi.MNam c,MessageSpi.MNam c) (h0 c M N)).
    intros.
    apply H.mem_2.
    rewrite <- HOp.name_left_synthesis.
    apply HOp.hedge_synthesis_spec_impl_hedge_in_synthesis.
    unfold h0.
    generalize HOp.h_le_h_irr_h.
    unfold HOp.le_h.
    intros.
    unfold HOp.hedge_synthesis_spec.
    apply H.
    apply HOp.SynInc.
    unfold HOp.relation_of_hedge.
    apply H.add_1.
    reflexivity.
  Qed.

  Lemma c_in_h0_unique : forall (c:Name.t)(M N:MessageSpi.message),forall (d:MessageSpi.message),(H.In (MessageSpi.MNam c,d) (h0 c M N))->d=(MessageSpi.MNam c).
    intros.
    elim (h0_is_left_consistent c M N);intros;and_elim.
    rewrite (H1 _ _ (c_in_h0 c M N) _ _ H);auto.
  Qed.

  Lemma c_left_name_h0 : forall (c:Name.t)(M N:MessageSpi.message),forall (x:nat),(deBruijn.mem_fn Name.deBruijn_fn_base x 0 c = true)->(HOp.is_left_name_spec x (h0 c M N)).
    intros.
    exists (MessageSpi.MNam c).
    exists (MessageSpi.MNam c).
    split;auto.
    unfold HOp.relation_of_hedge.
    apply c_in_h0.
  Qed.

  Lemma M_left_name_h0 : forall (c:Name.t)(M N:MessageSpi.message),forall (x:nat),(deBruijn.mem_fn MessageSpi.deBruijn_fn_base x 0 M = true)->(HOp.is_left_name_spec x (h0 c M N)).
    intros.
    unfold HOp.is_left_name_spec.
    unfold h0.
    eapply HOp.mem_fn_le_h.
    apply HOp.h_le_h_irr_h. 
    exists M.
    exists M.
    split.
    unfold HOp.relation_of_hedge.
    apply H.add_2.
    apply H.add_1.
    simpl;auto.
    auto.
  Qed.

  Lemma N_left_name_h0 : forall (c:Name.t)(M N:MessageSpi.message),forall (x:nat),(deBruijn.mem_fn MessageSpi.deBruijn_fn_base x 0 N = true)->(HOp.is_left_name_spec x (h0 c M N)).
    intros.
    unfold HOp.is_left_name_spec.
    eapply HOp.inclusion_is_left_name.
    elim (HOp.Equal_impl_same _ _ (h0_Equal c M N));intros.
    apply H1.
    apply M_left_name_h0.
    auto.
  Qed.

  Lemma h0_fresh : forall (c:Name.t)(M N:MessageSpi.message),forall (x:nat),(HOp.is_fresh_left_spec x (h0 c M N))<->(HOp.is_fresh_right_spec x (h0 c M N)).
    intros.
    unfold HOp.is_fresh_left_spec, HOp.is_fresh_right_spec.
    apply HOp.is_reflexive_fresh_left.
    apply h0_reflexive.
  Qed.

  Lemma fn_Ps_h0 : forall (c:Name.t)(M N:MessageSpi.message),forall (x:nat),(fn x (Ps c M))->(HOp.is_left_name_spec x (h0 c M N)).
    intros.
    unfold Ps in H.
    unfold fn in H.
    rewrite AgentSpi.P.fn_New in H.
    rewrite AgentSpi.P.fn_Output in H.
    rewrite AgentSpi.P.fn_Zero in H.
    rewrite orb_b_false in H.
    case (orb_true_elim _ _ H);clear H;intro H.
    simpl in H.
    generalize (deBruijn.fn_plus_lift Name.deBruijn_fn_base 0 1 (Name.lift 0 1 c) x).
    simpl;intros.
    rewrite <- H0 in H.
    generalize (deBruijn.lift_fn Name.deBruijn_fn_base 0 1 c x).
    simpl;intros.
    rewrite H1 in H.
    apply c_left_name_h0.
    auto.
    rewrite EvalSpi.mem_fn_expr_of_msg in H.
    rewrite MessageSpi.mem_fn_MEnc_s in H.
    assert (deBruijn.mem_fn MessageSpi.deBruijn_fn_base x 1 (MessageSpi.MNam (Name.Ref 0)) = false).
    reflexivity.
    rewrite H0 in H.
    rewrite orb_b_false in H.
    generalize (deBruijn.fn_plus_lift MessageSpi.deBruijn_fn_base 0 1 (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) x).
    replace (0+1) with 1;try reflexivity.
    intro.
    rewrite <- H1 in H.
    generalize (deBruijn.lift_fn MessageSpi.deBruijn_fn_base 0 1 M x).
    intros.
    apply M_left_name_h0.
    rewrite <- H2.
    auto.
  Qed.

  (** the bisimulation is the following *)
  Inductive perfect_encryption1 (c:Name.t)(M N:MessageSpi.message) : hedged_relation_type :=
  | C1 : (perfect_encryption1 c M N ((h0 c M N),(Ps c M),(Ps c N)))
  | C2 : forall (k:nat),HOp.is_fresh_left_spec k (h0 c M N)->(perfect_encryption1 c M N ((h1 c M N k k,AgentSpi.P.Zero,AgentSpi.P.Zero)))
  | C1sym : (perfect_encryption1 c M N ((h0 c N M),(Ps c N),(Ps c M)))
  | C2sym : forall (k:nat),HOp.is_fresh_left_spec k (h0 c N M)->(perfect_encryption1 c M N ((h1 c N M k k,AgentSpi.P.Zero,AgentSpi.P.Zero))).

  Lemma perfect_encryption1_symmetric_hedged : forall (c:Name.t)(M N:MessageSpi.message),(symmetric_hedged_relation (perfect_encryption1 c M N)).
    unfold symmetric_hedged_relation.
    intros.
    inversion H.
    subst h;subst p;subst q.
    exists (h0 c N M).
    split.
    apply HOp.HedgeProp.equal_sym.
    eapply HOp.HedgeProp.equal_trans.
    apply h0_transpose.
    apply h0_Equal.
    apply C1sym.
    subst h;subst p;subst q.
    exists (h1 c N M k k).
    split.
    apply HOp.HedgeProp.equal_sym.
    apply h1_transpose.
    apply C2sym.
    unfold HOp.is_fresh_left_spec.
    eapply HOp.inclusion_is_fresh_left.
    elim (HOp.Equal_impl_same _ _ (h0_Equal c M N)).
    intros.
    apply H2.
    auto.
    subst h;subst p;subst q.
    exists (h0 c M N).
    split.
    apply HOp.HedgeProp.equal_sym.
    eapply HOp.HedgeProp.equal_trans.
    apply h0_transpose.
    apply h0_Equal.
    apply C1.
    subst h;subst p;subst q.
    exists (h1 c M N k k).
    split.
    apply HOp.HedgeProp.equal_sym.
    apply h1_transpose.
    apply C2.
    unfold HOp.is_fresh_left_spec.
    eapply HOp.inclusion_is_fresh_left.
    elim (HOp.Equal_impl_same _ _ (h0_Equal c M N)).
    intros.
    apply H0.
    auto.
  Qed.

  Lemma perfect_encryption1_left_hedged : forall (c:Name.t)(M N:MessageSpi.message),(left_hedged_relation (perfect_encryption1 c M N)).
    unfold left_hedged_relation.
    intros.
    inversion H;subst h;subst p;subst q.
    apply fn_Ps_h0.
    auto.
    discriminate H0.
    apply fn_Ps_h0.
    auto.
    discriminate H0.
  Qed.

  Lemma perfect_encryption1_left_consistent : forall (c:Name.t)(M N:MessageSpi.message),(left_consistent_hedged_relation (perfect_encryption1 c M N)).
    intros.
    unfold left_consistent_hedged_relation.
    intros.
    inversion H;subst h;subst p;subst q.
    apply h0_is_left_consistent.
    apply h1_is_left_consistent.
    auto.
    apply h0_is_left_consistent.
    apply h1_is_left_consistent.
    auto.
  Qed.

  Lemma perfect_encryption1_late_hedged_bisimulation : forall (c:Name.t)(M N:MessageSpi.message),(late_hedged_bisimulation_condition (perfect_encryption1 c M N)).
    unfold late_hedged_bisimulation_condition.
    intros.
    inversion H;subst h;subst p;subst q.
      (** C1 *)
    repeat split;intros.
    elim (Ps_no_silent_trans _ _ _ H0).
    elim (Ps_no_input_trans _ _ _ _ H0).
    elim (Ps_output_trans_unique _ _ _ _ _ _ H0);intros;and_elim.
    subst a.
    subst c0.
    subst p'.
    exists 1.
    exists (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 N) (MessageSpi.MNam (Name.Ref 0))).
    exists AgentSpi.P.Zero.
    generalize (c_in_h0_unique _ _ _ _ H2).
    intro.
    injection H3;clear H3;intros;subst b.
    split.
    apply Ps_output_trans_exists.
    subst m.
    elim (HOp.n_fresh_left_single _ _ H1).
    intro k;intros.
    and_elim.
    exists tildec.
    split.
    apply HOp.is_reflexive_n_fresh_left.
    apply h0_reflexive.
    auto.
    exists (h1 c M N k k).
    split.
    subst tildec.
    repeat (rewrite instantiate_message_Ps).
    apply (irr_h1_equal_h1 c M N k k).
    auto.
    unfold instantiate_process.
    rewrite AgentSpi.P.low_subst_Zero.
    apply C2.
    auto.
      (** C2 *)
    repeat split;intros.
    inversion H0.
    inversion H0.
    inversion H0.
      (** C1sym *)
    repeat split;intros.
    elim (Ps_no_silent_trans _ _ _ H0).
    elim (Ps_no_input_trans _ _ _ _ H0).
    elim (Ps_output_trans_unique _ _ _ _ _ _ H0);intros;and_elim.
    subst a.
    subst c0.
    subst p'.
    exists 1.
    exists (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0))).
    exists AgentSpi.P.Zero.
    generalize (c_in_h0_unique _ _ _ _ H2).
    intro.
    injection H3;clear H3;intros;subst b.
    split.
    apply Ps_output_trans_exists.
    subst m.
    elim (HOp.n_fresh_left_single _ _ H1).
    intro k;intros.
    and_elim.
    exists tildec.
    split.
    apply HOp.is_reflexive_n_fresh_left.
    apply h0_reflexive.
    auto.
    exists (h1 c N M k k).
    split.
    subst tildec.
    repeat (rewrite instantiate_message_Ps).
    apply (irr_h1_equal_h1 c N M k k).
    auto.
    unfold instantiate_process.
    rewrite AgentSpi.P.low_subst_Zero.
    apply C2sym.
    auto.
      (** C2sym *)
    repeat split;intros.
    inversion H0.
    inversion H0.
    inversion H0.
  Qed.

  Theorem perfect_encryption1_is_late_hedged_bisimulation : forall (c:Name.t)(M N:MessageSpi.message),late_hedged_bisimulation (perfect_encryption1 c M N).
    unfold late_hedged_bisimulation.
    unfold symmetric_left_consistent_hedged_relation.
    intros.
    split.
    split.
    apply perfect_encryption1_symmetric_hedged.
    split.
    apply perfect_encryption1_left_hedged.
    apply perfect_encryption1_left_consistent.
    apply perfect_encryption1_late_hedged_bisimulation.
  Qed.

  Theorem perfect_encryption1_late_hedged_bisimilar : forall (c:Name.t)(M N:MessageSpi.message),late_hedged_bisimilar (h0 c M N) (Ps c M) (Ps c N).
    unfold late_hedged_bisimilar;intros.
    exists (perfect_encryption1 c M N).
    split.
    apply C1.
    apply perfect_encryption1_is_late_hedged_bisimulation.
  Qed.

  (** Example 2 *)
  (** We show that (^k)'c<enc_s(M,k)>'c<enc_s(M,k)> != (^k)'c<enc_s(N,k)>'c<enc_s(N',k)> *)
  (** provided that N != N' and c is a detectable channel *)
  Definition Qs (c:Name.t)(M M':MessageSpi.t) :=
    AgentSpi.P.New 
    (AgentSpi.P.Output 
      (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 c)) 
      (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0)))) 
      (AgentSpi.P.Output 
        (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 c)) 
        (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M') (MessageSpi.MNam (Name.Ref 0)))) 
        AgentSpi.P.Zero)).

  Definition Qs' (c:Name.t)(M':MessageSpi.t) :=       
    AgentSpi.P.Output 
    (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 c)) 
    (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M') (MessageSpi.MNam (Name.Ref 0)))) 
    AgentSpi.P.Zero.

  Definition Qs'inst (c:Name.t)(M':MessageSpi.t)(k:Name.t) :=
    AgentSpi.P.Output 
    (ExpressionSpi.ENam c) 
    (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s M' (MessageSpi.MNam k)))
    AgentSpi.P.Zero.

  Lemma instantiate_Qs' : forall (k:Name.t)(c:Name.t)(M':MessageSpi.t),(instantiate_process (cons k nil) (Qs' c M'))=(Qs'inst c M' k).
    unfold Qs', Qs'inst.
    intros.
    unfold instantiate_process.
    rewrite AgentSpi.P.low_subst_Output.
    rewrite <- low_subst_AgentSpi.
    generalize (deBruijn.low_subst_lift_nil ExpressionSpi.deBruijn_subst 0 0 0 (map ExpressionSpi.j (k :: nil)) (ExpressionSpi.ENam c)).
    fold ExpressionSpi.deBruijn_base.
    lazy beta delta [length map] iota.     
    repeat (rewrite plus_0_l).     
    intro.
    rewrite AgentSpi.P.low_subst_Zero.
    replace (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 c)) with (deBruijn.lift ExpressionSpi.deBruijn_base 0 1 (ExpressionSpi.ENam c));try reflexivity.
    rewrite H.
    rewrite deBruijn.lift_0.
    rewrite <- low_subst_AgentSpi.
    generalize (SubstSpi.low_subst_expr_of_msg (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M') (MessageSpi.MNam (Name.Ref 0))) 0 0 (cons (MessageSpi.j k) nil)).
    intros.
    replace (SubstSpi.msg_subst (cons (MessageSpi.j k) nil)) with (cons (ExpressionSpi.j k) nil) in H0;try reflexivity.
    rewrite <- H0.
    rewrite MessageSpi.low_subst_MEnc_s.
    clear H0.
    generalize (deBruijn.low_subst_lift_nil MessageSpi.deBruijn_subst 0 0 0 (map MessageSpi.j (k :: nil)) M').
    repeat (rewrite plus_0_l).
    lazy beta delta [map length] iota.
    fold MessageSpi.deBruijn_base.
    rewrite (deBruijn.lift_0 MessageSpi.deBruijn_base 0 M').
    intros.
    rewrite H0.
    assert (deBruijn.low_subst MessageSpi.deBruijn_subst 0 0 (MessageSpi.j k :: nil) (MessageSpi.MNam (Name.Ref 0)) = MessageSpi.MNam (deBruijn.lift Name.deBruijn_base 0 0 k)).
    reflexivity.
    rewrite (deBruijn.lift_0 Name.deBruijn_base) in H1.
    rewrite <- H1.
    auto.
  Qed.
  
  Lemma Qs_output_trans_unique : forall (c:Name.t)(M M':MessageSpi.t)(a:Name.t)(n:nat)(m:MessageSpi.t)(p':AgentSpi.P.t),(RedSpiCong.R.reduction (Qs c M M') (ActionSpi.Outp a) (AgentSpi.C (AgentSpi.Conc n (EvalSpi.expr_of_mesg m) p')))->a=c /\ n=1 /\ p'=(Qs' c M') /\ m=(MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0))).
    intros.
    unfold Qs in H.
    inversion H.
    inversion H2.
    unfold ActionSpi.mk_output in H10.
    simpl in H10.
    rewrite EvalSpi.evalc_msg in H10.
    injection H10;clear H10;intros.
    subst m0.
    subst p0.
    subst mu0.
    subst mu.
    unfold ActionSpi.mk_res in H4.
    simpl in H4.
    assert (0 < 1);try omega.
    generalize (deBruijn.lift_not_fn Name.deBruijn_fn_base 0 1 0 H6 c);intro.
    simpl in H8.
    rewrite H8 in H4.
    injection H4;clear H4;intros.
    clear H6.
    generalize (deBruijn.low1_lift_lift Name.deBruijn_base 0 0 0 c).
    intros.
    simpl in H6.
    rewrite H6 in H4.
    generalize (deBruijn.lift_0 Name.deBruijn_base 0 c).
    intros.
    simpl in H10.
    rewrite H10 in H4.
    subst a.
    clear H8;clear H6;clear H10.
    subst mu'.
    subst a0.
    clear H2.
    subst p.
    subst e.
    subst f.
    unfold ActionSpi.A.new in H1.
    rewrite ActionSpi.A.new_conc_def in H1.
    rewrite ExpressionSpi.mem_fn_EEnc_s in H1.
    assert (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 0 (ExpressionSpi.ENam (Name.Ref 0)) = true).
    reflexivity.
    rewrite H0 in H1.
    rewrite orb_b_true in H1.
    injection H1;clear H1;intros.
    subst n.
    subst p'.
    clear H.
    repeat (split;trivial).
    clear H0.
    assert (ExpressionSpi.EEnc_s (EvalSpi.expr_of_mesg (MessageSpi.t_map (fun d : nat => Name.lift d 1) 0 M)) (ExpressionSpi.ENam (Name.Ref 0)) = (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0)))));try reflexivity.
    rewrite H in H2;clear H.
    symmetry.
    apply EvalSpi.expr_of_mesg_injective.
    auto.
  Qed.

  Lemma Qs_output_trans_exists : forall (c:Name.t)(M M':MessageSpi.t),(RedSpiCong.R.reduction (Qs c M M') (ActionSpi.Outp c) (AgentSpi.C (AgentSpi.Conc 1 (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0)))) (Qs' c M')))).
    intros.
    replace (AgentSpi.C (AgentSpi.Conc 1 (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0)))) (Qs' c M'))) with (AgentSpi.new (AgentSpi.C (AgentSpi.Conc 0 (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0)))) (Qs' c M')))).
    unfold Qs, Qs'.
    eapply RedSpiCong.R.RedRes.
    eapply RedSpiCong.R.RedOutput.
    unfold ActionSpi.mk_output.
    lazy beta delta [EvalSpi.evalc] iota;fold EvalSpi.evalc.
    rewrite EvalSpi.evalc_msg.
    reflexivity.
    unfold ActionSpi.mk_res.
    assert (0<1);try omega.
    generalize (deBruijn.lift_not_fn Name.deBruijn_fn_base 0 1 0 H c);intro.
    simpl.
    simpl in H0.
    rewrite H0.
    generalize (deBruijn.low1_lift_lift Name.deBruijn_base 0 0 0 c).
    intros.
    simpl in H1.
    rewrite H1.
    rewrite (deBruijn.lift_0 Name.deBruijn_base).
    auto.
    unfold AgentSpi.new.
    rewrite AgentSpi.new_conc_def.
    lazy beta delta [EvalSpi.expr_of_mesg] iota;fold EvalSpi.expr_of_mesg.
    rewrite ExpressionSpi.mem_fn_EEnc_s.
    assert (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 0 (ExpressionSpi.ENam (Name.Ref 0)) = true).
    reflexivity.
    rewrite H.
    rewrite orb_b_true.
    auto.
  Qed.

  Lemma Qs'inst_output_trans_unique : forall (c:Name.t)(M':MessageSpi.t)(k:Name.t)(a:Name.t)(n:nat)(m:MessageSpi.t)(p':AgentSpi.P.t),(RedSpiCong.R.reduction (Qs'inst c M' k) (ActionSpi.Outp a) (AgentSpi.C (AgentSpi.Conc n (EvalSpi.expr_of_mesg m) p')))->a=c /\ n=0 /\ p'=AgentSpi.P.Zero /\ m=(MessageSpi.MEnc_s M' (MessageSpi.MNam k)).
    intros.
    unfold Qs'inst in H.
    inversion H.
    subst p;subst mu;subst n;subst m0.
    unfold ActionSpi.mk_output in H4.
    simpl in H4.
    rewrite EvalSpi.evalc_msg in H4.
    injection H4;clear H4;intros.
    subst e;subst a.
    split;auto.
    split;auto.
    split;auto.
    subst f.
    assert ((ExpressionSpi.EEnc_s (EvalSpi.expr_of_mesg M') (ExpressionSpi.ENam k)) = (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s M' (MessageSpi.MNam k))));try reflexivity.
    rewrite H0 in H1.
    symmetry.
    apply EvalSpi.expr_of_mesg_injective.
    auto.
  Qed.
  
  Lemma Qs'inst_output_trans_exists : forall (c:Name.t)(M':MessageSpi.t)(k:Name.t),(RedSpiCong.R.reduction (Qs'inst c M' k) (ActionSpi.Outp c) (AgentSpi.C (AgentSpi.Conc 0 (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s M' (MessageSpi.MNam k))) AgentSpi.P.Zero))).
    intros.
    unfold Qs'inst.
    eapply RedSpiCong.R.RedOutput.
    unfold ActionSpi.mk_output.
    simpl.
    rewrite EvalSpi.evalc_msg.
    reflexivity.
  Qed.

  Definition h2 (c:Name.t)(M N N':MessageSpi.t)(k l:nat) := HOp.hedge_irreducible (H.add (MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k)),MessageSpi.MEnc_s N' (MessageSpi.MNam (Name.Ref l))) (h1 c M N k l)).
  
  Lemma h2_is_left_consistent_iff : forall (c:Name.t)(M N N':MessageSpi.t)(k l:nat),(HOp.is_fresh_left_spec k (h0 c M N))->(N = N' <->(HOp.is_left_consistent (h2 c M N N' k l))).
    intros;split.
    intro.
    rewrite <- H0.
    unfold h2.
    unfold HOp.is_left_consistent.
    eapply HOp.same_impl_is_left_consistent_relation_impl.
    apply HOp.Equal_impl_same.
    eapply HOp.HedgeProp.equal_trans.
    apply (irr_h1_equal_h1 c M N k l).
    apply H.
    apply HOp.hedge_irreducible_morphism.
    unfold h1.
    split;intros.
    auto.
    elim (HOp.HedgeOtherProp.add_In_simpl (H.add (MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k)),MessageSpi.MEnc_s N (MessageSpi.MNam (Name.Ref l))) (h0 c M N)) (MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k)), MessageSpi.MEnc_s N (MessageSpi.MNam (Name.Ref l))) a).
    intros.
    case (H2 H1);clear H1;clear H2;clear H3;intros.
    apply H.add_1.
    auto.
    auto.
    apply h1_is_left_consistent.
    auto.
    intros.
    elim H0.
    intros;and_elim.
    generalize (HOp.inj_left_syn _ H2 H5 H7 H3 H4).
    clear H0;clear H1;clear H2;clear H3;clear H4;clear H5;clear H6;clear H7;clear H8;clear H9.
    intro.
    generalize (HOp.h_le_h_irr_h (H.add (MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k)),MessageSpi.MEnc_s N' (MessageSpi.MNam (Name.Ref l))) (h1 c M N k l))).
    fold (h2 c M N N' k l).
    unfold HOp.le_h.
    intros.
    assert (HOp.hedge_synthesis_relation (HOp.relation_of_hedge (h2 c M N N' k l))  (MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k))) (MessageSpi.MEnc_s N (MessageSpi.MNam (Name.Ref l)))).
    apply H1.
    apply HOp.SynInc.
    unfold HOp.relation_of_hedge.
    apply H.add_2.
    unfold h1.
    auto.
    assert (HOp.hedge_synthesis_relation (HOp.relation_of_hedge (h2 c M N N' k l))  (MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k))) (MessageSpi.MEnc_s N' (MessageSpi.MNam (Name.Ref l)))).
    apply H1.
    apply HOp.SynInc.
    unfold HOp.relation_of_hedge.
    auto.
    generalize (H0 _ _ H2 _ _ H3).
    intros.
    assert (MessageSpi.MEnc_s N (MessageSpi.MNam (Name.Ref l)) = MessageSpi.MEnc_s N' (MessageSpi.MNam (Name.Ref l))).
    apply H4;auto.
    clear H4.
    injection H5.
    auto.
  Qed.

  (** this axiom will be removed soon *)
  (** it just says that "there always exist a fresh name" *)
  Axiom exists_fresh : forall (h:H.t)(n:nat),exists tildec:list Name.t,HOp.n_fresh_left h n tildec.
  
  Lemma perfect_encryption2_not_bisimilar : forall (c:Name.t)(M N N':MessageSpi.t),(N<>N')->~exists h:H.t,(H.In (MessageSpi.MNam c,MessageSpi.MNam c) h)/\(late_hedged_bisimilar h (Qs c M M) (Qs c N N')).
    intros.
    red;intro.
    ex_elim H0 h.
    elim H1.
    intro R;intros;and_elim.
    clear H1.
    unfold late_hedged_bisimulation in H3.
    elim H3.
    clear H3;intro H3.
    unfold symmetric_left_consistent_hedged_relation in H3.
    and_elim.
    intro.
    elim (H5 _ _ _ H2).
    intros;and_elim.
    clear H6.
    clear H7.
    generalize (H8 _ _ _ _ (Qs_output_trans_exists c M M)).
    clear H8;intros.
    elim (exists_fresh h 1).
    intro tildec;intro.
    elim (H6 _ H7 _ H0).
    clear H6.
    intro d;intros.
    ex_elim H6 n.
    ex_elim H6 q'.
    ex_elim H8 tilded.
    ex_elim H9 h'.
    elim (Qs_output_trans_unique _ _ _ _ _ _ _ H6);intros;and_elim.
    subst d.
    clear H11.
    subst q'.
    subst n.
    clear H6.
    elim (HOp.n_fresh_left_single _ _ H7).
    intro k;intros.
    and_elim.
    subst tildec.
    assert (exists l:nat,tilded=cons (Name.Ref l) nil).
    elim H8;clear H8;intros.
    elim H6;clear H6;intros.
    destruct tilded;try (discriminate H6).
    destruct tilded;try (discriminate H6).
    destruct t.
    exists n;auto.
    ex_elim H6 l.
    subst tilded.
    repeat (rewrite instantiate_message_Ps in H9).
    repeat (rewrite instantiate_Qs' in H10).
    elim (H5 _ _ _ H10).
    intros;and_elim.
    clear H6.
    clear H12.
    generalize (H13 _ _ _ _ (Qs'inst_output_trans_exists c M (Name.Ref k))).
    clear H13.
    intros.
    assert (HOp.n_fresh_left h' 0 nil).
    split.
    apply HOp.n_pwd_nil.
    apply HOp.are_fresh_left_nil.
    generalize (H6 _ H12);clear H6;intros.
    assert (H.In (MessageSpi.MNam c,MessageSpi.MNam c) h').
    apply H.mem_2.
    rewrite <- HOp.name_left_synthesis.
    apply HOp.hedge_synthesis_spec_impl_hedge_in_synthesis.
    generalize (HOp.HedgeProp.subset_equal (HOp.HedgeProp.equal_sym H9)).
    intro.
    generalize (HOp.Subset_impl_inclusion _ _ H13);clear H13;intro.
    generalize (HOp.incl_impl_le_h _ _ H13).
    unfold HOp.le_h.
    intros.
    unfold HOp.hedge_synthesis_spec.
    apply H14;clear H14.
    apply HOp.h_le_h_irr_h.
    apply HOp.SynInc.
    unfold HOp.relation_of_hedge.
    apply H.add_2.
    auto.
    elim (H6 _ H13).
    intro d;intros.
    clear H6.
    ex_elim H14 n.
    ex_elim H14 q'.
    ex_elim H6 tilded.
    ex_elim H15 h''.
    elim (Qs'inst_output_trans_unique _ _ _ _ _ _ _ H14);intros;and_elim.
    subst d.
    subst q'.
    assert (tilded = nil).
    elim H6;intros.
    inversion H18.
    destruct tilded;try (discriminate H21).
    auto.
    subst tilded.
    subst n.
    repeat (rewrite instantiate_message_nil in H15).
    repeat (rewrite instantiate_process_nil in H16).
    generalize (H4 _ _ _ H16).
    intro.
    clear H0.
    clear H3.
    clear H1.
    clear H4.
    clear H2.
    clear H5.
    clear H11.
    clear H7.
    clear H8.
    clear H10.
    clear H12.
    clear H13.
    clear H17.
    clear H6.
    clear H16.
    clear H14.
    apply H.
    elim H18;clear H18;intros;and_elim.
    generalize (HOp.inj_left_syn _ H1 H4 H6 H2 H3).
    clear H0;clear H1;clear H2;clear H3;clear H4;clear H5;clear H6;clear H7;clear H8.
    intro.
    generalize (HOp.HedgeProp.subset_equal (HOp.HedgeProp.equal_sym H9));intro.
    generalize (HOp.HedgeProp.subset_equal (HOp.HedgeProp.equal_sym H15));intro.
    generalize (HOp.Subset_impl_inclusion _ _ H1);clear H1;intro.
    generalize (HOp.Subset_impl_inclusion _ _ H2);clear H2;intro.
    generalize (HOp.incl_impl_le_h _ _ H1);clear H1;intro.
    generalize (HOp.incl_impl_le_h _ _ H2);clear H2;intro.
    generalize (HOp.h_le_h_irr_h (H.add (MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k)), MessageSpi.MEnc_s N (MessageSpi.MNam (Name.Ref l))) h)).
    intro.
    generalize (HOp.le_h_trans _ _ _ H3 H1);clear H3;intro.
    generalize (HOp.h_le_h_irr_h (H.add (MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k)), MessageSpi.MEnc_s N' (MessageSpi.MNam (Name.Ref l))) h')).
    intro.
    generalize (HOp.le_h_trans _ _ _ H4 H2);clear H4;intro.
    assert (H.Subset h' (H.add (MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k)),MessageSpi.MEnc_s N' (MessageSpi.MNam (Name.Ref l))) h'));auto.
    red;intro.
    intro.
    apply H.add_2.
    auto.
    generalize (HOp.Subset_impl_inclusion _ _ H5);clear H5;intro.
    generalize (HOp.incl_impl_le_h _ _ H5);clear H5;intro.
    generalize (HOp.le_h_trans _ _ _ H3 H5);clear H3.
    intro.
    generalize (HOp.le_h_trans _ _ _ H3 H4);clear H3;intro.
    clear H2.
    clear H1.
    clear H5.
    assert (HOp.hedge_synthesis_spec h'' (MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k))) (MessageSpi.MEnc_s N' (MessageSpi.MNam (Name.Ref l)))).
    unfold HOp.hedge_synthesis_spec.
    apply H4.
    apply HOp.SynInc.
    unfold HOp.relation_of_hedge.
    apply H.add_1.
    auto.
    assert (HOp.hedge_synthesis_spec h'' (MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k))) (MessageSpi.MEnc_s N (MessageSpi.MNam (Name.Ref l)))).
    unfold HOp.hedge_synthesis_spec.
    apply H3.
    apply HOp.SynInc.
    unfold HOp.relation_of_hedge.
    apply H.add_1.
    auto.
    generalize (H0 _ _ H1 _ _ H2).
    intros.
    assert ((MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k)))=(MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k)))).
    auto.
    generalize (H5 H6);clear H5;clear H6.
    intro.
    injection H5;auto.
  Qed.

  (** Example 3 *)
  Definition R1s(c : Name.t)(M : MessageSpi.t) := 
    AgentSpi.P.New (AgentSpi.P.Output (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 c))  (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref (0))))) (AgentSpi.P.Input (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 c)) (AgentSpi.P.IfThen (FormulaeSpi.FEqual (ExpressionSpi.ENam (Name.Ref (0))) (ExpressionSpi.ENam (Name.Ref (1)))) (AgentSpi.P.Output (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 2 c)) (ExpressionSpi.ENam (Name.Ref (1))) (AgentSpi.P.Zero))))).

  Definition R2s(c : Name.t)(M : MessageSpi.t) := 
    AgentSpi.P.New (AgentSpi.P.Output (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 c))  (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref (0))))) (AgentSpi.P.Input (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 c)) (AgentSpi.P.Zero))).

  Definition R1s'(c : Name.t) := 
    (AgentSpi.P.Input (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 c)) (AgentSpi.P.IfThen (FormulaeSpi.FEqual (ExpressionSpi.ENam (Name.Ref (0))) (ExpressionSpi.ENam (Name.Ref (1)))) (AgentSpi.P.Output (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 2 c)) (ExpressionSpi.ENam (Name.Ref (1))) (AgentSpi.P.Zero)))).

  Definition R1s'inst (c:Name.t)(k:Name.t) :=
    AgentSpi.P.Input (ExpressionSpi.ENam c) (AgentSpi.P.IfThen (FormulaeSpi.FEqual (ExpressionSpi.ENam (Name.Ref (0))) (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 k))) (AgentSpi.P.Output (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 c)) (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 k)) (AgentSpi.P.Zero))).

  Definition R1s''(c:Name.t)(k:Name.t) := (AgentSpi.P.IfThen (FormulaeSpi.FEqual (ExpressionSpi.ENam (Name.Ref (0))) (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 k))) (AgentSpi.P.Output (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 c)) (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 k)) (AgentSpi.P.Zero))).

  Definition R1s''inst(c:Name.t)(k:Name.t)(M:MessageSpi.t) := (AgentSpi.P.IfThen (FormulaeSpi.FEqual (EvalSpi.expr_of_mesg M) (ExpressionSpi.ENam k)) (AgentSpi.P.Output (ExpressionSpi.ENam c) (ExpressionSpi.ENam k) (AgentSpi.P.Zero))).

  Definition R2s'(c:Name.t) := (AgentSpi.P.Input (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 c)) (AgentSpi.P.Zero)).

  Definition R2s'inst(c : Name.t) := 
    AgentSpi.P.Input (ExpressionSpi.ENam c) (AgentSpi.P.Zero).

  Definition R2s'' := AgentSpi.P.Zero.

  Definition R2s''inst := AgentSpi.P.Zero.

  Lemma R1s_no_silent_trans : forall (c:Name.t)(M:MessageSpi.t)(p':AgentSpi.P.t),~(RedSpiCong.R.reduction (R1s c M) ActionSpi.Tau (AgentSpi.Pr p')).
    intros.
    red;intro.
    unfold R1s in H.
    inversion H.
    inversion H2.
    subst a.
    unfold ActionSpi.A.new in H1.
    rewrite ActionSpi.A.new_conc_def in H1.
    generalize H1.
    case deBruijn.mem_fn;intro.
    discriminate H9.
    discriminate H9.
  Qed.

  Lemma R1s_no_input_trans : forall (c:Name.t)(M:MessageSpi.t)(a:Name.t)(p':AgentSpi.P.t),~(RedSpiCong.R.reduction (R1s c M) (ActionSpi.Inp a) (AgentSpi.F (AgentSpi.Abs p'))).
    intros.
    red;intro.
    unfold R1s in H.
    inversion H.
    inversion H2.
    subst a0.
    unfold ActionSpi.A.new in H1.
    rewrite ActionSpi.A.new_conc_def in H1.
    generalize H1.
    case deBruijn.mem_fn;intro.
    discriminate H9.
    discriminate H9.
  Qed.

  Lemma R1s_output_trans_unique : forall (c:Name.t)(M:MessageSpi.t)(a:Name.t)(n:nat)(m:MessageSpi.t)(p':AgentSpi.P.t),(RedSpiCong.R.reduction (R1s c M) (ActionSpi.Outp a) (AgentSpi.C (AgentSpi.Conc n (EvalSpi.expr_of_mesg m) p')))->a=c /\ n=1 /\ p'=(R1s' c) /\ m=(MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0))).
    intros.
    unfold R1s in H.
    inversion H.
    inversion H2.
    unfold ActionSpi.mk_output in H10.
    simpl in H10.
    rewrite EvalSpi.evalc_msg in H10.
    injection H10;clear H10;intros.
    subst m0.
    subst p0.
    subst mu0.
    subst mu.
    unfold ActionSpi.mk_res in H4.
    simpl in H4.
    assert (0 < 1);try omega.
    generalize (deBruijn.lift_not_fn Name.deBruijn_fn_base 0 1 0 H6 c);intro.
    simpl in H8.
    rewrite H8 in H4.
    injection H4;clear H4;intros.
    clear H6.
    generalize (deBruijn.low1_lift_lift Name.deBruijn_base 0 0 0 c).
    intros.
    simpl in H6.
    rewrite H6 in H4.
    generalize (deBruijn.lift_0 Name.deBruijn_base 0 c).
    intros.
    simpl in H10.
    rewrite H10 in H4.
    subst a.
    clear H8;clear H6;clear H10.
    subst mu'.
    subst a0.
    clear H2.
    subst p.
    subst e.
    subst f.
    unfold ActionSpi.A.new in H1.
    rewrite ActionSpi.A.new_conc_def in H1.
    rewrite ExpressionSpi.mem_fn_EEnc_s in H1.
    assert (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 0 (ExpressionSpi.ENam (Name.Ref 0)) = true).
    reflexivity.
    rewrite H0 in H1.
    rewrite orb_b_true in H1.
    injection H1;clear H1;intros.
    subst n.
    subst p'.
    clear H.
    repeat (split;trivial).
    clear H0.
    assert (ExpressionSpi.EEnc_s (EvalSpi.expr_of_mesg (MessageSpi.t_map (fun d : nat => Name.lift d 1) 0 M)) (ExpressionSpi.ENam (Name.Ref 0)) = (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0)))));try reflexivity.
    rewrite H in H2;clear H.
    symmetry.
    apply EvalSpi.expr_of_mesg_injective.
    auto.
  Qed.

  Lemma R1s_output_trans_exists : forall (c:Name.t)(M:MessageSpi.t),(RedSpiCong.R.reduction (R1s c M) (ActionSpi.Outp c) (AgentSpi.C (AgentSpi.Conc 1 (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0)))) (R1s' c)))).
    intros.
    replace (AgentSpi.C (AgentSpi.Conc 1 (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0)))) (R1s' c))) with (AgentSpi.new (AgentSpi.C (AgentSpi.Conc 0 (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0))))  (R1s' c)))).
    unfold R1s.
    eapply RedSpiCong.R.RedRes.
    eapply RedSpiCong.R.RedOutput.
    unfold ActionSpi.mk_output.
    lazy beta delta [EvalSpi.evalc] iota;fold EvalSpi.evalc.
    rewrite EvalSpi.evalc_msg.
    reflexivity.
    unfold ActionSpi.mk_res.
    assert (0<1);try omega.
    generalize (deBruijn.lift_not_fn Name.deBruijn_fn_base 0 1 0 H c);intro.
    simpl.
    simpl in H0.
    rewrite H0.
    generalize (deBruijn.low1_lift_lift Name.deBruijn_base 0 0 0 c).
    intros.
    simpl in H1.
    rewrite H1.
    rewrite (deBruijn.lift_0 Name.deBruijn_base).
    auto.
    unfold AgentSpi.new.
    rewrite AgentSpi.new_conc_def.
    lazy beta delta [EvalSpi.expr_of_mesg] iota;fold EvalSpi.expr_of_mesg.
    rewrite ExpressionSpi.mem_fn_EEnc_s.
    assert (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 0 (ExpressionSpi.ENam (Name.Ref 0)) = true).
    reflexivity.
    rewrite H.
    rewrite orb_b_true.
    auto.
  Qed.

  Lemma R2s_no_silent_trans : forall (c:Name.t)(M:MessageSpi.t)(p':AgentSpi.P.t),~(RedSpiCong.R.reduction (R2s c M) ActionSpi.Tau (AgentSpi.Pr p')).
    intros.
    red;intro.
    unfold R2s in H.
    inversion H.
    inversion H2.
    subst a.
    unfold ActionSpi.A.new in H1.
    rewrite ActionSpi.A.new_conc_def in H1.
    generalize H1.
    case deBruijn.mem_fn;intro.
    discriminate H9.
    discriminate H9.
  Qed.

  Lemma R2s_no_input_trans : forall (c:Name.t)(M:MessageSpi.t)(a:Name.t)(p':AgentSpi.P.t),~(RedSpiCong.R.reduction (R2s c M) (ActionSpi.Inp a) (AgentSpi.F (AgentSpi.Abs p'))).
    intros.
    red;intro.
    unfold R2s in H.
    inversion H.
    inversion H2.
    subst a0.
    unfold ActionSpi.A.new in H1.
    rewrite ActionSpi.A.new_conc_def in H1.
    generalize H1.
    case deBruijn.mem_fn;intro.
    discriminate H9.
    discriminate H9.
  Qed.

  Lemma R2s_output_trans_unique : forall (c:Name.t)(M:MessageSpi.t)(a:Name.t)(n:nat)(m:MessageSpi.t)(p':AgentSpi.P.t),(RedSpiCong.R.reduction (R2s c M) (ActionSpi.Outp a) (AgentSpi.C (AgentSpi.Conc n (EvalSpi.expr_of_mesg m) p')))->a=c /\ n=1 /\ p'=(R2s' c) /\ m=(MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0))).
    intros.
    unfold R2s in H.
    inversion H.
    inversion H2.
    unfold ActionSpi.mk_output in H10.
    simpl in H10.
    rewrite EvalSpi.evalc_msg in H10.
    injection H10;clear H10;intros.
    subst m0.
    subst p0.
    subst mu0.
    subst mu.
    unfold ActionSpi.mk_res in H4.
    simpl in H4.
    assert (0 < 1);try omega.
    generalize (deBruijn.lift_not_fn Name.deBruijn_fn_base 0 1 0 H6 c);intro.
    simpl in H8.
    rewrite H8 in H4.
    injection H4;clear H4;intros.
    clear H6.
    generalize (deBruijn.low1_lift_lift Name.deBruijn_base 0 0 0 c).
    intros.
    simpl in H6.
    rewrite H6 in H4.
    generalize (deBruijn.lift_0 Name.deBruijn_base 0 c).
    intros.
    simpl in H10.
    rewrite H10 in H4.
    subst a.
    clear H8;clear H6;clear H10.
    subst mu'.
    subst a0.
    clear H2.
    subst p.
    subst e.
    subst f.
    unfold ActionSpi.A.new in H1.
    rewrite ActionSpi.A.new_conc_def in H1.
    rewrite ExpressionSpi.mem_fn_EEnc_s in H1.
    assert (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 0 (ExpressionSpi.ENam (Name.Ref 0)) = true).
    reflexivity.
    rewrite H0 in H1.
    rewrite orb_b_true in H1.
    injection H1;clear H1;intros.
    subst n.
    subst p'.
    clear H.
    repeat (split;trivial).
    clear H0.
    assert (ExpressionSpi.EEnc_s (EvalSpi.expr_of_mesg (MessageSpi.t_map (fun d : nat => Name.lift d 1) 0 M)) (ExpressionSpi.ENam (Name.Ref 0)) = (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0)))));try reflexivity.
    rewrite H in H2;clear H.
    symmetry.
    apply EvalSpi.expr_of_mesg_injective.
    auto.
  Qed.

  Lemma R2s_output_trans_exists : forall (c:Name.t)(M:MessageSpi.t),(RedSpiCong.R.reduction (R2s c M) (ActionSpi.Outp c) (AgentSpi.C (AgentSpi.Conc 1 (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0)))) (R2s' c)))).
    intros.
    replace (AgentSpi.C (AgentSpi.Conc 1 (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0)))) (R2s' c))) with (AgentSpi.new (AgentSpi.C (AgentSpi.Conc 0 (EvalSpi.expr_of_mesg (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0)))) (R2s' c)))).
    unfold R2s.
    eapply RedSpiCong.R.RedRes.
    eapply RedSpiCong.R.RedOutput.
    unfold ActionSpi.mk_output.
    lazy beta delta [EvalSpi.evalc] iota;fold EvalSpi.evalc.
    rewrite EvalSpi.evalc_msg.
    reflexivity.
    unfold ActionSpi.mk_res.
    assert (0<1);try omega.
    generalize (deBruijn.lift_not_fn Name.deBruijn_fn_base 0 1 0 H c);intro.
    simpl.
    simpl in H0.
    rewrite H0.
    generalize (deBruijn.low1_lift_lift Name.deBruijn_base 0 0 0 c).
    intros.
    simpl in H1.
    rewrite H1.
    rewrite (deBruijn.lift_0 Name.deBruijn_base).
    auto.
    unfold AgentSpi.new.
    rewrite AgentSpi.new_conc_def.
    lazy beta delta [EvalSpi.expr_of_mesg] iota;fold EvalSpi.expr_of_mesg.
    rewrite ExpressionSpi.mem_fn_EEnc_s.
    assert (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 0 (ExpressionSpi.ENam (Name.Ref 0)) = true).
    reflexivity.
    rewrite H.
    rewrite orb_b_true.
    auto.
  Qed.

  Lemma instantiate_R1s' : forall (k:Name.t)(c:Name.t),(instantiate_process (cons k nil) (R1s' c))=(R1s'inst c k).
    unfold R1s', R1s'inst.
    intros.
    unfold instantiate_process.
    rewrite AgentSpi.P.low_subst_Input.
    rewrite <- low_subst_AgentSpi.
    generalize (deBruijn.low_subst_lift_nil ExpressionSpi.deBruijn_subst 0 0 0 (map ExpressionSpi.j (k :: nil)) (ExpressionSpi.ENam c)).
    fold ExpressionSpi.deBruijn_base.
    lazy beta delta [length map] iota.     
    repeat (rewrite plus_0_l).     
    intro.
    rewrite AgentSpi.P.low_subst_IfThen.
    replace (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 c)) with (deBruijn.lift ExpressionSpi.deBruijn_base 0 1 (ExpressionSpi.ENam c));try reflexivity.
    rewrite H.
    rewrite deBruijn.lift_0.
    rewrite AgentSpi.P.low_subst_Output.
    rewrite AgentSpi.P.low_subst_Zero.
    rewrite <- low_subst_AgentSpi.
    rewrite FormulaeSpi.low_subst_FEqual.
    assert (deBruijn.low_subst ExpressionSpi.deBruijn_subst 1 0 (ExpressionSpi.j k :: nil) (ExpressionSpi.ENam (Name.Ref 0)) = ExpressionSpi.ENam (Name.Ref 0)).
    reflexivity.
    rewrite H0.
    assert (deBruijn.low_subst ExpressionSpi.deBruijn_subst 1 0 (ExpressionSpi.j k :: nil) (ExpressionSpi.ENam (Name.Ref 1)) = ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 k)).
    reflexivity.
    rewrite H1.
    generalize (deBruijn.lift_compose Name.deBruijn_base 0 1 1 0).
    lazy beta delta [plus] iota.  
    intro.
    rewrite <- H2.
    generalize (deBruijn.low_subst_lift_nil ExpressionSpi.deBruijn_subst 1 0 0 (map ExpressionSpi.j (k :: nil)) (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 c))).
    fold ExpressionSpi.deBruijn_base.
    lazy beta delta [length map] iota.     
    repeat (rewrite plus_0_l).
    replace (deBruijn.lift ExpressionSpi.deBruijn_base 1 1 (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 c))) with (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 1 1 (deBruijn.lift Name.deBruijn_base 0 1 c)));try reflexivity.
    intro.
    rewrite H3.
    rewrite deBruijn.lift_0.
    rewrite <- low_subst_AgentSpi.
    rewrite H1.
    reflexivity.
  Qed.

  Lemma instantiate_R2s' : forall (k:Name.t)(c:Name.t),(instantiate_process (cons k nil) (R2s' c))=(R2s'inst c).
    unfold R2s', R2s'inst.
    intros.
    unfold instantiate_process.
    rewrite AgentSpi.P.low_subst_Input.
    rewrite <- low_subst_AgentSpi.
    generalize (deBruijn.low_subst_lift_nil ExpressionSpi.deBruijn_subst 0 0 0 (map ExpressionSpi.j (k :: nil)) (ExpressionSpi.ENam c)).
    fold ExpressionSpi.deBruijn_base.
    lazy beta delta [length map] iota.     
    repeat (rewrite plus_0_l).     
    intro.
    replace (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 c)) with (deBruijn.lift ExpressionSpi.deBruijn_base 0 1 (ExpressionSpi.ENam c));try reflexivity.
    rewrite H.
    rewrite AgentSpi.P.low_subst_Zero.
    rewrite deBruijn.lift_0.
    auto.
  Qed.

  Lemma R1s'inst_no_silent_trans : forall (c:Name.t)(k:Name.t)(p':AgentSpi.P.t),~(RedSpiCong.R.reduction (R1s'inst c k) ActionSpi.Tau (AgentSpi.Pr p')).
    intros.
    red;intro.
    unfold R1s'inst in H.
    inversion H.
  Qed.

  Lemma R1s'inst_no_output_trans : forall (c:Name.t)(k:Name.t)(a:Name.t)(n:nat)(m:MessageSpi.t)(p':AgentSpi.P.t),(~RedSpiCong.R.reduction (R1s'inst c k) (ActionSpi.Outp a) (AgentSpi.C (AgentSpi.Conc n (EvalSpi.expr_of_mesg m) p'))).
    intros.
    red;intro.
    unfold R1s'inst in H.
    inversion H.
  Qed.

  Lemma R1s'inst_input_trans_unique : forall (c:Name.t)(k:Name.t)(a:Name.t)(p':AgentSpi.P.t),(RedSpiCong.R.reduction (R1s'inst c k) (ActionSpi.Inp a) (AgentSpi.F (AgentSpi.Abs p')))->a=c /\ p'=(R1s'' c k).
    intros.
    unfold R1s'inst in H.
    inversion H.
    injection H3.
    intros.
    split.
    auto.
    subst p.
    auto.
  Qed.
    
  Lemma R1s'inst_input_trans_exists : forall (c:Name.t)(k:Name.t),(RedSpiCong.R.reduction (R1s'inst c k) (ActionSpi.Inp c) (AgentSpi.F (AgentSpi.Abs (R1s'' c k)))).
    intros.
    unfold R1s'inst, R1s''.
    eapply RedSpiCong.R.RedInput.
    reflexivity.
  Qed.

  Lemma R2s'inst_no_silent_trans : forall (c:Name.t)(p':AgentSpi.P.t),~(RedSpiCong.R.reduction (R2s'inst c) ActionSpi.Tau (AgentSpi.Pr p')).
    intros.
    red;intro.
    unfold R2s'inst in H.
    inversion H.
  Qed.

  Lemma R2s'inst_no_output_trans : forall (c:Name.t)(a:Name.t)(n:nat)(m:MessageSpi.t)(p':AgentSpi.P.t),(~RedSpiCong.R.reduction (R2s'inst c) (ActionSpi.Outp a) (AgentSpi.C (AgentSpi.Conc n (EvalSpi.expr_of_mesg m) p'))).
    intros.
    red;intro.
    unfold R2s'inst in H.
    inversion H.
  Qed.

  Lemma R2s'inst_input_trans_unique : forall (c:Name.t)(a:Name.t)(p':AgentSpi.P.t),(RedSpiCong.R.reduction (R2s'inst c) (ActionSpi.Inp a) (AgentSpi.F (AgentSpi.Abs p')))->a=c /\ p'=R2s''.
    intros.
    unfold R2s'inst in H.
    inversion H.
    injection H3.
    intros.
    split.
    auto.
    subst p.
    auto.
  Qed.
    
  Lemma R2s'inst_input_trans_exists : forall (c:Name.t),(RedSpiCong.R.reduction (R2s'inst c) (ActionSpi.Inp c) (AgentSpi.F (AgentSpi.Abs (R2s'')))).
    intros.
    unfold R2s'inst, R2s''.
    eapply RedSpiCong.R.RedInput.
    reflexivity.
  Qed.

  Lemma instantiate_R1s'' : forall (k:Name.t)(c:Name.t)(M:MessageSpi.t),(subst_input M (R1s'' c k))=(R1s''inst c k M).
    unfold R1s'', R1s''inst.
    intros.
    unfold subst_input.
    rewrite AgentSpi_low_subst_IfThen.
    rewrite AgentSpi_low_subst_Output.
    rewrite FormulaeSpi.low_subst_FEqual.
    generalize (deBruijn.low_subst_lift_nil ExpressionSpi.deBruijn_subst 0 0 0 (EvalSpi.expr_of_mesg M :: nil)).
    fold ExpressionSpi.deBruijn_base.
    lazy beta delta [length map] iota.     
    repeat (rewrite plus_0_l).     
    intro.
    replace (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 c)) with (deBruijn.lift ExpressionSpi.deBruijn_base 0 1 (ExpressionSpi.ENam c));try reflexivity.
    replace (ExpressionSpi.ENam (deBruijn.lift Name.deBruijn_base 0 1 k)) with (deBruijn.lift ExpressionSpi.deBruijn_base 0 1 (ExpressionSpi.ENam k));try reflexivity.
    rewrite H.
    rewrite deBruijn.lift_0.
    rewrite AgentSpi.P.low_subst_Zero.
    rewrite H.
    rewrite deBruijn.lift_0.
    replace (@deBruijn.low_subst ExpressionSpi.t ExpressionSpi.deBruijn_base ExpressionSpi.t ExpressionSpi.deBruijn_base ExpressionSpi.deBruijn_subst O O (@cons ExpressionSpi.t (EvalSpi.expr_of_mesg M) (@nil ExpressionSpi.t)) (ExpressionSpi.ENam (Name.Ref O))) with (deBruijn.lift ExpressionSpi.deBruijn_base 0 0 (EvalSpi.expr_of_mesg M));try reflexivity.
    rewrite deBruijn.lift_0.
    auto.
  Qed.
    
  Lemma instantiate_R2s'' : forall (M:MessageSpi.t),(subst_input M R2s'')=(R2s''inst).
    unfold R2s'', R2s''inst.
    intros.
    reflexivity.
  Qed.

  Lemma R1s''inst_no_silent_trans : forall (k:Name.t)(c:Name.t)(M:MessageSpi.t)(p':AgentSpi.P.t),~(RedSpiCong.R.reduction (R1s''inst c k M) ActionSpi.Tau (AgentSpi.Pr p')).
    intros.
    red;intro.
    unfold R1s''inst in H.
    inversion H.
    inversion H2.
  Qed.

  Lemma R1s''inst_no_input_trans : forall (k:Name.t)(c:Name.t)(M:MessageSpi.t)(a:Name.t)(p':AgentSpi.P.t),~(RedSpiCong.R.reduction (R1s''inst c k M) (ActionSpi.Inp a) (AgentSpi.F (AgentSpi.Abs p'))).
    intros.
    red;intro.
    unfold R1s''inst in H.
    inversion H.
    inversion H2.
  Qed.

  Lemma R1s''inst_no_output_trans : forall (k:Name.t)(c:Name.t)(M:MessageSpi.t)(a:Name.t)(n:nat)(m:MessageSpi.t)(p':AgentSpi.P.t),M<>MessageSpi.MNam k->(~RedSpiCong.R.reduction (R1s''inst c k M) (ActionSpi.Outp a) (AgentSpi.C (AgentSpi.Conc n (EvalSpi.expr_of_mesg m) p'))).
    intros.
    red;intro.
    unfold R1s''inst in H0.
    inversion H0.
    inversion H3.
    subst n.
    subst mu0.
    subst p0.
    subst m0.
    subst p'.
    injection H11.
    intros.
    subst mu.
    unfold ActionSpi.mk_if in H6.
    simpl in H6.
    rewrite EvalSpi.evalc_msg in H6.
    destruct (MessageSpi.Ord.compare M (MessageSpi.MNam k));try (discriminate H6).
    elim H.
    rewrite e0.
    auto.
  Qed.

  Lemma R2s''inst_no_silent_trans : forall (p':AgentSpi.P.t),~(RedSpiCong.R.reduction (R2s''inst) ActionSpi.Tau (AgentSpi.Pr p')).
    intros.
    red;intro.
    unfold R2s''inst in H.
    inversion H.
  Qed.

  Lemma R2s''inst_no_input_trans : forall (a:Name.t)(p':AgentSpi.P.t),~(RedSpiCong.R.reduction (R2s''inst) (ActionSpi.Inp a) (AgentSpi.F (AgentSpi.Abs p'))).
    intros.
    red;intro.
    unfold R2s''inst in H.
    inversion H.
  Qed.

  Lemma R2s''inst_no_output_trans : forall (a:Name.t)(n:nat)(m:MessageSpi.t)(p':AgentSpi.P.t),(~RedSpiCong.R.reduction (R2s''inst) (ActionSpi.Outp a) (AgentSpi.C (AgentSpi.Conc n (EvalSpi.expr_of_mesg m) p'))).
    intros.
    red;intro.
    unfold R2s''inst in H.
    inversion H.
  Qed.

  Definition h3 (c:Name.t)(M:MessageSpi.t) := (HOp.hedge_irreducible (H.add (MessageSpi.MNam c,MessageSpi.MNam c)(H.singleton (M,M)))).
    
  Definition h4 (c:Name.t)(M:MessageSpi.t)(k:nat) := (H.add (MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k)),MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k))) (h3 c M)).

  Lemma h3_Equal : forall (c:Name.t)(M:MessageSpi.t),(H.Equal (h0 c M M) (h3 c M)).
    intros.
    unfold h3.
    unfold h0.
    apply HOp.hedge_irreducible_morphism.
    split;intros.
    elim (HOp.HedgeProp.Add_add (H.add (M,M) (H.singleton (M,M))) (MessageSpi.MNam c,MessageSpi.MNam c) a);intros.
    case (H0 H);clear H0;clear H1;intro.
    apply H.add_1.
    auto.
    apply H.add_2.
    elim (HOp.HedgeProp.Add_add (H.singleton (M,M)) (M,M) a);intros.
    case (H1 H0);clear H1;clear H2;intro.
    apply H.singleton_2.
    auto.
    auto.
    elim (HOp.HedgeProp.Add_add (H.singleton (M,M)) (MessageSpi.MNam c,MessageSpi.MNam c) a);intros.
    case (H0 H);clear H0;clear H1;intro.
    apply H.add_1.
    auto.
    apply H.add_2.
    apply H.add_2.
    auto.
  Qed.

  Lemma h4_Equal : forall (c:Name.t)(M:MessageSpi.t)(k:nat),(H.Equal (h1 c M M k k) (h4 c M k)).
    unfold h4.
    unfold h1.
    intros.
    apply HOp.HedgeOtherProp.compat_op_add.
    auto.
    apply h3_Equal.
  Qed.

  Lemma h3_is_left_consistent : forall (c:Name.t)(M:MessageSpi.t),(HOp.is_left_consistent (h3 c M)).
    intros.
    unfold HOp.is_left_consistent.
    eapply HOp.same_impl_is_left_consistent_relation_impl.
    apply HOp.Equal_impl_same.
    apply h3_Equal.
    apply h0_is_left_consistent.
  Qed.

  Lemma h4_is_left_consistent : forall (c:Name.t)(M:MessageSpi.t)(k:nat),(HOp.is_fresh_left_spec k (h3 c M))->(HOp.is_left_consistent (h4 c M k)).
    intros.
    unfold HOp.is_left_consistent.
    eapply HOp.same_impl_is_left_consistent_relation_impl.
    apply HOp.Equal_impl_same.
    apply h4_Equal.
    apply h1_is_left_consistent.
    unfold HOp.is_fresh_left_spec.
    eapply HOp.inclusion_is_fresh_left.
    apply HOp.Subset_impl_inclusion.
    apply HOp.HedgeProp.subset_equal.
    apply h3_Equal.
    apply H.
  Qed.

  Lemma c_in_h3 : forall (c:Name.t)(M:MessageSpi.t),(H.In (MessageSpi.MNam c,MessageSpi.MNam c) (h3 c M)).
    intros.
    elim (h3_Equal c M (MessageSpi.MNam c,MessageSpi.MNam c)).
    intros.
    apply H.
    apply c_in_h0.
  Qed.
    
  Lemma c_in_h3_unique : forall (c:Name.t)(M:MessageSpi.t),forall (d:MessageSpi.message),(H.In (MessageSpi.MNam c,d) (h3 c M))->d=(MessageSpi.MNam c).
    intros.
    elim (h3_is_left_consistent c M);intros;and_elim.
    rewrite (H1 _ _ (c_in_h3 c M) _ _ H);auto.
  Qed.

  Lemma c_left_name_h3 : forall (c:Name.t)(M:MessageSpi.t),forall (x:nat),(deBruijn.mem_fn Name.deBruijn_fn_base x 0 c = true)->(HOp.is_left_name_spec x (h3 c M)).
    intros.
    exists (MessageSpi.MNam c).
    exists (MessageSpi.MNam c).
    split;auto.
    unfold HOp.relation_of_hedge.
    apply c_in_h3.
  Qed.

  Lemma M_left_name_h3 : forall (c:Name.t)(M:MessageSpi.t),forall (x:nat),(deBruijn.mem_fn MessageSpi.deBruijn_fn_base x 0 M = true)->(HOp.is_left_name_spec x (h3 c M)).
    intros.
    unfold HOp.is_left_name_spec.
    unfold h3.
    eapply HOp.mem_fn_le_h.
    apply HOp.h_le_h_irr_h. 
    exists M.
    exists M.
    split.
    unfold HOp.relation_of_hedge.
    apply H.add_2.
    apply H.singleton_2.
    auto.
    auto.
  Qed.

  Lemma fn_R1s_h3 : forall (c:Name.t)(M:MessageSpi.message),forall (x:nat),(fn x (R1s c M))->(HOp.is_left_name_spec x (h3 c M)).
    intros.
    unfold R1s in H.
    unfold fn in H.
    rewrite AgentSpi.P.fn_New in H.
    rewrite AgentSpi.P.fn_Output in H.
    rewrite AgentSpi.P.fn_Input in H.
    rewrite AgentSpi.P.fn_IfThen in H.
    rewrite AgentSpi.P.fn_Output in H.
    rewrite AgentSpi.P.fn_Zero in H.
    rewrite orb_b_false in H.
    case (orb_true_elim _ _ H);clear H;intro H.
    case (orb_true_elim _ _ H);clear H;intro H.
    simpl in H.
    generalize (deBruijn.fn_plus_lift Name.deBruijn_fn_base 0 1 (Name.lift 0 1 c) x).
    simpl;intros.
    rewrite <- H0 in H.
    generalize (deBruijn.lift_fn Name.deBruijn_fn_base 0 1 c x).
    simpl;intros.
    rewrite H1 in H.
    apply c_left_name_h3.
    auto.
    rewrite EvalSpi.mem_fn_expr_of_msg in H.
    rewrite MessageSpi.mem_fn_MEnc_s in H.
    assert (deBruijn.mem_fn MessageSpi.deBruijn_fn_base x 1 (MessageSpi.MNam (Name.Ref 0)) = false).
    reflexivity.
    rewrite H0 in H.
    rewrite orb_b_false in H.
    generalize (deBruijn.fn_plus_lift MessageSpi.deBruijn_fn_base 0 1 (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) x).
    replace (0+1) with 1;try reflexivity.
    intro.
    rewrite <- H1 in H.
    generalize (deBruijn.lift_fn MessageSpi.deBruijn_fn_base 0 1 M x).
    intros.
    apply M_left_name_h3.
    rewrite <- H2.
    auto.
    case (orb_true_elim _ _ H);clear H;intro H.
    simpl in H.
    generalize (deBruijn.fn_plus_lift Name.deBruijn_fn_base 0 1 (Name.lift 0 1 c) x).
    simpl;intros.
    rewrite <- H0 in H.
    generalize (deBruijn.lift_fn Name.deBruijn_fn_base 0 1 c x).
    simpl;intros.
    rewrite H1 in H.
    apply c_left_name_h3.
    auto.
    case (orb_true_elim _ _ H);clear H;intro H.
    discriminate H.
    case (orb_true_elim _ _ H);clear H;intro H.
    simpl in H.
    generalize (deBruijn.fn_plus_lift Name.deBruijn_fn_base 0 2 (Name.lift 0 2 c) x).
    simpl;intros.
    rewrite <- H0 in H.
    generalize (deBruijn.lift_fn Name.deBruijn_fn_base 0 2 c x).
    simpl;intros.
    rewrite H1 in H.
    apply c_left_name_h3.
    auto.
    discriminate H.
  Qed.

  Lemma fn_R2s_h3 : forall (c:Name.t)(M:MessageSpi.message),forall (x:nat),(fn x (R2s c M))->(HOp.is_left_name_spec x (h3 c M)).
    intros.
    unfold R2s in H.
    unfold fn in H.
    rewrite AgentSpi.P.fn_New in H.
    rewrite AgentSpi.P.fn_Output in H.
    rewrite AgentSpi.P.fn_Input in H.
    rewrite AgentSpi.P.fn_Zero in H.
    rewrite orb_b_false in H.
    case (orb_true_elim _ _ H);clear H;intro H.
    case (orb_true_elim _ _ H);clear H;intro H.
    simpl in H.
    generalize (deBruijn.fn_plus_lift Name.deBruijn_fn_base 0 1 (Name.lift 0 1 c) x).
    simpl;intros.
    rewrite <- H0 in H.
    generalize (deBruijn.lift_fn Name.deBruijn_fn_base 0 1 c x).
    simpl;intros.
    rewrite H1 in H.
    apply c_left_name_h3.
    auto.
    rewrite EvalSpi.mem_fn_expr_of_msg in H.
    rewrite MessageSpi.mem_fn_MEnc_s in H.
    assert (deBruijn.mem_fn MessageSpi.deBruijn_fn_base x 1 (MessageSpi.MNam (Name.Ref 0)) = false).
    reflexivity.
    rewrite H0 in H.
    rewrite orb_b_false in H.
    generalize (deBruijn.fn_plus_lift MessageSpi.deBruijn_fn_base 0 1 (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) x).
    replace (0+1) with 1;try reflexivity.
    intro.
    rewrite <- H1 in H.
    generalize (deBruijn.lift_fn MessageSpi.deBruijn_fn_base 0 1 M x).
    intros.
    apply M_left_name_h3.
    rewrite <- H2.
    auto.
    simpl in H.
    generalize (deBruijn.fn_plus_lift Name.deBruijn_fn_base 0 1 (Name.lift 0 1 c) x).
    simpl;intros.
    rewrite <- H0 in H.
    generalize (deBruijn.lift_fn Name.deBruijn_fn_base 0 1 c x).
    simpl;intros.
    rewrite H1 in H.
    apply c_left_name_h3.
    auto.
  Qed.

  Lemma c_in_h4 : forall (c:Name.t)(M:MessageSpi.t)(k:nat),(H.In (MessageSpi.MNam c,MessageSpi.MNam c) (h4 c M k)).
    intros.
    unfold h4.
    apply H.add_2.
    apply c_in_h3.
  Qed.
    
  Lemma c_in_h4_unique : forall (c:Name.t)(M:MessageSpi.t)(k:nat),forall (d:MessageSpi.message),(HOp.is_fresh_left_spec k (h3 c M))->(H.In (MessageSpi.MNam c,d) (h4 c M k))->d=(MessageSpi.MNam c).
    intros.
    elim (h4_is_left_consistent c M k H);intros;and_elim.
    rewrite (H2 _ _ (c_in_h4 c M k) _ _ H0);auto.
  Qed.

  Lemma c_left_name_h4 : forall (c:Name.t)(M:MessageSpi.t)(k:nat),forall (x:nat),(deBruijn.mem_fn Name.deBruijn_fn_base x 0 c = true)->(HOp.is_left_name_spec x (h4 c M k)).
    intros.
    exists (MessageSpi.MNam c).
    exists (MessageSpi.MNam c).
    split;auto.
    unfold HOp.relation_of_hedge.
    apply c_in_h4.
  Qed.

  Lemma M_left_name_h4 : forall (c:Name.t)(M:MessageSpi.t)(k:nat),forall (x:nat),(deBruijn.mem_fn MessageSpi.deBruijn_fn_base x 0 M = true)->(HOp.is_left_name_spec x (h4 c M k)).
    intros.
    unfold HOp.is_left_name_spec.
    unfold h4.
    unfold HOp.is_left_name.
    elim (M_left_name_h3 c M x H).
    intro m;intros.
    ex_elim H0 n.
    exists m.
    exists n.
    unfold HOp.relation_of_hedge.
    split.
    apply H.add_2.
    auto.
    auto.
  Qed.

  Lemma k_left_name_h4 : forall (c:Name.t)(M:MessageSpi.t)(k:nat),forall (x:nat),(deBruijn.mem_fn Name.deBruijn_fn_base x 0 (Name.Ref k) = true)->(HOp.is_left_name_spec x (h4 c M k)).
    intros.
    exists (MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k))).
    exists (MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k))).
    split.
    unfold h4.
    unfold HOp.relation_of_hedge.
    apply H.add_1.
    auto.
    rewrite MessageSpi.mem_fn_MEnc_s.
    simpl.
    simpl in H.
    rewrite H.
    rewrite orb_b_true.
    auto.
  Qed.

  Lemma fn_R1s'inst_h4 : forall (c:Name.t)(M:MessageSpi.t)(k:nat),forall (x:nat),(fn x (R1s'inst c (Name.Ref k)))->(HOp.is_left_name_spec x (h4 c M k)).
    intros.
    unfold R1s'inst in H.
    unfold fn in H.
    rewrite AgentSpi.P.fn_Input in H.
    rewrite AgentSpi.P.fn_IfThen in H.
    rewrite AgentSpi.P.fn_Output in H.
    rewrite AgentSpi.P.fn_Zero in H.
    rewrite orb_b_false in H.
    case (orb_true_elim _ _ H);clear H;intro H.
    apply c_left_name_h4.
    auto.
    case (orb_true_elim _ _ H);clear H;intro H.
    rewrite FormulaeSpi.mem_fn_FEqual in H.
    case (orb_true_elim _ _ H);clear H;intro H.
    discriminate H.
    simpl in H.
    generalize (deBruijn.fn_plus_lift Name.deBruijn_fn_base 0 1 (Name.lift 0 1 (Name.Ref k)) x).
    simpl;intros.
    rewrite <- H0 in H.
    generalize (deBruijn.lift_fn Name.deBruijn_fn_base 0 1 (Name.Ref k) x).
    simpl;intros.
    rewrite H1 in H.
    apply k_left_name_h4.
    auto.
    case (orb_true_elim _ _ H);clear H;intro H.
    simpl in H.
    generalize (deBruijn.fn_plus_lift Name.deBruijn_fn_base 0 1 (Name.lift 0 1 c) x).
    simpl;intros.
    rewrite <- H0 in H.
    generalize (deBruijn.lift_fn Name.deBruijn_fn_base 0 1 c x).
    simpl;intros.
    rewrite H1 in H.
    apply c_left_name_h4.
    auto.
    simpl in H.
    generalize (deBruijn.fn_plus_lift Name.deBruijn_fn_base 0 1 (Name.lift 0 1 (Name.Ref k)) x).
    simpl;intros.
    rewrite <- H0 in H.
    generalize (deBruijn.lift_fn Name.deBruijn_fn_base 0 1 (Name.Ref k) x).
    simpl;intros.
    rewrite H1 in H.
    apply k_left_name_h4.
    auto.
  Qed.

  Lemma fn_R2s'inst_h4 : forall (c:Name.t)(M:MessageSpi.t)(k:nat),forall (x:nat),(fn x (R2s'inst c))->(HOp.is_left_name_spec x (h4 c M k)).
    intros.
    unfold R2s'inst in H.
    unfold fn in H.
    rewrite AgentSpi.P.fn_Input in H.
    rewrite AgentSpi.P.fn_Zero in H.
    rewrite orb_b_false in H.
    apply c_left_name_h4.
    auto.
  Qed.

  Lemma fn_R1s''inst_h4_cup_B : forall (c:Name.t)(M:MessageSpi.t)(k:nat),forall (B:H.t)(m n:MessageSpi.t),(possible_inputs (h4 c M k) B m n)->forall (x:nat),(fn x (R1s''inst c (Name.Ref k) m))->(HOp.is_left_name_spec x (H.union (h4 c M k) B)).
    intros.
    elim H;clear H.
    intro Bfresh;intros.
    unfold R1s''inst in H0.
    unfold fn in H0.
    rewrite AgentSpi.P.fn_IfThen in H0.
    rewrite AgentSpi.P.fn_Output in H0.
    rewrite AgentSpi.P.fn_Zero in H0.
    rewrite orb_b_false in H0.
    case (orb_true_elim _ _ H0);clear H0;intro H0.
    rewrite FormulaeSpi.mem_fn_FEqual in H0.
    case (orb_true_elim _ _ H0);clear H0;intro H0.
    unfold HOp.is_left_name_spec.
    apply HOp.mem_fn_left_synthesis.
    elim H;intros.
    exists m.
    exists n.
    split;auto.
    rewrite EvalSpi.mem_fn_expr_of_msg in H0;auto.
    simpl in H0.
    elim (k_left_name_h4 c M k x H0).
    intro x0.
    intro.
    ex_elim H1 y0.
    exists x0.
    exists y0.
    split;auto.
    unfold HOp.relation_of_hedge.
    apply H.union_2.
    auto.
    case (orb_true_elim _ _ H0);clear H0;intro H0.
    simpl in H0.
    elim (c_left_name_h4 c M k x H0).
    intro x0;intro.
    ex_elim H1 y0.
    exists x0.
    exists y0.
    split;auto.
    unfold HOp.relation_of_hedge.
    apply H.union_2.
    auto.
    simpl in H0.
    elim (k_left_name_h4 c M k x H0).
    intro x0;intro.
    ex_elim H1 y0.
    exists x0.
    exists y0.
    split;auto.
    unfold HOp.relation_of_hedge.
    apply H.union_2.
    auto.
  Qed.

  Lemma fn_R2s''inst_h4_cup_B : forall (c:Name.t)(M:MessageSpi.t)(k:nat),forall (B:H.t)(m n:MessageSpi.t),(possible_inputs (h4 c M k) B m n)->forall (x:nat),(fn x (R2s''inst))->(HOp.is_left_name_spec x (H.union (h4 c M k) B)).
    intros.
    unfold R2s''inst in H0.
    unfold fn in H0.
    discriminate H0.
  Qed.

  Inductive possible_inputs1 (c:Name.t)(M:MessageSpi.t) : hedged_relation_type :=
  | C'1 : possible_inputs1 c M ((h3 c M),(R1s c M),(R2s c M))
  | C'1sym : possible_inputs1 c M ((h3 c M),(R2s c M),(R1s c M))
  | C'2 : forall (k:nat),HOp.is_fresh_left_spec k (h3 c M)->(possible_inputs1 c M ((h4 c M k,R1s'inst c (Name.Ref k),R2s'inst c)))
  | C'2sym : forall (k:nat),HOp.is_fresh_left_spec k (h3 c M)->(possible_inputs1 c M ((h4 c M k,R2s'inst c,R1s'inst c (Name.Ref k))))
  | C'3 : forall (k:nat),HOp.is_fresh_left_spec k (h3 c M)->forall (B:H.t)(m n:MessageSpi.t),possible_inputs (h4 c M k) B m n->(possible_inputs1 c M ((H.union (h4 c M k) B),R1s''inst c (Name.Ref k) m,R2s''inst))
  | C'3sym : forall (k:nat),HOp.is_fresh_left_spec k (h3 c M)->forall (B:H.t)(m n:MessageSpi.t),possible_inputs (h4 c M k) B m n->(possible_inputs1 c M ((H.union (h4 c M k) B),R2s''inst,R1s''inst c (Name.Ref k) n)).

  Lemma h3_transpose : forall (c:Name.t)(M:MessageSpi.t),(H.Equal (h3 c M) (HOp.hedge_transpose (h3 c M))).
    intros.
    apply HOp.HedgeProp.equal_sym.
    eapply HOp.HedgeProp.equal_trans.
    apply HOp.hedge_transpose_morphism.
    apply HOp.HedgeProp.equal_sym.
    apply h3_Equal.
    eapply HOp.HedgeProp.equal_trans.
    apply h0_transpose.
    apply h3_Equal.
  Qed.

  Lemma h4_transpose : forall (c:Name.t)(M:MessageSpi.t)(k:nat),(H.Equal (h4 c M k) (HOp.hedge_transpose (h4 c M k))).
    intros.
    apply HOp.HedgeProp.equal_sym.
    eapply HOp.HedgeProp.equal_trans.
    apply HOp.hedge_transpose_morphism.
    apply HOp.HedgeProp.equal_sym.
    apply h4_Equal.
    eapply HOp.HedgeProp.equal_trans.
    apply h1_transpose.
    apply h4_Equal.
  Qed.

  Lemma possible_inputs1_symmetric_hedged : forall (c:Name.t)(M:MessageSpi.message),(symmetric_hedged_relation (possible_inputs1 c M)).
    unfold symmetric_hedged_relation.
    intros.
    inversion H.
    subst h;subst p;subst q.
    exists (h3 c M).
    split.
    apply h3_transpose.
    apply C'1sym.
    subst h;subst p;subst q.
    exists (h3 c M).
    split.
    apply h3_transpose.
    apply C'1.
    subst h;subst p;subst q.
    exists (h4 c M k).
    split.
    apply h4_transpose.
    apply C'2sym.
    auto.
    subst h;subst p;subst q.
    exists (h4 c M k).
    split.
    apply h4_transpose.
    apply C'2.
    auto.
    subst h;subst p;subst q.
    exists (H.union (h4 c M k) (HOp.hedge_transpose B)).
    split.
    apply HOp.HedgeProp.equal_sym.
    eapply HOp.HedgeProp.equal_trans.
    apply HOp.hedge_transpose_union.
    apply HOp.HedgeOtherProp.Equal_union_Equal_Equal.
    apply HOp.HedgeProp.equal_sym.
    apply h4_transpose.
    apply HOp.HedgeProp.equal_refl.
    apply C'3sym with n.
    apply H2.
    eapply possible_inputs_Subset.
    apply HOp.HedgeProp.subset_equal.    
    apply HOp.HedgeProp.equal_sym.
    apply h4_transpose.
    eapply HOp.are_fresh_names_Subset.
    apply HOp.HedgeProp.subset_equal.
    apply h4_transpose.
    apply HOp.HedgeProp.subset_refl.
    apply HOp.are_fresh_names_transpose.
    elim H4;intros;auto.
    apply HOp.HedgeProp.equal_refl.
    apply possible_inputs_transpose.    
    apply H4.
    subst h;subst p;subst q.
    exists (H.union (h4 c M k) (HOp.hedge_transpose B)).
    split.
    apply HOp.HedgeProp.equal_sym.
    eapply HOp.HedgeProp.equal_trans.
    apply HOp.hedge_transpose_union.
    apply HOp.HedgeOtherProp.Equal_union_Equal_Equal.
    apply HOp.HedgeProp.equal_sym.
    apply h4_transpose.
    apply HOp.HedgeProp.equal_refl.
    apply C'3 with m.
    apply H2.
    eapply possible_inputs_Subset.
    apply HOp.HedgeProp.subset_equal.    
    apply HOp.HedgeProp.equal_sym.
    apply h4_transpose.
    eapply HOp.are_fresh_names_Subset.
    apply HOp.HedgeProp.subset_equal.
    apply h4_transpose.
    apply HOp.HedgeProp.subset_refl.
    apply HOp.are_fresh_names_transpose.
    elim H4;intros;auto.
    apply HOp.HedgeProp.equal_refl.
    apply possible_inputs_transpose.    
    apply H4.
  Qed.

  Lemma possible_inputs1_left_hedged : forall (c:Name.t)(M:MessageSpi.message),(left_hedged_relation (possible_inputs1 c M)).
    unfold left_hedged_relation.
    intros.
    inversion H;subst h;subst p;subst q.
    apply fn_R1s_h3;auto.
    apply fn_R2s_h3;auto.
    apply fn_R1s'inst_h4;auto.
    apply fn_R2s'inst_h4;auto.
    apply (fn_R1s''inst_h4_cup_B c M k B _ _ H5);auto.
    apply (fn_R2s''inst_h4_cup_B c M k B _ _ H5);auto.
  Qed.

  Lemma possible_inputs1_left_consistent : forall (c:Name.t)(M:MessageSpi.message),(left_consistent_hedged_relation (possible_inputs1 c M)).
    unfold left_consistent_hedged_relation.
    intros.
    inversion H;subst h;subst p;subst q.
    apply h3_is_left_consistent.
    apply h3_is_left_consistent.
    apply h4_is_left_consistent;auto.
    apply h4_is_left_consistent;auto.
    apply (HOp.are_fresh_names_left_consistent _ (h4_is_left_consistent _ _ _ H2) B).
    elim H4;intros;auto.
    apply (HOp.are_fresh_names_left_consistent _ (h4_is_left_consistent _ _ _ H2) B).
    elim H4;intros;auto.
  Qed.

  Lemma possible_inputs1_late_hedged_bisimulation : forall (c:Name.t)(M:MessageSpi.t),(late_hedged_bisimulation_condition (possible_inputs1 c M)).
    unfold late_hedged_bisimulation_condition.
    intros.
    inversion H;subst h;subst p;subst q.
    (* C'1 *)
    repeat split;intros.
    elim (R1s_no_silent_trans _ _ _ H0).
    elim (R1s_no_input_trans _ _ _ _ H0).
    elim (R1s_output_trans_unique _ _ _ _ _ _ H0);intros;and_elim.
    subst a.
    subst c0.
    subst p'.
    exists 1.
    exists (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0))).
    exists (R2s' c).
    generalize (c_in_h3_unique _ _ _ H2).
    intro.
    injection H3;clear H3;intros;subst b.
    split.
    apply R2s_output_trans_exists.
    subst m.
    elim (HOp.n_fresh_left_single _ _ H1).
    intro k;intros.
    and_elim.
    exists tildec.
    split.
    apply HOp.is_reflexive_n_fresh_left.
    eapply HOp.inclusion_is_reflexive.
    apply HOp.Subset_impl_inclusion.
    apply HOp.HedgeProp.subset_equal.
    apply HOp.HedgeProp.equal_sym.
    apply h3_Equal.
    apply h0_reflexive.
    auto.
    exists (h4 c M k).
    split.
    subst tildec.
    repeat (rewrite instantiate_message_Ps).
    apply HOp.cond_irr_impl_irreducible.
    unfold HOp.cond_irr.
    apply HOp.is_left_consistent_impl_cond_irr.
    generalize (h4_is_left_consistent c M k).
    unfold h4.
    intros.
    apply H3.
    auto.
    subst tildec.
    rewrite instantiate_R1s'.
    rewrite instantiate_R2s'.
    apply C'2.
    auto.
    (* C'1sym *)
    repeat split;intros.
    elim (R2s_no_silent_trans _ _ _ H0).
    elim (R2s_no_input_trans _ _ _ _ H0).
    elim (R2s_output_trans_unique _ _ _ _ _ _ H0);intros;and_elim.
    subst a.
    subst c0.
    subst p'.
    exists 1.
    exists (MessageSpi.MEnc_s (deBruijn.lift MessageSpi.deBruijn_base 0 1 M) (MessageSpi.MNam (Name.Ref 0))).
    exists (R1s' c).
    generalize (c_in_h3_unique _ _ _ H2).
    intro.
    injection H3;clear H3;intros;subst b.
    split.
    apply R1s_output_trans_exists.
    subst m.
    elim (HOp.n_fresh_left_single _ _ H1).
    intro k;intros.
    and_elim.
    exists tildec.
    split.
    apply HOp.is_reflexive_n_fresh_left.
    eapply HOp.inclusion_is_reflexive.
    apply HOp.Subset_impl_inclusion.
    apply HOp.HedgeProp.subset_equal.
    apply HOp.HedgeProp.equal_sym.
    apply h3_Equal.
    apply h0_reflexive.
    auto.
    exists (h4 c M k).
    split.
    subst tildec.
    repeat (rewrite instantiate_message_Ps).
    apply HOp.cond_irr_impl_irreducible.
    unfold HOp.cond_irr.
    apply HOp.is_left_consistent_impl_cond_irr.
    generalize (h4_is_left_consistent c M k).
    unfold h4.
    intros.
    apply H3.
    auto.
    subst tildec.
    rewrite instantiate_R1s'.
    rewrite instantiate_R2s'.
    apply C'2sym.
    auto.
    (* C'2 *)
    repeat split;intros.
    elim (R1s'inst_no_silent_trans _ _ _ H0).
    elim (R1s'inst_input_trans_unique _ _ _ _ H0);intros;and_elim.
    subst a;subst p'.
    exists R2s''.
    generalize (c_in_h4_unique _ _ _ _ H1 H2).
    intro.
    injection H3;clear H3;intros;subst b.
    split.
    apply R2s'inst_input_trans_exists.
    intros.
    exists (H.union (h4 c M k) B).
    split.
    apply HOp.HedgeProp.equal_refl.
    rewrite instantiate_R1s''.
    rewrite instantiate_R2s''.
    apply C'3 with n;auto.
    elim (R1s'inst_no_output_trans _ _ _ _ _ _ H0).
    (* C'2sym *)
    repeat split;intros.
    elim (R2s'inst_no_silent_trans _ _ H0).
    elim (R2s'inst_input_trans_unique _ _ _ H0);intros;and_elim.
    subst a;subst p'.
    exists (R1s'' c (Name.Ref k)).
    generalize (c_in_h4_unique _ _ _ _ H1 H2).
    intro.
    injection H3;clear H3;intros;subst b.
    split.
    apply R1s'inst_input_trans_exists.
    intros.
    exists (H.union (h4 c M k) B).
    split.
    apply HOp.HedgeProp.equal_refl.
    rewrite instantiate_R1s''.
    rewrite instantiate_R2s''.
    apply C'3sym with m;auto.
    elim (R2s'inst_no_output_trans _ _ _ _ _ H0).
    (* C'3 *)
    repeat split;intros.
    elim (R1s''inst_no_silent_trans _ _ _ _ H0).
    elim (R1s''inst_no_input_trans _ _ _ _ _ H0).
    assert (m <> MessageSpi.MNam (Name.Ref k)).
    intro.
    elim H4.
    intro Bfresh;intros;and_elim.
    generalize (HOp.hedge_synthesis_spec_impl_hedge_in_synthesis _ _ _ H6).
    rewrite H5.
    rewrite HOp.name_left_synthesis.
    intro.
    generalize (H.mem_2 H8).
    intro.
    case (H.union_1 H9).
    unfold h4.
    intro.
    generalize (HOp.HedgeProp.Add_add (h3 c M)  (MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k)),MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k)))).
    intro.
    elim (H11 (MessageSpi.MNam (Name.Ref k), n)).
    intros.
    generalize (H12 H10);clear H10;clear H11;clear H12;clear H13;intros.
    case H10;clear H10;intro.
    simpl in H10.
    and_elim.
    discriminate H10.
    elim H2.
    exists (MessageSpi.MNam (Name.Ref k)).
    exists n.
    split;auto.
    simpl.
    generalize (Name.fn_0_name k).
    simpl;auto.
    intro.
    elim Bfresh.
    intros.
    and_elim.
    elim (H11 _ _ H10).
    intro x;intros.
    ex_elim H14 l.
    injection H14;clear H14;intros;subst x.
    subst n.
    assert (H.In (MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k)),MessageSpi.MEnc_s M (MessageSpi.MNam (Name.Ref k))) (h4 c M k)).
    unfold h4.
    apply H.add_1.
    auto.
    destruct l as [l].
    generalize (H13 _ _ H10 _ _ H14).
    intros;and_elim.
    rewrite MessageSpi.mem_fn_MEnc_s in H15.
    elim (orb_false_elim _ _ H15).
    intros.
    simpl in H18.
    generalize (Name.fn_0_name k).
    simpl.
    intro.
    rewrite H19 in H18.
    discriminate H18.
    elim (R1s''inst_no_output_trans _ _ _ _ _ _ _ H5 H0).
    (* C'3sym *)
    repeat split;intros.
    elim (R2s''inst_no_silent_trans _ H0).
    elim (R2s''inst_no_input_trans _ _ H0).
    elim (R2s''inst_no_output_trans _ _ _ _ H0).
  Qed.

  Theorem possible_inputs1_is_late_hedged_bisimulation : forall (c:Name.t)(M:MessageSpi.t),late_hedged_bisimulation (possible_inputs1 c M).
    unfold late_hedged_bisimulation.
    unfold symmetric_left_consistent_hedged_relation.
    intros.
    split.
    split.
    apply possible_inputs1_symmetric_hedged.
    split.
    apply possible_inputs1_left_hedged.
    apply possible_inputs1_left_consistent.
    apply possible_inputs1_late_hedged_bisimulation.
  Qed.

  Theorem possible_inputs1_late_hedged_bisimilar : forall (c:Name.t)(M:MessageSpi.t),late_hedged_bisimilar (h3 c M) (R1s c M) (R2s c M).
    unfold late_hedged_bisimilar;intros.
    exists (possible_inputs1 c M).
    split.
    apply C'1.
    apply possible_inputs1_is_late_hedged_bisimulation.
  Qed.
End Example.




