(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import Arith.
Require Import name.
Require Import messageSpi.
Require Import expressionSpi.
Require Import formulaeSpi.
Require Import missing.
Require Import deBruijnType.
Require Import tactics.
Require Import opSpi.
Require Import evalSpi.
Require Import substSpi.

Module FnInMsg.
  Import OpSpi.

  Inductive fn_in_msg (k d:nat) : ExpressionSpi.t -> Prop :=
  | fnInNam : forall (x:Name.t),(deBruijn.mem_fn Name.deBruijn_fn_base k d x)=true->(fn_in_msg k d (ExpressionSpi.ENam x))
  | fnInFst : forall (e1:ExpressionSpi.t),(fn_in_msg k d e1)->forall (e2:ExpressionSpi.t),(fn_in_msg k d (ExpressionSpi.EPair e1 e2))
  | fnInSnd : forall (e2:ExpressionSpi.t),(fn_in_msg k d e2)->forall (e1:ExpressionSpi.t),(fn_in_msg k d (ExpressionSpi.EPair e1 e2))
  | fnInEnc_s_d : forall (e1:ExpressionSpi.t),(fn_in_msg k d e1)->forall (e2:ExpressionSpi.t),(fn_in_msg k d (ExpressionSpi.EEnc_s e1 e2))
  | fnInEnc_s_k : forall (e2:ExpressionSpi.t),(fn_in_msg k d e2)->forall (e1:ExpressionSpi.t),(fn_in_msg k d (ExpressionSpi.EEnc_s e1 e2))
  | fnInEnc_a_d : forall (e1:ExpressionSpi.t),(fn_in_msg k d e1)->forall (e2:ExpressionSpi.t),(fn_in_msg k d (ExpressionSpi.EEnc_a e1 e2))
  | fnInEnc_a_k : forall (e2:ExpressionSpi.t),(fn_in_msg k d e2)->forall (e1:ExpressionSpi.t),(fn_in_msg k d (ExpressionSpi.EEnc_a e1 e2))
  | fnInOp : forall (e:ExpressionSpi.t),(fn_in_msg k d e)->forall (o:op),(fn_in_msg k d (ExpressionSpi.EOp o e)).

  Lemma fn_in_msg_expr_of_msg : forall (d k:nat)(m:MessageSpi.t),(deBruijn.mem_fn MessageSpi.deBruijn_fn_base k d m = true)<->(fn_in_msg k d (EvalSpi.expr_of_mesg m)).
    split.
    induction m;lazy beta delta [EvalSpi.expr_of_mesg] iota;fold EvalSpi.expr_of_mesg.
    rewrite MessageSpi.mem_fn_MNam;intro.
    apply fnInNam;auto.
    rewrite MessageSpi.mem_fn_MPair;intro.
    case (orb_true_elim (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k d m1) (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k d m2) H);intro.
    apply fnInFst;auto.
    apply fnInSnd;auto.
    rewrite MessageSpi.mem_fn_MEnc_s;intro.
    case (orb_true_elim (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k d m1) (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k d m2) H);intro.
    apply fnInEnc_s_d;auto.
    apply fnInEnc_s_k;auto.
    rewrite MessageSpi.mem_fn_MEnc_a;intro.
    case (orb_true_elim (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k d m1) (deBruijn.mem_fn MessageSpi.deBruijn_fn_base k d m2) H);intro.
    apply fnInEnc_a_d;auto.
    apply fnInEnc_a_k;auto.
    rewrite MessageSpi.mem_fn_MOp;intro.
    apply fnInOp;auto.
    assert (exists e:ExpressionSpi.t,EvalSpi.expr_of_mesg m = e).
    exists (EvalSpi.expr_of_mesg m);auto.
    elim H;clear H;intro e;intros.
    rewrite H in H0.
    generalize H;clear H.
    generalize m;clear m.
    induction H0;intros.
    destruct m;try (discriminate H0).
    lazy beta delta [EvalSpi.expr_of_mesg] iota in H0;fold EvalSpi.expr_of_mesg in H0.
    injection H0;intros.
    subst t.
    rewrite MessageSpi.mem_fn_MNam;auto.
    destruct m;try (discriminate H).
    lazy beta delta [EvalSpi.expr_of_mesg] iota in H;fold EvalSpi.expr_of_mesg in H.
    injection H;intros.
    rewrite MessageSpi.mem_fn_MPair.
    rewrite (IHfn_in_msg m1);auto.
    destruct m;try (discriminate H).
    lazy beta delta [EvalSpi.expr_of_mesg] iota in H;fold EvalSpi.expr_of_mesg in H.
    injection H;intros.
    rewrite MessageSpi.mem_fn_MPair.
    rewrite (IHfn_in_msg m2);auto.
    rewrite orb_b_true;auto.
    destruct m;try (discriminate H).
    lazy beta delta [EvalSpi.expr_of_mesg] iota in H;fold EvalSpi.expr_of_mesg in H.
    injection H;intros.
    rewrite MessageSpi.mem_fn_MEnc_s.
    rewrite (IHfn_in_msg m1);auto.
    destruct m;try (discriminate H).
    lazy beta delta [EvalSpi.expr_of_mesg] iota in H;fold EvalSpi.expr_of_mesg in H.
    injection H;intros.
    rewrite MessageSpi.mem_fn_MEnc_s.
    rewrite (IHfn_in_msg m2);auto.
    rewrite orb_b_true;auto.
    destruct m;try (discriminate H).
    lazy beta delta [EvalSpi.expr_of_mesg] iota in H;fold EvalSpi.expr_of_mesg in H.
    injection H;intros.
    rewrite MessageSpi.mem_fn_MEnc_a.
    rewrite (IHfn_in_msg m1);auto.
    destruct m;try (discriminate H).
    lazy beta delta [EvalSpi.expr_of_mesg] iota in H;fold EvalSpi.expr_of_mesg in H.
    injection H;intros.
    rewrite MessageSpi.mem_fn_MEnc_a.
    rewrite (IHfn_in_msg m2);auto.
    rewrite orb_b_true;auto.
    destruct m;try (discriminate H).
    lazy beta delta [EvalSpi.expr_of_mesg] iota in H;fold EvalSpi.expr_of_mesg in H.
    injection H;intros.
    rewrite MessageSpi.mem_fn_MOp.
    auto.
  Qed.

  Lemma in_msg_name_not_supp : forall (i:nat)(d d' k:nat)(es:list ExpressionSpi.t)(x:Name.t),(i <= d')->(fn_in_msg i d (ExpressionSpi.ENam x))->(deBruijn.low_subst ExpressionSpi.deBruijn_subst (d+S d') k es (ExpressionSpi.ENam x))=(ExpressionSpi.ENam x).
    intros.
    destruct x.
    assert (n=d+i).
    inversion H0.
    simpl in H2.
    unfold DeBruijnNat.mem_fn in H2.
    generalize H2;clear H2.
    case eq_nat_dec;intro;auto.
    intros.
    discriminate H2.
    simpl.
    unfold Name.low_subst_generic.
    unfold DeBruijnNat.low_subst_gen.
    unfold DeBruijnNat.low_subst_generic.
    unfold DeBruijnNat.low_subst_generic_aux.
    case le_lt_dec;intro;try_absurd_arith.
    unfold ExpressionSpi.j.
    auto.
  Qed.

  Lemma fn_in_msg_not_supp : forall (i:nat)(d d' k:nat)(es:list ExpressionSpi.t)(x:ExpressionSpi.t),(i <= d')->(fn_in_msg i d x)->(fn_in_msg i d (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d+S d') k es x)).
    intros.
    induction H0.
    rewrite (in_msg_name_not_supp i);auto.
    apply fnInNam;auto.
    apply fnInNam;auto.
   rewrite ExpressionSpi.low_subst_EPair.
    apply fnInFst;auto.
    rewrite ExpressionSpi.low_subst_EPair.
    apply fnInSnd;auto.
    rewrite ExpressionSpi.low_subst_EEnc_s.
    apply fnInEnc_s_d;auto.
    rewrite ExpressionSpi.low_subst_EEnc_s.
    apply fnInEnc_s_k;auto.
    rewrite ExpressionSpi.low_subst_EEnc_a.
    apply fnInEnc_a_d;auto.
    rewrite ExpressionSpi.low_subst_EEnc_a.
    apply fnInEnc_a_k;auto.
    rewrite ExpressionSpi.low_subst_EOp.
    apply fnInOp;auto.
  Qed.

  Lemma evalc_subst_in_msg_mem_fn : forall (i:nat)(d d' k:nat)(es:list ExpressionSpi.t)(e:ExpressionSpi.t),(i <= d')->(fn_in_msg i d e)->forall (m:MessageSpi.t),(EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d+S d') k es e)=Some m)->(deBruijn.mem_fn MessageSpi.deBruijn_fn_base i d m = true).
    intros i d d' k es e H.
    induction 1.
    intro m.
    rewrite (in_msg_name_not_supp i);auto.
    intros.
    simpl in H1.
    injection H1;clear H1;intros.
    subst m.
    auto.
    apply fnInNam;auto.
    rewrite ExpressionSpi.low_subst_EPair.
    intros.
    rewrite EvalSpi.evalc_EPair in H1.
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e1));try (discriminate H1).
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e2));try (discriminate H1).
    injection H1;intros;subst m.
    rewrite MessageSpi.mem_fn_MPair.
    rewrite (IHfn_in_msg t);auto.
    rewrite ExpressionSpi.low_subst_EPair.
    intros.
    rewrite EvalSpi.evalc_EPair in H1.
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e1));try (discriminate H1).
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e2));try (discriminate H1).
    injection H1;intros;subst m.
    rewrite MessageSpi.mem_fn_MPair.
    rewrite (IHfn_in_msg t0);auto.
    rewrite orb_b_true;auto.
    rewrite ExpressionSpi.low_subst_EEnc_s.
    intros.
    rewrite EvalSpi.evalc_EEnc_s in H1.
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e1));try (discriminate H1).
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e2));try (discriminate H1).
    injection H1;intros;subst m.
    rewrite MessageSpi.mem_fn_MEnc_s.
    rewrite (IHfn_in_msg t);auto.
    rewrite ExpressionSpi.low_subst_EEnc_s.
    intros.
    rewrite EvalSpi.evalc_EEnc_s in H1.
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e1));try (discriminate H1).
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e2));try (discriminate H1).
    injection H1;intros;subst m.
    rewrite MessageSpi.mem_fn_MEnc_s.
    rewrite (IHfn_in_msg t0);auto.
    rewrite orb_b_true;auto.
    rewrite ExpressionSpi.low_subst_EEnc_a.
    intros.
    rewrite EvalSpi.evalc_EEnc_a in H1.
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e1));try (discriminate H1).
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e2));try (discriminate H1).
    injection H1;intros;subst m.
    rewrite MessageSpi.mem_fn_MEnc_a.
    rewrite (IHfn_in_msg t);auto.
    rewrite ExpressionSpi.low_subst_EEnc_a.
    intros.
    rewrite EvalSpi.evalc_EEnc_a in H1.
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e1));try (discriminate H1).
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e2));try (discriminate H1).
    injection H1;intros;subst m.
    rewrite MessageSpi.mem_fn_MEnc_a.
    rewrite (IHfn_in_msg t0);auto.
    rewrite orb_b_true;auto.
    rewrite ExpressionSpi.low_subst_EOp.
    intros.
    rewrite EvalSpi.evalc_EOp in H1.
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e));try (discriminate H1).
    injection H1;intros;subst m.
    rewrite MessageSpi.mem_fn_MOp.
    rewrite (IHfn_in_msg t);auto.
  Qed.

  Lemma evalc_subst_no_redex_mem_fn_in_msg : forall (i:nat)(d d' k:nat)(es:list ExpressionSpi.t)(e:ExpressionSpi.t),(i <= d')->forall (m:MessageSpi.t),(EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d+S d') k es e)=Some m)->(deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base i d e = true)->(EvalSpi.nb_redex e = 0%nat)->(fn_in_msg i d e).
    induction e;intros.
    rewrite ExpressionSpi.mem_fn_ENam in H1.
    apply fnInNam;auto.
    rewrite ExpressionSpi.mem_fn_EPair in H1.
    rewrite ExpressionSpi.low_subst_EPair in H0.
    rewrite EvalSpi.evalc_EPair in H0.
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e1));try (discriminate H0).
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e2));try (discriminate H0).
    injection H0;clear H0;intros.
    simpl in H2.
    case (orb_true_elim (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base i d e1) (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base i d e2) H1);intro.
    apply fnInFst.
    apply IHe1 with t;auto.
    omega.
    apply fnInSnd.
    apply IHe2 with t0;auto.
    omega.
    rewrite ExpressionSpi.mem_fn_EEnc_s in H1.
    rewrite ExpressionSpi.low_subst_EEnc_s in H0.
    rewrite EvalSpi.evalc_EEnc_s in H0.
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e1));try (discriminate H0).
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e2));try (discriminate H0).
    injection H0;clear H0;intros.
    simpl in H2.
    case (orb_true_elim (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base i d e1) (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base i d e2) H1);intro.
    apply fnInEnc_s_d.
    apply IHe1 with t;auto.
    omega.
    apply fnInEnc_s_k.
    apply IHe2 with t0;auto.
    omega.
    rewrite ExpressionSpi.mem_fn_EEnc_a in H1.
    rewrite ExpressionSpi.low_subst_EEnc_a in H0.
    rewrite EvalSpi.evalc_EEnc_a in H0.
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e1));try (discriminate H0).
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e2));try (discriminate H0).
    injection H0;clear H0;intros.
    simpl in H2.
    case (orb_true_elim (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base i d e1) (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base i d e2) H1);intro.
    apply fnInEnc_a_d.
    apply IHe1 with t;auto.
    omega.
    apply fnInEnc_a_k.
    apply IHe2 with t0;auto.
    omega.
    rewrite ExpressionSpi.low_subst_EFst in H0.
    rewrite EvalSpi.evalc_EFst in H0.
    simpl in H2. 
    case (Option_elim_option MessageSpi.t (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e)));intro H3;try (rewrite H3 in H0;discriminate H0).
    ex_elim H3 m'.
    rewrite H3 in H0.
    destruct m';try (discriminate H0).
    rewrite ExpressionSpi.mem_fn_EFst in H1.
    assert (EvalSpi.nb_redex e = 0).
    destruct e;try (discriminate H2);omega.
    generalize (IHe H (MessageSpi.MPair m'1 m'2) H3 H1 H4);intro.
    inversion H5;try (subst e);repeat (rewrite ExpressionSpi.low_subst_EPair in H3 || rewrite ExpressionSpi.low_subst_EEnc_s in H3 || rewrite ExpressionSpi.low_subst_EEnc_a in H3 || rewrite ExpressionSpi.low_subst_EOp in H3);repeat (rewrite EvalSpi.evalc_EPair in H3 || rewrite EvalSpi.evalc_EEnc_s in H3 || rewrite EvalSpi.evalc_EEnc_a in H3 || rewrite EvalSpi.evalc_EOp in H3);try (discriminate H2);try  (destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e1));destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e2));discriminate H3;fail).
    rewrite (in_msg_name_not_supp i) in H3;auto.
    discriminate H3.
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e0));try (discriminate H3).
    rewrite ExpressionSpi.low_subst_ESnd in H0.
    rewrite EvalSpi.evalc_ESnd in H0.
    simpl in H2. 
    case (Option_elim_option MessageSpi.t (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e)));intro H3;try (rewrite H3 in H0;discriminate H0).
    ex_elim H3 m'.
    rewrite H3 in H0.
    destruct m';try (discriminate H0).
    rewrite ExpressionSpi.mem_fn_ESnd in H1.
    assert (EvalSpi.nb_redex e = 0).
    destruct e;try (discriminate H2);omega.
    generalize (IHe H (MessageSpi.MPair m'1 m'2) H3 H1 H4);intro.
    inversion H5;try (subst e);repeat (rewrite ExpressionSpi.low_subst_EPair in H3 || rewrite ExpressionSpi.low_subst_EEnc_s in H3 || rewrite ExpressionSpi.low_subst_EEnc_a in H3 || rewrite ExpressionSpi.low_subst_EOp in H3);repeat (rewrite EvalSpi.evalc_EPair in H3 || rewrite EvalSpi.evalc_EEnc_s in H3 || rewrite EvalSpi.evalc_EEnc_a in H3 || rewrite EvalSpi.evalc_EOp in H3);try (discriminate H2);try  (destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e1));destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e2));discriminate H3;fail).
    rewrite (in_msg_name_not_supp i) in H3;auto.
    discriminate H3.
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e0));try (discriminate H3).
    rewrite ExpressionSpi.low_subst_EDec_s in H0.
    rewrite EvalSpi.evalc_EDec_s in H0.
    simpl in H2. 
    case (Option_elim_option MessageSpi.t (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e1)));intro H3;try (rewrite H3 in H0;discriminate H0).
    ex_elim H3 m'.
    rewrite H3 in H0.
    destruct m';try (discriminate H0).
    rewrite ExpressionSpi.mem_fn_EDec_s in H1.
    case (Option_elim_option MessageSpi.t (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e2)));intro H4;try (rewrite H4 in H0;discriminate H0).
    ex_elim H4 n'.
    rewrite H4 in H0.
    destruct (MessageSpi.Ord.compare m'2 n');try (discriminate H0).
    assert (EvalSpi.nb_redex e1 = 0).
    destruct e1;try (discriminate H2);omega.
    assert (EvalSpi.nb_redex e2 = 0).
    destruct e1;try (discriminate H2);omega.
    assert (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base i d e1 = true).
    case (orb_true_elim (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base i d e1) (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base i d e2) H1);clear H1;intro H1;auto.
    generalize (IHe2 H n' H4 H1 H6);intro.
    generalize (evalc_subst_in_msg_mem_fn i d d' k es e2 H H7 n' H4);intro.
    rewrite <- e in H8.
    generalize (EvalSpi.evalc_mem_fn_true (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e1) (MessageSpi.MEnc_s m'1 m'2) H3 d i).
    rewrite MessageSpi.mem_fn_MEnc_s.
    rewrite H8.
    rewrite orb_b_true.
    intro.
    generalize (H9 (refl_equal true));clear H9;intro.
    generalize (deBruijn.low_subst_fn_not_supp ExpressionSpi.deBruijn_fn_subst i d d' k es e1 H).
    fold ExpressionSpi.deBruijn_fn_base.
    fold ExpressionSpi.deBruijn_subst.
    intro.
    rewrite H9 in H10.
    auto.
    clear H1;generalize H7;clear H7;intro H1.
    generalize (IHe1 H (MessageSpi.MEnc_s m'1 m'2) H3 H1 H5);intro.
    inversion H7;try (subst e1);repeat (rewrite ExpressionSpi.low_subst_EPair in H3 || rewrite ExpressionSpi.low_subst_EEnc_s in H3 || rewrite ExpressionSpi.low_subst_EEnc_a in H3 || rewrite ExpressionSpi.low_subst_EOp in H3);repeat (rewrite EvalSpi.evalc_EPair in H3 || rewrite EvalSpi.evalc_EEnc_s in H3 || rewrite EvalSpi.evalc_EEnc_a in H3 || rewrite EvalSpi.evalc_EOp in H3);try (discriminate H2);try  (destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e0));destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e3));discriminate H3;fail).
    rewrite (in_msg_name_not_supp i) in H3;auto.
    discriminate H3.
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e0));try (discriminate H3).
    rewrite ExpressionSpi.low_subst_EDec_a in H0.
    rewrite EvalSpi.evalc_EDec_a in H0.
    simpl in H2. 
    case (Option_elim_option MessageSpi.t (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e1)));intro H3;try (rewrite H3 in H0;discriminate H0).
    ex_elim H3 m'.
    rewrite H3 in H0.
    destruct m';try (discriminate H0).
    rewrite ExpressionSpi.mem_fn_EDec_a in H1.
    case (Option_elim_option MessageSpi.t (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e2)));intro H4;try (rewrite H4 in H0;discriminate H0).
    ex_elim H4 n'.
    rewrite H4 in H0.
    case (Option_elim_option MessageSpi.t (MessageSpi.inverse_key m'2));intro H5;try (rewrite H5 in H0;discriminate H0).
    ex_elim H5 n'inv.
    rewrite H5 in H0.
    destruct (MessageSpi.Ord.compare n'inv n');try (discriminate H0).
    assert (EvalSpi.nb_redex e1 = 0).
    destruct e1;try (discriminate H2);omega.
    assert (EvalSpi.nb_redex e2 = 0).
    destruct e1;try (discriminate H2);omega.
    assert (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base i d e1 = true).
    case (orb_true_elim (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base i d e1) (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base i d e2) H1);clear H1;intro H1;auto.
    generalize (IHe2 H n' H4 H1 H7);intro.
    generalize (evalc_subst_in_msg_mem_fn i d d' k es e2 H H8 n' H4);intro.
    rewrite <- e in H9.
    rewrite <- (MessageSpi.mem_fn_inverse_key d i m'2 n'inv) in H9;auto.
    generalize (EvalSpi.evalc_mem_fn_true (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e1) (MessageSpi.MEnc_a m'1 m'2) H3 d i).
    rewrite MessageSpi.mem_fn_MEnc_a.
    rewrite H9.
    rewrite orb_b_true.
    intro.
    generalize (H10 (refl_equal true));clear H10;intro.
    generalize (deBruijn.low_subst_fn_not_supp ExpressionSpi.deBruijn_fn_subst i d d' k es e1 H).
    fold ExpressionSpi.deBruijn_fn_base.
    fold ExpressionSpi.deBruijn_subst.
    intro.
    rewrite H10 in H11.
    auto.
    clear H1;generalize H8;clear H8;intro H1.
    generalize (IHe1 H (MessageSpi.MEnc_a m'1 m'2) H3 H1 H6);intro.
    inversion H8;try (subst e1);repeat (rewrite ExpressionSpi.low_subst_EPair in H3 || rewrite ExpressionSpi.low_subst_EEnc_s in H3 || rewrite ExpressionSpi.low_subst_EEnc_a in H3 || rewrite ExpressionSpi.low_subst_EOp in H3);repeat (rewrite EvalSpi.evalc_EPair in H3 || rewrite EvalSpi.evalc_EEnc_s in H3 || rewrite EvalSpi.evalc_EEnc_a in H3 || rewrite EvalSpi.evalc_EOp in H3);try (discriminate H2);try  (destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e0));destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e3));discriminate H3;fail).
    rewrite (in_msg_name_not_supp i) in H3;auto.
    discriminate H3.
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e0));try (discriminate H3).
    rewrite ExpressionSpi.mem_fn_EOp in H1.
    rewrite ExpressionSpi.low_subst_EOp in H0.
    rewrite EvalSpi.evalc_EOp in H0.
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d + S d') k es e));try (discriminate H0).
    injection H0;clear H0;intros.
    simpl in H2.
    apply fnInOp.
    apply IHe with t;auto.
  Qed.
  
  Lemma fn_in_msg_mem_fn : forall (d k:nat)(e:ExpressionSpi.t),(fn_in_msg k d e)->(deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base k d e = true).
    induction 1;intros.
    rewrite ExpressionSpi.mem_fn_ENam;auto.
    rewrite ExpressionSpi.mem_fn_EPair.
    rewrite IHfn_in_msg;auto.
    rewrite ExpressionSpi.mem_fn_EPair.
    rewrite IHfn_in_msg;auto.
    rewrite orb_b_true;auto.
    rewrite ExpressionSpi.mem_fn_EEnc_s.
    rewrite IHfn_in_msg;auto.
    rewrite ExpressionSpi.mem_fn_EEnc_s.
    rewrite IHfn_in_msg;auto.
    rewrite orb_b_true;auto.
    rewrite ExpressionSpi.mem_fn_EEnc_a.
    rewrite IHfn_in_msg;auto.
    rewrite ExpressionSpi.mem_fn_EEnc_a.
    rewrite IHfn_in_msg;auto.
    rewrite orb_b_true;auto.
    rewrite ExpressionSpi.mem_fn_EOp.
    rewrite IHfn_in_msg;auto.
  Qed.

  Lemma evalc_subst_no_redex_mem_fn_in_evalc : forall (i:nat)(d d' k:nat)(es:list ExpressionSpi.t)(e:ExpressionSpi.t),(i <= d')->forall (m:MessageSpi.t),(EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d+S d') k es e)=Some m)->(deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base i d e = true)->(EvalSpi.nb_redex e = 0%nat)->(deBruijn.mem_fn MessageSpi.deBruijn_fn_base i d m = true).
    intros.
    generalize (evalc_subst_no_redex_mem_fn_in_msg i d d' k es e H m H0 H1 H2).
    intro.
    apply (evalc_subst_in_msg_mem_fn i d d' k es e H H3 m H0).
  Qed.

  Lemma evalc_subst_no_redex_mem_fn_in_evalc_false : forall (i:nat)(d d' k:nat)(es:list ExpressionSpi.t)(e:ExpressionSpi.t),(i <= d')->forall (m:MessageSpi.t),(EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d+S d') k es e)=Some m)->(deBruijn.mem_fn MessageSpi.deBruijn_fn_base i d m = false)->(EvalSpi.nb_redex e = 0%nat)->(deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base i d e = false).
    intros.
    case (Bool_elim_bool (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base i d e));intro;auto.
    rewrite (evalc_subst_no_redex_mem_fn_in_evalc i d d' k es e) in H1;auto.
    discriminate H1.
  Qed.

  Lemma evalc_subst_no_redex_mem_fn_same : forall (i:nat)(d d' k:nat)(es:list ExpressionSpi.t)(e:ExpressionSpi.t),(i <= d')->(EvalSpi.nb_redex e = 0%nat)->forall (m:MessageSpi.t),(EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d+S d') k es e)=Some m)->(deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base i d e = deBruijn.mem_fn MessageSpi.deBruijn_fn_base i d m).
    intros.
    case (Bool_elim_bool (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base i d e));intro.
    rewrite H2.
    symmetry.
    eapply evalc_subst_no_redex_mem_fn_in_evalc;eauto.
    generalize (deBruijn.low_subst_fn_not_supp ExpressionSpi.deBruijn_fn_subst i d d' k es e H).
    fold ExpressionSpi.deBruijn_fn_base.
    fold ExpressionSpi.deBruijn_subst.
    intros.
    rewrite H2;symmetry.
    rewrite H3 in H2.
    eapply EvalSpi.evalc_mem_fn_false;eauto.
  Qed.
End FnInMsg.

