(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import Arith.
Require Export expressionType.
Require Export Relations.
Require Import deBruijnType.

Module Type FormulaeType.
  Declare Module E : ExpressionType.
  Parameter t : Set.

  Declare Module Ord : OrderedType with Definition t := t with Definition eq := @eq t.

  Parameter deBruijn_base : deBruijn.deBruijn_base t.
  Parameter deBruijn_subst : deBruijn.deBruijn_subst t deBruijn_base E.t E.deBruijn_base.

  Parameter deBruijn_fn_base : deBruijn.deBruijn_fn t deBruijn_base.
  Parameter deBruijn_fn_subst : deBruijn.deBruijn_subst_fn t deBruijn_base deBruijn_fn_base E.t E.deBruijn_base E.deBruijn_fn_base deBruijn_subst.

  Parameter low_subst_lemma : forall (x:t)(d d' k k':nat)(es fs:list E.t),deBruijn.low_subst deBruijn_subst (d+d'+k) k' es (deBruijn.low_subst deBruijn_subst d k fs x)=deBruijn.low_subst deBruijn_subst d k (map (deBruijn.low_subst E.deBruijn_subst (d'+k) k' es) fs) (deBruijn.low_subst deBruijn_subst (d+d'+length fs) k' es x).

  Parameter low_subst_split_lemma : forall (x:t)(d k:nat)(es fs:list E.t),deBruijn.low_subst deBruijn_subst d 0 es (deBruijn.low_subst deBruijn_subst (d+length es) k fs x)=deBruijn.low_subst deBruijn_subst d k (app es fs) x.

  Parameter compute_fn : nat -> t -> ListSet.set nat. 
  Parameter compute_fn_ok : forall (k:nat)(d:nat)(x:t),(deBruijn.mem_fn deBruijn_fn_base k d x = ListSet.set_mem eq_nat_dec k (compute_fn d x)).
End FormulaeType.



