(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import FSets.

Module OpSpi.

  Inductive op : Set :=
    | Pub : op
    | Priv : op
    | Hash : op.

  Module Ord : OrderedType with Definition t := op with Definition eq := @eq op.
    Definition t := op.
    Definition eq := @eq t.
    Definition lt (o1 o2:t) :=
      match (o1,o2) with
	| (Pub,Pub) => False
	| (Pub,_) => True
	| (_,Pub) => False
	| (Priv,Priv) => False
	| (Priv,_) => True
	| (_,Priv) => False
	| (Hash,Hash) => False
      end.

    Lemma eq_refl : forall (x:t),eq x x.
      unfold eq;auto.
    Qed.

    Lemma eq_sym : forall (x y:t),(eq x y)->(eq y x).
      unfold eq;auto.
    Qed.

    Lemma eq_trans : forall (x y z:t),(eq x y)->(eq y z)->(eq x z).
      unfold eq;auto.
      intros;congruence.
    Qed.

    Lemma lt_trans : forall (x y z:t),(lt x y)->(lt y z)->(lt x z).
      unfold lt.
      destruct x;destruct y;destruct z;try tauto.
    Qed.
    
    Lemma lt_not_eq : forall (x y:t),(lt x y)->~(eq x y).
      unfold lt, eq.
      destruct x;destruct y;try tauto;intros;red;intro;discriminate H0.
    Qed.

    Definition compare : forall (x y:t),(Compare lt eq x y).
      unfold lt, eq.
      destruct x;destruct y;try (apply EQ;simpl;auto;fail);try (apply LT;simpl;auto;fail);try (apply GT;simpl;auto;fail).
    Defined.

    Hint Immediate eq_sym.
    Hint Resolve eq_refl eq_trans lt_not_eq lt_trans.
  End Ord.

  Theorem op_eq_dec : forall (o1 o2:op),{o1=o2}+{~o1=o2}.
    intros.
    case (Ord.compare o1 o2);intro;[right | left | right];try (red;intro).
    elim (Ord.lt_not_eq l);auto.
    rewrite e;auto.
    elim (Ord.lt_not_eq l);auto.
  Defined.

  Lemma op_eq_dec_refl : forall (o:op),(exists H:o=o,(op_eq_dec o o)=(left (o<>o) H)).
    intro o.
    case (op_eq_dec o o).
    intro.
    exists e.
    trivial.
    intro.
    elim n.
    trivial.
  Qed.
End OpSpi.
