(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import name.
Require Import expressionSpi.
Require Import messageSpi.
Require Import formulaeSpi.
Require Import deBruijnType.
Require Import actionSpi.
Require Import actionSpiNC.
Require Import actionSpiSymb.
Require Import missing.
Require Import tactics.
Require Import reduction.
Require Import calculi.
Require Import evalSpi.
Require Import Arith.

Require Import redcong.
Module RedSpiCong := RedCong(ActionSpi)(CongruenceSpi).

(* Module RedSpi := RedSpiCong.R. *)
Module RedSpiNC := Reduction(ActionSpiNC).
Module RedSpiSymb := Reduction(ActionSpiSymb).

(* Module TransSet := FSetAVL.Make(RedPi.Transition). *)
(* Module TransSet := FSetList.Make(RedSpiNC.Transition). *)

(* Module RedSpiComp := RedSpiNC.Make(TransSet). *)

(* Definition P1 := AgentSpi.P.Input (ExpressionSpi.ENam (Name.Ref 0)) (AgentSpi.P.Zero). *)
(* Definition P2 := AgentSpi.P.New (AgentSpi.P.Output (ExpressionSpi.ENam (Name.Ref 1)) (ExpressionSpi.EEnc_s (ExpressionSpi.ENam (Name.Ref 1)) (ExpressionSpi.ENam (Name.Ref 0)))  (AgentSpi.P.Zero)). *)

(* Eval vm_compute in (RedSpiComp.late_transitions P1). *)
(* Eval vm_compute in (RedSpiComp.late_transitions P2). *)

(* Definition P := AgentSpi.P.New (AgentSpi.P.Par P1 P2). *)

(* Eval vm_compute in (RedSpiComp.late_transitions P). *)

(* needed to tweak the type system problem with modules *)
Lemma reduction_spi_conc_message_aux : forall (p:SpiCalcul.t)(mu:ActionSpi.t)(a:AgentSpi.t),(RedSpiCong.R.reduction p mu a)->forall (n:nat)(e:ExpressionSpi.t)(q:SpiCalcul.t),a=(ActionSpi.A.C (ActionSpi.A.Conc n e q))->exists m:MessageSpi.t,e=(EvalSpi.expr_of_mesg m).
  induction 1;intros.
  discriminate H.
  discriminate H0.
  injection H0;intros.
  unfold ActionSpi.mk_output in H.
  destruct (EvalSpi.evalc e);try (discriminate H).
  destruct t;try (discriminate H).
  destruct (EvalSpi.evalc f);try (discriminate H).
  injection H;intros.
  subst m.
  subst e0.
  exists t0;auto.
  eapply IHreduction;eauto.
  eapply IHreduction;eauto.
  eapply IHreduction;eauto.
  destruct a.
  rewrite ActionSpi.A.new_Pr in H1;discriminate H1.
  rewrite ActionSpi.A.new_F in H1;discriminate H1.
  rewrite ActionSpi.A.new_C in H1.
  destruct c.
  rewrite ActionSpi.A.new_conc_def in H1.
  destruct (deBruijn.mem_fn ActionSpi.A.P.E.deBruijn_fn_base 0 n0 t).
  injection H1;intros.
  subst n;subst t;subst t0.
  eapply IHreduction.
  reflexivity.
  assert (n0=n).
  injection H1;intros;auto.
  assert ((deBruijn.low1 ActionSpi.A.P.E.deBruijn_base n0 t)=e).
  injection H1;intros;auto.
  assert ((ActionSpi.A.P.New (deBruijn.swapR ActionSpi.A.P.deBruijn_base 0 n0 t0))=q).
  injection H1;intros;auto.
  clear H1.
  subst n0.
  subst q.
  subst e.
  elim (IHreduction n t t0);auto.
  intro m;intros.
  subst t.
  simpl.
  rewrite EvalSpi.map_expr_of_msg.
  exists (deBruijn.low1 MessageSpi.deBruijn_base n m).
  reflexivity.
  destruct a.
  rewrite ActionSpi.A.par_r_Pr in H0;discriminate H0.
  rewrite ActionSpi.A.par_r_F in H0;discriminate H0.
  rewrite ActionSpi.A.par_r_C in H0.
  destruct c.
  rewrite ActionSpi.A.conc_par_r_def in H0.
  injection H0;clear H0;intros.
  subst t;subst n0.
  eauto.
  destruct a.
  rewrite ActionSpi.A.par_l_Pr in H0;discriminate H0.
  rewrite ActionSpi.A.par_l_F in H0;discriminate H0.
  rewrite ActionSpi.A.par_l_C in H0.
  destruct c.
  rewrite ActionSpi.A.conc_par_l_def in H0.
  injection H0;clear H0;intros.
  subst t;subst n0.
  eauto.
  destruct f;destruct c.
  rewrite ActionSpi.A.pseudo_appl_l_def in H2;discriminate H2.
  destruct f;destruct c.
  rewrite ActionSpi.A.pseudo_appl_r_def in H2;discriminate H2.
  destruct a.
  rewrite ActionSpi.A.par_r_Pr in H0;discriminate H0.
  rewrite ActionSpi.A.par_r_F in H0;discriminate H0.
  rewrite ActionSpi.A.par_r_C in H0.
  destruct c.
  rewrite ActionSpi.A.conc_par_r_def in H0.
  injection H0;intros.
  subst n0.
  subst t.
  eauto.
  destruct f;destruct c.
  rewrite ActionSpi.A.pseudo_appl_l_def in H2;discriminate H2.
Qed.

Lemma reduction_spi_conc_message : forall (p:SpiCalcul.t)(mu:ActionSpi.t)(a:AgentSpi.t),(RedSpiCong.R.reduction p mu a)->forall (n:nat)(e:ExpressionSpi.t)(q:SpiCalcul.t),a=(AgentSpi.C (AgentSpi.Conc n e q))->exists m:MessageSpi.t,e=(EvalSpi.expr_of_mesg m).
  intros.
  eapply reduction_spi_conc_message_aux;eauto.
Qed.

Lemma reduction_spi_impl_reduction_spi_nc : forall (p:SpiCalcul.t)(mu:ActionSpi.t)(a:AgentSpi.t),(RedSpiCong.R.reduction p mu a)->exists l:list Name.t,(RedSpiNC.reduction p (mu,l) a).
  induction 1.
  (* RedSilent *)
  exists (nil (A:=Name.t)).
  apply RedSpiNC.RedSilent.
  (* RedInput *)
  generalize (ActionSpi.mk_input_kind e mu H).
  intro.
  destruct mu;try (discriminate H0).
  exists (cons t nil).
  apply RedSpiNC.RedInput.
  unfold ActionSpiNC.mk_input.
  unfold ActionSpi.mk_input in H.
  destruct (evalSpi.EvalSpi.evalc e);try (discriminate H).
  destruct t0;try (discriminate H).
  injection H;intros.
  subst t;auto.
  (* RedOutput *)
  generalize (ActionSpi.mk_output_kind e f mu m H).
  intro.
  destruct mu;try (discriminate H0).
  exists (cons t nil).
  apply RedSpiNC.RedOutput.
  unfold ActionSpiNC.mk_output.
  unfold ActionSpi.mk_output in H.
  destruct (evalSpi.EvalSpi.evalc e);try (discriminate H).
  destruct t0;try (discriminate H).
  destruct (evalSpi.EvalSpi.evalc f);try (discriminate H).
  injection H;intros.
  subst m;subst t;auto.
  (* RedIfThen *)
  unfold ActionSpi.mk_if in H0.
  generalize (EvalSpi.eval_nc_eval phi).
  destruct (EvalSpi.eval phi);try (discriminate H0).
  intro.
  elim IHreduction;intro l;intros.
  exists (app l (snd (EvalSpi.eval_nc phi))).
  eapply RedSpiNC.RedIfThen;eauto.
  unfold ActionSpiNC.mk_if.
  destruct (EvalSpi.eval_nc phi).
  simpl in H1.
  subst b.
  injection H0;clear H0;intros;subst mu'.
  reflexivity.
  (* RedSumL *)
  elim IHreduction;intro l;intros.
  exists l.
  apply RedSpiNC.RedSumL;auto.
  (* RedSumR *)
  elim IHreduction;intro l;intros.
  exists l.
  apply RedSpiNC.RedSumR;auto.
  (* RedRes *)
  elim IHreduction;intro l;intros.
  exists (ActionSpiNC.new_c l).
  eapply RedSpiNC.RedRes;eauto.
  unfold ActionSpiNC.mk_res.
  rewrite H0.
  auto.
  (* RedParL *)
  elim IHreduction;intro l;intros.
  exists l.
  apply RedSpiNC.RedParL;auto.
  (* RedParR *)
  elim IHreduction;intro l;intros.
  exists l.
  apply RedSpiNC.RedParR;auto.
  (* RedCloseL *)
  elim IHreduction1;intro linp;intro.
  elim IHreduction2;intro lout;intro.
  exists (app linp lout).
  eapply RedSpiNC.RedCloseL;eauto.
  unfold ActionSpiNC.mk_interact.
  rewrite H1.
  rewrite (RedSpiCong.R.reduction_abstraction_input H f);auto.
  (* RedCloseR *)
  elim IHreduction1;intro lout;intro.
  elim IHreduction2;intro linp;intro.
  exists (app linp lout).
  eapply RedSpiNC.RedCloseR;eauto.
  unfold ActionSpiNC.mk_interact.
  rewrite H1.
  rewrite (RedSpiCong.R.reduction_concretion_output H c);auto.
  (* RedBangAct *)
  elim IHreduction;intro l;intro.
  exists l.
  eapply RedSpiNC.RedBangAct;eauto.
  (* RedBangCloseL *)
  elim IHreduction1;intro linp;intro.
  elim IHreduction2;intro lout;intro.
  exists (app linp lout).
  eapply RedSpiNC.RedBangCloseL;eauto.
  unfold ActionSpiNC.mk_interact.
  rewrite H1.
  rewrite (RedSpiCong.R.reduction_abstraction_input H f);auto.
Qed.

Lemma reduction_spi_nc_impl_reduction_spi : forall (p:SpiCalcul.t)(mu:ActionSpiNC.t)(a:AgentSpi.t),(RedSpiNC.reduction p mu a)->RedSpiCong.R.reduction p (fst mu) a.
  induction 1;intros.
  (* RedSilent *)
  unfold ActionSpiNC.mk_silent.
  simpl.
  apply RedSpiCong.R.RedSilent.
  (* RedInput *)
  apply RedSpiCong.R.RedInput.
  unfold ActionSpiNC.mk_input in H.
  unfold ActionSpi.mk_input.
  destruct (EvalSpi.evalc e);try (discriminate H).
  destruct t;try (discriminate H).
  injection H;clear H;intros;subst mu;try reflexivity.
  (* RedOutput *)
  apply RedSpiCong.R.RedOutput.
  unfold ActionSpiNC.mk_output in H.
  unfold ActionSpi.mk_output.
  destruct (EvalSpi.evalc e);try (discriminate H).
  destruct t;try (discriminate H).
  destruct (EvalSpi.evalc f);try (discriminate H).
  injection H;intros.
  subst mu.
  subst m.
  reflexivity.
  (* RedIfThen *)
  eapply RedSpiCong.R.RedIfThen;eauto.
  unfold ActionSpiNC.mk_if in H0.
  unfold ActionSpi.mk_if.
  generalize (EvalSpi.eval_nc_eval phi).
  destruct mu.
  destruct (EvalSpi.eval_nc phi);try (discriminate H0).
  destruct b;try (discriminate H0).
  simpl.
  intros.
  rewrite <- H1.
  injection H0;intros;subst mu'.
  reflexivity.
  (* RedSumL *)
  apply RedSpiCong.R.RedSumL;auto.
  (* RedSumR *)
  apply RedSpiCong.R.RedSumR;auto.
  (* RedRes *)
  eapply RedSpiCong.R.RedRes;eauto.
  unfold ActionSpiNC.mk_res in H0.
  destruct mu.
  simpl.
  destruct (ActionSpi.mk_res t);try (discriminate H0).
  injection H0;intros;subst mu';reflexivity.
  (* RedParL *)
  apply RedSpiCong.R.RedParL;auto.
  (* RedParR *)
  apply RedSpiCong.R.RedParR;auto.
  (* RedCloseL *)
  eapply RedSpiCong.R.RedCloseL;eauto.
  unfold ActionSpiNC.mk_interact in H1.
  destruct inp;destruct out.
  simpl.
  destruct (ActionSpi.mk_interact t t0);try (discriminate H1).
  destruct (ActionSpi.get_kind t);try (discriminate H1).
  injection H1;intros;subst tau;reflexivity.
  injection H1;intros;subst tau;reflexivity.
  (* RedCloseR *)
  eapply RedSpiCong.R.RedCloseR;eauto.
  unfold ActionSpiNC.mk_interact in H1.
  destruct inp;destruct out.
  simpl.
  destruct (ActionSpi.mk_interact t0 t);try (discriminate H1).
  destruct (ActionSpi.get_kind t0);try (discriminate H1).
  injection H1;intros;subst tau;reflexivity.
  injection H1;intros;subst tau;reflexivity.
  (* RedBangAct *)
  eapply RedSpiCong.R.RedBangAct;eauto.
  (* RedBangCloseL *)
  eapply RedSpiCong.R.RedBangCloseL;eauto.
  unfold ActionSpiNC.mk_interact in H1.
  destruct inp;destruct out.
  simpl.
  destruct (ActionSpi.mk_interact t t0);try (discriminate H1).
  destruct (ActionSpi.get_kind t);try (discriminate H1).
  injection H1;intros;subst tau;reflexivity.
  injection H1;intros;subst tau;reflexivity.
Qed.

Lemma reduction_spi_nc_conc_message : forall (p:SpiCalcul.t)(mu:ActionSpiNC.t)(a:AgentSpi.t),(RedSpiNC.reduction p mu a)->forall (n:nat)(e:ExpressionSpi.t)(q:SpiCalcul.t),a=(AgentSpi.C (AgentSpi.Conc n e q))->exists m:MessageSpi.t,e=(EvalSpi.expr_of_mesg m).
  intros.
  eapply reduction_spi_conc_message;eauto.
  apply reduction_spi_nc_impl_reduction_spi.
  eauto.
Qed.

(* Module TransSet := FSetList.Make(RedSpiCong.R.Transition).  *)
(* Module RedSpiComp := RedSpiCong.R.Make(TransSet). *)

(* Module TransSetNC := FSetList.Make(RedSpiNC.Transition).  *)
(* Module RedSpiCompNC := RedSpiNC.Make(TransSetNC). *)

(* Definition P1 := SpiCalcul.Input (ExpressionSpi.ENam (Name.Ref 0)) (SpiCalcul.Zero). *)
(* Definition P2 := SpiCalcul.Output (ExpressionSpi.ENam (Name.Ref 0)) (ExpressionSpi.ENam (Name.Ref 0))  (SpiCalcul.Zero). *)

(* Eval compute in (RedSpiCompNC.late_transitions P1). *)
(* Eval compute in (RedSpiCompNC.late_transitions P2). *)

(* Definition P := SpiCalcul.New (SpiCalcul.Par P1 P2). *)

(* Eval compute in (RedSpiCompNC.late_transitions P). *)

Lemma reduction_spi_congruence_silent : forall (p p':SpiCalcul.t),(RedSpiCong.R.reduction p ActionSpi.Tau (AgentSpi.Pr p'))->forall (q:SpiCalcul.t),(CongruenceSpi.Rcongruence CongruenceSpi.StructuralCongruenceAxioms p q)->exists q':SpiCalcul.t,(RedSpiCong.R.reduction q ActionSpi.Tau (AgentSpi.Pr q'))/\(CongruenceSpi.Rcongruence CongruenceSpi.StructuralCongruenceAxioms p' q').
  intros.
  elim (RedSpiCong.StructuralCongruence_simulation _ _ H0 _ _ H). 
  intro mu;intros.
  ex_elim H1 b.
  unfold ActionSpi.equal in H1.
  subst mu.
  assert (inclusion _ CongruenceSpi.NewSwap CongruenceSpi.StructuralCongruenceAxioms).
  unfold inclusion.
  intros.
  apply CongruenceSpi.SCNewSwap;auto.
  elim (CongruenceSpi.Acongruence_Pr H1 H3 p');auto.
  intro q';intros;and_elim.
  subst b.
  exists q'.
  split;auto.
Qed.

Lemma reduction_spi_congruence_input : forall (p:SpiCalcul.t)(a:Name.t)(p':SpiCalcul.t),(RedSpiCong.R.reduction p (ActionSpi.Inp a) (AgentSpi.F (AgentSpi.Abs p')))->forall (q:SpiCalcul.t),(CongruenceSpi.Rcongruence CongruenceSpi.StructuralCongruenceAxioms p q)->exists q':SpiCalcul.t,(RedSpiCong.R.reduction q (ActionSpi.Inp a) (AgentSpi.F (AgentSpi.Abs q')))/\(CongruenceSpi.Rcongruence CongruenceSpi.StructuralCongruenceAxioms p' q').
  intros.
  elim (RedSpiCong.StructuralCongruence_simulation _ _ H0 _ _ H). 
  intro mu;intros.
  ex_elim H1 b.
  unfold ActionSpi.equal in H1.
  subst mu.
  assert (inclusion _ CongruenceSpi.NewSwap CongruenceSpi.StructuralCongruenceAxioms).
  unfold inclusion.
  intros.
  apply CongruenceSpi.SCNewSwap;auto.
  assert (CongruenceSpi.R_preserves_swapR (CongruenceSpi.Rcongruence CongruenceSpi.StructuralCongruenceAxioms)).
  apply CongruenceSpi.Rcongruence_swapR.
  apply CongruenceSpi.StructuralCongruenceAxioms_preserves_swapR.
  elim (CongruenceSpi.Acongruence_F H1 H4 H3 p');auto.
  intro q';intros;and_elim.
  subst b.
  exists q'.
  split;auto.
Qed.

Require Import workAround.

Lemma reduction_spi_congruence_output : forall (p:SpiCalcul.t)(a:Name.t)(n:nat)(m:ExpressionSpi.t)(p':SpiCalcul.t),(RedSpiCong.R.reduction p (ActionSpi.Outp a) (AgentSpi.C (AgentSpi.Conc n m p')))->forall (q:SpiCalcul.t),(CongruenceSpi.Rcongruence CongruenceSpi.StructuralCongruenceAxioms p q)->exists q':SpiCalcul.t,exists m':ExpressionSpi.t,(RedSpiCong.R.reduction q (ActionSpi.Outp a) (AgentSpi.C (AgentSpi.Conc n m' q')))/\forall es : list ExpressionSpi.t,length es = n ->exists fs : list ExpressionSpi.t,permutation.is_permutation _ es fs /\deBruijn.low_subst ExpressionSpi.deBruijn_subst 0 0 es m = deBruijn.low_subst ExpressionSpi.deBruijn_subst 0 0 fs m' /\ CongruenceSpi.Rcongruence CongruenceSpi.StructuralCongruenceAxioms (deBruijn.low_subst AgentSpi.P.deBruijn_subst 0 0 es p') (deBruijn.low_subst AgentSpi.P.deBruijn_subst 0 0 fs q').
  intros.
  elim (RedSpiCong.StructuralCongruence_simulation _ _ H0 _ _ H).
  intro mu;intros.
  ex_elim H1 b.
  unfold ActionSpi.equal in H1.
  subst mu.
  assert (inclusion _ CongruenceSpi.NewSwap CongruenceSpi.StructuralCongruenceAxioms).
  unfold inclusion.
  intros.
  apply CongruenceSpi.SCNewSwap;auto.
  assert (CongruenceSpi.R_preserves_swapR (CongruenceSpi.Rcongruence CongruenceSpi.StructuralCongruenceAxioms)).
  apply CongruenceSpi.Rcongruence_swapR.
  apply CongruenceSpi.StructuralCongruenceAxioms_preserves_swapR.
  assert (CongruenceSpi.R_preserves_low_subst (CongruenceSpi.Rcongruence CongruenceSpi.StructuralCongruenceAxioms)).
  apply CongruenceSpi.Rcongruence_low_subst.
  apply CongruenceSpi.StructuralCongruenceAxioms_preserves_low_subst.
  elim (CongruenceSpi.Acongruence_C_permutation H1 H4 H5 H3 n m p');auto.
  intro m';intros.
  ex_elim H6 q'.
  subst b.
  exists q'.
  exists m'.
  split.
  auto.
  intros.
  elim (H8 _ H6).
  intro fs;intros;and_elim.
  exists fs.
  split.
  auto.
  split.
  generalize (low_subst_AgentSpi_bis 0 0 es m).
  intro.
  rewrite H12.
  generalize (low_subst_AgentSpi_bis 0 0 fs m').
  intros.
  rewrite H13.
  auto.
  apply H11.
Qed.

