(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import name.
Require Import expressionSpi.
Require Import messageSpi.
Require Import formulaeSpi.
Require Import deBruijnType.
Require Import actionSpi.
Require Import actionSpiNC.
Require Import missing.
Require Import tactics.
Require Import reduction.
Require Import calculi.
Require Import evalSpi.
Require Import Arith.
Require Import reductionSpi.
Require Import substSpi.
Require Import workAround.

Module ReductionSpiNCprops.
  Fixpoint low_subst_constraint (d k:nat)(es:list ExpressionSpi.t)(xss:list Name.t) {struct xss} : option (list Name.t) :=
    match xss with
      | nil => Some nil
      | cons x xs => 
	(match deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es (ExpressionSpi.j x) with
	   | ExpressionSpi.ENam y => 
	     (match low_subst_constraint d k es xs with
		| None => None
		| Some(ys) => Some(cons y ys)
	      end)
	   | _ => None
	 end)
    end.

  Lemma low_subst_constraint_nil : forall (d k:nat)(es:list ExpressionSpi.t),(low_subst_constraint d k es nil)=Some nil.
    intros;reflexivity.
  Qed.

  Lemma low_subst_constraint_cons : forall (d k:nat)(es:list ExpressionSpi.t)(x:Name.t)(xs:list Name.t),(low_subst_constraint d k es (cons x xs))=(match deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es (ExpressionSpi.j x) with ExpressionSpi.ENam y => option_map (fun ys => cons y ys) (low_subst_constraint d k es xs) | _ => None end).
    intros.
    lazy beta delta [low_subst_constraint] iota.
    fold low_subst_constraint.
    destruct (low_subst_constraint d k es xs);reflexivity.
  Qed.

  Lemma low_subst_S : forall (d k:nat)(es:list ExpressionSpi.t)(x:nat),deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es (ExpressionSpi.j (Name.Ref x))=deBruijn.low1 ExpressionSpi.deBruijn_base 0 (deBruijn.low_subst ExpressionSpi.deBruijn_subst (S d) k es (ExpressionSpi.j (Name.Ref (S x)))).
    intros.
    simpl.
    unfold Name.low_subst_generic.
    unfold DeBruijnNat.low_subst_gen.
    unfold DeBruijnNat.low_subst_generic.
    unfold DeBruijnNat.low_subst_generic_aux.
    replace (S x - S d) with (x - d);try omega.
    elim (List_nth_Some ExpressionSpi.t es (x-d));intros.
    case le_lt_dec;intro;case le_lt_dec;intro;try_absurd_arith.
    destruct (nth_error es (x-d)).
    generalize (deBruijn.low1_lift_simpl ExpressionSpi.deBruijn_base 0 d t);simpl.
    intro.
    congruence.
    assert (length es <= x-d).
    case (le_lt_dec (length es) (x-d));auto.
    intros.
    elim H;auto.
    intros.
    discriminate H1.
    replace (S x - length es + k) with (S (x-length es+k));try omega.
    reflexivity.
    reflexivity.
  Qed.

  Lemma low_subst_name_mem_fn : forall (d k:nat)(es:list ExpressionSpi.t)(x:Name.t),(deBruijn.mem_fn Name.deBruijn_fn_base 0 0 x = deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 0 (deBruijn.low_subst ExpressionSpi.deBruijn_subst (S d) k es (ExpressionSpi.j x))).
    intros.
    generalize (deBruijn.low_subst_fn_not_supp ExpressionSpi.deBruijn_fn_subst 0 0 d k es (ExpressionSpi.j x) (le_O_n d)).
    simpl;auto.
  Qed.

  Lemma low_subst_constraint_new_c : forall (xs:list Name.t)(d k:nat)(es:list ExpressionSpi.t),(low_subst_constraint d k es (ActionSpiNC.new_c xs))=(option_map ActionSpiNC.new_c (low_subst_constraint (S d) k es xs)).
    induction xs.
    intros.
    reflexivity.
    intros.
    destruct a as [x].
    destruct x.
    rewrite ActionSpiNC.new_c_cons_0.
    rewrite IHxs.
    simpl.
    destruct (low_subst_constraint (S d) k es xs);try reflexivity.
    rewrite ActionSpiNC.new_c_cons_S.
    repeat (rewrite low_subst_constraint_cons).
    rewrite low_subst_S.
    assert (deBruijn.mem_fn Name.deBruijn_fn_base 0 0 (Name.Ref (S x)) = false).
    reflexivity.
    generalize (low_subst_name_mem_fn d k es (Name.Ref (S x))).
    rewrite H;intro.
    symmetry in H0.
    destruct (deBruijn.low_subst ExpressionSpi.deBruijn_subst (S d) k es (ExpressionSpi.j (Name.Ref (S x))));try reflexivity.
    rewrite IHxs.
    destruct (low_subst_constraint (S d) k es xs);try reflexivity.
    simpl.
    destruct t.
    destruct n.
    discriminate H0.
    rewrite ActionSpiNC.new_c_cons_S.
    reflexivity.
  Qed.

  Lemma low_subst_constraint_app : forall (xs ys:list Name.t)(d k:nat)(es:list ExpressionSpi.t),(low_subst_constraint d k es (app xs ys))=match (low_subst_constraint d k es xs) with None => None | Some(xs) => option_map (fun ys => app xs ys) (low_subst_constraint d k es ys) end.
    induction xs;intros.
    simpl.
    unfold option_map.
    destruct (low_subst_constraint d k es ys);auto.
    replace (app (cons a xs) ys) with (cons a (app xs ys));try reflexivity.
    rewrite low_subst_constraint_cons.
    rewrite low_subst_constraint_cons.
    rewrite IHxs.
    destruct (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es (ExpressionSpi.j a));auto.
    destruct (low_subst_constraint d k es xs);auto.
    simpl.
    destruct (low_subst_constraint d k es ys);auto.
  Qed.

  Lemma low_subst_constraint_eval_nc : forall (phi:FormulaeSpi.t)(l:list Name.t),(EvalSpi.eval_nc phi = (true,l))->forall (ms:list MessageSpi.t)(d k:nat)(l':list Name.t),(low_subst_constraint d k (SubstSpi.msg_subst ms) l = Some l')->(EvalSpi.eval_nc (deBruijn.low_subst FormulaeSpi.deBruijn_subst d k (SubstSpi.msg_subst ms) phi) = (true,l')).
    induction phi;intros.
    simpl in H.
    rewrite FormulaeSpi.low_subst_FName.
    unfold EvalSpi.eval_nc.
    generalize (SubstSpi.low_subst_msg_evalc ms t);intros.
    destruct (EvalSpi.evalc t);try (discriminate H).
    rewrite (H1 t0);auto.
    destruct t0;try (discriminate H).
    injection H;clear H;intros;subst l.
    rewrite low_subst_constraint_cons in H0.
    rewrite low_subst_constraint_nil in H0.
    unfold option_map in H0.
    generalize (SubstSpi.low_subst_expr_of_msg (MessageSpi.MNam t0) d k ms);intro.
    replace (EvalSpi.expr_of_mesg (MessageSpi.MNam t0)) with (ExpressionSpi.j t0) in H;try reflexivity.
    rewrite <- H in H0.
    destruct (deBruijn.low_subst MessageSpi.deBruijn_subst d k ms (MessageSpi.MNam t0));try (discriminate H0).
    simpl in H0.
    injection H0;intros.
    subst l';auto.
    simpl in H.
    rewrite FormulaeSpi.low_subst_FEqual.
    unfold EvalSpi.eval_nc.
    generalize (SubstSpi.low_subst_msg_evalc ms t);intros.
    generalize (SubstSpi.low_subst_msg_evalc ms t0);intros.
    destruct (EvalSpi.evalc t);destruct (EvalSpi.evalc t0);try (discriminate H).
    rewrite (H1 t1);auto.
    rewrite (H2 t2);auto.
    generalize H;clear H.
    destruct (MessageSpi.Ord.compare t1 t2);intro H;try (discriminate H).
    rewrite e.
    case MessageSpi.Ord.compare;intros.
    elim (MessageSpi.Ord.lt_not_eq (deBruijn.low_subst MessageSpi.deBruijn_subst d k ms t2) (deBruijn.low_subst MessageSpi.deBruijn_subst d k ms t2));auto.
    injection H;clear H;intros;subst l.
    simpl in H0.
    injection H0;clear H0;intros;subst l';auto.
    elim (MessageSpi.Ord.lt_not_eq (deBruijn.low_subst MessageSpi.deBruijn_subst d k ms t2) (deBruijn.low_subst MessageSpi.deBruijn_subst d k ms t2));auto.
  Qed.
  
  Definition low_subst_action (d k:nat)(es:list ExpressionSpi.t)(a:ActionSpi.t) : option ActionSpi.t :=
    match a with
      | ActionSpi.Tau => Some ActionSpi.Tau
      | ActionSpi.Inp a => 
	(match deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es (ExpressionSpi.j a) with
	   | ExpressionSpi.ENam b =>  Some (ActionSpi.Inp b)
	   | _ => None
	 end)
      | ActionSpi.Outp a => 
	(match deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es (ExpressionSpi.j a) with
	   | ExpressionSpi.ENam b =>  Some (ActionSpi.Outp b)
	   | _ => None
	 end)
    end.

  Lemma low_subst_action_mem_fn : forall (a:ActionSpi.t),deBruijn.mem_fn ActionSpi.deBruijn_fn_base 0 0 a = false ->forall (d k:nat)(es:list ExpressionSpi.t)(b:ActionSpi.t),(low_subst_action (S d) k es a = Some b)->deBruijn.mem_fn ActionSpi.deBruijn_fn_base 0 0 b = false.
    destruct a;unfold low_subst_action.
    intros.
    injection H0;intros;subst b.
    reflexivity.
    intros.
    generalize (low_subst_name_mem_fn d k es t).
    destruct (deBruijn.low_subst ExpressionSpi.deBruijn_subst (S d) k es (ExpressionSpi.j t));try (discriminate H0).
    injection H0;intros;subst b.
    simpl.
    simpl in H2.
    simpl in H.
    rewrite <- H2;auto.
    intros.
    generalize (low_subst_name_mem_fn d k es t).
    destruct (deBruijn.low_subst ExpressionSpi.deBruijn_subst (S d) k es (ExpressionSpi.j t));try (discriminate H0).
    injection H0;intros;subst b.
    simpl.
    simpl in H2.
    simpl in H.
    rewrite <- H2;auto.
  Qed.

  Lemma mk_res_low_subst_action : forall (mu mu':ActionSpi.t),(ActionSpi.mk_res mu = Some mu')->forall (d k:nat)(es:list ExpressionSpi.t)(mu0:ActionSpi.t),(low_subst_action (S d) k es mu = Some mu0)->exists mu1:ActionSpi.t,(ActionSpi.mk_res mu0 = Some mu1 /\ (low_subst_action d k es mu' = Some mu1)).
    intros.
    unfold ActionSpi.mk_res in H.
    unfold ActionSpi.mk_res.
    generalize (low_subst_action_mem_fn mu);intro.
    case (Bool_elim_bool (deBruijn.mem_fn ActionSpi.deBruijn_fn_base 0 0 mu));intro H2;rewrite H2 in H;try (discriminate H).
    rewrite (H1 H2 d k es);auto.
    exists (deBruijn.low1 ActionSpi.deBruijn_base 0 mu0).
    split;auto.
    assert (mu' = (deBruijn.low1 ActionSpi.deBruijn_base 0 mu)).
    injection H;auto.
    rewrite H3.
    destruct mu.
    simpl in H0.
    injection H0;intros;subst mu0.
    reflexivity.
    generalize (deBruijn.low_subst_low1_comm ExpressionSpi.deBruijn_fn_subst 0 d k es (ExpressionSpi.j t)).
    fold ExpressionSpi.deBruijn_fn_base;fold ExpressionSpi.deBruijn_subst;fold ExpressionSpi.deBruijn_base.
    intros.
    simpl in H2.
    generalize (H4 H2);clear H4;intro.
    repeat (rewrite plus_0_l in H4).
    simpl.
    simpl in H4.
    rewrite H4.
    simpl in H0.
    destruct (Name.low_subst_generic ExpressionSpi.t (fun d k : nat => ExpressionSpi.t_map (fun d0 : nat => Name.lift d0 k) d) ExpressionSpi.j (S d) k es t);try (discriminate H0).
    injection H0;intros.
    subst mu0.
    reflexivity.
    generalize (deBruijn.low_subst_low1_comm ExpressionSpi.deBruijn_fn_subst 0 d k es (ExpressionSpi.j t)).
    fold ExpressionSpi.deBruijn_fn_base;fold ExpressionSpi.deBruijn_subst;fold ExpressionSpi.deBruijn_base.
    intros.
    simpl in H2.
    generalize (H4 H2);clear H4;intro.
    repeat (rewrite plus_0_l in H4).
    simpl.
    simpl in H4.
    rewrite H4.
    simpl in H0.
    destruct (Name.low_subst_generic ExpressionSpi.t (fun d k : nat => ExpressionSpi.t_map (fun d0 : nat => Name.lift d0 k) d) ExpressionSpi.j (S d) k es t);try (discriminate H0).
    injection H0;intros.
    subst mu0.
    reflexivity.
  Qed.

  Lemma mk_interact_low_subst_action : forall (mu1 mu2 mu:ActionSpi.t),(ActionSpi.mk_interact mu1 mu2 = Some mu)->forall (d k:nat)(es:list ExpressionSpi.t)(mu1' mu2':ActionSpi.t),(low_subst_action d k es mu1 = Some mu1')->(low_subst_action d k es mu2 = Some mu2')->exists mu':ActionSpi.t,(ActionSpi.mk_interact mu1' mu2' = Some mu' /\ (low_subst_action d k es mu = Some mu')).
    intros.
    destruct mu1;destruct mu2;try (discriminate H);unfold ActionSpi.mk_interact in H.
    unfold ActionSpi.mk_interact.
    lazy beta delta [low_subst_action] iota in H0.
    lazy beta delta [low_subst_action] iota in H1.
    destruct (Name.Ord.compare t t0);try (discriminate H).
    rewrite <- e in H1.
    destruct (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es (ExpressionSpi.j t));try (discriminate H0);try (discriminate H1).
    injection H0;injection H1;intros.
    subst mu1';subst mu2'.
    injection H;intros;subst mu.
    simpl.
    exists ActionSpi.Tau.
    split;auto.
    case Name.Ord.compare;intros;auto;elim (Name.Ord.lt_not_eq t1 t1);auto.
    unfold ActionSpi.mk_interact.
    lazy beta delta [low_subst_action] iota in H0.
    lazy beta delta [low_subst_action] iota in H1.
    destruct (Name.Ord.compare t0 t);try (discriminate H).
    rewrite e in H1.
    destruct (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es (ExpressionSpi.j t));try (discriminate H0);try (discriminate H1).
    injection H0;injection H1;intros.
    subst mu1';subst mu2'.
    injection H;intros;subst mu.
    simpl.
    exists ActionSpi.Tau.
    split;auto.
    case Name.Ord.compare;intros;auto;elim (Name.Ord.lt_not_eq t1 t1);auto.
  Qed.

  Lemma reductionNC_low_subst : forall (p:SpiCalcul.t)(mu0:ActionSpiNC.t)(a:AgentSpi.t),(RedSpiNC.reduction p mu0 a)->forall (mu:ActionSpi.t)(c:list Name.t),mu0=(mu,c)->forall (d k:nat)(sigma:list MessageSpi.t)(c':list Name.t),(low_subst_constraint d k (SubstSpi.msg_subst sigma) c = Some c')->exists mu':ActionSpi.t,((low_subst_action d k (SubstSpi.msg_subst sigma) mu = Some mu')/\RedSpiNC.reduction (deBruijn.low_subst SpiCalcul.deBruijn_subst d k (SubstSpi.msg_subst sigma) p) (mu',c') (deBruijn.low_subst AgentSpi.deBruijn_subst d k (SubstSpi.msg_subst sigma) a)).
    induction 1;intros.
    (* RedSilent *)
    rewrite SpiCalcul.low_subst_Tau.
    exists ActionSpi.mk_silent.
    unfold ActionSpiNC.mk_silent in H.
    injection H;clear H;intros.
    subst mu;subst c.
    rewrite AgentSpi.low_subst_Pr.
    injection H0;clear H0;intros;subst c'.
    split.
    reflexivity.
    apply RedSpiNC.RedSilent.
    (* RedInput *)
    unfold ActionSpiNC.mk_input in H.
    generalize (SubstSpi.low_subst_msg_evalc sigma e);intro.
    destruct (EvalSpi.evalc e);try (discriminate H).
    destruct t;try (discriminate H).
    injection H;clear H;intros;subst mu.
    injection H;clear H;intros;subst mu0;subst c.
    unfold low_subst_action.
    rewrite low_subst_constraint_cons in H1.
    rewrite low_subst_constraint_nil in H1.
    generalize (H2 (MessageSpi.MNam t) (refl_equal (Some (MessageSpi.MNam t))) d k);clear H2;intro.
    generalize (SubstSpi.low_subst_expr_of_msg (MessageSpi.MNam t) d k sigma).
    replace (EvalSpi.expr_of_mesg (MessageSpi.MNam t)) with (ExpressionSpi.j t);try reflexivity.
    intro.
    destruct (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k (SubstSpi.msg_subst sigma) (ExpressionSpi.j t));try (discriminate H1).
    simpl in H1.
    injection H1;clear H1;intros;subst c'.
    rewrite SpiCalcul_low_subst_Input_aux.
    rewrite AgentSpi.low_subst_F.
    exists (ActionSpi.Inp t0).
    split;auto.
    apply RedSpiNC.RedInput.
    unfold ActionSpiNC.mk_input.
    rewrite H.
    destruct (deBruijn.low_subst MessageSpi.deBruijn_subst d k sigma (MessageSpi.MNam t));try (discriminate H0).
    injection H0;clear H0;intros.
    subst t1;auto.
    (* RedOutput *)
    unfold ActionSpiNC.mk_output in H.
    generalize (SubstSpi.low_subst_msg_evalc sigma e);intro.
    generalize (SubstSpi.low_subst_msg_evalc sigma f);intro.
    destruct (EvalSpi.evalc e);destruct (EvalSpi.evalc f);try (discriminate H).
    destruct t;try (discriminate H).
    injection H;clear H;intros;subst mu.
    injection H4;clear H4;intros;subst mu0;subst c.
    unfold low_subst_action.
    rewrite low_subst_constraint_cons in H1.
    rewrite low_subst_constraint_nil in H1.
    generalize (H2 (MessageSpi.MNam t) (refl_equal (Some (MessageSpi.MNam t))) d k);clear H2;intro.
    generalize (SubstSpi.low_subst_expr_of_msg (MessageSpi.MNam t) d k sigma).
    replace (EvalSpi.expr_of_mesg (MessageSpi.MNam t)) with (ExpressionSpi.j t);try reflexivity.
    intro.
    destruct (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k (SubstSpi.msg_subst sigma) (ExpressionSpi.j t));try (discriminate H1).
    simpl in H1.
    injection H1;clear H1;intros;subst c'.
    rewrite SpiCalcul_low_subst_Output_aux.
    rewrite AgentSpi_low_subst_C_aux.
    rewrite plus_0_r.
    exists (ActionSpi.Outp t1).
    split;auto.
    apply RedSpiNC.RedOutput.
    unfold ActionSpiNC.mk_output.
    rewrite H0.
    destruct (deBruijn.low_subst MessageSpi.deBruijn_subst d k sigma (MessageSpi.MNam t));try (discriminate H2).
    injection H2;clear H2;intros.
    subst t1.
    rewrite (H3 t0);auto.
    subst m.
    rewrite (SubstSpi.low_subst_expr_of_msg t0 d k sigma).
    auto.
    destruct t;discriminate H.
    (* RedIfThen *)
    destruct mu as [mu1 c1].
    unfold ActionSpiNC.mk_if in H0.
    generalize (low_subst_constraint_eval_nc phi);intro.
    destruct (EvalSpi.eval_nc phi).
    destruct b;try (discriminate H0).
    generalize (H3 l (refl_equal (true,l)));clear H3;intro.
    injection H0;clear H0;intros;subst mu'.
    injection H0;clear H0;intros;subst mu1;subst c.
    rewrite low_subst_constraint_app in H2.
    generalize (IHreduction mu0 c1 (refl_equal (mu0,c1)));clear IHreduction;intro.
    generalize (H0 d k sigma);clear H0;intro.
    destruct (low_subst_constraint d k (SubstSpi.msg_subst sigma) c1);try (discriminate H2).
    elim (H0 l0);clear H0;auto.
    intro mu1;intros;and_elim.
    exists mu1.
    split;auto.
    rewrite SpiCalcul_low_subst_IfThen_aux.
    eapply RedSpiNC.RedIfThen;eauto.
    unfold ActionSpiNC.mk_if.
    generalize (H3 sigma d k);clear H3;intro.
    destruct (low_subst_constraint d k (SubstSpi.msg_subst sigma) l);try (discriminate H2).
    injection H2;clear H2;intros;subst c'.
    rewrite (H3 l1);auto.
    (* RedSumL *)
    elim (IHreduction mu0 c H0 d k sigma c' H1).
    intro mu';intros;and_elim.
    exists mu'.
    split;auto.
    rewrite SpiCalcul.low_subst_Sum.
    apply RedSpiNC.RedSumL;auto.
    (* RedSumR *)
    elim (IHreduction mu0 c H0 d k sigma c' H1).
    intro mu';intros;and_elim.
    exists mu'.
    split;auto.
    rewrite SpiCalcul.low_subst_Sum.
    apply RedSpiNC.RedSumR;auto.
    (* RedRes *)
    rewrite SpiCalcul.low_subst_New.
    rewrite AgentSpi.low_subst_new.
    unfold ActionSpiNC.mk_res in H0.
    destruct mu as [mu1 c1].
    case (Option_elim_option ActionSpi.t (ActionSpi.mk_res mu1));intro.
    ex_elim s mu'1.
    rewrite s in H0.
    injection H0;clear H0;intros.
    subst mu'.
    injection H0;clear H0;intros.
    subst mu'1;subst c.
    rewrite low_subst_constraint_new_c in H2.
    generalize (IHreduction mu1 c1 (refl_equal (mu1,c1)) (S d) k sigma);clear IHreduction;intros.
    destruct (low_subst_constraint (S d) k (SubstSpi.msg_subst sigma) c1);try (discriminate H2).
    elim (H0 l);auto;clear H0.
    intro mu'1;intros;and_elim.
    elim (mk_res_low_subst_action mu1 mu0 s d k (SubstSpi.msg_subst sigma) mu'1 H0).
    intro mu'0;intros;and_elim.
    exists mu'0.
    split;auto.
    eapply RedSpiNC.RedRes;eauto.
    unfold ActionSpiNC.mk_res.
    rewrite H3.
    injection H2;intros;subst c'.
    auto.
    rewrite e in H0;discriminate H0.
    (* RedParL *)
    elim (IHreduction mu0 c H0 d k sigma c' H1).
    intro mu'0;intros;and_elim.
    rewrite SpiCalcul.low_subst_Par.
    rewrite AgentSpi.low_subst_par_r.
    exists mu'0.
    split;auto.
    apply RedSpiNC.RedParL;auto.
    (* RedParR *)
    elim (IHreduction mu0 c H0 d k sigma c' H1).
    intro mu'0;intros;and_elim.
    rewrite SpiCalcul.low_subst_Par.
    rewrite AgentSpi.low_subst_par_l.
    exists mu'0.
    split;auto.
    apply RedSpiNC.RedParR;auto.
    (* RedCloseL *)
    destruct inp as [inp cinp].
    destruct out as [out cout].
    unfold ActionSpiNC.mk_interact in H1.
    subst tau.
    case (Option_elim_option ActionSpi.t (ActionSpi.mk_interact inp out));intro H2;try (rewrite H2 in H1;discriminate H1).
    ex_elim H2 tau.
    rewrite H2 in H1.
    generalize (RedSpiNC.reduction_abstraction_input H f (refl_equal (ActionSpiNC.A.F f)));intro.
    unfold ActionSpiNC.get_kind in H4.
    destruct (ActionSpi.get_kind inp);try (discriminate H4).
    clear H4.
    injection H1;clear H1;intros.
    subst mu;subst c0.
    rewrite low_subst_constraint_app in H3.
    generalize (IHreduction1 inp cinp (refl_equal (inp,cinp)) d k sigma);clear IHreduction1.
    generalize (IHreduction2 out cout (refl_equal (out,cout)) d k sigma);clear IHreduction2.
    intros.
    destruct (low_subst_constraint d k (SubstSpi.msg_subst sigma) cinp);try (discriminate H3).
    destruct (low_subst_constraint d k (SubstSpi.msg_subst sigma) cout);try (discriminate H3).
    injection H3;clear H3;intros.
    subst c'.
    elim (H1 l0);auto;clear H1.
    intro out';intros;and_elim.
    elim (H4 l);auto;clear H4.
    intro inp';intros;and_elim.
    rewrite SpiCalcul.low_subst_Par.
    destruct f as [p'].
    destruct c as [n m q'].
    rewrite AgentSpi.low_subst_pseudo_appl_l.
    elim (mk_interact_low_subst_action inp out tau H2 d k (SubstSpi.msg_subst sigma) inp' out' H4 H1).
    intro tau';intros;and_elim.
    exists tau'.
    split;auto.
    eapply RedSpiNC.RedCloseL;eauto.
    unfold ActionSpiNC.mk_interact.
    rewrite H6.
    rewrite AgentSpi.low_subst_F in H5.
    generalize (RedSpiNC.reduction_abstraction_input H5 (AgentSpi.Abs (deBruijn.low_subst AgentSpi.P.deBruijn_subst (S d) k (SubstSpi.msg_subst sigma) p')) (refl_equal (AgentSpi.F (AgentSpi.Abs (deBruijn.low_subst AgentSpi.P.deBruijn_subst (S d) k (SubstSpi.msg_subst sigma) p')))));intro.
    unfold ActionSpiNC.get_kind in H8.
    destruct (ActionSpi.get_kind inp');try (discriminate H8).
    auto.
    (* RedCloseR *)
    destruct inp as [inp cinp].
    destruct out as [out cout].
    unfold ActionSpiNC.mk_interact in H1.
    subst tau.
    case (Option_elim_option ActionSpi.t (ActionSpi.mk_interact out inp));intro H2;try (rewrite H2 in H1;discriminate H1).
    ex_elim H2 tau.
    rewrite H2 in H1.
    generalize (RedSpiNC.reduction_concretion_output H c (refl_equal (ActionSpiNC.A.C c)));intro.
    unfold ActionSpiNC.get_kind in H4.
    destruct (ActionSpi.get_kind out);try (discriminate H4).
    clear H4.
    injection H1;clear H1;intros.
    subst mu;subst c0.
    rewrite low_subst_constraint_app in H3.
    generalize (IHreduction2 inp cinp (refl_equal (inp,cinp)) d k sigma);clear IHreduction2.
    generalize (IHreduction1 out cout (refl_equal (out,cout)) d k sigma);clear IHreduction1.
    intros.
    destruct (low_subst_constraint d k (SubstSpi.msg_subst sigma) cinp);try (discriminate H3).
    destruct (low_subst_constraint d k (SubstSpi.msg_subst sigma) cout);try (discriminate H3).
    injection H3;clear H3;intros.
    subst c'.
    elim (H1 l0);auto;clear H1.
    intro out';intros;and_elim.
    elim (H4 l);auto;clear H4.
    intro inp';intros;and_elim.
    rewrite SpiCalcul.low_subst_Par.
    destruct f as [p'].
    destruct c as [n m q'].
    rewrite AgentSpi.low_subst_pseudo_appl_r.
    elim (mk_interact_low_subst_action out inp tau H2 d k (SubstSpi.msg_subst sigma) out' inp' H1 H4).
    intro tau';intros;and_elim.
    exists tau'.
    split;auto.
    eapply RedSpiNC.RedCloseR;eauto.
    unfold ActionSpiNC.mk_interact.
    rewrite H6.
    rewrite AgentSpi.low_subst_C in H3.
    generalize (RedSpiNC.reduction_concretion_output H3 (AgentSpi.Conc n (deBruijn.low_subst AgentSpi.P.E.deBruijn_subst (d + n) k (SubstSpi.msg_subst sigma) m) (deBruijn.low_subst AgentSpi.P.deBruijn_subst (d + n) k (SubstSpi.msg_subst sigma) q')) (refl_equal (AgentSpi.C (AgentSpi.Conc n (deBruijn.low_subst AgentSpi.P.E.deBruijn_subst (d + n) k (SubstSpi.msg_subst sigma) m) (deBruijn.low_subst AgentSpi.P.deBruijn_subst (d + n) k (SubstSpi.msg_subst sigma) q')))));intro.
    unfold ActionSpiNC.get_kind in H8.
    destruct (ActionSpi.get_kind out');try (discriminate H8).
    auto.
    (* RedBangAct *)
    elim (IHreduction mu0 c H0 d k sigma c' H1).
    intro mu'0;intros;and_elim.
    rewrite SpiCalcul.low_subst_Bang.
    rewrite AgentSpi.low_subst_par_r.
    rewrite SpiCalcul.low_subst_Bang.
    exists mu'0.
    split;auto.
    apply RedSpiNC.RedBangAct;auto.
    (* RedBangCloseL *)
    destruct inp as [inp cinp].
    destruct out as [out cout].
    unfold ActionSpiNC.mk_interact in H1.
    subst tau.
    case (Option_elim_option ActionSpi.t (ActionSpi.mk_interact inp out));intro H2;try (rewrite H2 in H1;discriminate H1).
    ex_elim H2 tau.
    rewrite H2 in H1.
    generalize (RedSpiNC.reduction_abstraction_input H f (refl_equal (ActionSpiNC.A.F f)));intro.
    unfold ActionSpiNC.get_kind in H4.
    destruct (ActionSpi.get_kind inp);try (discriminate H4).
    clear H4.
    injection H1;clear H1;intros.
    subst mu;subst c0.
    rewrite low_subst_constraint_app in H3.
    generalize (IHreduction1 inp cinp (refl_equal (inp,cinp)) d k sigma);clear IHreduction1.
    generalize (IHreduction2 out cout (refl_equal (out,cout)) d k sigma);clear IHreduction2.
    intros.
    destruct (low_subst_constraint d k (SubstSpi.msg_subst sigma) cinp);try (discriminate H3).
    destruct (low_subst_constraint d k (SubstSpi.msg_subst sigma) cout);try (discriminate H3).
    injection H3;clear H3;intros.
    subst c'.
    elim (H1 l0);auto;clear H1.
    intro out';intros;and_elim.
    elim (H4 l);auto;clear H4.
    intro inp';intros;and_elim.
    rewrite SpiCalcul.low_subst_Bang.
    rewrite AgentSpi.low_subst_par_r.
    rewrite SpiCalcul.low_subst_Bang.
    destruct f as [p'].
    destruct c as [n m q'].
    rewrite AgentSpi.low_subst_pseudo_appl_l.
    elim (mk_interact_low_subst_action inp out tau H2 d k (SubstSpi.msg_subst sigma) inp' out' H4 H1).
    intro tau';intros;and_elim.
    exists tau'.
    split;auto.
    eapply RedSpiNC.RedBangCloseL;eauto.
    unfold ActionSpiNC.mk_interact.
    rewrite H6.
    rewrite AgentSpi.low_subst_F in H5.
    generalize (RedSpiNC.reduction_abstraction_input H5 (AgentSpi.Abs (deBruijn.low_subst AgentSpi.P.deBruijn_subst (S d) k (SubstSpi.msg_subst sigma) p')) (refl_equal (AgentSpi.F (AgentSpi.Abs (deBruijn.low_subst AgentSpi.P.deBruijn_subst (S d) k (SubstSpi.msg_subst sigma) p')))));intro.
    unfold ActionSpiNC.get_kind in H8.
    destruct (ActionSpi.get_kind inp');try (discriminate H8).
    auto.
  Qed.
End ReductionSpiNCprops.


