(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import name.
Require Import expressionSpi.
Require Import messageSpi.
Require Import formulaeSpi.
Require Import deBruijnType.
Require Import actionSpi.
Require Import actionSpiNC.
Require Import actionSpiSymb.
Require Import missing.
Require Import tactics.
Require Import reduction.
Require Import calculi.
Require Import evalSpi.
Require Import Arith.
Require Import reductionSpi.
Require Import substSpi.
Require Import workAround.
Require Import reductionSpiNCprops.
Require Import roSpi.
Require Import fnInMsg.

Module ReductionSpiSymbprops.
  Definition nb_redex_action_symb (a:ActionSpiSymb.t) :=
    let (a,_) := a in
      match a with
        | ActionSpiSymb.Tau => 0%nat
        | ActionSpiSymb.Inp e => EvalSpi.nb_redex e
        | ActionSpiSymb.Outp e => EvalSpi.nb_redex e
      end.

  Definition nb_redex_agent (a:AgentSpi.t) :=
    match a with
      | AgentSpi.Pr _ => 0%nat
      | AgentSpi.F _ => 0%nat
      | AgentSpi.C (AgentSpi.Conc _ e _) => EvalSpi.nb_redex e
    end. 

  Lemma reduction_spi_symb_nb_redex : forall (p:SpiCalcul.t)(mu:ActionSpiSymb.t)(a:AgentSpi.t),(RedSpiSymb.reduction p mu a)->(nb_redex_action_symb mu = 0%nat)/\(nb_redex_agent a = 0%nat).
    induction 1;intros;and_elim;simpl;auto.
    split;auto.
    unfold ActionSpiSymb.mk_input in H.
    injection H;intros.
    subst mu;simpl.
    apply EvalSpi.evala_no_redex.
    unfold ActionSpiSymb.mk_output in H.
    injection H;intros.
    subst mu;subst m.
    simpl.
    split;apply EvalSpi.evala_no_redex.
    unfold ActionSpiSymb.mk_if in H0.
    destruct mu.
    injection H0;intros.
    subst mu'.
    simpl in IHreduction.
    simpl.
    auto.
    unfold ActionSpiSymb.mk_res in H0.
    destruct mu.
    destruct (deBruijn.mem_fn ActionSpiSymb.deBruijn_fn_base_action 0 0 a0);try (discriminate H0).
    injection H0;intros;subst mu'.
    simpl.
    simpl in IHreduction.
    split.
    destruct a0;simpl;auto.
    rewrite EvalSpi.nb_redex_map;auto.
    rewrite EvalSpi.nb_redex_map;auto.
    clear H0.
    destruct a;simpl;auto.
    destruct c0;simpl in H1.
    rewrite ActionSpiSymb.A.new_conc_def.
    case deBruijn.mem_fn;auto.
    simpl.
    rewrite EvalSpi.nb_redex_map;auto.
    split;auto.
    destruct a;simpl;auto.
    destruct c;simpl in H0;simpl;auto.
    split;auto.
    destruct a;simpl;auto.
    destruct c;simpl in H0;simpl;auto.
    unfold ActionSpiSymb.mk_interact in H1.
    destruct inp;destruct out.
    destruct a;destruct a0;try (discriminate H1);injection H1;clear H1;intros;subst tau;simpl;split;auto.
    destruct f;destruct c;simpl;auto.
    destruct f;destruct c;simpl;auto.
    unfold ActionSpiSymb.mk_interact in H1.
    destruct inp;destruct out.
    destruct a;destruct a0;try (discriminate H1);injection H1;clear H1;intros;subst tau;simpl;split;auto.
    destruct f;destruct c;simpl;auto.
    destruct f;destruct c;simpl;auto.
    split;auto.
    destruct a;simpl;auto.
    destruct c;simpl;auto.
    unfold ActionSpiSymb.mk_interact in H1.
    destruct inp;destruct out.
    destruct a;destruct a0;try (discriminate H1);injection H1;clear H1;intros;subst tau;simpl;split;auto.
    destruct f;destruct c;simpl;auto.
    destruct f;destruct c;simpl;auto.
  Qed.

  Definition low_subst_constraint (d k:nat)(es:list ExpressionSpi.t)(c:ActionSpiSymb.constraint) :=
    let (n,lphi) := c in (n,deBruijn.low_subst ActionSpiSymb.deBruijn_subst_list (d+n) k es lphi).

  Lemma low_subst_constraint_cons : forall (d k:nat)(es:list ExpressionSpi.t)(n:nat)(phi:FormulaeSpi.t)(lphi:list FormulaeSpi.t),(low_subst_constraint d k es (n,cons phi lphi))=(n,cons (deBruijn.low_subst FormulaeSpi.deBruijn_subst (d+n) k es phi) (deBruijn.low_subst ActionSpiSymb.deBruijn_subst_list (d+n) k es lphi)).
    intros.
    reflexivity.
  Qed.

  Lemma low_subst_list_cons : forall (d k:nat)(es:list ExpressionSpi.t)(phi:FormulaeSpi.t)(lphi:list FormulaeSpi.t),(deBruijn.low_subst ActionSpiSymb.deBruijn_subst_list d k es (cons phi lphi))=(cons (deBruijn.low_subst FormulaeSpi.deBruijn_subst d k es phi) (deBruijn.low_subst ActionSpiSymb.deBruijn_subst_list d k es lphi)).
    intros.
    reflexivity.
  Qed.

  Lemma low_subst_list_nil : forall (d k:nat)(es:list ExpressionSpi.t),(deBruijn.low_subst ActionSpiSymb.deBruijn_subst_list d k es nil)=nil.
    intros.
    reflexivity.
  Qed.

  Lemma low_subst_constraint_union_c : forall (d k:nat)(es:list ExpressionSpi.t)(c1 c2:ActionSpiSymb.constraint),(low_subst_constraint d k es (ActionSpiSymb.union_c c1 c2))=(ActionSpiSymb.union_c (low_subst_constraint d k es c1) (low_subst_constraint d k es c2)).
    destruct c1 as [n lphi].
    destruct c2 as [m lpsi].
    unfold ActionSpiSymb.union_c.
    unfold low_subst_constraint.
    generalize (deBruijn.low_subst_lift_comm ActionSpiSymb.deBruijn_subst_list 0 (d+n) m k es lphi).
    fold ActionSpiSymb.deBruijn_subst_list.
    fold ActionSpiSymb.deBruijn_base_list.
    rewrite plus_0_l.
    rewrite plus_assoc.
    intro.
    rewrite <- H.
    clear H.
    generalize (deBruijn.low_subst_lift_comm ActionSpiSymb.deBruijn_subst_list m d n k es lpsi).
    fold ActionSpiSymb.deBruijn_subst_list.
    fold ActionSpiSymb.deBruijn_base_list.
    replace (m+d+n) with (d+n+m);try omega.
    rewrite (plus_comm m d).
    intro.
    rewrite <- H.
    simpl.
    rewrite deBruijn.map_list_app.
    auto.
  Qed.

  Lemma low_subst_constraint_new_c : forall (d k:nat)(es:list ExpressionSpi.t)(c:ActionSpiSymb.constraint),(low_subst_constraint d k es (ActionSpiSymb.new_c c))=(ActionSpiSymb.new_c (low_subst_constraint (S d) k es c)).
    intros.
    destruct c as [n lphi].
    unfold ActionSpiSymb.new_c.
    unfold low_subst_constraint.
    assert (deBruijn.mem_fn ActionSpiSymb.deBruijn_fn_base_list 0 n lphi = deBruijn.mem_fn ActionSpiSymb.deBruijn_fn_base_list 0 n (deBruijn.low_subst ActionSpiSymb.deBruijn_subst_list (n+S d) k es lphi)).
    apply (deBruijn.low_subst_fn_not_supp ActionSpiSymb.deBruijn_fn_subst_list 0 n d k es lphi (le_O_n d)).
    case (Bool_elim_bool (deBruijn.mem_fn ActionSpiSymb.deBruijn_fn_base_list 0 n lphi));intro H0;rewrite H0.
    rewrite H0 in H.
    rewrite (plus_comm (S d) n).
    rewrite <- H.
    rewrite (plus_comm n (S d)).
    rewrite <- plus_n_Sm.
    auto.
    rewrite (plus_comm (S d) n).
    rewrite <- H.
    rewrite H0.
    generalize (deBruijn.low_subst_low1_comm ActionSpiSymb.deBruijn_fn_subst_list n d k es lphi H0).
    intros.
    simpl in H1.
    simpl.
    rewrite <- H1.
    rewrite plus_comm;auto.
  Qed.

  Fixpoint nc_list (lphi:list FormulaeSpi.t) : list Name.t :=
    match lphi with
      | nil => nil
      | cons phi lphi =>
        let (_,l) := EvalSpi.eval_nc phi in
          app l (nc_list lphi)
    end.
  
  Lemma nc_list_cons_FEqual : forall (e f:ExpressionSpi.t)(lphi:list FormulaeSpi.t),(nc_list (cons (FormulaeSpi.FEqual e f) lphi))=(nc_list lphi).
    intros.
    lazy beta delta [nc_list] iota;fold nc_list.
    unfold EvalSpi.eval_nc.
    destruct (EvalSpi.evalc e);destruct (EvalSpi.evalc f);try reflexivity.
    destruct (MessageSpi.Ord.compare t t0);try reflexivity.
  Qed.

  Lemma nc_list_app : forall (lphi lpsi:list FormulaeSpi.t),(nc_list (app lphi lpsi))=(app (nc_list lphi) (nc_list lpsi)).
    induction lphi;intros;simpl.
    auto.
    destruct (EvalSpi.eval_nc a).
    rewrite app_ass.
    rewrite IHlphi.
    auto.
  Qed.

  Lemma nc_list_mem_fn_true : forall (lphi:list FormulaeSpi.t)(d k:nat),(deBruijn.mem_fn Name.deBruijn_fn_base_list k d (nc_list lphi) = true)->(deBruijn.mem_fn ActionSpiSymb.deBruijn_fn_base_list k d lphi = true).
    induction lphi.
    intros.
    discriminate H.
    lazy beta delta [nc_list] iota;fold nc_list.
    intros.
    generalize (EvalSpi.eval_nc_mem_fn_true a);intro.
    destruct (EvalSpi.eval_nc a).
    unfold Name.deBruijn_fn_base_list in H.
    rewrite deBruijn.mem_fn_app in H.
    fold Name.deBruijn_fn_base_list in H.
    unfold snd in H0.
    case (orb_prop (deBruijn.mem_fn Name.deBruijn_fn_base_list k d l) (deBruijn.mem_fn Name.deBruijn_fn_base_list k d (nc_list lphi)) H);intro.
    unfold ActionSpiSymb.deBruijn_fn_base_list.
    rewrite deBruijn.mem_fn_cons.
    rewrite H0;auto.
    unfold ActionSpiSymb.deBruijn_fn_base_list.
    rewrite deBruijn.mem_fn_cons.
    rewrite IHlphi;auto.
    rewrite orb_b_true;auto.
  Qed.

  Lemma nc_list_mem_fn_false : forall (lphi:list FormulaeSpi.t)(d k:nat),(deBruijn.mem_fn ActionSpiSymb.deBruijn_fn_base_list k d lphi = false)->(deBruijn.mem_fn Name.deBruijn_fn_base_list k d (nc_list lphi) = false).
    intros.
    case (Bool_elim_bool (deBruijn.mem_fn Name.deBruijn_fn_base_list k d (nc_list lphi)));intro;auto.
    rewrite nc_list_mem_fn_true in H;auto.
    discriminate H.
  Qed.

  Fixpoint mk_nu_nc_list (n:nat)(l:list Name.t) {struct n} : list Name.t :=
    match n with
      | O => l
      | S n => ActionSpiNC.new_c (mk_nu_nc_list n l)
    end.

  Lemma mk_nu_nc_list_app : forall (n:nat)(l l':list Name.t),(mk_nu_nc_list n (app l l'))=(app (mk_nu_nc_list n l) (mk_nu_nc_list n l')).
    induction n.
    simpl;intros.
    auto.
    intros.
    simpl.
    rewrite IHn.
    rewrite ActionSpiNC.new_c_app.
    auto.
  Qed.

  Definition nc_of_constraint (c:ActionSpiSymb.constraint) :=
    let (n,lphi) := c in
      mk_nu_nc_list n (nc_list lphi).

  Lemma mk_nu_nc_list_nil : forall (n:nat),(mk_nu_nc_list n nil)=nil.
    induction n;simpl;auto.
    rewrite IHn.
    rewrite ActionSpiNC.new_c_nil.
    auto.
  Qed.

  Lemma nc_of_constraint_nil : forall (n:nat),(nc_of_constraint (n,nil))=nil.
    simpl;intros.
    apply mk_nu_nc_list_nil.
  Qed.

  Lemma nc_of_constraint_cons_FEqual : forall (n:nat)(e f:ExpressionSpi.t)(lphi:list FormulaeSpi.t),(nc_of_constraint (n,(cons (FormulaeSpi.FEqual e f) lphi))) = (nc_of_constraint (n,lphi)).
    intros.
    unfold nc_of_constraint.
    rewrite nc_list_cons_FEqual.
    auto.
  Qed.

  Lemma mk_nu_nc_cons_keep_aux : forall (n:nat)(k:nat)(l:list Name.t),(mk_nu_nc_list n (cons (Name.Ref (n+k)) l))=(cons (Name.Ref k) (mk_nu_nc_list n l)).
    induction n.
    simpl;intros.
    auto.
    intros.
    simpl.
    rewrite plus_n_Sm.
    rewrite IHn.
    rewrite ActionSpiNC.new_c_cons_S.
    auto.
  Qed.

  Lemma mk_nu_nc_cons_keep : forall (n:nat)(i:nat)(l:list Name.t),(n <= i)->(mk_nu_nc_list n (cons (Name.Ref i) l))=(cons (Name.Ref (i-n)) (mk_nu_nc_list n l)).
    intros.
    assert (i=n+(i-n)).
    omega.
    rewrite H0.
    rewrite mk_nu_nc_cons_keep_aux.
    rewrite <- H0.
    auto.
  Qed.

  Lemma mk_nu_nc_cons_remove : forall (n:nat)(i:nat)(l:list Name.t),(i < n)->(mk_nu_nc_list n (cons (Name.Ref i) l))=(mk_nu_nc_list n l).
    induction n.
    intros.
    inversion H.
    intros.
    simpl.
    inversion H.
    generalize (mk_nu_nc_cons_keep_aux n 0 l).
    rewrite plus_0_r.
    intros.
    rewrite H0.
    rewrite ActionSpiNC.new_c_cons_0.
    auto.
    rewrite IHn;try omega.
    auto.
  Qed.

  Lemma mk_nu_nc_list_lift : forall (d k n:nat)(l:list Name.t),(mk_nu_nc_list (d+k+n) (deBruijn.map_list (fun d => deBruijn.lift Name.deBruijn_base d k) d l))=(mk_nu_nc_list (d+n) l).
    induction l.
    simpl.
    repeat (rewrite mk_nu_nc_list_nil);auto.
    simpl in IHl.
    simpl.
    destruct a.
    simpl.
    unfold DeBruijnNat.lift.
    case le_lt_dec;intros.
    case (le_lt_dec (d+n) n0);intro.
    repeat (rewrite mk_nu_nc_cons_keep);try omega.
    replace (n0+k-(d+k+n)) with (n0-(d+n));try omega.
    rewrite IHl;auto.
    repeat (rewrite mk_nu_nc_cons_remove);try omega;auto.
    repeat (rewrite mk_nu_nc_cons_remove);try omega;auto.
  Qed.

  Lemma nc_list_lift : forall (d k:nat)(lphi:list FormulaeSpi.t),(nc_list (deBruijn.lift ActionSpiSymb.deBruijn_base_list d k lphi))=(deBruijn.map_list (fun d => deBruijn.lift Name.deBruijn_base d k) d (nc_list lphi)).
    induction lphi.
    reflexivity.
    generalize (EvalSpi.eval_nc_map (fun d => Name.lift d k));simpl;intros.
    rewrite H.
    simpl in IHlphi.
    rewrite IHlphi.
    destruct (EvalSpi.eval_nc a);simpl.
    rewrite deBruijn.map_list_app.
    auto.
    intros.
    apply (deBruijn.lift_injective Name.deBruijn_base).
  Qed.

  Lemma nc_of_constraint_union_c : forall (c1 c2:ActionSpiSymb.constraint),app (nc_of_constraint c1) (nc_of_constraint c2) = nc_of_constraint (ActionSpiSymb.union_c c1 c2).
    destruct c1 as [n lphi].
    destruct c2 as [m lpsi].
    unfold ActionSpiSymb.union_c.
    unfold nc_of_constraint.
    rewrite nc_list_app.
    rewrite mk_nu_nc_list_app.
    repeat (rewrite nc_list_lift).
    generalize (mk_nu_nc_list_lift 0 m n).
    generalize (mk_nu_nc_list_lift m n 0);intros.
    repeat (rewrite plus_0_r in H).
    repeat (rewrite plus_0_l in H0).
    rewrite plus_comm.
    rewrite H0.
    rewrite H.
    auto.
  Qed.

  Lemma nc_list_low1 : forall (d:nat)(lphi:list FormulaeSpi.t),(deBruijn.mem_fn ActionSpiSymb.deBruijn_fn_base_list 0 d lphi = false)->(nc_list (deBruijn.low1 ActionSpiSymb.deBruijn_base_list d lphi))=(deBruijn.map_list (fun d => deBruijn.low1 Name.deBruijn_base d) d (nc_list lphi)).
    induction lphi;intros;try reflexivity.
    assert (deBruijn.mem_fn ActionSpiSymb.deBruijn_fn_base_list 0 d lphi = false).
    unfold ActionSpiSymb.deBruijn_fn_base_list in H.
    rewrite deBruijn.mem_fn_cons in H.
    fold ActionSpiSymb.deBruijn_fn_base_list in H.
    elim (orb_false_elim (deBruijn.mem_fn FormulaeSpi.deBruijn_fn_base 0 d a) (deBruijn.mem_fn ActionSpiSymb.deBruijn_fn_base_list 0 d lphi) H);auto.
    assert (deBruijn.mem_fn FormulaeSpi.deBruijn_fn_base 0 d a = false).
    unfold ActionSpiSymb.deBruijn_fn_base_list in H.
    rewrite deBruijn.mem_fn_cons in H.
    fold ActionSpiSymb.deBruijn_fn_base_list in H.
    elim (orb_false_elim (deBruijn.mem_fn FormulaeSpi.deBruijn_fn_base 0 d a) (deBruijn.mem_fn ActionSpiSymb.deBruijn_fn_base_list 0 d lphi) H);auto.
    clear H.
    simpl.
    generalize (EvalSpi.eval_nc_low1 d a H1);simpl;intros.
    rewrite H.
    destruct (EvalSpi.eval_nc a).
    simpl.
    simpl in H.
    generalize (IHlphi H0);simpl;intros.
    rewrite H2.
    simpl.
    rewrite deBruijn.map_list_app;auto.
  Qed.

  Lemma mk_nu_nc_list_mem_fn : forall (d k n:nat)(l:list Name.t),(deBruijn.mem_fn Name.deBruijn_fn_base_list k d (mk_nu_nc_list n l))=(deBruijn.mem_fn Name.deBruijn_fn_base_list k (d+n) l).
    intros d k n.
    generalize k;clear k.
    generalize d;clear d.
    induction n;intros.
    rewrite plus_0_r.
    auto.
    lazy beta delta [mk_nu_nc_list] iota;fold mk_nu_nc_list.    
    rewrite ActionSpiNC.mem_fn_new_c.
    rewrite IHn.
    replace (S d+n) with (d+S n);try omega.
    auto.
  Qed.

  Lemma mk_nu_nc_list_low1 : forall (d n:nat)(l:list Name.t),(deBruijn.mem_fn Name.deBruijn_fn_base_list 0 (d+n) l = false)->(deBruijn.low1 Name.deBruijn_base_list d (mk_nu_nc_list n l))=(mk_nu_nc_list n (deBruijn.low1 Name.deBruijn_base_list (d+n) l)).
    intros d n.
    generalize d;clear d.
    induction n;intros.
    rewrite plus_0_r.
    reflexivity.
    lazy beta delta [mk_nu_nc_list] iota;fold mk_nu_nc_list.
    rewrite ActionSpiNC.low1_new_c.
    rewrite IHn.
    replace (S d+n) with (d+S n);try omega.
    auto.
    replace (S d+n) with (d+S n);try omega.
    auto.
    rewrite mk_nu_nc_list_mem_fn.
    replace (S d+n) with (d+S n);try omega.
    auto.
  Qed.

  Lemma nc_of_constraint_new_c : forall (c:ActionSpiSymb.constraint),(ActionSpiNC.new_c (nc_of_constraint c))=(nc_of_constraint (ActionSpiSymb.new_c c)).
    destruct c as [n lphi].
    unfold ActionSpiSymb.new_c.
    case (Bool_elim_bool (deBruijn.mem_fn ActionSpiSymb.deBruijn_fn_base_list 0 n lphi));intro H0;rewrite H0.
    unfold nc_of_constraint.
    reflexivity.
    unfold nc_of_constraint.
    rewrite nc_list_low1;auto.
    unfold ActionSpiNC.new_c.
    rewrite ActionSpiNC.remove_nothing.
    rewrite mk_nu_nc_list_low1.
    simpl.
    auto.
    rewrite plus_0_l.
    apply nc_list_mem_fn_false;auto.
    rewrite mk_nu_nc_list_mem_fn.
    rewrite plus_0_l.
    apply nc_list_mem_fn_false;auto.
  Qed.

  Fixpoint eval_list (lphi:list FormulaeSpi.t) : bool :=
    match lphi with
      | nil => true
      | cons phi lphi => andb (EvalSpi.eval phi) (eval_list lphi)
    end.

  Lemma eval_list_app : forall (lphi lpsi:list FormulaeSpi.t),(eval_list (app lphi lpsi))=(andb (eval_list lphi) (eval_list lpsi)).
    induction lphi;intros;try reflexivity.
    simpl.
    rewrite IHlphi.
    rewrite andb_assoc.
    auto.
  Qed.

  Lemma eval_list_lift : forall (lphi:list FormulaeSpi.t)(d k:nat),(eval_list lphi)=(eval_list (deBruijn.lift ActionSpiSymb.deBruijn_base_list d k lphi)).
    induction lphi;intros;try reflexivity.
    lazy beta delta [eval_list] iota;fold eval_list.
    rewrite (IHlphi d k).
    rewrite <- (EvalSpi.eval_lift d k).
    reflexivity.
  Qed.

  Lemma eval_list_swapR : forall (lphi:list FormulaeSpi.t)(d k:nat),(eval_list lphi)=(eval_list (deBruijn.swapR ActionSpiSymb.deBruijn_base_list d k lphi)).
    induction lphi;intros;try reflexivity.
    lazy beta delta [eval_list] iota;fold eval_list.
    rewrite (IHlphi d k).
    rewrite <- (EvalSpi.eval_swapR d k).
    reflexivity.
  Qed.

  Lemma eval_list_low1 : forall (lphi:list FormulaeSpi.t)(d:nat),(deBruijn.mem_fn ActionSpiSymb.deBruijn_fn_base_list 0 d lphi = false)->(eval_list lphi)=(eval_list (deBruijn.low1 ActionSpiSymb.deBruijn_base_list d lphi)).
    induction lphi;intros;try reflexivity.
    unfold ActionSpiSymb.deBruijn_fn_base_list in H.
    rewrite deBruijn.mem_fn_cons in H.
    fold ActionSpiSymb.deBruijn_fn_base_list in H.
    elim (orb_false_elim (deBruijn.mem_fn FormulaeSpi.deBruijn_fn_base 0 d a) (deBruijn.mem_fn ActionSpiSymb.deBruijn_fn_base_list 0 d lphi) H);clear H;intros.
    lazy beta delta [eval_list] iota;fold eval_list.
    rewrite (IHlphi d);auto.
    rewrite <- (EvalSpi.eval_low1 d);auto.
  Qed.

  Definition eval_constraint (c:ActionSpiSymb.constraint) :=
    let (_,lphi) := c in eval_list lphi.

  Lemma eval_constraint_union_c : forall (c1 c2:ActionSpiSymb.constraint),(eval_constraint (ActionSpiSymb.union_c c1 c2))=(andb (eval_constraint c1) (eval_constraint c2)).
    unfold ActionSpiSymb.union_c.
    destruct c1 as [n lphi].
    destruct c2 as [m lpsi].
    unfold eval_constraint.
    rewrite eval_list_app.
    repeat (rewrite <- eval_list_lift).
    auto.
  Qed.

  Lemma eval_constraint_new_c : forall (c:ActionSpiSymb.constraint),(eval_constraint (ActionSpiSymb.new_c c))=(eval_constraint c).
    unfold ActionSpiSymb.new_c.
    destruct c as [n lphi].
    unfold eval_constraint.
    case (Bool_elim_bool (deBruijn.mem_fn ActionSpiSymb.deBruijn_fn_base_list 0 n lphi));intro H;rewrite H.
    auto.
    rewrite <- eval_list_low1;auto.
  Qed.

  Inductive aroEval : AgentSpi.t -> AgentSpi.t -> Prop :=
  | AROEproc : forall (p q:SpiCalcul.t),(RoSpi.pro  p q)->(aroEval (AgentSpi.Pr p) (AgentSpi.Pr q))
  | AROEabs : forall (p q:SpiCalcul.t),(RoSpi.pro p q)->(aroEval (AgentSpi.F (AgentSpi.Abs p)) (AgentSpi.F (AgentSpi.Abs q)))
  | AROEconc : forall (n:nat)(e:ExpressionSpi.t)(m:MessageSpi.t)(p q:SpiCalcul.t),(EvalSpi.evalc e = Some m)->(RoSpi.pro p q)->(aroEval (AgentSpi.C (AgentSpi.Conc n e p)) (AgentSpi.C (AgentSpi.Conc n (EvalSpi.expr_of_mesg m) q))).

  Lemma aroE_par_l : forall (a b:AgentSpi.t)(p q:SpiCalcul.t),(aroEval a b)->(RoSpi.pro p q)->(aroEval (AgentSpi.par_l p a) (AgentSpi.par_l q b)).
    induction 1;intros.
    repeat (rewrite AgentSpi.par_l_Pr).
    apply AROEproc.
    apply RoSpi.PROpar;auto.
    repeat (rewrite AgentSpi.par_l_F).
    repeat (rewrite AgentSpi.abs_par_l_def).
    apply AROEabs.
    apply RoSpi.PROpar;auto.
    apply RoSpi.pro_lift;auto.
    repeat (rewrite AgentSpi.par_l_C).
    repeat (rewrite AgentSpi.conc_par_l_def).
    apply AROEconc;auto.
    apply RoSpi.PROpar;auto.
    apply RoSpi.pro_lift;auto.
  Qed.
  
  Lemma aroE_par_r : forall (a b:AgentSpi.t)(p q:SpiCalcul.t),(aroEval a b)->(RoSpi.pro p q)->(aroEval (AgentSpi.par_r a p) (AgentSpi.par_r b q)).
    induction 1;intros.
    repeat (rewrite AgentSpi.par_r_Pr).
    apply AROEproc.
    apply RoSpi.PROpar;auto.
    repeat (rewrite AgentSpi.par_r_F).
    repeat (rewrite AgentSpi.abs_par_r_def).
    apply AROEabs.
    apply RoSpi.PROpar;auto.
    apply RoSpi.pro_lift;auto.
    repeat (rewrite AgentSpi.par_r_C).
    repeat (rewrite AgentSpi.conc_par_r_def).
    apply AROEconc;auto.
    apply RoSpi.PROpar;auto.
    apply RoSpi.pro_lift;auto.
  Qed.

  Lemma aroE_pseudo_appl_l : forall (p1 q1 p2 q2:SpiCalcul.t)(n:nat)(e:ExpressionSpi.t)(m:MessageSpi.t),(EvalSpi.evalc e = Some m)->(RoSpi.pro p1 p2)->(RoSpi.pro q1 q2)->(aroEval (AgentSpi.pseudo_appl_l (AgentSpi.Abs p1) (AgentSpi.Conc n e q1)) (AgentSpi.pseudo_appl_l (AgentSpi.Abs p2) (AgentSpi.Conc n (EvalSpi.expr_of_mesg m) q2))).
    intros.
    repeat (rewrite AgentSpi.pseudo_appl_l_def).
    apply AROEproc.
    apply RoSpi.PROmk_nu.
    apply RoSpi.PROpar;auto.
    apply RoSpi.pro_low_subst_list;auto.
    apply List_forall_cons.
    unfold EvalSpi.evalc_succeeds.
    exists m;auto.
    apply List_forall_nil.
    apply List_relation_cons.
    apply RoSpi.evalc_impl_ero;auto.
    apply List_relation_nil.
  Qed.

  Lemma aroE_pseudo_appl_r : forall (p1 q1 p2 q2:SpiCalcul.t)(n:nat)(e:ExpressionSpi.t)(m:MessageSpi.t),(EvalSpi.evalc e = Some m)->(RoSpi.pro p1 p2)->(RoSpi.pro q1 q2)->(aroEval (AgentSpi.pseudo_appl_r (AgentSpi.Conc n e p1) (AgentSpi.Abs q1)) (AgentSpi.pseudo_appl_r (AgentSpi.Conc n (EvalSpi.expr_of_mesg m) p2) (AgentSpi.Abs q2))).
    intros.
    repeat (rewrite AgentSpi.pseudo_appl_r_def).
    apply AROEproc.
    apply RoSpi.PROmk_nu.
    apply RoSpi.PROpar;auto.
    apply RoSpi.pro_low_subst_list;auto.
    apply List_forall_cons.
    unfold EvalSpi.evalc_succeeds.
    exists m;auto.
    apply List_forall_nil.
    apply List_relation_cons.
    apply RoSpi.evalc_impl_ero;auto.
    apply List_relation_nil.
  Qed.

  Inductive agent_new_c_cond : AgentSpi.t -> AgentSpi.t -> Prop :=
  | AEproc : forall (p q:SpiCalcul.t),(agent_new_c_cond (AgentSpi.Pr p) (AgentSpi.Pr q))
  | AEabs : forall (p q:SpiCalcul.t),(agent_new_c_cond (AgentSpi.F (AgentSpi.Abs p)) (AgentSpi.F (AgentSpi.Abs q)))
  | AEconc : forall (n:nat)(e f:ExpressionSpi.t)(p q:SpiCalcul.t),(deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 n e)=(deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 n f)->(agent_new_c_cond (AgentSpi.C (AgentSpi.Conc n e p)) (AgentSpi.C (AgentSpi.Conc n f q))).

  Lemma aroE_new_c : forall (a b:AgentSpi.t),(aroEval a b)->(agent_new_c_cond a b)->(aroEval (AgentSpi.new a) (AgentSpi.new b)).
    induction 1;intros.
    simpl.
    apply AROEproc.
    apply RoSpi.PROnew.
    auto.
    repeat (rewrite AgentSpi.new_F).
    repeat (rewrite AgentSpi.new_abs_def).
    apply AROEabs.
    apply RoSpi.PROnew.
    apply RoSpi.pro_swapR.
    auto.
    inversion H1.
    subst n0;subst e0;subst p0;subst q0.
    repeat (rewrite AgentSpi.new_C).
    repeat (rewrite AgentSpi.new_conc_def).
    assert (deBruijn.mem_fn AgentSpi.P.E.deBruijn_fn_base 0 n (EvalSpi.expr_of_mesg m) = deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 n (EvalSpi.expr_of_mesg m)).
    simpl;auto.
    assert (deBruijn.mem_fn AgentSpi.P.E.deBruijn_fn_base 0 n e = deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 n e).
    simpl;auto.
    rewrite H2.
    rewrite H3 in H4.
    rewrite <- H4.
    case (Bool_elim_bool (deBruijn.mem_fn AgentSpi.P.E.deBruijn_fn_base 0 n e));intro H10;rewrite H10.
    apply AROEconc;auto.
    assert (deBruijn.low1 AgentSpi.P.E.deBruijn_base n (EvalSpi.expr_of_mesg m) = EvalSpi.expr_of_mesg (deBruijn.low1 MessageSpi.deBruijn_base n m)).
    simpl.
    rewrite EvalSpi.map_expr_of_msg.
    auto.
    rewrite H5.
    apply AROEconc;auto.
    rewrite EvalSpi.evalc_low1;auto.
    rewrite H.
    simpl.
    auto.
    apply RoSpi.PROnew.
    apply RoSpi.pro_swapR.
    auto.
  Qed.

  Lemma aroE_F_inv_NC : forall (p q:SpiCalcul.t),(aroEval (AgentSpi.F (AgentSpi.Abs p)) (ActionSpiNC.A.F (ActionSpiNC.A.Abs q)))->(RoSpi.pro p q).
    intros.
    inversion H.
    auto.
  Qed.

  Implicit Arguments aroE_F_inv_NC [p q].

  Lemma aroE_C_inv_NC : forall (n n':nat)(e f:ExpressionSpi.t)(p q:SpiCalcul.t),(aroEval (AgentSpi.C (AgentSpi.Conc n e p)) (ActionSpiNC.A.C (ActionSpiNC.A.Conc n' f q)))->exists m:MessageSpi.t,(EvalSpi.expr_of_mesg m = f)/\(EvalSpi.evalc e = Some m)/\(RoSpi.pro p q)/\(n=n').
    intros.
    inversion H.
    exists m.
    split;auto.
  Qed.

  Implicit Arguments aroE_C_inv_NC [n n' e f p q].

  Definition low_subst_action_symb (d k:nat)(es:list ExpressionSpi.t)(a:ActionSpiSymb.action) :=
    match a with
      | ActionSpiSymb.Tau => ActionSpiSymb.Tau
      | ActionSpiSymb.Inp e => ActionSpiSymb.Inp (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es e)
      | ActionSpiSymb.Outp e => ActionSpiSymb.Outp (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es e)
    end.

  Lemma low_subst_action_symb_fn_not_supp : forall (i:nat)(d d' k:nat)(es:list ExpressionSpi.t)(x:ActionSpiSymb.action),(i <= d')->(deBruijn.mem_fn ActionSpiSymb.deBruijn_fn_base_action i d x = deBruijn.mem_fn ActionSpiSymb.deBruijn_fn_base_action i d (low_subst_action_symb (d+S d') k es x)).
    destruct x;unfold low_subst_action_symb;intros.
    reflexivity.
    generalize (deBruijn.low_subst_fn_not_supp ExpressionSpi.deBruijn_fn_subst i d d' k es t H).
    simpl;intros;auto.
    generalize (deBruijn.low_subst_fn_not_supp ExpressionSpi.deBruijn_fn_subst i d d' k es t H).
    simpl;intros;auto.
  Qed.

  Lemma low_subst_action_symb_low1_comm : forall (d d' k:nat)(es:list ExpressionSpi.t)(x:ActionSpiSymb.action),(deBruijn.mem_fn ActionSpiSymb.deBruijn_fn_base_action 0 d x = false)->(low_subst_action_symb (d+d') k es (deBruijn.low1 ActionSpiSymb.deBruijn_base_action d x))=(deBruijn.low1 ActionSpiSymb.deBruijn_base_action d (low_subst_action_symb (d+S d') k es x)).
    destruct x;intros.
    reflexivity.
    simpl.
    simpl in H.
    generalize (deBruijn.low_subst_low1_comm ExpressionSpi.deBruijn_fn_subst d d' k es t);simpl.
    intros.
    rewrite H0;auto.
    simpl.
    simpl in H.
    generalize (deBruijn.low_subst_low1_comm ExpressionSpi.deBruijn_fn_subst d d' k es t);simpl.
    intros.
    rewrite H0;auto.
  Qed.

  Definition evalc_action_symb (a:ActionSpiSymb.action) :=
    match a with
      | ActionSpiSymb.Tau => Some ActionSpi.Tau
      | ActionSpiSymb.Inp e => 
        (match EvalSpi.evalc e with
           | Some (MessageSpi.MNam n) => Some (ActionSpi.Inp n)
           | _ => None
         end)
      | ActionSpiSymb.Outp e => 
        (match EvalSpi.evalc e with
           | Some (MessageSpi.MNam n) => Some (ActionSpi.Outp n)
           | _ => None
         end)
    end.

  Lemma evalc_action_symb_mem_fn_true : forall (a:ActionSpiSymb.action)(b:ActionSpi.t),(evalc_action_symb a = Some b)->forall (d k:nat),(deBruijn.mem_fn ActionSpi.deBruijn_fn_base k d b)=true->(deBruijn.mem_fn ActionSpiSymb.deBruijn_fn_base_action k d a = true).
    destruct a;unfold evalc_action_symb;intros.
    injection H;intros;subst b.
    discriminate H0.
    generalize (EvalSpi.evalc_mem_fn_true t);intro.
    destruct (EvalSpi.evalc t);try (discriminate H).
    destruct t0;try (discriminate H).
    injection H;intros;subst b.
    generalize (H1 (MessageSpi.MNam t0) (refl_equal (Some (MessageSpi.MNam t0))) d k).
    intros.
    simpl in H0;simpl in H2;simpl;auto.
    generalize (EvalSpi.evalc_mem_fn_true t);intro.
    destruct (EvalSpi.evalc t);try (discriminate H).
    destruct t0;try (discriminate H).
    injection H;intros;subst b.
    generalize (H1 (MessageSpi.MNam t0) (refl_equal (Some (MessageSpi.MNam t0))) d k).
    intros.
    simpl in H0;simpl in H2;simpl;auto.
  Qed.

  Lemma evalc_action_symb_mem_fn_false : forall (a:ActionSpiSymb.action)(b:ActionSpi.t),(evalc_action_symb a = Some b)->forall (d k:nat),(deBruijn.mem_fn ActionSpiSymb.deBruijn_fn_base_action k d a = false)->(deBruijn.mem_fn ActionSpi.deBruijn_fn_base k d b)=false.
    intros.
    case (Bool_elim_bool (deBruijn.mem_fn ActionSpi.deBruijn_fn_base k d b));intro;auto.
    rewrite (evalc_action_symb_mem_fn_true a b H) in H0;auto.
    discriminate H0.
  Qed.

  Lemma evalc_action_symb_low1 : forall (a:ActionSpiSymb.action)(b:ActionSpi.t),(evalc_action_symb a = Some b)->forall (d:nat),(deBruijn.mem_fn ActionSpiSymb.deBruijn_fn_base_action 0 d a = false)->(evalc_action_symb (deBruijn.low1 ActionSpiSymb.deBruijn_base_action d a))=(Some (deBruijn.low1 ActionSpi.deBruijn_base d b)).
    intros.
    destruct a.
    simpl in H;injection H;intros;subst b.
    reflexivity.
    generalize (EvalSpi.evalc_low1 d t H0);simpl;intros.
    rewrite H1.
    simpl in H.
    destruct (EvalSpi.evalc t);try (discriminate H);simpl.
    destruct t0;try (discriminate H);simpl.
    injection H;intros;subst b;reflexivity.
    generalize (EvalSpi.evalc_low1 d t H0);simpl;intros.
    rewrite H1.
    simpl in H.
    destruct (EvalSpi.evalc t);try (discriminate H);simpl.
    destruct t0;try (discriminate H);simpl.
    injection H;intros;subst b;reflexivity.
  Qed.

  Lemma reduction_nc_mk_res_symb : forall (mu mu':ActionSpi.t),(ActionSpi.mk_res mu = Some mu')->forall (d k:nat)(es:list ExpressionSpi.t)(mus:ActionSpiSymb.action),(evalc_action_symb (low_subst_action_symb (S d) k es mus))=(Some mu)->forall (c:ActionSpiSymb.constraint),(nb_redex_action_symb (mus,c) = 0)->(deBruijn.mem_fn ActionSpiSymb.deBruijn_fn_base_action 0 0 mus=false).
    intros.
    unfold ActionSpi.mk_res in H.
    case (Bool_elim_bool (deBruijn.mem_fn ActionSpi.deBruijn_fn_base 0 0 mu)).
    intro.
    rewrite H2 in H;discriminate H.
    intro.
    destruct mus;try reflexivity.
    unfold low_subst_action_symb in H0.
    unfold evalc_action_symb in H0.
    simpl in H1.
    generalize (FnInMsg.evalc_subst_no_redex_mem_fn_same 0 0 d k es t (le_O_n d) H1);rewrite plus_0_l.    
    intros.
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (S d) k es t));try (discriminate H0).
    simpl.
    simpl in H3.
    rewrite (H3 t0);auto.
    rewrite H2 in H.
    destruct t0;try (discriminate H0).
    injection H0;intros;subst mu.
    auto.
    unfold low_subst_action_symb in H0.
    unfold evalc_action_symb in H0.
    simpl in H1.
    generalize (FnInMsg.evalc_subst_no_redex_mem_fn_same 0 0 d k es t (le_O_n d) H1);rewrite plus_0_l.    
    intros.
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst (S d) k es t));try (discriminate H0).
    simpl.
    simpl in H3.
    rewrite (H3 t0);auto.
    rewrite H2 in H.
    destruct t0;try (discriminate H0).
    injection H0;intros;subst mu.
    auto.
  Qed.
  
  Implicit Arguments reduction_nc_mk_res_symb [mu mu' d k es mus c].

  Lemma aroE_new_c_res : forall (a b:AgentSpi.t),(aroEval a b)->forall (d k:nat)(es:list ExpressionSpi.t)(a':AgentSpi.t),a=(deBruijn.low_subst AgentSpi.deBruijn_subst (S d) k es a')->(nb_redex_agent a' = 0)->(aroEval (AgentSpi.new (deBruijn.low_subst AgentSpi.deBruijn_subst (S d) k es a')) (AgentSpi.new b)).
    induction 1;intros.
    destruct a'.
    apply aroE_new_c.
    rewrite AgentSpi.low_subst_Pr.
    rewrite AgentSpi.low_subst_Pr in H0.
    rewrite <- H0.
    apply AROEproc.
    auto.
    rewrite AgentSpi.low_subst_Pr.
    apply AEproc.
    destruct a;rewrite AgentSpi.low_subst_F in H0;discriminate H0.
    destruct c;rewrite AgentSpi.low_subst_C in H0;discriminate H0.
    destruct a';try (discriminate H0).
    destruct a;rewrite AgentSpi.low_subst_F in H0.
    rewrite AgentSpi.low_subst_F.
    apply aroE_new_c.
    rewrite <- H0.
    apply AROEabs.
    auto.
    apply AEabs.
    destruct c;rewrite AgentSpi.low_subst_C in H0;discriminate H0.
    destruct a';try (discriminate H1).
    destruct a;rewrite AgentSpi.low_subst_F in H1;discriminate H1.
    destruct c;rewrite AgentSpi.low_subst_C in H1.
    simpl in H2.
    rewrite AgentSpi.low_subst_C.
    apply aroE_new_c.
    rewrite <- H1.
    apply AROEconc;auto.
    assert (n=n0).
    injection H1;intros;auto.
    subst n0.
    apply AEconc.
    assert (e=(deBruijn.low_subst AgentSpi.P.E.deBruijn_subst (S d + n) k es t)).
    injection H1;intros;auto.
    rewrite H3 in H.
    rewrite plus_comm in H.
    generalize (FnInMsg.evalc_subst_no_redex_mem_fn_same 0 n d k es t (le_O_n d) H2 m).
    intros.
    assert (deBruijn.mem_fn ExpressionSpi.deBruijn_fn_base 0 n t = deBruijn.mem_fn MessageSpi.deBruijn_fn_base 0 n m).
    apply H4.
    simpl.
    simpl in H.
    auto.
    rewrite EvalSpi.mem_fn_expr_of_msg.
    rewrite <- H5.
    generalize (deBruijn.low_subst_fn_not_supp ExpressionSpi.deBruijn_fn_subst 0 n d k es t (le_O_n d)).
    fold ExpressionSpi.deBruijn_fn_base.
    fold ExpressionSpi.deBruijn_subst.
    intros.
    rewrite plus_comm.
    rewrite H6.
    simpl;auto.
    Opaque deBruijn.deBruijn_map_struct.
  Qed.
  Transparent deBruijn.deBruijn_map_struct.

  Lemma reduction_nc_impl_reduction_symb : forall (q:SpiCalcul.t)(muc:ActionSpiNC.t)(b:AgentSpi.t),(RedSpiNC.reduction q muc b)->forall (d k:nat)(es:list ExpressionSpi.t)(p:SpiCalcul.t)(Hevalcsucceeds:List_forall EvalSpi.evalc_succeeds es),(q = deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p)->forall (mu:ActionSpi.t)(l:list Name.t),muc=(mu,l)->exists mu0:ActionSpiSymb.action,exists c:ActionSpiSymb.constraint,exists a:AgentSpi.t,RedSpiSymb.reduction p (mu0,c) a /\ (evalc_action_symb (low_subst_action_symb d k es mu0) = Some mu) /\ (nc_of_constraint (low_subst_constraint d k es c) = l) /\ (eval_constraint (low_subst_constraint d k es c) = true) /\ (aroEval (deBruijn.low_subst AgentSpi.deBruijn_subst d k es a) b).
    induction 1;intros.
    (* RedSilent *)
    destruct p0;try (discriminate H).
    rewrite SpiCalcul_low_subst_Tau_NC in H.
    assert (p = (deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p0)).
    injection H;clear H;auto.
    clear H.
    exists ActionSpiSymb.Tau.
    exists (0%nat,nil (A:=FormulaeSpi.t)).
    injection H0;clear H0;intros.
    subst mu;subst l.
    exists (AgentSpi.Pr p0).
    split.
    apply RedSpiSymb.RedSilent.
    split.
    reflexivity.
    split.
    reflexivity.
    split.
    reflexivity.
    rewrite AgentSpi.low_subst_Pr.
    apply AROEproc.
    rewrite H1.
    apply RoSpi.PROrefl.
    (* RedInput *)
    destruct p0;try (discriminate H0).
    rewrite SpiCalcul_low_subst_Input_NC in H0.
    assert (e=(deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es t)).
    injection H0;auto.
    assert (p=(deBruijn.low_subst SpiCalcul.deBruijn_subst (S d) k es p0)).
    injection H0;auto.
    clear H0.
    exists (ActionSpiSymb.Inp (EvalSpi.evala t)).
    exists (0%nat,cons (FormulaeSpi.FName t) nil).
    exists (AgentSpi.F (AgentSpi.Abs p0)).
    split.
    apply RedSpiSymb.RedInput.
    reflexivity.
    unfold low_subst_action_symb.
    rewrite low_subst_constraint_cons.
    rewrite plus_0_r.
    unfold ActionSpiNC.mk_input in H.
    case (Option_elim_option MessageSpi.t (EvalSpi.evalc e));intro H0;try (rewrite H0 in H;discriminate H).
    ex_elim H0 m.
    rewrite H0 in H.
    destruct m;try (discriminate H).
    injection H;clear H;intros;subst mu.
    injection H;clear H;intros;subst mu0;subst l.
    split.
    unfold evalc_action_symb.
    rewrite (SubstSpi.evalc_subst_evala d k es t (MessageSpi.MNam t0));auto.
    rewrite <- H2;auto.
    rewrite FormulaeSpi.low_subst_FName.
    rewrite low_subst_list_nil.
    rewrite <- H2.
    rewrite AgentSpi.low_subst_F.
    split.
    simpl.
    rewrite H0.
    rewrite <- app_nil_end.
    reflexivity.
    split.
    simpl.
    rewrite H0.
    reflexivity.
    apply AROEabs.
    rewrite H3.
    apply RoSpi.PROrefl.
    (* RedOutput *)
    destruct p0;try (discriminate H0).
    rewrite SpiCalcul_low_subst_Output_NC in H0.
    assert (e=(deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es t)).
    injection H0;auto.
    assert (f=(deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es t0)).
    injection H0;auto.
    assert (p=(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p0)).
    injection H0;auto.
    clear H0.
    exists (ActionSpiSymb.Outp (EvalSpi.evala t)).
    exists (0%nat,cons (FormulaeSpi.FName t) (cons (FormulaeSpi.Wff t0) nil)).
    exists (AgentSpi.C (AgentSpi.Conc 0%nat (EvalSpi.evala t0) p0)).
    split.
    apply RedSpiSymb.RedOutput.
    reflexivity.
    unfold low_subst_action_symb.
    rewrite low_subst_constraint_cons.
    rewrite plus_0_r.
    unfold ActionSpiNC.mk_output in H.
    case (Option_elim_option MessageSpi.t (EvalSpi.evalc e));intro H0;try (rewrite H0 in H;discriminate H).
    ex_elim H0 m'.
    rewrite H0 in H.
    destruct m';try (discriminate H).
    case (Option_elim_option MessageSpi.t (EvalSpi.evalc f));intro H5;try (rewrite H5 in H;discriminate H).
    ex_elim H5 m''.
    rewrite H5 in H.
    injection H;clear H;intros.
    subst mu;subst m.
    injection H6;clear H6;intros.
    subst mu0;subst l.
    split.
    unfold evalc_action_symb.
    rewrite (SubstSpi.evalc_subst_evala d k es t (MessageSpi.MNam t1));auto.
    rewrite <- H2;auto.
    rewrite FormulaeSpi.low_subst_FName.
    rewrite <- H2.
    rewrite AgentSpi.low_subst_C.
    repeat (rewrite low_subst_list_cons).
    repeat (rewrite FormulaeSpi.low_subst_Wff).
    rewrite <- H3.
    rewrite low_subst_list_nil.
    rewrite plus_0_r.
    split.
    simpl.
    rewrite H0.
    rewrite H5.
    destruct MessageSpi.Ord.compare;try (elim (MessageSpi.Ord.lt_not_eq m'' m'');auto;fail).
    repeat (rewrite <- app_nil_end).
    auto.
    split.
    simpl.
    rewrite H0.
    rewrite H5.
    destruct MessageSpi.Ord.compare;try (elim (MessageSpi.Ord.lt_not_eq m'' m'');auto;fail).
    reflexivity.
    apply AROEconc.
    generalize (SubstSpi.evalc_subst_evala d k es t0 m'').
    rewrite <- H3.
    intro.
    generalize (H H5).
    rewrite low_subst_AgentSpi.
    auto.
    rewrite H4.
    apply RoSpi.PROrefl.
    (* RedIfThen *)
    destruct p0;try (discriminate H1).
    rewrite SpiCalcul_low_subst_IfThen_NC in H1.
    assert (phi=(deBruijn.low_subst FormulaeSpi.deBruijn_subst d k es t)).
    injection H1;auto.
    assert (p=(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p0)).
    injection H1;auto.
    clear H1.
    destruct mu as [mu1 l1].
    elim (IHreduction d k es p0 Hevalcsucceeds H4 mu1 l1 (refl_equal (mu1,l1)));clear IHreduction.
    intro mu1';intros.
    ex_elim H1 c1.
    ex_elim H1 b.
    unfold ActionSpiNC.mk_if in H0.
    exists mu1'.
    exists (ActionSpiSymb.union_c c1 (0%nat,cons t nil)).
    exists b.
    split.
    eapply RedSpiSymb.RedIfThen;eauto.
    repeat (rewrite low_subst_constraint_union_c).
    repeat (rewrite low_subst_constraint_cons).
    rewrite plus_0_r.
    repeat (rewrite eval_constraint_union_c).
    repeat (rewrite <- nc_of_constraint_union_c).
    rewrite H6;rewrite H7.
    assert (mu' = (mu1,app l1 (snd (EvalSpi.eval_nc phi)))).
    destruct (EvalSpi.eval_nc phi).
    destruct b0;try (discriminate H0).
    injection H0;auto.
    subst mu'.
    injection H9;clear H9;intros.
    subst mu1.
    subst l.
    split;auto.
    rewrite H3.
    split.
    unfold nc_of_constraint.
    unfold nc_list.
    destruct (EvalSpi.eval_nc (deBruijn.low_subst FormulaeSpi.deBruijn_subst d k es t)).
    simpl.
    rewrite app_nil_end.
    rewrite app_ass.
    auto.
    split;auto.
    rewrite <- H3.
    simpl.
    rewrite andb_b_true.
    rewrite <- EvalSpi.eval_nc_eval.
    destruct (EvalSpi.eval_nc phi).
    destruct b0;try (discriminate H0);auto.
    (* RedSumL *)
    destruct p0;try (discriminate H0).
    rewrite SpiCalcul_low_subst_Sum_NC in H0.
    assert (p=(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p0_1)).
    injection H0;auto.
    assert (q=(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p0_2)).
    injection H0;auto.
    clear H0.
    elim (IHreduction d k es p0_1 Hevalcsucceeds H2 mu0 l H1);clear IHreduction.
    intro mu1;intros.
    ex_elim H0 c.
    ex_elim H0 b.
    exists mu1.
    exists c.
    exists b.
    split;auto.
    apply RedSpiSymb.RedSumL;auto.
    (* RedSumR *)
    destruct p0;try (discriminate H0).
    rewrite SpiCalcul_low_subst_Sum_NC in H0.
    assert (q=(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p0_1)).
    injection H0;auto.
    assert (p=(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p0_2)).
    injection H0;auto.
    clear H0.
    elim (IHreduction d k es p0_2 Hevalcsucceeds H3 mu0 l H1);clear IHreduction.
    intro mu1;intros.
    ex_elim H0 c.
    ex_elim H0 b.
    exists mu1.
    exists c.
    exists b.
    split;auto.
    apply RedSpiSymb.RedSumR;auto.
    (* RedRes *)
    destruct p0;try (discriminate H1).
    rewrite SpiCalcul_low_subst_New_NC in H1.
    assert (p=(deBruijn.low_subst SpiCalcul.deBruijn_subst (S d) k es p0)).
    injection H1;auto.
    clear H1.
    destruct mu as [mu1 l1].
    elim (IHreduction (S d) k es p0 Hevalcsucceeds H3 mu1 l1 (refl_equal (mu1,l1)));clear IHreduction.
    intro mu1';intros.
    ex_elim H1 c1.
    ex_elim H1 b.
    elim (reduction_spi_symb_nb_redex p0 (mu1',c1) b H1);intros.
    unfold ActionSpiNC.mk_res in H0.
    case (Option_elim_option ActionSpi.t (ActionSpi.mk_res mu1));intro H10;try (rewrite H10 in H0;discriminate H0).
    ex_elim H10 mu'1.
    rewrite H10 in H0.
    injection H0;clear H0;intros.
    subst mu'.
    injection H0;clear H0;intros;subst mu0;subst l.
    generalize (reduction_nc_mk_res_symb H10 H4 H8);intro.
    exists (deBruijn.low1 ActionSpiSymb.deBruijn_base_action 0 mu1').
    exists (ActionSpiSymb.new_c c1).
    exists (AgentSpi.new b).
    split.
    eapply RedSpiSymb.RedRes;eauto.
    unfold ActionSpiSymb.mk_res.
    rewrite H0;auto.
    rewrite low_subst_constraint_new_c.
    rewrite <- nc_of_constraint_new_c.
    rewrite H5.
    rewrite eval_constraint_new_c.
    rewrite H6.
    split.
    generalize (low_subst_action_symb_low1_comm 0 d k es mu1' H0).
    repeat (rewrite plus_0_l).
    intro.
    rewrite H2.
    rewrite (evalc_action_symb_low1 (low_subst_action_symb (S d) k es mu1') mu1 H4);auto.
    unfold ActionSpi.mk_res in H10.
    case (Bool_elim_bool (deBruijn.mem_fn ActionSpi.deBruijn_fn_base 0 0 mu1));intro H11;rewrite H11 in H10;try (discriminate H10).
    injection H10;clear H10;intros;subst mu'1.
    simpl;auto.
    generalize (low_subst_action_symb_fn_not_supp 0 0 d k es mu1' (le_O_n d)).
    rewrite H0.
    auto.
    split;auto.
    split;auto.
    rewrite AgentSpi.low_subst_new.
    eapply aroE_new_c_res;eauto.
    (* RedParL *)
    destruct p0;try (discriminate H0).
    rewrite SpiCalcul_low_subst_Par_NC in H0.
    assert (p=(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p0_1)).
    injection H0;auto.
    assert (q=(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p0_2)).
    injection H0;auto.
    clear H0.
    elim (IHreduction d k es p0_1 Hevalcsucceeds H2 mu0 l H1);clear IHreduction.
    intro mu1;intros.
    ex_elim H0 c.
    ex_elim H0 b.
    exists mu1.
    exists c.
    exists (AgentSpi.par_r b p0_2).
    split.
    apply RedSpiSymb.RedParL;auto.    
    split;auto.
    split;auto.
    split;auto.
    rewrite AgentSpi.low_subst_par_r.
    apply aroE_par_r;auto.
    rewrite H3.
    apply RoSpi.pro_low_subst.
    apply RoSpi.PROrefl.
    apply Hevalcsucceeds.
    (* RedParR *)
    destruct p0;try (discriminate H0).
    rewrite SpiCalcul_low_subst_Par_NC in H0.
    assert (p=(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p0_1)).
    injection H0;auto.
    assert (q=(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p0_2)).
    injection H0;auto.
    clear H0.
    elim (IHreduction d k es p0_2 Hevalcsucceeds H3 mu0 l H1);clear IHreduction.
    intro mu1;intros.
    ex_elim H0 c.
    ex_elim H0 b.
    exists mu1.
    exists c.
    exists (AgentSpi.par_l p0_1 b).
    split.
    apply RedSpiSymb.RedParR;auto.    
    split;auto.
    split;auto.
    split;auto.
    rewrite AgentSpi.low_subst_par_l.
    apply aroE_par_l;auto.
    rewrite H2.
    apply RoSpi.pro_low_subst.
    apply RoSpi.PROrefl.
    apply Hevalcsucceeds.
    (* RedCloseL *)
    destruct p0;try (discriminate H2).
    rewrite SpiCalcul_low_subst_Par_NC in H2.
    assert (p=(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p0_1)).
    injection H2;auto.
    assert (q=(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p0_2)).
    injection H2;auto.
    clear H2.
    destruct inp as [inp linp].
    destruct out as [out lout].
    elim (IHreduction1 d k es p0_1 Hevalcsucceeds H4 inp linp (refl_equal (inp,linp)));clear IHreduction1.
    intro inp';intros.
    ex_elim H2 cinp.
    ex_elim H2 f1.
    elim (IHreduction2 d k es p0_2 Hevalcsucceeds H5 out lout (refl_equal (out,lout)));clear IHreduction2.
    intro out';intros.
    ex_elim H10 cout.
    ex_elim H10 c1.
    destruct f1.
    inversion H9.
    destruct c1.
    inversion H14.
    destruct a0;rewrite AgentSpi.low_subst_F in H14.
    inversion H14.
    destruct a as [p0'].
    destruct f as [p'].
    rewrite AgentSpi.low_subst_F in H9.
    generalize (aroE_F_inv_NC H9);clear H9;intro H9.
    destruct c0 as [n e q0].
    destruct c as [n' f q'].
    rewrite AgentSpi.low_subst_C in H14.
    elim (aroE_C_inv_NC H14);clear H14.
    intro m;intros;and_elim.
    subst n'.
    generalize (RedSpiNC.reduction_abstraction_input H (ActionSpiNC.A.Abs p') (refl_equal (ActionSpiNC.A.F (ActionSpiNC.A.Abs p'))));intro.
    generalize (RedSpiNC.reduction_concretion_output H0 (ActionSpiNC.A.Conc n f q') (refl_equal (ActionSpiNC.A.C (ActionSpiNC.A.Conc n f q'))));intro.
    generalize (RedSpiSymb.reduction_abstraction_input H2 (AgentSpi.Abs p0') (refl_equal (AgentSpi.F (AgentSpi.Abs p0'))));intro.
    generalize (RedSpiSymb.reduction_concretion_output H10 (AgentSpi.Conc n e q0) (refl_equal (AgentSpi.C (AgentSpi.Conc n e q0))));intro.
    destruct inp';try (discriminate H19).
    destruct out';try (discriminate H20).
    destruct inp;try (discriminate H17).
    destruct out;try (discriminate H18).
    exists ActionSpiSymb.Tau.
    exists (ActionSpiSymb.union_c (0%nat,cons (FormulaeSpi.FEqual t t0) nil) (ActionSpiSymb.union_c cinp cout)).
    exists (AgentSpi.pseudo_appl_l (AgentSpi.Abs p0') (AgentSpi.Conc n e q0)).
    split.
    eapply (RedSpiSymb.RedCloseL);eauto.
    unfold ActionSpiNC.mk_interact in H1.
    unfold ActionSpi.mk_interact in H1.
    destruct (Name.Ord.compare t1 t2);try (discriminate H1).
    simpl in H1.
    injection H1;clear H1;intros.
    subst tau.
    injection H1;clear H1;intros.
    subst mu.
    subst l.
    split.
    reflexivity.
    repeat (rewrite low_subst_constraint_union_c).
    split.
    repeat (rewrite <- nc_of_constraint_union_c).
    rewrite H7;rewrite H12.
    unfold low_subst_constraint.
    rewrite plus_0_r.
    Opaque nc_of_constraint.
    simpl.
    Transparent nc_of_constraint.
    rewrite nc_of_constraint_cons_FEqual.
    rewrite nc_of_constraint_nil.
    reflexivity.
    split.
    repeat (rewrite eval_constraint_union_c).
    rewrite H8;rewrite H13.
    rewrite andb_b_true.
    rewrite low_subst_constraint_cons.
    rewrite plus_0_r.
    rewrite FormulaeSpi.low_subst_FEqual.
    unfold eval_constraint.
    unfold eval_list.
    rewrite andb_b_true.
    unfold EvalSpi.eval.
    unfold evalc_action_symb in H6.
    unfold low_subst_action_symb in H6.
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es t));try (discriminate H6).
    unfold evalc_action_symb in H11.
    unfold low_subst_action_symb in H11.
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es t0));try (discriminate H11).
    destruct t3;try (discriminate H6).
    destruct t4;try (discriminate H11).
    injection H6;clear H6;intros;subst t3.
    injection H11;clear H11;intros;subst t4.
    rewrite e0.
    destruct (MessageSpi.Ord.compare);auto;elim (MessageSpi.Ord.lt_not_eq (MessageSpi.MNam t2) (MessageSpi.MNam t2));auto.
    rewrite <- H14.
    rewrite AgentSpi.low_subst_pseudo_appl_l.
    apply aroE_pseudo_appl_l;auto.
    destruct c0;rewrite AgentSpi.low_subst_C in H9.
    inversion H9.
    (* RedCloseR *)
    destruct p0;try (discriminate H2).
    rewrite SpiCalcul_low_subst_Par_NC in H2.
    assert (p=(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p0_1)).
    injection H2;auto.
    assert (q=(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p0_2)).
    injection H2;auto.
    clear H2.
    destruct inp as [inp linp].
    destruct out as [out lout].
    elim (IHreduction2 d k es p0_2 Hevalcsucceeds H5 inp linp (refl_equal (inp,linp)));clear IHreduction2.
    intro inp';intros.
    ex_elim H2 cinp.
    ex_elim H2 f1.
    elim (IHreduction1 d k es p0_1 Hevalcsucceeds H4 out lout (refl_equal (out,lout)));clear IHreduction1.
    intro out';intros.
    ex_elim H10 cout.
    ex_elim H10 c1.
    destruct f1.
    inversion H9.
    destruct c1.
    inversion H14.
    destruct a0;rewrite AgentSpi.low_subst_F in H14.
    inversion H14.
    destruct a as [q0'].
    destruct f as [q'].
    rewrite AgentSpi.low_subst_F in H9.
    generalize (aroE_F_inv_NC H9);clear H9;intro H9.
    destruct c0 as [n e p0].
    destruct c as [n' f p'].
    rewrite AgentSpi.low_subst_C in H14.
    elim (aroE_C_inv_NC H14);clear H14.
    intro m;intros;and_elim.
    subst n'.
    generalize (RedSpiNC.reduction_abstraction_input H0 (ActionSpiNC.A.Abs q') (refl_equal (ActionSpiNC.A.F (ActionSpiNC.A.Abs q'))));intro.
    generalize (RedSpiNC.reduction_concretion_output H (ActionSpiNC.A.Conc n f p') (refl_equal (ActionSpiNC.A.C (ActionSpiNC.A.Conc n f p'))));intro.
    generalize (RedSpiSymb.reduction_abstraction_input H2 (AgentSpi.Abs q0') (refl_equal (AgentSpi.F (AgentSpi.Abs q0'))));intro.
    generalize (RedSpiSymb.reduction_concretion_output H10 (AgentSpi.Conc n e p0) (refl_equal (AgentSpi.C (AgentSpi.Conc n e p0))));intro.
    destruct inp';try (discriminate H19).
    destruct out';try (discriminate H20).
    destruct inp;try (discriminate H17).
    destruct out;try (discriminate H18).
    exists ActionSpiSymb.Tau.
    exists (ActionSpiSymb.union_c (0%nat,cons (FormulaeSpi.FEqual t t0) nil) (ActionSpiSymb.union_c cinp cout)).
    exists (AgentSpi.pseudo_appl_r (AgentSpi.Conc n e p0) (AgentSpi.Abs q0')).
    split.
    eapply RedSpiSymb.RedCloseR;eauto.
    unfold ActionSpiNC.mk_interact in H1.
    unfold ActionSpi.mk_interact in H1.
    destruct (Name.Ord.compare t1 t2);try (discriminate H1).
    simpl in H1.
    injection H1;clear H1;intros.
    subst tau.
    injection H1;clear H1;intros.
    subst mu.
    subst l.
    split.
    reflexivity.
    repeat (rewrite low_subst_constraint_union_c).
    split.
    repeat (rewrite <- nc_of_constraint_union_c).
    rewrite H7;rewrite H12.
    unfold low_subst_constraint.
    rewrite plus_0_r.
    Opaque nc_of_constraint.
    simpl.
    Transparent nc_of_constraint.
    rewrite nc_of_constraint_cons_FEqual.
    rewrite nc_of_constraint_nil.
    reflexivity.
    split.
    repeat (rewrite eval_constraint_union_c).
    rewrite H8;rewrite H13.
    rewrite andb_b_true.
    rewrite low_subst_constraint_cons.
    rewrite plus_0_r.
    rewrite FormulaeSpi.low_subst_FEqual.
    unfold eval_constraint.
    unfold eval_list.
    rewrite andb_b_true.
    unfold EvalSpi.eval.
    unfold evalc_action_symb in H6.
    unfold low_subst_action_symb in H6.
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es t));try (discriminate H6).
    unfold evalc_action_symb in H11.
    unfold low_subst_action_symb in H11.
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es t0));try (discriminate H11).
    destruct t3;try (discriminate H6).
    destruct t4;try (discriminate H11).
    injection H6;clear H6;intros;subst t3.
    injection H11;clear H11;intros;subst t4.
    rewrite e0.
    destruct (MessageSpi.Ord.compare);auto;elim (MessageSpi.Ord.lt_not_eq (MessageSpi.MNam t2) (MessageSpi.MNam t2));auto.
    rewrite <- H14.
    rewrite AgentSpi.low_subst_pseudo_appl_r.
    apply aroE_pseudo_appl_r;auto.
    destruct c0;rewrite AgentSpi.low_subst_C in H9.
    inversion H9.
    (* RedBangAct *)
    destruct p0;try (discriminate H0).
    rewrite SpiCalcul_low_subst_Bang_NC in H0.
    assert (p=(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p0)).
    injection H0;auto.
    clear H0.
    elim (IHreduction d k es p0 Hevalcsucceeds H2 mu0 l H1);clear IHreduction.
    intro mu1;intros.
    ex_elim H0 c.
    ex_elim H0 b.
    exists mu1.
    exists c.
    exists (AgentSpi.par_r b (SpiCalcul.Bang p0)).
    split.
    apply RedSpiSymb.RedBangAct;auto.    
    split;auto.
    split;auto.
    split;auto.
    rewrite AgentSpi.low_subst_par_r.
    rewrite SpiCalcul.low_subst_Bang.
    apply aroE_par_r;auto.
    rewrite H2.
    apply RoSpi.PRObang.
    apply RoSpi.pro_low_subst.
    apply RoSpi.PROrefl.
    apply Hevalcsucceeds.
    (* RedBangCloseL *)
    destruct p0;try (discriminate H2).
    rewrite SpiCalcul_low_subst_Bang_NC in H2.
    assert (p=(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p0)).
    injection H2;auto.
    clear H2.
    destruct inp as [inp linp].
    destruct out as [out lout].
    elim (IHreduction1 d k es p0 Hevalcsucceeds H4 inp linp (refl_equal (inp,linp)));clear IHreduction1.
    intro inp';intros.
    ex_elim H2 cinp.
    ex_elim H2 f1.
    elim (IHreduction2 d k es p0 Hevalcsucceeds H4 out lout (refl_equal (out,lout)));clear IHreduction2.
    intro out';intros.
    ex_elim H9 cout.
    ex_elim H9 c1.
    destruct f1.
    inversion H8.
    destruct c1.
    inversion H13.
    destruct a0;rewrite AgentSpi.low_subst_F in H13.
    inversion H13.
    destruct a as [p0'].
    destruct f as [p'].
    rewrite AgentSpi.low_subst_F in H8.
    generalize (aroE_F_inv_NC H8);clear H8;intro H8.
    destruct c0 as [n e q0].
    destruct c as [n' f q'].
    rewrite AgentSpi.low_subst_C in H13.
    elim (aroE_C_inv_NC H13);clear H13.
    intro m;intros;and_elim.
    subst n'.
    generalize (RedSpiNC.reduction_abstraction_input H (ActionSpiNC.A.Abs p') (refl_equal (ActionSpiNC.A.F (ActionSpiNC.A.Abs p'))));intro.
    generalize (RedSpiNC.reduction_concretion_output H0 (ActionSpiNC.A.Conc n f q') (refl_equal (ActionSpiNC.A.C (ActionSpiNC.A.Conc n f q'))));intro.
    generalize (RedSpiSymb.reduction_abstraction_input H2 (AgentSpi.Abs p0') (refl_equal (AgentSpi.F (AgentSpi.Abs p0'))));intro.
    generalize (RedSpiSymb.reduction_concretion_output H9 (AgentSpi.Conc n e q0) (refl_equal (AgentSpi.C (AgentSpi.Conc n e q0))));intro.
    destruct inp';try (discriminate H18).
    destruct out';try (discriminate H19).
    destruct inp;try (discriminate H16).
    destruct out;try (discriminate H17).
    exists ActionSpiSymb.Tau.
    exists (ActionSpiSymb.union_c (0%nat,cons (FormulaeSpi.FEqual t t0) nil) (ActionSpiSymb.union_c cinp cout)).
    exists (AgentSpi.par_r (AgentSpi.pseudo_appl_l (AgentSpi.Abs p0') (AgentSpi.Conc n e q0)) (SpiCalcul.Bang p0)).
    split.
    eapply (RedSpiSymb.RedBangCloseL);eauto.
    unfold ActionSpiNC.mk_interact in H1.
    unfold ActionSpi.mk_interact in H1.
    destruct (Name.Ord.compare t1 t2);try (discriminate H1).
    simpl in H1.
    injection H1;clear H1;intros.
    subst tau.
    injection H1;clear H1;intros.
    subst mu.
    subst l.
    split.
    reflexivity.
    repeat (rewrite low_subst_constraint_union_c).
    split.
    repeat (rewrite <- nc_of_constraint_union_c).
    rewrite H6;rewrite H11.
    unfold low_subst_constraint.
    rewrite plus_0_r.
    Opaque nc_of_constraint.
    simpl.
    Transparent nc_of_constraint.
    rewrite nc_of_constraint_cons_FEqual.
    rewrite nc_of_constraint_nil.
    reflexivity.
    split.
    repeat (rewrite eval_constraint_union_c).
    rewrite H7;rewrite H12.
    rewrite andb_b_true.
    rewrite low_subst_constraint_cons.
    rewrite plus_0_r.
    rewrite FormulaeSpi.low_subst_FEqual.
    unfold eval_constraint.
    unfold eval_list.
    rewrite andb_b_true.
    unfold EvalSpi.eval.
    unfold evalc_action_symb in H5.
    unfold low_subst_action_symb in H5.
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es t));try (discriminate H5).
    unfold evalc_action_symb in H10.
    unfold low_subst_action_symb in H10.
    destruct (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es t0));try (discriminate H10).
    destruct t3;try (discriminate H5).
    destruct t4;try (discriminate H10).
    injection H5;clear H5;intros;subst t3.
    injection H10;clear H10;intros;subst t4.
    rewrite e0.
    destruct (MessageSpi.Ord.compare);auto;elim (MessageSpi.Ord.lt_not_eq (MessageSpi.MNam t2) (MessageSpi.MNam t2));auto.
    rewrite <- H13.
    rewrite AgentSpi.low_subst_par_r.
    rewrite AgentSpi.low_subst_pseudo_appl_l.
    apply aroE_par_r;auto.
    apply aroE_pseudo_appl_l;auto.
    rewrite H4.
    rewrite SpiCalcul.low_subst_Bang.
    apply RoSpi.PRObang.
    apply RoSpi.pro_low_subst.
    apply RoSpi.PROrefl.
    unfold SubstSpi.msg_subst.
    apply Hevalcsucceeds.
    destruct c0;rewrite AgentSpi.low_subst_C in H8.
    inversion H8.
  Qed.

  Theorem reduction_nc_subst_impl_reduction_symb : forall (d k:nat)(sigma:list MessageSpi.t)(p:SpiCalcul.t)(mu:ActionSpi.t)(l:list Name.t)(b:AgentSpi.t),(RedSpiNC.reduction (deBruijn.low_subst SpiCalcul.deBruijn_subst d k (SubstSpi.msg_subst sigma) p) (mu,l) b)->exists mus:ActionSpiSymb.action,exists c:ActionSpiSymb.constraint,exists a:AgentSpi.t,RedSpiSymb.reduction p (mus,c) a /\ (evalc_action_symb (low_subst_action_symb d k (SubstSpi.msg_subst sigma) mus) = Some mu) /\ (nc_of_constraint (low_subst_constraint d k (SubstSpi.msg_subst sigma) c) = l) /\ (eval_constraint (low_subst_constraint d k (SubstSpi.msg_subst sigma) c) = true) /\ (aroEval (deBruijn.low_subst AgentSpi.deBruijn_subst d k (SubstSpi.msg_subst sigma) a) b).
    intros.
    eapply reduction_nc_impl_reduction_symb;eauto.
    unfold SubstSpi.msg_subst.
    apply EvalSpi.evalc_succeeds_msg.
  Qed.

  Lemma reduction_symb_impl_reduction_nc : forall (p:SpiCalcul.t)(musc:ActionSpiSymb.t)(a:AgentSpi.t),(RedSpiSymb.reduction p musc a)->forall (mus:ActionSpiSymb.action)(c:ActionSpiSymb.constraint),musc=(mus,c)->forall (d k:nat)(es:list ExpressionSpi.t)(Hevalcsucceeds:List_forall EvalSpi.evalc_succeeds es),(eval_constraint (low_subst_constraint d k es c))=true->exists mu:ActionSpi.t,exists b:AgentSpi.t,(RedSpiNC.reduction (deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p) (mu,nc_of_constraint (low_subst_constraint d k es c)) b) /\ (evalc_action_symb (low_subst_action_symb d k es mus) = Some mu) /\ (aroEval (deBruijn.low_subst AgentSpi.deBruijn_subst d k es a) b).
    induction 1;intros.
    (* RedSilent *)
    exists ActionSpi.mk_silent.
    rewrite SpiCalcul.low_subst_Tau.
    exists (AgentSpi.Pr (deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p)).
    unfold ActionSpiSymb.mk_silent in H.
    injection H;clear H;intros;subst mus;subst c.
    split.
    eapply RedSpiNC.RedSilent.
    split;try reflexivity.
    rewrite AgentSpi.low_subst_Pr.
    apply AROEproc.
    apply RoSpi.PROrefl.
    (* RedInput *)
    rewrite SpiCalcul_low_subst_Input_aux.
    unfold ActionSpiSymb.mk_input in H.
    injection H;clear H;intros;subst mu.
    injection H;clear H;intros;subst mus;subst c.
    rewrite low_subst_constraint_cons in H1.
    rewrite low_subst_list_nil in H1.
    rewrite plus_0_r in H1.
    rewrite FormulaeSpi.low_subst_FName in H1.
    unfold eval_constraint in H1.
    unfold eval_list in H1.
    rewrite andb_b_true in H1.
    unfold EvalSpi.eval in H1.
    case (Option_elim_option MessageSpi.t (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es e)));intro H2;try (rewrite H2 in H1;discriminate H1).
    ex_elim H2 a.
    rewrite H2 in H1.
    destruct a;try (discriminate H1).
    exists (ActionSpi.Inp t).
    exists (AgentSpi.F (AgentSpi.Abs (deBruijn.low_subst SpiCalcul.deBruijn_subst (S d) k es p))).
    rewrite low_subst_constraint_cons.
    rewrite plus_0_r.
    rewrite low_subst_list_nil.
    rewrite FormulaeSpi.low_subst_FName.
    unfold low_subst_action_symb.
    split.
    apply RedSpiNC.RedInput.
    unfold ActionSpiNC.mk_input.
    rewrite H2.
    unfold nc_of_constraint.
    unfold nc_list.
    unfold EvalSpi.eval_nc.
    rewrite H2.
    rewrite <- app_nil_end.
    reflexivity.
    split.
    unfold evalc_action_symb.
    rewrite (SubstSpi.evalc_subst_evala d k es e (MessageSpi.MNam t));auto.
    rewrite AgentSpi.low_subst_F.
    apply AROEabs.
    apply RoSpi.PROrefl.
    (* RedOutput *)
    rewrite SpiCalcul_low_subst_Output_aux.
    unfold ActionSpiSymb.mk_output in H.
    injection H;clear H;intros;subst mu;subst m.
    injection H2;clear H2;intros;subst mus;subst c.
    rewrite low_subst_constraint_cons in H1.
    rewrite plus_0_r in H1.
    rewrite low_subst_list_cons in H1.
    rewrite low_subst_list_nil in H1.
    unfold eval_constraint in H1.
    unfold eval_list in H1.
    rewrite andb_b_true in H1.
    elim (andb_prop (EvalSpi.eval (deBruijn.low_subst FormulaeSpi.deBruijn_subst d k es (FormulaeSpi.FName e))) (EvalSpi.eval (deBruijn.low_subst FormulaeSpi.deBruijn_subst d k es (FormulaeSpi.Wff f))) H1);clear H1;intros.
    rewrite FormulaeSpi.low_subst_FName in H.
    rewrite FormulaeSpi.low_subst_Wff in H0.
    rewrite low_subst_constraint_cons.
    rewrite plus_0_r.
    rewrite low_subst_list_cons.
    rewrite low_subst_list_nil.
    rewrite FormulaeSpi.low_subst_FName.
    rewrite FormulaeSpi.low_subst_Wff.
    unfold low_subst_action_symb.
    unfold EvalSpi.eval in H.
    case (Option_elim_option MessageSpi.t (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es e)));intro H1;try (rewrite H1 in H;discriminate H).
    ex_elim H1 a.
    rewrite H1 in H.
    destruct a;try (discriminate H).
    exists (ActionSpi.Outp t).
    unfold EvalSpi.eval in H0.
    unfold FormulaeSpi.Wff in H0.
    case (Option_elim_option MessageSpi.t (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es f)));intro H2;try (rewrite H2 in H0;discriminate H0).
    ex_elim H2 m.
    rewrite H2 in H0.
    exists (AgentSpi.C (AgentSpi.Conc 0 (EvalSpi.expr_of_mesg m) (deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p))).
    split.
    apply RedSpiNC.RedOutput.
    unfold ActionSpiNC.mk_output.
    rewrite H1.
    rewrite H2.
    unfold nc_of_constraint.
    lazy beta delta [nc_list] iota;fold nc_list. 
    unfold EvalSpi.eval_nc.
    rewrite H1.
    unfold FormulaeSpi.Wff.
    rewrite H2.
    destruct MessageSpi.Ord.compare;try (discriminate H0).
    repeat (rewrite <- app_nil_end).
    reflexivity.
    split.
    unfold evalc_action_symb.
    rewrite (SubstSpi.evalc_subst_evala d k es e (MessageSpi.MNam t));auto.
    rewrite AgentSpi_low_subst_C_aux.
    rewrite plus_0_r.
    apply AROEconc.
    rewrite (SubstSpi.evalc_subst_evala d k es f m);auto.
    apply RoSpi.PROrefl.
    (* RedIfThen *)
    unfold ActionSpiSymb.mk_if in H0.
    destruct mu.
    injection H0;clear H0;intros;subst mu'.
    injection H0;clear H0;intros;subst a0;subst c.
    rewrite low_subst_constraint_union_c in H2.
    rewrite eval_constraint_union_c in H2.
    elim (andb_prop (eval_constraint (low_subst_constraint d k es c0)) (eval_constraint (low_subst_constraint d k es (0, phi :: nil))) H2);clear H2;intros.
    elim (IHreduction mus c0 (refl_equal (mus,c0)) d k es Hevalcsucceeds H0).
    intro mu;intros.
    ex_elim H2 b.
    rewrite low_subst_constraint_cons in H1.
    rewrite plus_0_r in H1.
    rewrite low_subst_list_nil in H1.
    exists mu.
    exists b.
    rewrite SpiCalcul_low_subst_IfThen_aux.
    clear IHreduction.
    rewrite low_subst_constraint_union_c.
    rewrite <- nc_of_constraint_union_c.
    split.
    eapply RedSpiNC.RedIfThen;eauto.
    unfold ActionSpiNC.mk_if.
    unfold eval_constraint in H1.
    unfold eval_list in H1.
    rewrite andb_b_true in H1.
    rewrite low_subst_constraint_cons. 
    rewrite low_subst_list_nil.
    rewrite plus_0_r.
    lazy beta delta [nc_list nc_of_constraint] iota;fold nc_of_constraint;fold nc_list. 
    generalize (EvalSpi.eval_nc_eval (deBruijn.low_subst FormulaeSpi.deBruijn_subst d k es phi)).
    rewrite H1.
    intro.
    destruct (EvalSpi.eval_nc (deBruijn.low_subst FormulaeSpi.deBruijn_subst d k es phi)).
    simpl in H5.
    subst b0.
    rewrite <- app_nil_end.
    reflexivity.
    split;auto.
    (* RedSumL *)
    elim (IHreduction mus c H0 d k es Hevalcsucceeds H1);clear IHreduction.
    intro mu0;intros.
    ex_elim H2 b.
    exists mu0.
    exists b.
    split;auto.
    rewrite SpiCalcul.low_subst_Sum.
    apply RedSpiNC.RedSumL;auto.
    (* RedSumR *)
    elim (IHreduction mus c H0 d k es Hevalcsucceeds H1);clear IHreduction.
    intro mu0;intros.
    ex_elim H2 b.
    exists mu0.
    exists b.
    split;auto.
    rewrite SpiCalcul.low_subst_Sum.
    apply RedSpiNC.RedSumR;auto.
    (* RedRes *)
    rewrite SpiCalcul.low_subst_New.
    rewrite AgentSpi.low_subst_new.
    unfold ActionSpiSymb.mk_res in H0.
    destruct mu.
    case (Bool_elim_bool (deBruijn.mem_fn ActionSpiSymb.deBruijn_fn_base_action 0 0 a0));intro H3;rewrite H3 in H0;try (discriminate H0).
    subst mu'.
    assert (mus = deBruijn.low1 ActionSpiSymb.deBruijn_base_action 0 a0).
    injection H0;auto.
    rewrite <- H1 in H0.
    injection H0;clear H0;intros.
    subst c.
    rewrite low_subst_constraint_new_c in H2.
    rewrite eval_constraint_new_c in H2.
    elim (IHreduction a0 c0 (refl_equal (a0,c0)) (S d) k es Hevalcsucceeds H2);clear IHreduction.
    intro mu;intros.
    ex_elim H0 b.
    generalize (low_subst_action_symb_fn_not_supp 0 0 d k es a0 (le_O_n d)).
    rewrite plus_0_l.
    rewrite H3.
    intro.
    symmetry in H6.
    generalize (evalc_action_symb_mem_fn_false (low_subst_action_symb (S d) k es a0) mu H4 0 0 H6);intro.
    exists (deBruijn.low1 ActionSpi.deBruijn_base 0 mu).
    exists (AgentSpi.new b).
    split.
    rewrite low_subst_constraint_new_c.
    rewrite <- nc_of_constraint_new_c.
    eapply RedSpiNC.RedRes;eauto.
    unfold ActionSpiNC.mk_res.
    unfold ActionSpi.mk_res.
    rewrite H7.
    reflexivity.
    split.
    subst mus.
    generalize (low_subst_action_symb_low1_comm 0 d k es a0 H3).
    repeat (rewrite plus_0_l).
    intro.
    rewrite H1.
    apply (evalc_action_symb_low1 (low_subst_action_symb (S d) k es a0) mu H4 0 H6).
    eapply aroE_new_c_res;eauto.
    elim (reduction_spi_symb_nb_redex p (a0,c0) a H);auto.
    (* RedParL *)
    elim (IHreduction mus c H0 d k es Hevalcsucceeds H1);clear IHreduction.
    intro mu0;intros.
    ex_elim H2 b.
    exists mu0.
    rewrite SpiCalcul.low_subst_Par.
    exists (AgentSpi.par_r b (deBruijn.low_subst SpiCalcul.deBruijn_subst d k es q)).
    split.
    eapply RedSpiNC.RedParL;auto.
    split;auto.
    rewrite AgentSpi.low_subst_par_r.
    apply aroE_par_r;auto.
    apply RoSpi.PROrefl.
    (* RedParR *)
    elim (IHreduction mus c H0 d k es Hevalcsucceeds H1);clear IHreduction.
    intro mu0;intros.
    ex_elim H2 b.
    exists mu0.
    rewrite SpiCalcul.low_subst_Par.
    exists (AgentSpi.par_l (deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p) b).
    split.
    eapply RedSpiNC.RedParR;auto.
    split;auto.
    rewrite AgentSpi.low_subst_par_l.
    apply aroE_par_l;auto.
    apply RoSpi.PROrefl.
    (* RedCloseL *)
    generalize (RedSpiSymb.reduction_abstraction_input H f (refl_equal (AgentSpi.F f)));intro Hinp.
    generalize (RedSpiSymb.reduction_concretion_output H0 c (refl_equal (AgentSpi.C c)));intro Hout.
    unfold ActionSpiSymb.mk_interact in H1.
    destruct inp as [inp cinp].
    destruct inp;try (discriminate Hinp).
    destruct out as [out cout].
    destruct out;try (discriminate Hout).
    injection H1;clear H1;intros;subst tau.
    injection H1;clear H1;intros;subst mus;subst c0.
    repeat (rewrite low_subst_constraint_union_c in H3).
    repeat (rewrite eval_constraint_union_c in H3).
    rewrite low_subst_constraint_cons in H3.
    rewrite low_subst_list_nil in H3.
    rewrite plus_0_r in H3.
    elim (andb_prop (eval_constraint (0,deBruijn.low_subst FormulaeSpi.deBruijn_subst d k es (FormulaeSpi.FEqual t t0) :: nil)) (eval_constraint (low_subst_constraint d k es cinp) && eval_constraint (low_subst_constraint d k es cout)) H3);clear H3;intros.
    elim (andb_prop (eval_constraint (low_subst_constraint d k es cinp)) (eval_constraint (low_subst_constraint d k es cout)) H2);clear H2;intros.
    rewrite FormulaeSpi.low_subst_FEqual in H1.
    elim (IHreduction1 (ActionSpiSymb.Inp t) cinp (refl_equal (ActionSpiSymb.Inp t,cinp)) d k es Hevalcsucceeds H2);clear IHreduction1;intro inp';intros.
    ex_elim H4 f'.
    elim (IHreduction2 (ActionSpiSymb.Outp t0) cout (refl_equal (ActionSpiSymb.Outp t0,cout)) d k es Hevalcsucceeds H3);clear IHreduction2;intro out';intros.
    ex_elim H7 c'.
    exists ActionSpi.Tau.
    rewrite SpiCalcul.low_subst_Par.
    rewrite low_subst_constraint_union_c.
    destruct f.
    destruct c.
    rewrite AgentSpi.low_subst_F in H6.
    rewrite AgentSpi.low_subst_C in H9.
    rewrite AgentSpi.low_subst_pseudo_appl_l.
    unfold low_subst_action_symb in H5.
    unfold low_subst_action_symb in H8.
    unfold evalc_action_symb in H5.
    unfold evalc_action_symb in H8.
    case (Option_elim_option MessageSpi.t (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es t)));intro H10;try (rewrite H10 in H5;discriminate H5).
    ex_elim H10 a;rewrite H10 in H5.
    case (Option_elim_option MessageSpi.t (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es t0)));intro H11;try (rewrite H11 in H8;discriminate H8).
    ex_elim H11 b;rewrite H11 in H8.
    destruct a;try (discriminate H5).
    destruct b;try (discriminate H8).
    injection H5;clear H5;intros;subst inp'.
    injection H8;clear H8;intros;subst out'.
    elim (RedSpiNC.reduction_input_abstraction H4);try reflexivity.
    intro f1;intro.
    subst f'.
    elim (RedSpiNC.reduction_output_concretion H7);try reflexivity.
    intro c1;intro.
    subst c'.
    destruct f1 as [p'].
    destruct c1 as [n' f' q'].
    generalize (aroE_F_inv_NC H6);clear H6;intros.
    elim (aroE_C_inv_NC H9);clear H9.
    intro m;intros;and_elim.
    rewrite low_subst_constraint_union_c.
    rewrite low_subst_constraint_cons.
    rewrite plus_0_r.
    repeat (rewrite <- nc_of_constraint_union_c).
    rewrite FormulaeSpi.low_subst_FEqual.
    rewrite nc_of_constraint_cons_FEqual.
    rewrite nc_of_constraint_nil.
    unfold low_subst_action_symb.
    unfold evalc_action_symb.
    subst n'.
    exists (AgentSpi.pseudo_appl_l (ActionSpiNC.A.Abs p') (ActionSpiNC.A.Conc n f' q')).
    split.
    eapply RedSpiNC.RedCloseL;eauto.
    unfold ActionSpiNC.mk_interact.
    unfold ActionSpi.mk_interact.
    unfold eval_constraint in H1.
    unfold eval_list in H1.
    rewrite andb_b_true in H1.
    unfold EvalSpi.eval in H1.
    rewrite H10 in H1.
    rewrite H11 in H1.
    simpl in H1.
    destruct (Name.Ord.compare t4 t5);try (discriminate H1).
    reflexivity.
    split;auto.
    subst f'.
    apply aroE_pseudo_appl_l;auto.
    (* RedCloseR *)
    generalize (RedSpiSymb.reduction_abstraction_input H0 f (refl_equal (AgentSpi.F f)));intro Hinp.
    generalize (RedSpiSymb.reduction_concretion_output H c (refl_equal (AgentSpi.C c)));intro Hout.
    unfold ActionSpiSymb.mk_interact in H1.
    destruct inp as [inp cinp].
    destruct inp;try (discriminate Hinp).
    destruct out as [out cout].
    destruct out;try (discriminate Hout).
    injection H1;clear H1;intros;subst tau.
    injection H1;clear H1;intros;subst mus;subst c0.
    repeat (rewrite low_subst_constraint_union_c in H3).
    repeat (rewrite eval_constraint_union_c in H3).
    rewrite low_subst_constraint_cons in H3.
    rewrite low_subst_list_nil in H3.
    rewrite plus_0_r in H3.
    elim (andb_prop (eval_constraint (0,deBruijn.low_subst FormulaeSpi.deBruijn_subst d k es (FormulaeSpi.FEqual t t0) :: nil)) (eval_constraint (low_subst_constraint d k es cinp) && eval_constraint (low_subst_constraint d k es cout)) H3);clear H3;intros.
    elim (andb_prop (eval_constraint (low_subst_constraint d k es cinp)) (eval_constraint (low_subst_constraint d k es cout)) H2);clear H2;intros.
    rewrite FormulaeSpi.low_subst_FEqual in H1.
    elim (IHreduction2 (ActionSpiSymb.Inp t) cinp (refl_equal (ActionSpiSymb.Inp t,cinp)) d k es Hevalcsucceeds H2);clear IHreduction2;intro inp';intros.
    ex_elim H4 f'.
    elim (IHreduction1 (ActionSpiSymb.Outp t0) cout (refl_equal (ActionSpiSymb.Outp t0,cout)) d k es Hevalcsucceeds H3);clear IHreduction1;intro out';intros.
    ex_elim H7 c'.
    exists ActionSpi.Tau.
    rewrite SpiCalcul.low_subst_Par.
    rewrite low_subst_constraint_union_c.
    destruct f.
    destruct c.
    rewrite AgentSpi.low_subst_F in H6.
    rewrite AgentSpi.low_subst_C in H9.
    rewrite AgentSpi.low_subst_pseudo_appl_r.
    unfold low_subst_action_symb in H5.
    unfold low_subst_action_symb in H8.
    unfold evalc_action_symb in H5.
    unfold evalc_action_symb in H8.
    case (Option_elim_option MessageSpi.t (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es t)));intro H10;try (rewrite H10 in H5;discriminate H5).
    ex_elim H10 a;rewrite H10 in H5.
    case (Option_elim_option MessageSpi.t (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es t0)));intro H11;try (rewrite H11 in H8;discriminate H8).
    ex_elim H11 b;rewrite H11 in H8.
    destruct a;try (discriminate H5).
    destruct b;try (discriminate H8).
    injection H5;clear H5;intros;subst inp'.
    injection H8;clear H8;intros;subst out'.
    elim (RedSpiNC.reduction_input_abstraction H4);try reflexivity.
    intro f1;intro.
    subst f'.
    elim (RedSpiNC.reduction_output_concretion H7);try reflexivity.
    intro c1;intro.
    subst c'.
    destruct f1 as [p'].
    destruct c1 as [n' f' q'].
    generalize (aroE_F_inv_NC H6);clear H6;intros.
    elim (aroE_C_inv_NC H9);clear H9.
    intro m;intros;and_elim.
    rewrite low_subst_constraint_union_c.
    rewrite low_subst_constraint_cons.
    rewrite plus_0_r.
    repeat (rewrite <- nc_of_constraint_union_c).
    rewrite FormulaeSpi.low_subst_FEqual.
    rewrite nc_of_constraint_cons_FEqual.
    rewrite nc_of_constraint_nil.
    unfold low_subst_action_symb.
    unfold evalc_action_symb.
    subst n'.
    exists (AgentSpi.pseudo_appl_r (ActionSpiNC.A.Conc n f' q') (ActionSpiNC.A.Abs p')).
    split.
    eapply RedSpiNC.RedCloseR;eauto.
    unfold ActionSpiNC.mk_interact.
    unfold ActionSpi.mk_interact.
    unfold eval_constraint in H1.
    unfold eval_list in H1.
    rewrite andb_b_true in H1.
    unfold EvalSpi.eval in H1.
    rewrite H10 in H1.
    rewrite H11 in H1.
    simpl in H1.
    destruct (Name.Ord.compare t4 t5);try (discriminate H1).
    reflexivity.
    split;auto.
    subst f'.
    apply aroE_pseudo_appl_r;auto.
    (* RedBangAct *)
    elim (IHreduction mus c H0 d k es Hevalcsucceeds H1);clear IHreduction.
    intro mu0;intros.
    ex_elim H2 b.
    exists mu0.
    rewrite SpiCalcul.low_subst_Bang.
    exists (AgentSpi.par_r b (SpiCalcul.Bang (deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p))).
    split.
    eapply RedSpiNC.RedBangAct;auto.
    split;auto.
    rewrite AgentSpi.low_subst_par_r.
    apply aroE_par_r;auto.
    rewrite SpiCalcul.low_subst_Bang.
    apply RoSpi.PROrefl.
    (* RedBangCloseL *)
    generalize (RedSpiSymb.reduction_abstraction_input H f (refl_equal (AgentSpi.F f)));intro Hinp.
    generalize (RedSpiSymb.reduction_concretion_output H0 c (refl_equal (AgentSpi.C c)));intro Hout.
    unfold ActionSpiSymb.mk_interact in H1.
    destruct inp as [inp cinp].
    destruct inp;try (discriminate Hinp).
    destruct out as [out cout].
    destruct out;try (discriminate Hout).
    injection H1;clear H1;intros;subst tau.
    injection H1;clear H1;intros;subst mus;subst c0.
    repeat (rewrite low_subst_constraint_union_c in H3).
    repeat (rewrite eval_constraint_union_c in H3).
    rewrite low_subst_constraint_cons in H3.
    rewrite low_subst_list_nil in H3.
    rewrite plus_0_r in H3.
    elim (andb_prop (eval_constraint (0,deBruijn.low_subst FormulaeSpi.deBruijn_subst d k es (FormulaeSpi.FEqual t t0) :: nil)) (eval_constraint (low_subst_constraint d k es cinp) && eval_constraint (low_subst_constraint d k es cout)) H3);clear H3;intros.
    elim (andb_prop (eval_constraint (low_subst_constraint d k es cinp)) (eval_constraint (low_subst_constraint d k es cout)) H2);clear H2;intros.
    rewrite FormulaeSpi.low_subst_FEqual in H1.
    elim (IHreduction1 (ActionSpiSymb.Inp t) cinp (refl_equal (ActionSpiSymb.Inp t,cinp)) d k es Hevalcsucceeds H2);clear IHreduction1;intro inp';intros.
    ex_elim H4 f'.
    elim (IHreduction2 (ActionSpiSymb.Outp t0) cout (refl_equal (ActionSpiSymb.Outp t0,cout)) d k es Hevalcsucceeds H3);clear IHreduction2;intro out';intros.
    ex_elim H7 c'.
    exists ActionSpi.Tau.
    rewrite SpiCalcul.low_subst_Bang.
    rewrite low_subst_constraint_union_c.
    destruct f.
    destruct c.
    rewrite AgentSpi.low_subst_F in H6.
    rewrite AgentSpi.low_subst_C in H9.
    rewrite AgentSpi.low_subst_par_r.
    rewrite AgentSpi.low_subst_pseudo_appl_l.
    unfold low_subst_action_symb in H5.
    unfold low_subst_action_symb in H8.
    unfold evalc_action_symb in H5.
    unfold evalc_action_symb in H8.
    case (Option_elim_option MessageSpi.t (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es t)));intro H10;try (rewrite H10 in H5;discriminate H5).
    ex_elim H10 a;rewrite H10 in H5.
    case (Option_elim_option MessageSpi.t (EvalSpi.evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es t0)));intro H11;try (rewrite H11 in H8;discriminate H8).
    ex_elim H11 b;rewrite H11 in H8.
    destruct a;try (discriminate H5).
    destruct b;try (discriminate H8).
    injection H5;clear H5;intros;subst inp'.
    injection H8;clear H8;intros;subst out'.
    elim (RedSpiNC.reduction_input_abstraction H4);try reflexivity.
    intro f1;intro.
    subst f'.
    elim (RedSpiNC.reduction_output_concretion H7);try reflexivity.
    intro c1;intro.
    subst c'.
    destruct f1 as [p'].
    destruct c1 as [n' f' q'].
    generalize (aroE_F_inv_NC H6);clear H6;intros.
    elim (aroE_C_inv_NC H9);clear H9.
    intro m;intros;and_elim.
    rewrite low_subst_constraint_union_c.
    rewrite low_subst_constraint_cons.
    rewrite plus_0_r.
    repeat (rewrite <- nc_of_constraint_union_c).
    rewrite FormulaeSpi.low_subst_FEqual.
    rewrite nc_of_constraint_cons_FEqual.
    rewrite nc_of_constraint_nil.
    unfold low_subst_action_symb.
    unfold evalc_action_symb.
    subst n'.
    exists (AgentSpi.par_r (AgentSpi.pseudo_appl_l (ActionSpiNC.A.Abs p') (ActionSpiNC.A.Conc n f' q')) (SpiCalcul.Bang (deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p))).
    split.
    eapply RedSpiNC.RedBangCloseL;eauto.
    unfold ActionSpiNC.mk_interact.
    unfold ActionSpi.mk_interact.
    unfold eval_constraint in H1.
    unfold eval_list in H1.
    rewrite andb_b_true in H1.
    unfold EvalSpi.eval in H1.
    rewrite H10 in H1.
    rewrite H11 in H1.
    simpl in H1.
    destruct (Name.Ord.compare t4 t5);try (discriminate H1).
    reflexivity.
    split;auto.
    subst f'.
    apply aroE_par_r;auto.
    apply aroE_pseudo_appl_l;auto.
    rewrite SpiCalcul.low_subst_Bang.
    apply RoSpi.PROrefl.
  Qed.

  Theorem reduction_symb_impl_reduction_nc_subst : forall (p:SpiCalcul.t)(mus:ActionSpiSymb.action)(c:ActionSpiSymb.constraint)(a:AgentSpi.t),(RedSpiSymb.reduction p (mus,c) a)->forall (d k:nat)(sigma:list MessageSpi.t),(eval_constraint (low_subst_constraint d k (SubstSpi.msg_subst sigma) c))=true->exists mu:ActionSpi.t,exists b:AgentSpi.t,(RedSpiNC.reduction (deBruijn.low_subst SpiCalcul.deBruijn_subst d k (SubstSpi.msg_subst sigma) p) (mu,nc_of_constraint (low_subst_constraint d k (SubstSpi.msg_subst sigma) c)) b) /\ (evalc_action_symb (low_subst_action_symb d k (SubstSpi.msg_subst sigma) mus) = Some mu) /\ (aroEval (deBruijn.low_subst AgentSpi.deBruijn_subst d k (SubstSpi.msg_subst sigma) a) b).
    intros.
    eapply reduction_symb_impl_reduction_nc;eauto.
    unfold SubstSpi.msg_subst.
    apply EvalSpi.evalc_succeeds_msg.
  Qed.
End ReductionSpiSymbprops.



