(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import Arith.
Require Import FSets.
Require Import ordmissing.
Require Import missing.
Require Import tactics.
Require Import Relation_Operators.
Require Import Relations_2.
Require Import Wf_nat.
Require Import Multiset.
(* Require Import MultisetAuxiliary. *)
Require Import RelExtras.
Require Import MultisetCore.
Require Import MultisetList.
Require Import MultisetOrder.
Require Import Wellfounded.

Module SetOrder(setModule:S).
  Module setEqProps := FSetEqProperties.EqProperties setModule.
  Module setProps := FSetProperties.Properties setModule.
  Module setOtherProps := MissingSetProperties setModule.
  Module setEltProps := OrderedTypeFacts setModule.E.

  Definition is_lighter (weight:setModule.elt->nat)(x y:setModule.elt) := if (le_lt_dec (weight x) (weight y)) then false else true.

  Lemma compat_bool_is_lighter : forall (weight:setModule.elt->nat)(weight_refl:forall (x y:setModule.elt),(setModule.E.eq x y)->(weight x = weight y))(x:setModule.elt),(compat_bool setModule.E.eq (is_lighter weight x)).
    unfold compat_bool.
    intros.
    unfold is_lighter.
    rewrite (weight_refl x0 y H).
    trivial.
  Qed.

  Lemma is_lighter_carac : forall (weight:setModule.elt->nat)(x y:setModule.elt),(is_lighter weight x y)=true<->(weight y < weight x).
    unfold is_lighter.
    intros.
    case le_lt_dec;intro.
    split;intros.
    discriminate H.
    elim (lt_irrefl (weight y));omega.
    split;intros;auto.
  Qed.

  Definition is_heavier (weight:setModule.elt->nat)(x y:setModule.elt) := (weight y <= weight x).

  Lemma is_heavier_total_dec : forall (weight:setModule.elt->nat)(x y:setModule.elt),{is_heavier weight x y}+{is_heavier weight y x}.
    unfold is_heavier;intros.
    case (le_lt_dec (weight x) (weight y));intro.
    right;trivial.
    left;auto with arith.
  Defined.

  Lemma is_heavier_refl : forall (weight:setModule.elt->nat)(weight_refl:forall (x y:setModule.elt),(setModule.E.eq x y)->(weight x = weight y))(x y:setModule.elt),(setModule.E.eq x y)->(is_heavier weight x y).
    unfold is_heavier;intros.
    rewrite (weight_refl x y H);trivial.
  Qed.

  Lemma is_heavier_trans : forall (weight:setModule.elt->nat)(x y z:setModule.elt),(is_heavier weight x y)->(is_heavier weight y z)->(is_heavier weight x z).
    unfold is_heavier;intros.
    eapply le_trans;eauto.
  Qed.

  Lemma select_heavier : forall (weight:setModule.elt->nat)(weight_refl:forall (x y:setModule.elt),(setModule.E.eq x y)->(weight x = weight y))(s:setModule.t),(~setModule.Empty s)->{x:setModule.elt | (setModule.In x s)/\(forall (y:setModule.elt),(setModule.In y s)->(weight y <= weight x))}.
    intros.
    case (setOtherProps.select_such_that (is_heavier weight) (is_heavier_total_dec weight) (is_heavier_refl weight weight_refl) (is_heavier_trans weight) s);unfold is_heavier;intro;trivial.
    elim H;trivial.
  Defined.

  Definition lt_set1 (weight:setModule.elt->nat)(s t:setModule.t) := (exists x:setModule.t,exists y:setModule.t,(setOtherProps.strict_Subset x t)/\(setModule.Equal s (setModule.union x y))/\(forall (a:setModule.elt),(setModule.In a y)->(exists b:setModule.elt,(setModule.In b (setModule.diff t x))/\(weight a)<(weight b)))).

  Definition lt_set1_bis (weight:setModule.elt->nat)(s t:setModule.t) :=  (exists x:setModule.t,exists y:setModule.t,(setModule.Empty (setModule.inter t y))/\(setOtherProps.strict_Subset x t)/\(setModule.Equal s (setModule.union x y))/\(forall (a:setModule.elt),(setModule.In a y)->(exists b:setModule.elt,(setModule.In b (setModule.diff t x))/\(weight a)<(weight b)))).

  Lemma lt_set1_iff_lt_set1_bis : forall (weight:setModule.elt->nat)(weight_refl:forall (x y:setModule.elt),(setModule.E.eq x y)->(weight x = weight y))(s t:setModule.t),(lt_set1 weight s t)<->(lt_set1_bis weight s t).
    unfold lt_set1, lt_set1_bis;split;intros.
    ex_elim H x.
    ex_elim H y.
    exists (setModule.union x (setModule.inter y t)).
    exists (setModule.diff y t).
    split.
    intro;intro.
    generalize (setModule.inter_1 H2);intro.
    generalize (setModule.inter_2 H2);intro.
    elim (setModule.diff_2 H4).
    trivial.
    split.
    split.
    apply setProps.union_subset_3.
    red in H;tauto.
    apply setProps.inter_subset_2.
    intro.
    elim (setOtherProps.strict_Subset_iff_1 x t);intros.
    elim (H3 H);intros.
    clear H3;clear H4.
    assert (setModule.Subset (setModule.diff t x) (setModule.inter y t)).
    intro;intro.
    elim (H2 a);intros.
    generalize (setModule.diff_1 H3);intro.
    case (setModule.union_1 (H7 H8));intro;trivial.
    elim (setModule.diff_2 H3);trivial.
    assert (~setModule.Empty (setModule.inter y t)).
    intro.
    apply H6.
    intro;intro.
    elim (H4 a).
    apply H3;trivial.
    elim (select_heavier weight weight_refl (setModule.inter y t) H4);intro u;intro.
    and_elim.
    generalize (setModule.inter_1 p);intro.
    elim (H1 u H8);intro v;intro.
    and_elim.
    generalize (H3 v H9);intro.
    generalize (H7 v H11);intro.
    elim (lt_irrefl (weight v));omega.
    split.
    eapply setProps.equal_trans.
    apply H0.
    apply setProps.equal_sym.
    eapply setProps.equal_trans.
    apply setProps.union_assoc.
    apply setProps.union_equal_2.
    eapply setProps.equal_trans.
    apply setProps.union_sym.
    apply setProps.diff_inter_all.
    intros.
    generalize (setModule.diff_1 H2);intro.
    elim (H1 a H3);intro b;intros;and_elim.
    elim (setOtherProps.strict_Subset_iff_1 x t);intros.
    elim (H6 H);clear H6;clear H7;intros.
    elim (select_heavier weight weight_refl (setModule.diff t x) H7);intros;and_elim.
    assert (~setModule.In x0 y).
    intro.
    elim (H1 x0 H9);intros.
    and_elim.
    generalize (H8 x1 H10);intro.
    elim (lt_irrefl (weight x0));omega.
    exists x0.
    split.
    apply setModule.diff_3.
    eapply setModule.diff_1.
    apply p.
    intro.
    case (setModule.union_1 H10);intro.
    elim (setModule.diff_2 p).
    trivial.
    elim H9.
    eapply setModule.inter_1.
    apply H11.
    eapply lt_le_trans.
    apply H5.
    apply H8;auto.
    ex_elim H x.
    ex_elim H y.
    exists x.
    exists y.
    split;trivial.
    split;trivial.
  Qed.

  Lemma Subset_lt_set1_lt_set1 : forall (weight:setModule.elt->nat)(s t u:setModule.t),(setModule.Subset s t)->(lt_set1 weight t u)->(lt_set1 weight s u).
    unfold lt_set1;intros.
    ex_elim H0 x.
    ex_elim H0 y.
    exists (setModule.inter x s).
    exists (setModule.inter y s).
    split.
    eapply setOtherProps.Subset_strict_Subset.
    apply setProps.inter_subset_1.
    trivial.
    split.
    apply setProps.equal_sym.
    eapply setProps.equal_trans.
    apply setProps.equal_sym.
    apply setProps.union_inter_1.
    eapply setProps.equal_trans.
    apply setProps.equal_sym.
    apply setProps.inter_equal_1.
    apply H1.
    eapply setProps.equal_trans.
    apply setProps.inter_sym.
    apply setProps.inter_subset_equal.
    trivial.
    intros.
    generalize (setModule.inter_1 H3);intro.
    elim (H2 a H4).
    intro b;intros;and_elim.
    exists b.
    split;trivial.
    apply setModule.diff_3.
    eapply setModule.diff_1.
    apply H5.
    intro.
    apply (setModule.diff_2 H5).
    apply (setModule.inter_1 H7).
  Qed.

  Lemma cond_lt_set1 : forall (weight:setModule.elt->nat)(s t:setModule.t),(exists t':setModule.t,(setOtherProps.strict_Subset t' t)/\(forall (x:setModule.elt),(setModule.In x s)->(setModule.In x t')\/(exists y:setModule.elt,(setModule.In y (setModule.diff t t'))/\(weight x < weight y))))->(lt_set1 weight s t).
    intros.
    ex_elim H t'.
    unfold lt_set1.
    exists (setModule.inter t' s).
    exists (setModule.diff s (setModule.inter t' s)).
    split.
    eapply setOtherProps.Subset_strict_Subset.
    apply setProps.inter_subset_1.
    apply H.
    split.
    apply setProps.equal_sym.
    eapply setProps.equal_trans.
    apply setProps.union_sym.
    eapply setProps.equal_trans.
    apply setProps.union_equal_2.
    eapply setProps.equal_trans.
    apply setProps.equal_sym.
    apply setProps.inter_subset_equal.
    apply setProps.inter_subset_2.
    apply setProps.inter_sym.
    apply setProps.diff_inter_all.
    intros.
    generalize (setModule.diff_1 H1);intro.
    case (H0 a H2);intro.
    elim (setModule.diff_2 H1);auto.
    ex_elim H3 y.
    exists y.
    split;trivial.
    apply setModule.diff_3.
    eapply setModule.diff_1.
    apply H3.
    intro.
    elim (setModule.diff_2 H3).
    eapply setModule.inter_1.
    apply H5.
  Qed.

  Module IntEqset <: Eqset.
    Definition A := nat.
    Definition eqA := @eq nat.
    
    Theorem sid_theoryA : Setoid_Theory A eqA.
      unfold eqA;split;auto.
      intros.
      congruence.
    Qed.
  End IntEqset.
  
  Module msetModule := FiniteMultisetList IntEqset.
  Module msetModuleOrder := MultisetOrder msetModule.

  Definition msetSetoid : Setoid_Theory msetModule.Multiset msetModule.meq.
    split;auto with multisets.
    intros;eauto with multisets.
  Qed.

  Definition mset_add (n:nat) (m:msetModule.Multiset) := msetModule.union (msetModule.singleton n) m.

  Definition mset_add_weight (weight:setModule.elt->nat)(x:setModule.elt) (m:msetModule.Multiset) := mset_add (weight x) m.

  Lemma compat_op_mset_add_weight : forall (weight:setModule.elt->nat)(weight_refl:forall (x y:setModule.elt),(setModule.E.eq x y)->(weight x = weight y)),compat_op setModule.E.eq msetModule.meq (mset_add_weight weight).
    unfold compat_op;intros.
    unfold mset_add_weight.
    rewrite (weight_refl x x' H).
    unfold mset_add.
    apply msetModule.multeq_meq;intros.
    repeat (rewrite msetModule.union_mult).
    generalize (msetModule.meq_multeq H0);intro.
    rewrite (H1 x0).
    trivial.
  Qed.

  Lemma transpose_mset_add_weight : forall (weight:setModule.elt->nat),transpose msetModule.meq (mset_add_weight weight).
    unfold transpose;intros.
    unfold mset_add_weight.
    unfold mset_add.
    apply msetModule.multeq_meq;intros.
    repeat (rewrite msetModule.union_mult).
    rewrite plus_assoc.
    rewrite (plus_comm (msetModule.mult x0 (msetModule.singleton (weight x)))).
    rewrite plus_assoc;trivial.
  Qed.

  Lemma meq_refl : forall (s:msetModule.Multiset),(msetModule.meq s s).
    auto with multisets.
  Qed.

  Lemma meq_sym : forall (s t:msetModule.Multiset),(msetModule.meq s t)->(msetModule.meq t s).
    auto with multisets.
  Qed.

  Lemma meq_trans : forall (s t u:msetModule.Multiset),(msetModule.meq s t)->(msetModule.meq t u)->(msetModule.meq s u).
    intros.
    eauto with multisets.
  Qed.

  Lemma meq_empty : forall (s:msetModule.Multiset),(msetModule.meq (msetModule.union s (msetModule.empty)) s).
    intros.
    apply msetModule.multeq_meq;intros.
    rewrite msetModule.union_mult.
    rewrite msetModule.empty_mult.
    auto with arith.
  Qed.

  Lemma meq_union_equal_1 : forall (s t u:msetModule.Multiset),(msetModule.meq t u)->(msetModule.meq (msetModule.union s t) (msetModule.union s u)).
    intros.
    apply msetModule.multeq_meq;intros.
    generalize (msetModule.meq_multeq H);intros.
    repeat (rewrite msetModule.union_mult).
    rewrite (H0 x);trivial.
  Qed.

  Lemma meq_union_sym : forall (s t:msetModule.Multiset),(msetModule.meq (msetModule.union s t) (msetModule.union t s)).
    intros.
    apply msetModule.multeq_meq;intros.
    repeat (rewrite msetModule.union_mult).
    auto with arith.
  Qed.

  Lemma meq_union_assoc : forall (s t u:msetModule.Multiset),(msetModule.meq (msetModule.union (msetModule.union s t) u) (msetModule.union s (msetModule.union t u))).
    intros.
    apply msetModule.multeq_meq;intros.
    repeat (rewrite msetModule.union_mult).
    auto with arith.
  Qed.

  Lemma meq_union_equal_2 : forall (s t u:msetModule.Multiset),(msetModule.meq t u)->(msetModule.meq (msetModule.union t s) (msetModule.union u s)).
    intros.
    eapply meq_trans.
    apply meq_union_sym.
    apply meq_sym.
    eapply meq_trans.
    apply meq_union_sym.
    apply meq_sym.
    apply meq_union_equal_1.
    trivial.
  Qed.

  Definition weight_set (weight:setModule.elt->nat) (s:setModule.t) := setModule.fold (mset_add_weight weight) s msetModule.empty.

  Lemma Equal_weight_set_meq : forall (weight:setModule.elt->nat)(weight_refl:forall (x y:setModule.elt),(setModule.E.eq x y)->(weight x = weight y))(s t:setModule.t),(setModule.Equal s t)->(msetModule.meq (weight_set weight s) (weight_set weight t)).
    intros.
    unfold weight_set.
    apply (setProps.fold_equal msetSetoid (compat_op_mset_add_weight weight weight_refl) (transpose_mset_add_weight weight) msetModule.empty H).
  Qed.

  Lemma Empty_weight_set_empty : forall (weight:setModule.elt->nat)(weight_refl:forall (x y:setModule.elt),(setModule.E.eq x y)->(weight x = weight y))(s:setModule.t),(setModule.Empty s)->(msetModule.meq (weight_set weight s) msetModule.empty).
    intros.
    unfold weight_set.
    generalize (setOtherProps.empty_is_empty s H);intro.
    eapply meq_trans.
    apply (setProps.fold_equal msetSetoid (compat_op_mset_add_weight weight weight_refl) (transpose_mset_add_weight weight) msetModule.empty H0).
    apply (setProps.fold_empty msetSetoid (mset_add_weight weight) msetModule.empty).
  Qed.

  Lemma weight_set_Add : forall (weight:setModule.elt->nat)(weight_refl:forall (x y:setModule.elt),(setModule.E.eq x y)->(weight x = weight y))(s s':setModule.t)(x:setModule.elt),(~setModule.In x s)->(setProps.Add x s s')->(msetModule.meq (weight_set weight s') (msetModule.union (msetModule.singleton (weight x)) (weight_set weight s))).
    intros.
    eapply meq_trans.
    unfold weight_set.
    apply meq_refl.
    generalize (setOtherProps.add_is_add s s' x H0);intro.
    eapply meq_trans.
    apply (setProps.fold_equal msetSetoid (compat_op_mset_add_weight weight weight_refl) (transpose_mset_add_weight weight) msetModule.empty H1).
    eapply meq_trans.
    apply (setProps.fold_add msetSetoid (compat_op_mset_add_weight weight weight_refl) (transpose_mset_add_weight weight) msetModule.empty H).
    fold (weight_set weight s).
    unfold mset_add_weight.
    unfold mset_add.
    apply meq_refl.
  Qed.

  Lemma weight_set_empty_Empty : forall (weight:setModule.elt->nat)(weight_refl:forall (x y:setModule.elt),(setModule.E.eq x y)->(weight x = weight y))(s:setModule.t),(msetModule.meq (weight_set weight s) msetModule.empty)->(setModule.Empty s).
    intros weight weight_refl.
    intro s.
    unfold weight_set.
    apply setProps.set_induction with (P:=(fun s => msetModule.meq (setModule.fold (mset_add_weight weight) s msetModule.empty) msetModule.empty -> setModule.Empty s));clear s;intros.
    trivial.
    fold (weight_set weight s') in H2.
    generalize (weight_set_Add weight weight_refl s s' x H0 H1);intro.
    assert (msetModule.meq (msetModule.union (msetModule.singleton (weight x)) (weight_set weight s)) (msetModule.empty)).
    eapply meq_trans.
    apply meq_sym.
    apply H3.
    trivial.
    generalize (msetModule.meq_multeq H4);intro.
    generalize (H5 (weight x));intro.
    rewrite msetModule.union_mult in H6.
    rewrite msetModule.empty_mult in H6.
    rewrite msetModule.singleton_mult_in in H6.
    discriminate H6.
    auto with multisets.
  Qed.
    
  Lemma weight_set_disjoint_union : forall (weight:setModule.elt->nat)(weight_refl:forall (x y:setModule.elt),(setModule.E.eq x y)->(weight x = weight y))(s t:setModule.t),(setModule.Empty (setModule.inter s t))->(msetModule.meq (weight_set weight (setModule.union s t)) (msetModule.union (weight_set weight s) (weight_set weight t))).
    intros weight weight_refl.
    intro s.
    apply setProps.set_induction with (P:=(fun s => forall (t:setModule.t),(setModule.Empty (setModule.inter s t))->(msetModule.meq (weight_set weight (setModule.union s t)) (msetModule.union (weight_set weight s) (weight_set weight t)))));clear s;intros.
    assert (setModule.Equal (setModule.union s t) t).
    apply setProps.union_subset_equal.
    intro;intro.
    elim (H a);trivial.
    generalize (Equal_weight_set_meq weight weight_refl (setModule.union s t) t H1);intro.
    eapply meq_trans.
    apply H2.
    apply msetModule.multeq_meq;intro.
    rewrite msetModule.union_mult.
    generalize (Empty_weight_set_empty weight weight_refl s H);intro.
    generalize (msetModule.meq_multeq H3);intro.
    generalize (H4 x);intro.
    rewrite msetModule.empty_mult in H5.
    rewrite H5.
    auto with arith.
    assert (setModule.Empty (setModule.inter s t)).
    intro;intro.
    elim (H2 a).
    generalize (setModule.inter_1 H3);intro.
    generalize (setModule.inter_2 H3);intro.
    apply setModule.inter_3;trivial.
    elim (H1 a);auto.
    assert (~setModule.In x (setModule.union s t)).
    intro.
    case (setModule.union_1 H4);intro;auto.
    elim (H2 x).
    apply setModule.inter_3;trivial.
    elim (H1 x);auto.
    assert (setProps.Add x (setModule.union s t) (setModule.union s' t)).
    split;intros.
    case (setModule.union_1 H5);intro.
    elim (H1 y);intros.
    case (H7 H6);intro.
    left;trivial.
    right.
    apply setModule.union_2;trivial.
    right.
    apply setModule.union_3;trivial.
    case H5;intro.
    apply setModule.union_2.
    elim (H1 y);intros.
    apply H8.
    left;trivial.
    case (setModule.union_1 H6);intro.
    apply setModule.union_2.
    elim (H1 y);auto.
    apply setModule.union_3.
    trivial.
    eapply meq_trans.
    apply (weight_set_Add weight weight_refl).
    apply H4.
    trivial.
    apply meq_sym.
    eapply meq_trans.
    apply meq_union_equal_2.
    apply (weight_set_Add weight weight_refl).
    apply H0.
    trivial.
    apply meq_sym.
    eapply meq_trans.
    apply meq_union_equal_1.
    apply H.
    trivial.
    apply meq_sym.
    apply meq_union_assoc.
  Qed.

  Lemma weight_set_image : forall (weight:setModule.elt->nat)(weight_refl:forall (x y:setModule.elt),(setModule.E.eq x y)->(weight x = weight y))(s:setModule.t)(x:setModule.elt),(setModule.In x s)->(msetModuleOrder.MSet.member (weight x) (weight_set weight s)).
    intros.
    assert (~(setModule.In x (setModule.remove x s)));auto.
    assert (setProps.Add x (setModule.remove x s) s);auto.
    unfold msetModuleOrder.MSet.member.
    generalize (weight_set_Add weight weight_refl (setModule.remove x s) s x H0 H1);intro.
    generalize (msetModule.meq_multeq H2);intros.
    rewrite (H3 (weight x)).
    rewrite msetModule.union_mult.
    rewrite msetModule.singleton_mult_in.
    simpl;auto with arith.
    unfold msetModule.Sid.eqA.
    unfold IntEqset.eqA.
    trivial.
  Qed.

  Lemma weight_set_inversion : forall (weight:setModule.elt->nat)(weight_refl:forall (x y:setModule.elt),(setModule.E.eq x y)->(weight x = weight y))(s:setModule.t)(n:nat),(msetModuleOrder.MSet.member n (weight_set weight s))->(exists x:setModule.elt,(setModule.In x s)/\(weight x)=n).
    intros weight weight_refl.
    intro.
    unfold msetModuleOrder.MSet.member.
    apply setProps.set_induction with (P:=(fun s => forall (n:nat),(msetModule.mult n (weight_set weight s) > 0)->(exists x:setModule.elt,(setModule.In x s)/\(weight x)=n)));clear s;intros.
    generalize (Empty_weight_set_empty weight weight_refl s H);intro.
    generalize (msetModule.meq_multeq H1);intro.
    rewrite (H2 n) in H0.
    rewrite msetModule.empty_mult in H0.
    inversion H0.
    generalize (weight_set_Add weight weight_refl s s' x H0 H1);intro.
    generalize (msetModule.meq_multeq H3);intro.
    rewrite (H4 n) in H2.
    rewrite msetModule.union_mult in H2.
    case (case_plus_gt_0 (msetModule.mult n (msetModule.singleton (weight x))) (msetModule.mult n (weight_set weight s)) H2);intro.
    case (eq_nat_dec n (weight x));intro.
    exists x.
    split.
    elim (H1 x);auto.
    symmetry;trivial.
    rewrite msetModule.singleton_mult_notin in H5.
    inversion H5.
    intro.
    apply n0.
    tauto.
    elim (H n H5);intro y;intros.
    and_elim.
    exists y.
    split.
    elim (H1 y);auto.
    trivial.
  Qed.

  Lemma lt_set1_impl_lt_weight_mul : forall (weight:setModule.elt->nat)(weight_refl:forall (x y:setModule.elt),(setModule.E.eq x y)->(weight x = weight y))(s t:setModule.t),(lt_set1 weight s t)->(msetModuleOrder.MultisetLT gt (weight_set weight s) (weight_set weight t)).
    intros.
    elim (lt_set1_iff_lt_set1_bis weight weight_refl s t);intros.
    generalize (H0 H);clear H;clear H0;clear H1;intros.
    ex_elim H x.
    ex_elim H y.
    assert (setModule.Empty (setModule.inter x y)).
    intro;intro.
    elim H0;intros.
    elim (H a).
    apply setModule.inter_3.
    apply H4.
    eapply setModule.inter_1.
    apply H3.
    eapply setModule.inter_2.
    apply H3.
    unfold msetModuleOrder.MultisetLT.
    unfold transp.
    elim H0;intros.
    generalize (setOtherProps.Subset_diff_union x t H4);intro.
    assert (setModule.Empty (setModule.inter (setModule.diff t x) x)).
    intro;intro.
    generalize (setModule.inter_1 H7);intro.
    generalize (setModule.inter_2 H7);intro.
    elim (setModule.diff_2 H8);trivial.
    elim (setOtherProps.strict_Subset_iff_1 x t);intros.
    elim (H8 H0);intros.
    eapply msetModuleOrder.MSetGT.
    intro.
    apply H11.
    apply (weight_set_empty_Empty weight weight_refl).
    apply H12.
    eapply meq_trans.
    apply (Equal_weight_set_meq weight weight_refl).
    apply setProps.equal_sym.
    apply H6.
    apply meq_sym.
    eapply meq_trans.
    apply meq_union_sym.
    apply meq_sym.
    apply (weight_set_disjoint_union weight weight_refl).
    trivial.
    eapply meq_trans.
    apply (Equal_weight_set_meq weight weight_refl).
    apply H1.
    apply (weight_set_disjoint_union weight weight_refl).
    trivial.
    intros.
    elim (weight_set_inversion weight weight_refl y y0 H12).
    intro a;intros.
    and_elim.
    elim (H2 a H13);intro b;intros;and_elim.
    exists (weight b).
    apply (weight_set_image weight weight_refl).
    trivial.
    rewrite <- H14.
    auto with arith.
  Qed.

  Definition lt_set2 (weight:setModule.elt->nat)(s t:setModule.t) := msetModuleOrder.MultisetLT gt (weight_set weight s) (weight_set weight t).

  Definition lt_set3 (weight:setModule.elt->nat)(s t:setModule.t) := msetModuleOrder.MultisetLt gt (weight_set weight s) (weight_set weight t).

  Lemma lt_set2_impl_lt_set3 : forall (weight:setModule.elt->nat)(s t:setModule.t),(lt_set2 weight s t)->(lt_set3 weight s t).
    intros s t.
    unfold lt_set2, lt_set3.
    unfold msetModuleOrder.MultisetLt.
    unfold msetModuleOrder.MultisetLT.
    unfold transp.
    intros.
    eapply msetModuleOrder.direct_subset_red.
    intros.
    unfold msetModule.Sid.eqA, IntEqset.eqA in H0.
    unfold msetModule.Sid.eqA, IntEqset.eqA in H1.
    subst x;subst y;auto.
    auto.
  Qed.

  Lemma well_founded_lt_set2 : forall (weight:setModule.elt->nat),well_founded (lt_set2 weight).
    intro.
    eapply wf_incl.
    unfold Relation_Definitions.inclusion.
    apply lt_set2_impl_lt_set3.
    apply (wf_inverse_image setModule.t msetModule.Multiset (msetModuleOrder.MultisetLt gt) (fun s => weight_set weight s)).
    apply msetModuleOrder.mord_wf;intros.
    unfold msetModule.Sid.eqA, IntEqset.eqA in H.
    unfold msetModule.Sid.eqA, IntEqset.eqA in H0.
    subst x;subst y;auto.
    generalize lt_wf.
    apply wf_incl.
    unfold Relation_Definitions.inclusion.
    unfold transp.
    intros;auto.
  Qed.

  Lemma lt_set1_impl_lt_set2 : forall (weight:setModule.elt->nat)(weight_refl:forall (x y:setModule.elt),(setModule.E.eq x y)->(weight x = weight y))(s t:setModule.t),(lt_set1 weight s t)->(lt_set2 weight s t).
    unfold lt_set2.
    apply (lt_set1_impl_lt_weight_mul).
  Qed.
    
  Theorem well_founded_lt_set1 : forall (weight:setModule.elt->nat)(weight_refl:forall (x y:setModule.elt),(setModule.E.eq x y)->(weight x = weight y)),well_founded (lt_set1 weight).
    intros.
    eapply wf_incl.
    unfold Relation_Definitions.inclusion.
    apply (lt_set1_impl_lt_set2 weight weight_refl).
    apply well_founded_lt_set2.
  Qed.
End SetOrder.



