(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import deBruijnType.
Require Import name.
Require Import missing.
Require Import Arith.
Require Import tactics.

Module SubstLemmaMap.
  Section subst_lemma.
    Hypothesis t : Set.
    Hypothesis deBruijn_t : deBruijn.deBruijn_base t.
    Hypothesis to : Set.
    Hypothesis deBruijn_to : deBruijn.deBruijn_base to.
    Hypothesis t_subst_to : deBruijn.deBruijn_subst t deBruijn_t to deBruijn_to.
    Hypothesis t' : Set.
    Hypothesis map_fun: (nat->t->t)->nat->t'->t'.
    Hypothesis map_ext:deBruijn.map_struct t t' map_fun.
    Definition deBruijn_t' := deBruijn.deBruijn_map_struct t deBruijn_t t' map_fun map_ext.
    Definition t'_subst_to := deBruijn.deBruijn_s_map_struct t deBruijn_t to deBruijn_to t_subst_to t' map_fun map_ext.
    Hypothesis to_subst_to : deBruijn.deBruijn_subst to deBruijn_to to deBruijn_to.
    
    Hypothesis t_low_subst_lemma : forall (x:t)(d d' k k':nat)(es fs:list to),deBruijn.low_subst t_subst_to (d+d'+k) k' es (deBruijn.low_subst t_subst_to d k fs x)=deBruijn.low_subst t_subst_to d k (map (deBruijn.low_subst to_subst_to (d'+k) k' es) fs) (deBruijn.low_subst t_subst_to (d+d'+length fs) k' es x).

    Lemma low_subst_lemma : forall (x:t')(d d' k k':nat)(es fs:list to),deBruijn.low_subst t'_subst_to (d+d'+k) k' es (deBruijn.low_subst t'_subst_to d k fs x)=deBruijn.low_subst t'_subst_to d k (map (deBruijn.low_subst to_subst_to (d'+k) k' es) fs) (deBruijn.low_subst t'_subst_to (d+d'+length fs) k' es x).
      intros.
      simpl.
      generalize (deBruijn.map_compose map_ext (d'+k) 0 0 (d'+length fs) (fun d0 => deBruijn.low_subst t_subst_to d0 k' es) (fun d0 : nat => deBruijn.low_subst t_subst_to d0 k fs) (fun d0 : nat => deBruijn.low_subst t_subst_to d0 k (map (deBruijn.low_subst to_subst_to (d' + k) k' es) fs)) (fun d0 : nat => deBruijn.low_subst t_subst_to d0 k' es)).
      simpl;intros.
      replace (d+d'+k) with (d'+k+d);try omega.
      replace (d+d'+length fs) with (d'+length fs+d);try omega.
      apply H;clear H.
      intros.
      replace (d'+k+d0) with (d0+d'+k);try omega.
      replace (d'+length fs+d0) with (d0+d'+length fs);try omega.
      apply t_low_subst_lemma.
    Qed.
  End subst_lemma.
End SubstLemmaMap.

Module SubstLemmaMap2.
  Section subst_lemma.
    Hypothesis t1 : Set.
    Hypothesis deBruijn_t1 : deBruijn.deBruijn_base t1.
    Hypothesis t2 : Set.
    Hypothesis deBruijn_t2 : deBruijn.deBruijn_base t2.
    Hypothesis to : Set.
    Hypothesis deBruijn_to : deBruijn.deBruijn_base to.
    Hypothesis t1_subst_to : deBruijn.deBruijn_subst t1 deBruijn_t1 to deBruijn_to.
    Hypothesis t2_subst_to : deBruijn.deBruijn_subst t2 deBruijn_t2 to deBruijn_to.
    Hypothesis t : Set.
    Hypothesis map2_fun: (nat->t1->t1)->(nat->t2->t2)->nat->t->t.
    Hypothesis map2_ext:deBruijn.map2_struct t1 t2 t map2_fun.
    Definition deBruijn_t := deBruijn.deBruijn_map2_struct t1 deBruijn_t1 t2 deBruijn_t2 t map2_fun map2_ext.
    Definition t_subst_to := deBruijn.deBruijn_s_map2_struct t1 deBruijn_t1 t2 deBruijn_t2 to deBruijn_to t1_subst_to t2_subst_to t map2_fun map2_ext.
    Hypothesis to_subst_to : deBruijn.deBruijn_subst to deBruijn_to to deBruijn_to.
    
    Hypothesis t1_low_subst_lemma : forall (x:t1)(d d' k k':nat)(es fs:list to),deBruijn.low_subst t1_subst_to (d+d'+k) k' es (deBruijn.low_subst t1_subst_to d k fs x)=deBruijn.low_subst t1_subst_to d k (map (deBruijn.low_subst to_subst_to (d'+k) k' es) fs) (deBruijn.low_subst t1_subst_to (d+d'+length fs) k' es x).
    Hypothesis t2_low_subst_lemma : forall (x:t2)(d d' k k':nat)(es fs:list to),deBruijn.low_subst t2_subst_to (d+d'+k) k' es (deBruijn.low_subst t2_subst_to d k fs x)=deBruijn.low_subst t2_subst_to d k (map (deBruijn.low_subst to_subst_to (d'+k) k' es) fs) (deBruijn.low_subst t2_subst_to (d+d'+length fs) k' es x).

    Lemma low_subst_lemma : forall (x:t)(d d' k k':nat)(es fs:list to),deBruijn.low_subst t_subst_to (d+d'+k) k' es (deBruijn.low_subst t_subst_to d k fs x)=deBruijn.low_subst t_subst_to d k (map (deBruijn.low_subst to_subst_to (d'+k) k' es) fs) (deBruijn.low_subst t_subst_to (d+d'+length fs) k' es x).
      intros.
      simpl.
      generalize (deBruijn.map2_compose map2_ext (d'+k) 0 0 (d'+length fs) (fun d0 => deBruijn.low_subst t1_subst_to d0 k' es) (fun d0 : nat => deBruijn.low_subst t1_subst_to d0 k fs) (fun d0 : nat => deBruijn.low_subst t1_subst_to d0 k (map (deBruijn.low_subst to_subst_to (d' + k) k' es) fs)) (fun d0 : nat => deBruijn.low_subst t1_subst_to d0 k' es) (fun d0 => deBruijn.low_subst t2_subst_to d0 k' es) (fun d0 : nat => deBruijn.low_subst t2_subst_to d0 k fs) (fun d0 : nat => deBruijn.low_subst t2_subst_to d0 k (map (deBruijn.low_subst to_subst_to (d' + k) k' es) fs)) (fun d0 : nat => deBruijn.low_subst t2_subst_to d0 k' es)).
      simpl;intros.
      replace (d+d'+k) with (d'+k+d);try omega.
      replace (d+d'+length fs) with (d'+length fs+d);try omega.
      apply H;clear H.
      intros.
      replace (d'+k+d0) with (d0+d'+k);try omega.
      replace (d'+length fs+d0) with (d0+d'+length fs);try omega.
      apply t1_low_subst_lemma.
      intros.
      replace (d'+k+d0) with (d0+d'+k);try omega.
      replace (d'+length fs+d0) with (d0+d'+length fs);try omega.
      apply t2_low_subst_lemma.
    Qed.
  End subst_lemma.
End SubstLemmaMap2.

Module SubstLemmaGen.
  Section subst_lemma.
    Hypothesis to : Set.
    Hypothesis map_fun : (nat->Name.t->Name.t)->nat->to->to.
    Hypothesis map_str : deBruijn.map_struct Name.t to map_fun.
    Definition deBruijn_to :=  (deBruijn.deBruijn_map_struct Name.t Name.deBruijn_base to map_fun map_str).
    Hypothesis Hextension : deBruijn.deBruijn_extension Name.t Name.deBruijn_base to deBruijn_to.
    Hypothesis deBruijn_ext : deBruijn.deBruijn_extension Name.t Name.deBruijn_base to deBruijn_to.
    Hypothesis map_gen_fun : (nat->Name.t->to)->nat->to->to.
    Hypothesis map_ext : deBruijn.map_extension Name.t Name.deBruijn_base to deBruijn_to deBruijn_ext map_fun map_str map_gen_fun.
    Definition t_subst_to := Name.deBruijn_subst_gen to deBruijn_to deBruijn_ext.
    Definition to_subst_to := deBruijn.deBruijn_subst_gen_subst Name.t Name.deBruijn_base to map_fun map_str deBruijn_ext t_subst_to map_gen_fun map_ext.
    
    Lemma low_subst_name : forall (d k:nat)(es:list to)(x:Name.t),(deBruijn.low_subst_gen t_subst_to d k es x)=(Name.low_subst_generic to (fun d0 k0 : nat => map_fun (fun d1 : nat => Name.lift d1 k0) d0) (deBruijn.j deBruijn_ext) d k es x).
      intros.
      reflexivity.
    Qed.
    
    Lemma low_subst_j : forall (d k:nat)(es:list to)(x:Name.t),deBruijn.low_subst to_subst_to d k es (deBruijn.j deBruijn_ext x) =  Name.low_subst_generic to (fun d0 k0 : nat => map_fun (fun d1 : nat => Name.lift d1 k0) d0) (deBruijn.j deBruijn_ext) d k es x. 
      simpl.
      intros.
      rewrite (deBruijn.map_gen_j map_ext);auto.
    Qed.

    Lemma low_subst_lemma : forall (x:Name.t)(d d' k k':nat)(es:list to)(fs:list to),deBruijn.low_subst to_subst_to (d+d'+k) k' es (deBruijn.low_subst_gen t_subst_to d k fs x)=deBruijn.low_subst to_subst_to d k (map (deBruijn.low_subst to_subst_to (d'+k) k' es) fs) (deBruijn.low_subst_gen t_subst_to (d+d'+length fs) k' es x).
      intros.
      destruct x as [x].
      repeat (rewrite low_subst_name).
      unfold Name.low_subst_generic.
      unfold DeBruijnNat.low_subst_gen.
      unfold DeBruijnNat.low_subst_generic.
      unfold DeBruijnNat.low_subst_generic_aux.
      case le_lt_dec;intro.
      case le_lt_dec;intro.
      elim (List_nth_Some to fs (x-d));intros.
      case (Option_elim_option to (nth_error fs (x-d)));intro H1;try (rewrite H1).
      ex_elim H1 f.
      rewrite H1.
      assert (x-d < length fs).
      apply H0.
      exists f.
      auto.
      try_absurd_arith.
      assert (length fs <= x-d).
      case (le_lt_dec (length fs) (x-d));auto.
      intros.
      elim (H l1).
      intros.
      rewrite H1 in H2.
      discriminate H2.
      elim (List_nth_Some to es (x-(d+d'+length fs)));intros.
      case (Option_elim_option to (nth_error es (x-(d+d'+length fs))));intro H5;try (rewrite H5).
      ex_elim H5 e.
      rewrite H5.
      rewrite low_subst_j.
      unfold Name.low_subst_generic.
      unfold DeBruijnNat.low_subst_gen.
      unfold DeBruijnNat.low_subst_generic.
      unfold DeBruijnNat.low_subst_generic_aux.
      case le_lt_dec;intro;try_absurd_arith.
      clear H;clear H0;clear H3;clear H4.
      replace (x-length fs+k-(d+d'+k)) with (x-(d+d'+length fs));try omega.
      rewrite H5.
      generalize (deBruijn.lift_compose deBruijn_to 0 d (d'+length fs) 0).
      intro H6.
      rewrite plus_assoc in H6.
      repeat (rewrite plus_0_r in H6).
      repeat (rewrite plus_0_l in H6).      
      simpl in H6.
      rewrite <- H6.
      generalize (deBruijn.low_subst_lift_nil to_subst_to d k d' (map (deBruijn.low_subst to_subst_to (d' + k) k' es) fs)).
      intro H7.
      rewrite List_length_map in H7.
      simpl in H7.
      simpl.
      rewrite H7.
      generalize (deBruijn.lift_compose deBruijn_to 0 d (k+d') 0);simpl;intro H8.
      rewrite plus_0_r in H8.
      rewrite H8.
      rewrite (plus_comm k d').
      rewrite plus_assoc.
      rewrite plus_0_r;auto.
      assert (length es <= x-(d+d'+length fs)).
      case (le_lt_dec (length es) (x-(d+d'+length fs)));intro;auto.
      elim H3;intros;auto.
      rewrite H5 in H6;discriminate H6.
      rewrite low_subst_j.
      unfold Name.low_subst_generic.
      unfold DeBruijnNat.low_subst_gen.
      unfold DeBruijnNat.low_subst_generic.
      unfold DeBruijnNat.low_subst_generic_aux.
      case le_lt_dec;intro;try_absurd_arith.
      clear H;clear H0;clear H3;clear H4.
      replace (x-length fs+k-(d+d'+k)) with (x-(d+d'+length fs));try omega.
      rewrite H5.
      rewrite low_subst_j.
      unfold Name.low_subst_generic.
      unfold DeBruijnNat.low_subst_gen.
      unfold DeBruijnNat.low_subst_generic.
      unfold DeBruijnNat.low_subst_generic_aux.
      case le_lt_dec;intro;try_absurd_arith.
      rewrite List_length_map.
      replace (x - length fs + k - length es+k') with (x-length es+k'-length fs+k);try omega.
      rewrite List_nth_map.
      assert (nth_error fs (x-length es+k'-d) = None).
      elim (List_nth_Some to fs (x-length es+k'-d));intros.
      destruct (nth_error fs (x-length es+k'-d));auto.
      assert (x-length es + k' - d < length fs).
      apply H0.
      exists t;auto.
      clear H;clear H0.
      try_absurd_arith.
      rewrite H.
      reflexivity.
      elim (List_nth_Some to fs (x-d));intros.
      rewrite low_subst_j.
      unfold Name.low_subst_generic.
      unfold DeBruijnNat.low_subst_gen.
      unfold DeBruijnNat.low_subst_generic.
      unfold DeBruijnNat.low_subst_generic_aux.
      case le_lt_dec;intro;try_absurd_arith.
      case (Option_elim_option to (nth_error fs (x-d)));intro H1;try (rewrite H1).
      ex_elim H1 f.
      rewrite H1.
      assert (x-d < length fs).
      apply H0.
      exists f;auto.
      rewrite List_nth_map.
      rewrite H1.
      unfold option_map.
      generalize (deBruijn.low_subst_lift_comm to_subst_to 0 (d'+k) d k');simpl;intros.
      rewrite <- H3.
      replace (d+d'+k) with (d'+k+d);try omega.
      auto.
      assert (length fs <= x-d).
      case (le_lt_dec (length fs) (x-d));intros;auto.
      elim (H l2);intros.
      rewrite H1 in H2;discriminate H2.
      rewrite List_nth_map.
      rewrite H1.
      unfold option_map.
      rewrite List_length_map.
      rewrite low_subst_j.
      unfold Name.low_subst_generic.
      unfold DeBruijnNat.low_subst_gen.
      unfold DeBruijnNat.low_subst_generic.
      unfold DeBruijnNat.low_subst_generic_aux.
      case le_lt_dec;intro;try_absurd_arith.
      auto.
      case le_lt_dec;intro;try_absurd_arith.
      repeat (rewrite low_subst_j).
      unfold Name.low_subst_generic.
      unfold DeBruijnNat.low_subst_gen.
      unfold DeBruijnNat.low_subst_generic.
      unfold DeBruijnNat.low_subst_generic_aux.
      case le_lt_dec;intro;try_absurd_arith.
      case le_lt_dec;intro;try_absurd_arith.
      auto.
      Opaque deBruijn_to.
      Opaque t_subst_to.
      Opaque to_subst_to.
    Qed.
    Transparent deBruijn_to.
    Transparent t_subst_to.
    Transparent to_subst_to.
  End subst_lemma.
End SubstLemmaGen.
