(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import name.
Require Import deBruijnType.
Require Import evalSpi.
Require Import opSpi.
Require Import expressionSpi.
Require Import messageSpi.
Require Import formulaeSpi.
Require Import calculi.
Require Import missing.
Require Import tactics.
Require Import Arith.

Module SubstSpi.
  Import EvalSpi.
  Import OpSpi.

  Definition msg_subst (sigma:list MessageSpi.t) := map expr_of_mesg sigma.

  (* substitutions *)
  Lemma low_subst_id : forall (d:nat)(e:ExpressionSpi.t),(deBruijn.low_subst ExpressionSpi.deBruijn_subst  d 0 nil e)=e.
    induction e;intros.
    simpl.
    destruct t;simpl.
    unfold Name.low_subst_generic.
    unfold DeBruijnNat.low_subst_gen.
    unfold DeBruijnNat.low_subst_generic.
    unfold DeBruijnNat.low_subst_generic_aux.
    case le_lt_dec;try reflexivity.
    intros.
    rewrite List_nth_nil.
    simpl.
    rewrite <- minus_n_O.
    rewrite plus_0_r.
    auto.
    rewrite ExpressionSpi.low_subst_EPair;congruence.
    rewrite ExpressionSpi.low_subst_EEnc_s;congruence.
    rewrite ExpressionSpi.low_subst_EEnc_a;congruence.
    rewrite ExpressionSpi.low_subst_EFst;congruence.
    rewrite ExpressionSpi.low_subst_ESnd;congruence.
    rewrite ExpressionSpi.low_subst_EDec_s;congruence.
    rewrite ExpressionSpi.low_subst_EDec_a;congruence.
    rewrite ExpressionSpi.low_subst_EOp;congruence.
  Qed.

  Lemma low_subst_evalc : forall (sigma:list ExpressionSpi.t),(List_forall evalc_succeeds sigma)->forall (e:ExpressionSpi.t)(m:MessageSpi.t),(evalc e = Some m)->forall (d k:nat),(evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma e))=(Some (deBruijn.low_subst MessageSpi.deBruijn_subst d k (evalc_list sigma) m)).
    induction e;intros.
    simpl in H0.
    injection H0;intros.
    destruct t. 
    subst m.
    simpl.
    unfold Name.low_subst_generic.
    unfold DeBruijnNat.low_subst_gen.
    unfold DeBruijnNat.low_subst_generic.
    unfold DeBruijnNat.low_subst_generic_aux.
    case le_lt_dec.
    case (Option_elim_option ExpressionSpi.t (nth_error sigma (n-d))).
    intros.
    ex_elim s e.
    elim (evalc_succeeds_list_Some sigma H (n-d) e);auto;intros;and_elim.
    rewrite s.
    rewrite evalc_map.
    rewrite H2.
    simpl.
    rewrite H1.
    auto.
    intros.
    apply (deBruijn.lift_injective Name.deBruijn_base).
    intros.
    rewrite e.
    rewrite (evalc_succeeds_list_None sigma H (n-d));auto.
    rewrite <- evalc_succeeds_list_length;auto.
    intros.
    reflexivity.
    rewrite ExpressionSpi.low_subst_EPair.
    rewrite evalc_EPair in H0.
    destruct (evalc e1);try (discriminate H0).
    destruct (evalc e2);try (discriminate H0).
    injection H0;intros;subst m.
    rewrite evalc_EPair.
    rewrite (IHe1 t (refl_equal (Some t)) d k).
    rewrite (IHe2 t0 (refl_equal (Some t0)) d k).
    reflexivity.
    rewrite ExpressionSpi.low_subst_EEnc_s.
    rewrite evalc_EEnc_s in H0.
    destruct (evalc e1);try (discriminate H0).
    destruct (evalc e2);try (discriminate H0).
    injection H0;intros;subst m.
    rewrite evalc_EEnc_s.
    rewrite (IHe1 t (refl_equal (Some t)) d k).
    rewrite (IHe2 t0 (refl_equal (Some t0)) d k).
    reflexivity.
    rewrite ExpressionSpi.low_subst_EEnc_a.
    rewrite evalc_EEnc_a in H0.
    destruct (evalc e1);try (discriminate H0).
    destruct (evalc e2);try (discriminate H0).
    injection H0;intros;subst m.
    rewrite evalc_EEnc_a.
    rewrite (IHe1 t (refl_equal (Some t)) d k).
    rewrite (IHe2 t0 (refl_equal (Some t0)) d k).
    reflexivity.
    rewrite ExpressionSpi.low_subst_EFst.
    rewrite evalc_EFst in H0.
    destruct (evalc e);try (discriminate H0).
    rewrite evalc_EFst.
    rewrite (IHe  t (refl_equal (Some t)) d k).
    destruct t;try (discriminate H0).
    injection H0;intros;subst m.
    reflexivity.
    rewrite ExpressionSpi.low_subst_ESnd.
    rewrite evalc_ESnd in H0.
    destruct (evalc e);try (discriminate H0).
    rewrite evalc_ESnd.
    rewrite (IHe t);auto.
    destruct t;try (discriminate H0).
    injection H0;intros;subst m.
    reflexivity.
    rewrite ExpressionSpi.low_subst_EDec_s.
    rewrite evalc_EDec_s in H0.
    rewrite evalc_EDec_s.
    destruct (evalc e1);try (discriminate H0).
    rewrite (IHe1 t);auto.
    destruct t;try (discriminate H0).
    destruct (evalc e2);try (discriminate H0).
    rewrite (IHe2 t);auto.
    destruct (MessageSpi.Ord.compare t2 t);try (discriminate H0).
    rewrite MessageSpi.low_subst_MEnc_s.
    rewrite e.
    injection H0;intros;subst t1.
    destruct (MessageSpi.Ord.compare);auto.
    elim (MessageSpi.Ord.lt_not_eq (deBruijn.low_subst MessageSpi.deBruijn_subst d k (evalc_list sigma) t) (deBruijn.low_subst MessageSpi.deBruijn_subst d k (evalc_list sigma) t));auto.
    elim (MessageSpi.Ord.lt_not_eq (deBruijn.low_subst MessageSpi.deBruijn_subst d k (evalc_list sigma) t) (deBruijn.low_subst MessageSpi.deBruijn_subst d k (evalc_list sigma) t));auto.
    rewrite ExpressionSpi.low_subst_EDec_a.
    rewrite evalc_EDec_a in H0.
    rewrite evalc_EDec_a.
    destruct (evalc e1);try (discriminate H0).
    rewrite (IHe1 t);auto.
    destruct t;try (discriminate H0).
    destruct (evalc e2);try (discriminate H0).
    rewrite (IHe2 t);auto.
    rewrite MessageSpi.low_subst_MEnc_a.
    destruct t2;try (simpl in H0;discriminate H0).
    rewrite MessageSpi.low_subst_MOp.
    destruct o;try (simpl in H0;discriminate H0);unfold MessageSpi.inverse_key in H0.
    unfold MessageSpi.inverse_key.
    destruct (MessageSpi.Ord.compare (MessageSpi.MOp Priv t2) t);try (discriminate H0).
    rewrite <- e.
    rewrite MessageSpi.low_subst_MOp.
    injection H0;intros;subst t1.
    destruct (MessageSpi.Ord.compare);auto.
    elim (MessageSpi.Ord.lt_not_eq (MessageSpi.MOp Priv (deBruijn.low_subst MessageSpi.deBruijn_subst d k (evalc_list sigma) t2)) (MessageSpi.MOp Priv (deBruijn.low_subst MessageSpi.deBruijn_subst d k (evalc_list sigma) t2)));auto.
    elim (MessageSpi.Ord.lt_not_eq (MessageSpi.MOp Priv (deBruijn.low_subst MessageSpi.deBruijn_subst d k (evalc_list sigma) t2)) (MessageSpi.MOp Priv (deBruijn.low_subst MessageSpi.deBruijn_subst d k (evalc_list sigma) t2)));auto.
    unfold MessageSpi.inverse_key.
    destruct (MessageSpi.Ord.compare (MessageSpi.MOp Pub t2) t);try (discriminate H0).
    rewrite <- e.
    rewrite MessageSpi.low_subst_MOp.
    injection H0;intros;subst t1.
    destruct (MessageSpi.Ord.compare);auto.
    elim (MessageSpi.Ord.lt_not_eq (MessageSpi.MOp Pub (deBruijn.low_subst MessageSpi.deBruijn_subst d k (evalc_list sigma) t2)) (MessageSpi.MOp Pub (deBruijn.low_subst MessageSpi.deBruijn_subst d k (evalc_list sigma) t2)));auto.
    elim (MessageSpi.Ord.lt_not_eq (MessageSpi.MOp Pub (deBruijn.low_subst MessageSpi.deBruijn_subst d k (evalc_list sigma) t2)) (MessageSpi.MOp Pub (deBruijn.low_subst MessageSpi.deBruijn_subst d k (evalc_list sigma) t2)));auto.
    rewrite ExpressionSpi.low_subst_EOp.
    rewrite evalc_EOp in H0.
    destruct (evalc e);try (discriminate H0).
    injection H0;intros;subst m.
    rewrite evalc_EOp.
    rewrite (IHe t);auto.
  Qed.

  Lemma evalc_succeeds_list_msg : forall (sigma:list MessageSpi.t),(List_forall evalc_succeeds (msg_subst sigma)).
    unfold msg_subst.
    induction sigma.
    simpl.
    apply List_forall_nil.
    simpl.
    apply List_forall_cons;auto.
    unfold evalc_succeeds.
    exists a.
    apply evalc_msg.
  Qed.

  Lemma evalc_list_msg : forall (sigma:list MessageSpi.t),(evalc_list (msg_subst sigma))=sigma.
    unfold msg_subst.
    induction sigma;try reflexivity.
    simpl.
    rewrite evalc_msg.
    rewrite IHsigma.
    auto.
  Qed.

  Lemma low_subst_msg_evalc : forall (sigma:list MessageSpi.t)(e:ExpressionSpi.t)(m:MessageSpi.t),(evalc e = Some m)->forall (d k:nat),(evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k (msg_subst sigma) e))=(Some (deBruijn.low_subst MessageSpi.deBruijn_subst d k sigma m)).
    intros.
    rewrite (low_subst_evalc (map expr_of_mesg sigma) (evalc_succeeds_list_msg sigma) e m H).
    fold (msg_subst sigma).
    rewrite evalc_list_msg.
    auto.
  Qed.

  Lemma low_subst_expr_of_msg : forall (m:MessageSpi.t)(d k:nat)(ms:list MessageSpi.t),(expr_of_mesg (deBruijn.low_subst MessageSpi.deBruijn_subst d k ms m))=(deBruijn.low_subst ExpressionSpi.deBruijn_subst d k (msg_subst ms) (expr_of_mesg m)).
    unfold msg_subst.
    induction m;intros;try reflexivity.
    simpl.
    destruct t;simpl.
    unfold Name.low_subst_generic.
    unfold DeBruijnNat.low_subst_gen.
    unfold DeBruijnNat.low_subst_generic.
    unfold DeBruijnNat.low_subst_generic_aux.
    case (le_lt_dec d n);intro;try reflexivity.
    rewrite List_nth_map.
    rewrite List_length_map.
    destruct (nth_error ms (n-d));try reflexivity.
    simpl.
    rewrite map_expr_of_msg;auto.
    rewrite MessageSpi.low_subst_MPair.
    lazy beta delta [expr_of_mesg] iota;fold expr_of_mesg.
    rewrite ExpressionSpi.low_subst_EPair.
    rewrite IHm1;rewrite IHm2;auto.
    rewrite MessageSpi.low_subst_MEnc_s.
    lazy beta delta [expr_of_mesg] iota;fold expr_of_mesg.
    rewrite ExpressionSpi.low_subst_EEnc_s.
    rewrite IHm1;rewrite IHm2;auto.
    rewrite MessageSpi.low_subst_MEnc_a.
    lazy beta delta [expr_of_mesg] iota;fold expr_of_mesg.
    rewrite ExpressionSpi.low_subst_EEnc_a.
    rewrite IHm1;rewrite IHm2;auto.
    rewrite MessageSpi.low_subst_MOp.
    lazy beta delta [expr_of_mesg] iota;fold expr_of_mesg.
    rewrite ExpressionSpi.low_subst_EOp.
    rewrite IHm;auto.
  Qed.

  (* substitutions and evala *)
  Lemma evala_subst_evala : forall (d k:nat)(sigma:list ExpressionSpi.t)(e:ExpressionSpi.t),(evala (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma (evala e)))=(evala (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma e)).
    induction e;intros.
    reflexivity.
    repeat (rewrite ExpressionSpi.low_subst_EPair || rewrite evala_EPair).
    rewrite IHe1;auto.
    rewrite IHe2;auto.
    repeat (rewrite ExpressionSpi.low_subst_EEnc_s || rewrite evala_EEnc_s).
    rewrite IHe1;auto.
    rewrite IHe2;auto.
    repeat (rewrite ExpressionSpi.low_subst_EEnc_a || rewrite evala_EEnc_a).
    rewrite IHe1;auto.
    rewrite IHe2;auto.
    rewrite ExpressionSpi.low_subst_EFst.
    rewrite evala_EFst.
    destruct (evala e);try (rewrite ExpressionSpi.low_subst_EFst;rewrite evala_EFst;rewrite IHe;auto;rewrite evala_EFst;auto).
    rewrite ExpressionSpi.low_subst_EPair in IHe.
    rewrite evala_EPair in IHe.
    rewrite evala_EFst.
    rewrite <- IHe;auto;trivial.
    rewrite ExpressionSpi.low_subst_ESnd.
    rewrite evala_ESnd.
    destruct (evala e);try (rewrite ExpressionSpi.low_subst_ESnd;rewrite evala_ESnd;rewrite IHe;auto;rewrite evala_ESnd;auto).
    rewrite ExpressionSpi.low_subst_EPair in IHe.
    rewrite evala_EPair in IHe.
    rewrite evala_ESnd.
    rewrite <- IHe;auto;trivial.
    rewrite ExpressionSpi.low_subst_EDec_s.
    rewrite evala_EDec_s.
    destruct (evala e1);try (rewrite ExpressionSpi.low_subst_EDec_s;rewrite evala_EDec_s;rewrite IHe1;auto;rewrite IHe2;auto;rewrite evala_EDec_s;auto).
    rewrite ExpressionSpi.low_subst_EEnc_s in IHe1.
    rewrite evala_EEnc_s in IHe1.
    rewrite evala_EDec_s.
    rewrite <- IHe1;auto.
    rewrite ExpressionSpi.low_subst_EDec_a.
    rewrite evala_EDec_a.
    destruct (evala e1);try (rewrite ExpressionSpi.low_subst_EDec_a;rewrite evala_EDec_a;rewrite IHe1;auto;rewrite IHe2;auto;rewrite evala_EDec_a;auto).
    rewrite ExpressionSpi.low_subst_EEnc_a in IHe1.
    rewrite evala_EEnc_a in IHe1.
    rewrite evala_EDec_a.
    rewrite <- IHe1;auto.
    repeat (rewrite ExpressionSpi.low_subst_EOp || rewrite evala_EOp).
    rewrite IHe;auto.
  Qed.

  Lemma evala_idempotent : forall (e:ExpressionSpi.t),(evala (evala e))=(evala e).
    intros.
    generalize (evala_subst_evala 0 0 nil e).
    repeat (rewrite low_subst_id).
    auto.
  Qed.

  Lemma evalc_subst_evala : forall (d k:nat)(sigma:list ExpressionSpi.t)(e:ExpressionSpi.t)(m:MessageSpi.t),(evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma e))=(Some m)->(evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma (evala e)))=(Some m).
    induction e;intros m.
    simpl.
    auto.
    rewrite evala_EPair.
    repeat (rewrite ExpressionSpi.low_subst_EPair).
    intro H.
    rewrite evalc_EPair in H.
    destruct (evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma e1));try (discriminate H).
    destruct (evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma e2));try (discriminate H).
    injection H;clear H;intros;subst m.
    rewrite evalc_EPair.
    rewrite (IHe1 t);auto.
    rewrite (IHe2 t0);auto.
    rewrite evala_EEnc_s.
    repeat (rewrite ExpressionSpi.low_subst_EEnc_s).
    intro H.
    rewrite evalc_EEnc_s in H.
    destruct (evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma e1));try (discriminate H).
    destruct (evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma e2));try (discriminate H).
    injection H;clear H;intros;subst m.
    rewrite evalc_EEnc_s.
    rewrite (IHe1 t);auto.
    rewrite (IHe2 t0);auto.
    rewrite evala_EEnc_a.
    repeat (rewrite ExpressionSpi.low_subst_EEnc_a).
    intro H.
    rewrite evalc_EEnc_a in H.
    destruct (evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma e1));try (discriminate H).
    destruct (evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma e2));try (discriminate H).
    injection H;clear H;intros;subst m.
    rewrite evalc_EEnc_a.
    rewrite (IHe1 t);auto.
    rewrite (IHe2 t0);auto.
    rewrite ExpressionSpi.low_subst_EFst.
    intro H.
    rewrite evalc_EFst in H.
    case (Option_elim_option MessageSpi.t (evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma e)));intro.
    ex_elim s t.
    rewrite s in H.
    destruct t;try discriminate H.
    injection H;clear H;intros;subst m.
    rewrite evala_EFst.
    destruct (evala e);try (rewrite ExpressionSpi.low_subst_EFst);try (rewrite evalc_EFst;rewrite (IHe (MessageSpi.MPair t1 t2) s);trivial).
    generalize (IHe (MessageSpi.MPair t1 t2) s).
    rewrite ExpressionSpi.low_subst_EPair.
    rewrite evalc_EPair.
    intro.
    destruct (evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma t3));try (discriminate H).
    destruct (evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma t4));try (discriminate H).
    injection H;intros;congruence.
    rewrite e0 in H;discriminate H.
    rewrite ExpressionSpi.low_subst_ESnd.
    intro H.
    rewrite evalc_ESnd in H.
    case (Option_elim_option MessageSpi.t (evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma e)));intro.
    ex_elim s t.
    rewrite s in H.
    destruct t;try discriminate H.
    injection H;clear H;intros;subst m.
    rewrite evala_ESnd.
    destruct (evala e);try (rewrite ExpressionSpi.low_subst_ESnd);try (rewrite evalc_ESnd;rewrite (IHe (MessageSpi.MPair t1 t2) s);trivial).
    generalize (IHe (MessageSpi.MPair t1 t2) s).
    rewrite ExpressionSpi.low_subst_EPair.
    rewrite evalc_EPair.
    intro.
    destruct (evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma t3));try (discriminate H).
    destruct (evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma t4));try (discriminate H).
    injection H;intros;congruence.
    rewrite e0 in H;discriminate H.
    rewrite ExpressionSpi.low_subst_EDec_s.
    intro H.
    rewrite evalc_EDec_s in H.
    case (Option_elim_option MessageSpi.t (evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma e1)));intro.
    ex_elim s t.
    rewrite s in H.
    destruct t;try discriminate H.
    case (Option_elim_option MessageSpi.t (evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma e2)));intro.
    ex_elim s0 t0.
    rewrite s0 in H;try (discriminate H).
    destruct (MessageSpi.Ord.compare t2 t0);try (discriminate H).
    rewrite <- e in s0.
    injection H;clear H;intros;subst m.
    rewrite evala_EDec_s.
    destruct (evala e1);try (rewrite ExpressionSpi.low_subst_EDec_s);try (rewrite evalc_EDec_s;rewrite (IHe1 (MessageSpi.MEnc_s t1 t2) s);rewrite (IHe2 t2 s0);case (MessageSpi.Ord.compare t2 t2);trivial;intro;elim (MessageSpi.Ord.lt_not_eq t2 t2);auto).   
    generalize (IHe1 (MessageSpi.MEnc_s t1 t2) s).
    rewrite ExpressionSpi.low_subst_EEnc_s.
    rewrite evalc_EEnc_s.
    intro H.
    destruct (evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma t3));try (discriminate H).
    destruct (evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma t4));try (discriminate H).
    injection H;intros;congruence.
    rewrite e in H;discriminate H.
    rewrite e in H;discriminate H.
    rewrite ExpressionSpi.low_subst_EDec_a.
    intro H.
    rewrite evalc_EDec_a in H.
    case (Option_elim_option MessageSpi.t (evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma e1)));intro.
    ex_elim s t.
    rewrite s in H.
    destruct t;try discriminate H.
    case (Option_elim_option MessageSpi.t (evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma e2)));intro.
    ex_elim s0 t0.
    rewrite s0 in H;try (discriminate H).
    case (Option_elim_option MessageSpi.t (MessageSpi.inverse_key t2));intro.
    ex_elim s1 t'2.
    rewrite s1 in H.
    destruct (MessageSpi.Ord.compare t'2 t0);try (discriminate H).
    injection H;clear H;intros;subst m.
    rewrite <- e in s0.
    rewrite evala_EDec_a.
    destruct (evala e1);try (rewrite ExpressionSpi.low_subst_EDec_a);try (rewrite evalc_EDec_a;rewrite (IHe1 (MessageSpi.MEnc_a t1 t2) s);rewrite (IHe2 t'2 s0);rewrite s1;case (MessageSpi.Ord.compare t'2 t'2);trivial;intro;elim (MessageSpi.Ord.lt_not_eq t'2 t'2);auto).
    generalize (IHe1 (MessageSpi.MEnc_a t1 t2) s).
    rewrite ExpressionSpi.low_subst_EEnc_a.
    rewrite evalc_EEnc_a.
    intro H.
    destruct (evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma t3));try (discriminate H).
    destruct (evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma t4));try (discriminate H).
    injection H;intros;congruence.
    rewrite e in H;discriminate H.
    rewrite e in H;discriminate H.
    rewrite e in H;discriminate H.
    rewrite evala_EOp.
    repeat (rewrite ExpressionSpi.low_subst_EOp).
    intro H.
    rewrite evalc_EOp in H.
    destruct (evalc (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k sigma e));try (discriminate H).
    injection H;clear H;intros;subst m.
    rewrite evalc_EOp.
    rewrite (IHe t);auto.
  Qed.
End SubstSpi.




