(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import deBruijnType.
Require Import name.
Require Import missing.
Require Import Arith.
Require Import tactics.

Module SubstLemmaFacts.
  Section subst_lemma_facts.
    Hypothesis t : Set.
    Hypothesis deBruijn_t : deBruijn.deBruijn_base t.
    Hypothesis to : Set.
    Hypothesis deBruijn_to : deBruijn.deBruijn_base to.
    Hypothesis t_subst_to : deBruijn.deBruijn_subst t deBruijn_t to deBruijn_to.
    
    Hypothesis t_low_subst_split_lemma : forall (x:t)(d k:nat)(es fs:list to),deBruijn.low_subst t_subst_to d 0 es (deBruijn.low_subst t_subst_to (d+length es) k fs x)=deBruijn.low_subst t_subst_to d k (app es fs) x.
    
    Lemma low_subst_snoc_lemma : forall (x:t)(d k:nat)(e:to)(es:list to),deBruijn.low_subst t_subst_to d k (app es (cons e nil)) x =  deBruijn.low_subst t_subst_to d 0 (e :: nil) (deBruijn.swapR deBruijn_t d k (deBruijn.low_subst t_subst_to d k (map (deBruijn.lift deBruijn_to k 1) es) x)).
      intros.
      assert ((length (app es (cons e nil)))=(S (length es))).
      rewrite List_length_append.
      rewrite plus_comm;reflexivity.
      assert (exists e0:to,exists es0:list to,(app es (cons e nil))=(cons e0 es0)).
      destruct es.
      exists e.
      exists (nil (A:=to)).
      reflexivity.
      exists t0.
      exists (app es (cons e nil)).
      simpl.
      auto.
      ex_elim H0 e0.
      ex_elim H0 es0.
      rewrite H0.
      generalize (deBruijn.low_subst_swapR_rotate t_subst_to d k e0 es0 x).
      rewrite <- H0.
      rewrite List_right_rotate_simpl.
      rewrite H0 in H.
      simpl in H.
      injection H;clear H;intros.
      rewrite H in H1.
      generalize (t_low_subst_split_lemma (deBruijn.swapR deBruijn_t d (length es) x) d k (cons e nil) es).
      intro.
      simpl in H2.
      rewrite plus_comm in H2;simpl in H2.
      rewrite H1 in H2.
      symmetry in H2.
      generalize (deBruijn.low_subst_swapR_simpl t_subst_to d k es x).
      intro.
      rewrite H3 in H2.
      auto.
    Qed.
  End subst_lemma_facts.
End SubstLemmaFacts.

Module SubstSplitLemmaMap.
  Section subst_split_lemma.
    Hypothesis t : Set.
    Hypothesis deBruijn_t : deBruijn.deBruijn_base t.
    Hypothesis to : Set.
    Hypothesis deBruijn_to : deBruijn.deBruijn_base to.
    Hypothesis t_subst_to : deBruijn.deBruijn_subst t deBruijn_t to deBruijn_to.
    Hypothesis t' : Set.
    Hypothesis map_fun: (nat->t->t)->nat->t'->t'.
    Hypothesis map_ext:deBruijn.map_struct t t' map_fun.
    Definition deBruijn_t' := deBruijn.deBruijn_map_struct t deBruijn_t t' map_fun map_ext.
    Definition t'_subst_to := deBruijn.deBruijn_s_map_struct t deBruijn_t to deBruijn_to t_subst_to t' map_fun map_ext.
    Hypothesis to_subst_to : deBruijn.deBruijn_subst to deBruijn_to to deBruijn_to.
    
    Hypothesis t_low_subst_split_lemma : forall (x:t)(d k:nat)(es fs:list to),deBruijn.low_subst t_subst_to d 0 es (deBruijn.low_subst t_subst_to (d+length es) k fs x)=deBruijn.low_subst t_subst_to d k (app es fs) x.

    Lemma low_subst_split_lemma : forall (x:t')(d k:nat)(es fs:list to),deBruijn.low_subst t'_subst_to d 0 es (deBruijn.low_subst t'_subst_to (d+length es) k fs x)=deBruijn.low_subst t'_subst_to d k (app es fs) x.
      intros.
      simpl.
      generalize (deBruijn.map_compose_simpl map_ext 0 (length es) 0 (fun d => deBruijn.low_subst t_subst_to d 0 es) (fun d => deBruijn.low_subst t_subst_to d k fs) (fun d => deBruijn.low_subst t_subst_to d k (app es fs))).
      simpl;intros.
      rewrite plus_comm.
      rewrite H.
      auto.
      intros.
      rewrite plus_comm;auto.
    Qed.
  End subst_split_lemma.
End SubstSplitLemmaMap.

Module SubstSplitLemmaMap2.
  Section subst_split_lemma.
    Hypothesis t1 : Set.
    Hypothesis deBruijn_t1 : deBruijn.deBruijn_base t1.
    Hypothesis t2 : Set.
    Hypothesis deBruijn_t2 : deBruijn.deBruijn_base t2.
    Hypothesis to : Set.
    Hypothesis deBruijn_to : deBruijn.deBruijn_base to.
    Hypothesis t1_subst_to : deBruijn.deBruijn_subst t1 deBruijn_t1 to deBruijn_to.
    Hypothesis t2_subst_to : deBruijn.deBruijn_subst t2 deBruijn_t2 to deBruijn_to.
    Hypothesis t : Set.
    Hypothesis map2_fun: (nat->t1->t1)->(nat->t2->t2)->nat->t->t.
    Hypothesis map2_ext:deBruijn.map2_struct t1 t2 t map2_fun.
    Definition deBruijn_t := deBruijn.deBruijn_map2_struct t1 deBruijn_t1 t2 deBruijn_t2 t map2_fun map2_ext.
    Definition t_subst_to := deBruijn.deBruijn_s_map2_struct t1 deBruijn_t1 t2 deBruijn_t2 to deBruijn_to t1_subst_to t2_subst_to t map2_fun map2_ext.
    Hypothesis to_subst_to : deBruijn.deBruijn_subst to deBruijn_to to deBruijn_to.
    
    Hypothesis t1_low_subst_split_lemma : forall (x:t1)(d k:nat)(es fs:list to),deBruijn.low_subst t1_subst_to d 0 es (deBruijn.low_subst t1_subst_to (d+length es) k fs x)=deBruijn.low_subst t1_subst_to d k (app es fs) x.
    Hypothesis t2_low_subst_split_lemma : forall (x:t2)(d k:nat)(es fs:list to),deBruijn.low_subst t2_subst_to d 0 es (deBruijn.low_subst t2_subst_to (d+length es) k fs x)=deBruijn.low_subst t2_subst_to d k (app es fs) x.

    Lemma low_subst_split_lemma : forall (x:t)(d k:nat)(es fs:list to),deBruijn.low_subst t_subst_to d 0 es (deBruijn.low_subst t_subst_to (d+length es) k fs x)=deBruijn.low_subst t_subst_to d k (app es fs) x.
      intros.
      simpl.
      generalize (deBruijn.map2_compose_simpl map2_ext 0 (length es) 0 (fun d => deBruijn.low_subst t1_subst_to d 0 es) (fun d => deBruijn.low_subst t1_subst_to d k fs) (fun d => deBruijn.low_subst t1_subst_to d k (app es fs)) (fun d => deBruijn.low_subst t2_subst_to d 0 es) (fun d => deBruijn.low_subst t2_subst_to d k fs) (fun d => deBruijn.low_subst t2_subst_to d k (app es fs))).
      simpl;intros.
      rewrite plus_comm.
      rewrite H.
      auto.
      intros.
      rewrite plus_comm;auto.
      intros.
      rewrite plus_comm;auto.
    Qed.
  End subst_split_lemma.
End SubstSplitLemmaMap2.

Module SubstSplitLemmaGen.
  Section subst_split_lemma.
    Hypothesis to : Set.
    Hypothesis map_fun : (nat->Name.t->Name.t)->nat->to->to.
    Hypothesis map_str : deBruijn.map_struct Name.t to map_fun.
    Definition deBruijn_to :=  (deBruijn.deBruijn_map_struct Name.t Name.deBruijn_base to map_fun map_str).
    Hypothesis Hextension : deBruijn.deBruijn_extension Name.t Name.deBruijn_base to deBruijn_to.
    Hypothesis deBruijn_ext : deBruijn.deBruijn_extension Name.t Name.deBruijn_base to deBruijn_to.
    Hypothesis map_gen_fun : (nat->Name.t->to)->nat->to->to.
    Hypothesis map_ext : deBruijn.map_extension Name.t Name.deBruijn_base to deBruijn_to deBruijn_ext map_fun map_str map_gen_fun.
    Definition t_subst_to := Name.deBruijn_subst_gen to deBruijn_to deBruijn_ext.
    Definition to_subst_to := deBruijn.deBruijn_subst_gen_subst Name.t Name.deBruijn_base to map_fun map_str deBruijn_ext t_subst_to map_gen_fun map_ext.
    
    Lemma low_subst_name : forall (d k:nat)(es:list to)(x:Name.t),(deBruijn.low_subst_gen t_subst_to d k es x)=(Name.low_subst_generic to (fun d0 k0 : nat => map_fun (fun d1 : nat => Name.lift d1 k0) d0) (deBruijn.j deBruijn_ext) d k es x).
      intros.
      reflexivity.
    Qed.
    
    Lemma low_subst_j : forall (d k:nat)(es:list to)(x:Name.t),deBruijn.low_subst to_subst_to d k es (deBruijn.j deBruijn_ext x) =  Name.low_subst_generic to (fun d0 k0 : nat => map_fun (fun d1 : nat => Name.lift d1 k0) d0) (deBruijn.j deBruijn_ext) d k es x. 
      simpl.
      intros.
      rewrite (deBruijn.map_gen_j map_ext);auto.
    Qed.

    Lemma low_subst_split_lemma : forall (x:Name.t)(d k:nat)(es:list to)(fs:list to),deBruijn.low_subst to_subst_to d 0 es (deBruijn.low_subst_gen t_subst_to (d+length es) k fs x)=deBruijn.low_subst_gen t_subst_to d k (app es fs) x.
      intros.
      destruct x as [x].
      repeat (rewrite low_subst_name).
      unfold Name.low_subst_generic.
      unfold DeBruijnNat.low_subst_gen.
      unfold DeBruijnNat.low_subst_generic.
      unfold DeBruijnNat.low_subst_generic_aux.
      case le_lt_dec;intro.
      case le_lt_dec;intro.
      elim (List_nth_Some to fs (x-(d+length es)));intros.
      case (Option_elim_option to (nth_error fs (x-(d+length es))));intro H1;try (rewrite H1).
      ex_elim H1 f.
      rewrite H1.
      assert (x-(d+length es) < length fs).
      apply H0.
      exists f.
      auto.
      assert ((x-(d+length es)+length es) = (x-d)).
      omega.
      rewrite <- H3.
      rewrite List_nth_app_right.
      rewrite H1.
      generalize (deBruijn.low_subst_lift_nil to_subst_to d 0 0 es (map_fun (fun d0 => Name.lift d0 d) 0 f)).
      simpl.
      generalize (deBruijn.lift_compose deBruijn_to 0 d (length es) 0 f).
      simpl.
      repeat (rewrite plus_0_r).
      intro.
      rewrite H4.
      intro.
      rewrite H5.
      generalize (deBruijn.lift_0 deBruijn_to).
      simpl.
      intro.
      rewrite H6.
      auto.
      assert (length fs <= x-(d+length es)).
      case (le_lt_dec (length fs) (x-(d+length es)));auto.
      intros.
      elim (H l1).
      intros.
      rewrite H1 in H2.
      discriminate H2.
      assert ((x-(d+length es)+length es) = (x-d)).
      omega.
      rewrite <- H3.
      rewrite List_nth_app_right.
      rewrite H1.
      rewrite low_subst_j.      
      unfold Name.low_subst_generic.
      unfold DeBruijnNat.low_subst_gen.
      unfold DeBruijnNat.low_subst_generic.
      unfold DeBruijnNat.low_subst_generic_aux.
      case le_lt_dec;intro;try_absurd_arith.      
      assert (length es <= x - length fs - d).
      clear H;clear H0.
      clear H1.
      generalize (plus_le_compat_r _ _ (length es) H2).
      rewrite H3.
      intro.
      rewrite plus_comm in H.
      generalize (Arith_le_minus_le _ _ _ H).
      intro.
      rewrite Arith_minus_minus in H0.
      rewrite plus_comm in H0.
      rewrite <- Arith_minus_minus in H0.
      apply H0.
      assert (length es <= x - length fs + k - d).
      eapply le_trans.
      apply H4.
      apply Arith_le_plus_minus.
      elim (List_nth_Some to es (x-length fs + k -d));intros.
      case (Option_elim_option to (nth_error es (x-length fs+k-d)));intro H8;try (rewrite H8).
      ex_elim H8 e.
      rewrite H8.
      assert (x - length fs + k - d < length es).
      apply H7.
      exists e.
      auto.
      elim (lt_irrefl (length es)).
      eapply le_lt_trans.
      apply H5.
      auto.
      rewrite List_length_append.
      rewrite plus_0_r.
      rewrite <- Arith_minus_minus.
      assert ((x - length fs + k - length es)=(x - length es - length fs + k)).
      clear H;clear H0.
      clear H6;clear H7.
      clear H1;clear H8.
      rewrite Arith_minus_minus.
      rewrite (plus_comm (length es) (length fs)).
      rewrite <- Arith_minus_minus.
      assert (length es <= x - length fs).
      eapply le_trans.
      apply H4.
      apply le_minus.
      generalize (le_plus_minus _ _ H).
      intro.
      rewrite H0.
      rewrite <- plus_assoc.
      rewrite minus_plus.
      rewrite minus_plus.
      trivial.
      rewrite H9;auto.
      try_absurd_arith.
      case le_lt_dec;intro.
      rewrite List_nth_app_left;try omega.
      rewrite low_subst_j.
      unfold Name.low_subst_generic.
      unfold DeBruijnNat.low_subst_gen.
      unfold DeBruijnNat.low_subst_generic.
      unfold DeBruijnNat.low_subst_generic_aux.
      case le_lt_dec;intro;try_absurd_arith.
      elim (List_nth_Some to es (x-d));intros.
      assert (x-d < length es);try omega.
      elim (H H1).
      intro e;intros.
      rewrite H2.
      auto.
      rewrite low_subst_j.
      unfold Name.low_subst_generic.
      unfold DeBruijnNat.low_subst_gen.
      unfold DeBruijnNat.low_subst_generic.
      unfold DeBruijnNat.low_subst_generic_aux.
      case le_lt_dec;intro;try_absurd_arith.
      auto.
    Qed.
    Transparent deBruijn_to.
    Transparent t_subst_to.
    Transparent to_subst_to.
  End subst_split_lemma.
End SubstSplitLemmaGen.

