(* A formalisation of the spi calculus in Coq *)
(* Copyright (C) 2005-2007 Sebastien Briais *)
(* http://lamp.epfl.ch/~sbriais/ *)

(* This library is free software; you can redistribute it and/or modify *)
(* it under the terms of the GNU Lesser General Public License as *)
(* published by the Free Software Foundation; either version 2.1 of the *)
(* License, or (at your option) any later version. *)

(* This library is distributed in the hope that it will be useful, but *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *)
(* Lesser General Public License for more details. *)

(* You should have received a copy of the GNU Lesser General Public *)
(* License along with this library; if not, write to the Free Software *)
(* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA *)
(* 02110-1301 USA *)

Require Import name.
Require Import expressionSpi.
Require Import messageSpi.
Require Import formulaeSpi.
Require Import deBruijnType.
Require Import calculi.
Require Import List.
Require Import actionSpiNC.

Lemma SpiCalcul_low_subst_Input_aux : forall (e:ExpressionSpi.t)(p:SpiCalcul.t)(d k:nat)(es:list ExpressionSpi.t),(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es (SpiCalcul.Input e p))=(SpiCalcul.Input (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es e) (deBruijn.low_subst SpiCalcul.deBruijn_subst (S d) k es p)).
  intros.
  rewrite SpiCalcul.low_subst_Input.
  simpl.
  auto.
Qed.

Lemma SpiCalcul_low_subst_Output_aux : forall (d k:nat)(es:list ExpressionSpi.t)(e1 e2:ExpressionSpi.t)(p:SpiCalcul.t),(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es (SpiCalcul.Output e1 e2 p))=(SpiCalcul.Output (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es e1) (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es e2) (deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p)).
  intros.
  rewrite SpiCalcul.low_subst_Output.
  simpl.
  auto.
Qed.

Lemma SpiCalcul_low_subst_IfThen_aux : forall (d k:nat)(es:list ExpressionSpi.t)(phi:FormulaeSpi.t)(p:SpiCalcul.t),(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es (SpiCalcul.IfThen phi p))=(SpiCalcul.IfThen (deBruijn.low_subst FormulaeSpi.deBruijn_subst d k es phi) (deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p)).
  intros.
  rewrite SpiCalcul.low_subst_IfThen.
  simpl.
  auto.
Qed.

Lemma AgentSpi_low_subst_C_aux : forall (d k:nat)(es:list ExpressionSpi.t)(n:nat)(m:ExpressionSpi.t)(p:SpiCalcul.t),deBruijn.low_subst AgentSpi.deBruijn_subst d k es (AgentSpi.C (AgentSpi.Conc n m p)) = AgentSpi.C (AgentSpi.Conc n (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d+n) k es m) (deBruijn.low_subst SpiCalcul.deBruijn_subst (d+n) k es p)).
  intros.
  rewrite AgentSpi.low_subst_C.
  simpl.
  auto.
Qed.

Lemma SpiCalcul_low_subst_Tau_NC : forall (d k:nat)(es:list ExpressionSpi.t)(p:SpiCalcul.t),(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es (SpiCalcul.Tau p))=(ActionSpiNC.A.P.Tau (deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p)).
  intros.
  reflexivity.
Qed.

Lemma SpiCalcul_low_subst_Input_NC : forall (d k:nat)(es:list ExpressionSpi.t)(e:ExpressionSpi.t)(p:SpiCalcul.t),(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es (SpiCalcul.Input e p))=(ActionSpiNC.A.P.Input (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es e) (deBruijn.low_subst SpiCalcul.deBruijn_subst (S d) k es p)).
  intros.
  simpl.
  auto.
Qed.

Lemma SpiCalcul_low_subst_Output_NC : forall (d k:nat)(es:list ExpressionSpi.t)(e f:ExpressionSpi.t)(p:SpiCalcul.t),(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es (SpiCalcul.Output e f p))=(ActionSpiNC.A.P.Output (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es e) (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es f) (deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p)).
  intros.
  simpl.
  auto.
Qed.

Lemma SpiCalcul_low_subst_IfThen_NC : forall (d k:nat)(es:list ExpressionSpi.t)(phi:FormulaeSpi.t)(p:SpiCalcul.t),(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es (SpiCalcul.IfThen phi p))=(ActionSpiNC.A.P.IfThen (deBruijn.low_subst FormulaeSpi.deBruijn_subst d k es phi) (deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p)).
  intros.
  simpl.
  auto.
Qed.

Lemma SpiCalcul_low_subst_Sum_NC : forall (d k:nat)(es:list ExpressionSpi.t)(p q:SpiCalcul.t),(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es (SpiCalcul.Sum p q))=(ActionSpiNC.A.P.Sum (deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p) (deBruijn.low_subst SpiCalcul.deBruijn_subst d k es q)).
  intros.
  simpl.
  auto.
Qed.

Lemma SpiCalcul_low_subst_Par_NC : forall (d k:nat)(es:list ExpressionSpi.t)(p q:SpiCalcul.t),(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es (SpiCalcul.Par p q))=(ActionSpiNC.A.P.Par (deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p) (deBruijn.low_subst SpiCalcul.deBruijn_subst d k es q)).
  intros.
  simpl.
  auto.
Qed.

Lemma SpiCalcul_low_subst_New_NC : forall (d k:nat)(es:list ExpressionSpi.t)(p:SpiCalcul.t),(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es (SpiCalcul.New p))=(ActionSpiNC.A.P.New (deBruijn.low_subst SpiCalcul.deBruijn_subst (S d) k es p)).
  intros.
  simpl.
  auto.
Qed.

Lemma SpiCalcul_low_subst_Bang_NC : forall (d k:nat)(es:list ExpressionSpi.t)(p:SpiCalcul.t),(deBruijn.low_subst SpiCalcul.deBruijn_subst d k es (SpiCalcul.Bang p))=(ActionSpiNC.A.P.Bang (deBruijn.low_subst SpiCalcul.deBruijn_subst d k es p)).
  intros.
  simpl.
  auto.
Qed.

Lemma AgentSpi_low_subst_F_NC : forall (d k:nat)(es:list ExpressionSpi.t)(p:SpiCalcul.t),(deBruijn.low_subst AgentSpi.deBruijn_subst d k es (AgentSpi.F (AgentSpi.Abs p)))=(ActionSpiNC.A.F (ActionSpiNC.A.Abs (deBruijn.low_subst SpiCalcul.deBruijn_subst (S d) k es p))).
  intros.
  simpl.
  auto.
Qed.

Lemma AgentSpi_low_subst_C_NC : forall (d k:nat)(es:list ExpressionSpi.t)(n:nat)(e:ExpressionSpi.t)(p:SpiCalcul.t),(deBruijn.low_subst AgentSpi.deBruijn_subst d k es (AgentSpi.C (AgentSpi.Conc n e p)))=(ActionSpiNC.A.C (ActionSpiNC.A.Conc n (deBruijn.low_subst ExpressionSpi.deBruijn_subst (d+n) k es e) (deBruijn.low_subst SpiCalcul.deBruijn_subst (d+n) k es p))).
  intros.
  simpl.
  auto.
Qed.

Lemma Abs_NC_Abs_Spi : forall (p:SpiCalcul.t),(ActionSpiNC.A.Abs p)=(AgentSpi.Abs p).
  intros.
  auto.
Qed.

Lemma Conc_NC_Conc_Spi : forall (n:nat)(e:ExpressionSpi.t)(p:SpiCalcul.t),(ActionSpiNC.A.Conc n e p)=(AgentSpi.Conc n e p).
  intros.
  auto.
Qed.

Lemma F_NC_F_Spi: forall (f:AgentSpi.abstraction),(ActionSpiNC.A.F f)=(AgentSpi.F f).
  intros;auto.
Qed.

Lemma C_NC_C_Spi: forall (c:AgentSpi.concretion),(ActionSpiNC.A.C c)=(AgentSpi.C c).
  intros;auto.
Qed.

Lemma low_subst_AgentSpi : forall (d k:nat)(es:list ExpressionSpi.t)(e:ExpressionSpi.t),(deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es e)=(deBruijn.low_subst AgentSpi.P.E.deBruijn_subst d k es e).
  intros.
  simpl.
  auto.
Qed.

Lemma low_subst_AgentSpi_bis : forall (d k:nat)(es:list ExpressionSpi.t)(e:ExpressionSpi.t),(deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es e)=(deBruijn.low_subst CongruenceSpi.A.P.E.deBruijn_subst d k es e).
  intros.
  simpl.
  auto.
Qed.

Lemma AgentSpi_low_subst_Input : forall (d k:nat)(es:list ExpressionSpi.t)(e:ExpressionSpi.t)(p:AgentSpi.P.t),(deBruijn.low_subst AgentSpi.P.deBruijn_subst d k es (AgentSpi.P.Input e p))=(AgentSpi.P.Input (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es e) (deBruijn.low_subst AgentSpi.P.deBruijn_subst (S d) k es p)).
  intros.
  rewrite AgentSpi.P.low_subst_Input.
  rewrite <- low_subst_AgentSpi.
  auto.
Qed.

Lemma AgentSpi_low_subst_Output : forall (d k:nat)(es:list ExpressionSpi.t)(e f:ExpressionSpi.t)(p:AgentSpi.P.t),(deBruijn.low_subst AgentSpi.P.deBruijn_subst d k es (AgentSpi.P.Output e f p))=(AgentSpi.P.Output (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es e) (deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es f)  (deBruijn.low_subst AgentSpi.P.deBruijn_subst d k es p)).
  intros.
  rewrite AgentSpi.P.low_subst_Output.
  rewrite <- low_subst_AgentSpi.
  assert ((deBruijn.low_subst AgentSpi.P.E.deBruijn_subst d k es f)=(deBruijn.low_subst ExpressionSpi.deBruijn_subst d k es f)).
  rewrite <- low_subst_AgentSpi.
  auto.
  rewrite H.
  auto.
Qed.

Lemma AgentSpi_low_subst_IfThen : forall (d k:nat)(es:list ExpressionSpi.t)(phi:FormulaeSpi.t)(p:AgentSpi.P.t),(deBruijn.low_subst AgentSpi.P.deBruijn_subst d k es (AgentSpi.P.IfThen phi p))=(AgentSpi.P.IfThen (deBruijn.low_subst FormulaeSpi.deBruijn_subst d k es phi) (deBruijn.low_subst AgentSpi.P.deBruijn_subst d k es p)).
  intros.
  rewrite AgentSpi.P.low_subst_IfThen.
  auto.
Qed.






